Module:       llvm-tablegen
Author:       Peter S. Housel
Copyright:    Original Code is Copyright 2009-2011 Gwydion Dylan Maintainers
              All rights reserved.
License:      See License.txt in this distribution for details.
Warranty:     Distributed WITHOUT WARRANTY OF ANY KIND

define class <tablegen-class-reference> (<object>)
  constant slot class-reference-class :: type-union(<tablegen-class>, <tablegen-multi-class>),
    required-init-keyword: class:;
  constant slot class-reference-template-arguments :: <sequence>,
    init-value: #(), init-keyword: template-arguments:;
end class;

define variable *tablegen-overrides* = make(<string-table>);

define function add-initializer
    (sequence :: <stretchy-sequence>,
     environment :: <string-table>,
     initializer :: <tablegen-initializer>)
 => (value :: <tablegen-value>);
  let value
    = element(*tablegen-overrides*, initializer.initializer-name, default: #f)
    | resolve-value(environment, initializer.initializer-value);
  add!(sequence, make(<tablegen-initializer>,
                      name: initializer.initializer-name,
                      type: initializer.initializer-type,
                      value: value));
  value
end function;

define function instantiate-class-reference
    (sequence :: <stretchy-sequence>,
     environment :: <string-table>,
     reference :: <tablegen-class-reference>,
     resolve? :: <boolean>)
 => ();
  let class = reference.class-reference-class;
  let template-arguments
    = reference.class-reference-template-arguments;
  for (parameter in class.record-template-parameters, index from 0)
    let parameter-name
      = format-to-string("%s:%s",
                         class.record-name,
                         parameter.initializer-name);
    let parameter-value
      = element(template-arguments, index, default: #f);
    if (parameter-value)
      environment[parameter-name] := parameter-value;
    else
      let value = resolve-value(environment, parameter.initializer-value);
      add!(sequence,
           make(<tablegen-initializer>,
                name: parameter-name,
                type: parameter.initializer-type,
                value: value));
      if (resolve?)
        environment[parameter-name] := value;
      end if;
    end if;
  end for;

  for (initializer in class.record-initializers)
    let value = add-initializer(sequence, environment, initializer);
    if (resolve?)
      environment[initializer.initializer-name] := value;
    end if;
  end for;
end function;

define variable *tablegen-anonymous-counter* :: <integer> = 0;
define variable *tablegen-anonymous-value-counter* :: <integer> = 0;

define constant $EOF-token = 0;
define constant $%LBRACE-token = 1;
define constant $%RBRACE-token = 2;
define constant $%CLASS-token = 3;
define constant $%ID-token = 4;
define constant $%LANGLE-token = 5;
define constant $%RANGLE-token = 6;
define constant $%COMMA-token = 7;
define constant $%FIELD-token = 8;
define constant $%STRING-token = 9;
define constant $%BIT-token = 10;
define constant $%BITS-token = 11;
define constant $%INTVAL-token = 12;
define constant $%INT-token = 13;
define constant $%LIST-token = 14;
define constant $%CODE-token = 15;
define constant $%DAG-token = 16;
define constant $%EQ-token = 17;
define constant $%STRVAL-token = 18;
define constant $%QUEST-token = 19;
define constant $%LBRACK-token = 20;
define constant $%RBRACK-token = 21;
define constant $%DOT-token = 22;
define constant $%LPAREN-token = 23;
define constant $%RPAREN-token = 24;
define constant $%CARTOK-token = 25;
define constant $%CDRTOK-token = 26;
define constant $%NULLTOK-token = 27;
define constant $%CASTTOK-token = 28;
define constant $%CONCATTOK-token = 29;
define constant $%SHLTOK-token = 30;
define constant $%SRATOK-token = 31;
define constant $%SRLTOK-token = 32;
define constant $%EQTOK-token = 33;
define constant $%STRCONCATTOK-token = 34;
define constant $%NAMECONCATTOK-token = 35;
define constant $%IFTOK-token = 36;
define constant $%FOREACHTOK-token = 37;
define constant $%SUBSTTOK-token = 38;
define constant $%MINUS-token = 39;
define constant $%COLON-token = 40;
define constant $%VARNAME-token = 41;
define constant $%SEMI-token = 42;
define constant $%LET-token = 43;
define constant $%DEF-token = 44;
define constant $%MULTICLASS-token = 45;
define constant $%DEFM-token = 46;
define constant $%IN-token = 47;

define function tablegen-parser-action0 (arg$1, arg$2) => (value)
  #f
end tablegen-parser-action0;

define function tablegen-parser-action1 (arg$1) => (value)
  #f
end tablegen-parser-action1;

define function tablegen-parser-action3 (arg$1) => (value)
  arg$1("", make(<string-table>))
end tablegen-parser-action3;

define function tablegen-parser-action7 (arg$1, arg$2, arg$3, arg$4) => (value)
  *tablegen-overrides* := arg$1
end tablegen-parser-action7;

define function tablegen-parser-action8 (arg$1, arg$2) => (value)
  *tablegen-overrides* := arg$1
end tablegen-parser-action8;

define function tablegen-parser-action9 (arg$1, arg$2, arg$3, arg$4) => (value)
  method (prefix :: <string>, environment :: <string-table>)
         // Alpha-convert template parameters
         for (parameter in arg$3)
           environment[parameter.initializer-name]
             := make(<tablegen-named-value>,
                     name: format-to-string("%s:%s",
                                            arg$2,
                                            parameter.initializer-name));
         end for;

         let initializers = make(<stretchy-object-vector>);

         for (reference in arg$4.head)
           instantiate-class-reference
             (initializers, environment, reference, #f);
         end for;

         do(curry(add-initializer, initializers, environment), arg$4.tail);
         $tablegen-classes[arg$2]
           := make(<tablegen-class>,
                   name: arg$2,
                   template-parameters: arg$3,
                   superclasses: map(class-reference-class, arg$4.head),
                   initializers: initializers);
       end
end tablegen-parser-action9;

define function tablegen-parser-action10 (arg$1) => (value)
  arg$1
end tablegen-parser-action10;

define function tablegen-parser-action12 () => (value)
  begin
         let name
           = format-to-string("anonymous.%d", *tablegen-anonymous-counter*);
         *tablegen-anonymous-counter* := *tablegen-anonymous-counter* + 1;
         name
       end
end tablegen-parser-action12;

define function tablegen-parser-action14 () => (value)
  #()
end tablegen-parser-action14;

define function tablegen-parser-action16 (arg$1, arg$2, arg$3) => (value)
  reverse!(arg$2)
end tablegen-parser-action16;

define function tablegen-parser-action17 (arg$1) => (value)
  list(arg$1)
end tablegen-parser-action17;

define function tablegen-parser-action18 (arg$1, arg$2, arg$3) => (value)
  add(arg$1, arg$3)
end tablegen-parser-action18;

define function tablegen-parser-action19 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<tablegen-initializer>, type: arg$2, name: arg$3, value: arg$4)
end tablegen-parser-action19;

define function tablegen-parser-action20 () => (value)
  #f
end tablegen-parser-action20;

define function tablegen-parser-action22 (arg$1) => (value)
  make(<tablegen-simple-type>, kind: #"STRING")
end tablegen-parser-action22;

define function tablegen-parser-action23 (arg$1) => (value)
  make(<tablegen-simple-type>, kind: #"BIT")
end tablegen-parser-action23;

define function tablegen-parser-action24 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<tablegen-bits-type>, size: arg$3)
end tablegen-parser-action24;

define function tablegen-parser-action25 (arg$1) => (value)
  make(<tablegen-simple-type>, kind: #"INT")
end tablegen-parser-action25;

define function tablegen-parser-action26 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<tablegen-list-type>, of: arg$3)
end tablegen-parser-action26;

define function tablegen-parser-action27 (arg$1) => (value)
  make(<tablegen-simple-type>, kind: #"CODE")
end tablegen-parser-action27;

define function tablegen-parser-action28 (arg$1) => (value)
  make(<tablegen-simple-type>, kind: #"DAG")
end tablegen-parser-action28;

define function tablegen-parser-action29 (arg$1) => (value)
  make(<tablegen-class-type>, class: arg$1)
end tablegen-parser-action29;

define function tablegen-parser-action30 (arg$1) => (value)
  $tablegen-classes[arg$1]
end tablegen-parser-action30;

define function tablegen-parser-action31 () => (value)
  make(<tablegen-unknown-value>)
end tablegen-parser-action31;

define function tablegen-parser-action32 (arg$1, arg$2) => (value)
  arg$2
end tablegen-parser-action32;

define function tablegen-parser-action33 (arg$1) => (value)
  make(<tablegen-named-value>, name: arg$1)
end tablegen-parser-action33;

define function tablegen-parser-action34 (arg$1) => (value)
  make(<tablegen-simple-value>, value: arg$1)
end tablegen-parser-action34;

define function tablegen-parser-action36 (arg$1) => (value)
  make(<tablegen-unknown-value>)
end tablegen-parser-action36;

define function tablegen-parser-action37 (arg$1, arg$2, arg$3) => (value)
  error("bits-value")
end tablegen-parser-action37;

define function tablegen-parser-action38 (arg$1, arg$2, arg$3, arg$4) => (value)
  begin
         let def-name
           = format-to-string("anonymous.val.%d",
                              *tablegen-anonymous-value-counter*);
         *tablegen-anonymous-value-counter*
           := *tablegen-anonymous-value-counter* + 1;

         let initializers = make(<stretchy-object-vector>);
         let environment = make(<string-table>);
         let reference
           = make(<tablegen-class-reference>,
                  class: arg$1, template-arguments: reverse!(arg$3));
         instantiate-class-reference(initializers, environment, reference, #t);

         $tablegen-definitions[def-name]
           := make(<tablegen-definition>,
                   name: def-name,
                   superclasses: vector(arg$1),
                   initializers: initializers);

         make(<tablegen-named-value>, name: def-name);
       end
end tablegen-parser-action38;

define function tablegen-parser-action39 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("bit-list value")
end tablegen-parser-action39;

define function tablegen-parser-action40 (arg$1, arg$2, arg$3) => (value)
  make(<tablegen-list-value>, list: arg$2)
end tablegen-parser-action40;

define function tablegen-parser-action41 (arg$1, arg$2, arg$3) => (value)
  error("field %s", arg$3)
end tablegen-parser-action41;

define function tablegen-parser-action42 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("dag")
end tablegen-parser-action42;

define function tablegen-parser-action43 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("slice")
end tablegen-parser-action43;

define function tablegen-parser-action44 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("!car")
end tablegen-parser-action44;

define function tablegen-parser-action45 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("!cdr")
end tablegen-parser-action45;

define function tablegen-parser-action46 (arg$1, arg$2, arg$3, arg$4) => (value)
  error("!null")
end tablegen-parser-action46;

define function tablegen-parser-action47 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  error("!cast")
end tablegen-parser-action47;

define function tablegen-parser-action48 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("!concat")
end tablegen-parser-action48;

define function tablegen-parser-action49 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("!shl")
end tablegen-parser-action49;

define function tablegen-parser-action50 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("!sra")
end tablegen-parser-action50;

define function tablegen-parser-action51 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("!srl")
end tablegen-parser-action51;

define function tablegen-parser-action52 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("!eq")
end tablegen-parser-action52;

define function tablegen-parser-action53 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<tablegen-operator-value>,
            kind: #"strconcat",
            operands: vector(arg$3, arg$5))
end tablegen-parser-action53;

define function tablegen-parser-action54 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  error("!nameconcat")
end tablegen-parser-action54;

define function tablegen-parser-action55 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("!if")
end tablegen-parser-action55;

define function tablegen-parser-action56 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("!foreach")
end tablegen-parser-action56;

define function tablegen-parser-action57 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("!subst")
end tablegen-parser-action57;

define function tablegen-parser-action60 (arg$1) => (value)
  reverse!(arg$1)
end tablegen-parser-action60;

define function tablegen-parser-action65 (arg$1, arg$2, arg$3) => (value)
  #f
end tablegen-parser-action65;

define function tablegen-parser-action68 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  #f
end tablegen-parser-action68;

define function tablegen-parser-action69 (arg$1, arg$2, arg$3, arg$4) => (value)
  #f
end tablegen-parser-action69;

define function tablegen-parser-action76 (arg$1, arg$2) => (value)
  pair(arg$1, arg$2)
end tablegen-parser-action76;

define function tablegen-parser-action78 (arg$1, arg$2) => (value)
  reverse!(arg$2)
end tablegen-parser-action78;

define function tablegen-parser-action81 (arg$1) => (value)
  make(<tablegen-class-reference>, class: arg$1)
end tablegen-parser-action81;

define function tablegen-parser-action82 (arg$1, arg$2, arg$3, arg$4) => (value)
  begin
         make(<tablegen-class-reference>,
              class: arg$1,
              template-arguments: reverse!(arg$3))
       end
end tablegen-parser-action82;

define function tablegen-parser-action83 (arg$1) => (value)
  #()
end tablegen-parser-action83;

define function tablegen-parser-action86 (arg$1, arg$2) => (value)
  add(arg$1, arg$2)
end tablegen-parser-action86;

define function tablegen-parser-action87 (arg$1, arg$2) => (value)
  arg$1
end tablegen-parser-action87;

define function tablegen-parser-action88 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("body-item %%LET")
end tablegen-parser-action88;

define function tablegen-parser-action90 (arg$1, arg$2, arg$3) => (value)
  error("opt-bit-list")
end tablegen-parser-action90;

define function tablegen-parser-action91 (arg$1, arg$2, arg$3) => (value)
  method (prefix :: <string>, environment :: <string-table>)
         let initializers = make(<stretchy-object-vector>);
         for (reference in arg$3.head)
           instantiate-class-reference
             (initializers, environment, reference, #t);
         end for;
         do(curry(add-initializer, initializers, environment), arg$3.tail);

         $tablegen-definitions[arg$2]
           := make(<tablegen-definition>,
                   name: concatenate(prefix, arg$2),
                   superclasses: map(class-reference-class, arg$3.head),
                   initializers: initializers);
       end
end tablegen-parser-action91;

define function tablegen-parser-action93 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         $tablegen-classes[arg$2]
           := make(<tablegen-multi-class>,
                   name: arg$2,
                   template-parameters: arg$3,
                   definitions: arg$5);
       end
end tablegen-parser-action93;

define function tablegen-parser-action99 (arg$1, arg$2, arg$3, arg$4) => (value)
  arg$2
end tablegen-parser-action99;

define function tablegen-parser-action100 () => (value)
  begin
         let previous-overrides = *tablegen-overrides*;
         // FIXME
         //*tablegen-overrides* := shallow-copy(*tablegen-overrides*);
         *tablegen-overrides* := make(<string-table>);
         map-into(*tablegen-overrides*, identity, previous-overrides);
         previous-overrides
       end
end tablegen-parser-action100;

define function tablegen-parser-action103 (arg$1, arg$2, arg$3, arg$4) => (value)
  *tablegen-overrides*[arg$1] := arg$4
end tablegen-parser-action103;

define constant tablegen-parser :: <parser>
  = make(<parser>,
  action-table:
      #[#[46, -11, 43, -13, 44, -8, 45, -10, 3, -4],
	#[65535, 4, 0, 4, 2, 4, 3, 4, 46, 4, 45, 4, 44, 4, 43, 4],
	#[65535, 6, 0, 6, 2, 6, 3, 6, 43, 6, 44, 6, 45, 6, 46, 6],
	#[40, 12, 42, 12, 1, 12, 4, -183, 5, 12],
	#[65535, 5, 3, 5, 0, 5, 2, 5, 46, 5, 43, 5, 44, 5, 45, 5],
	#[3, -4, 1, -243, 44, -8, 43, -13, 46, -11, 45, -10],
	#[65535, 1, 0, 1, 2, 1, 3, 1, 46, 1, 43, 1, 44, 1, 45, 1],
	#[1, 12, 4, -183, 40, 12, 42, 12],
	#[65535, 3, 0, 3, 2, 3, 3, 3, 46, 3, 45, 3, 43, 3, 44, 3],
	#[4, -231],
	#[4, -226],
	#[0, -181, 3, -4, 43, -13, 44, -8, 45, -10, 46, -11],
	#[65535, 100, 4, 100],
	#[4, -17],
	#[65535, 101, 7, 101, 47, 101],
	#[7, -177, 47, -178],
	#[1, -19, 17, 89],
	#[17, -32],
	#[12, -21],
	#[2, -31],
	#[21, 64, 2, 64, 7, 64, 39, -28, 12, -29],
	#[2, 63, 7, -23, 21, 63],
	#[12, -24],
	#[2, 67, 21, 67, 7, 67, 12, -26, 39, -25],
	#[12, -27],
	#[65535, 69, 2, 69, 21, 69, 7, 69],
	#[65535, 68, 2, 68, 7, 68, 21, 68],
	#[12, -30],
	#[65535, 66, 2, 66, 7, 66, 21, 66],
	#[65535, 65, 2, 65, 21, 65, 7, 65],
	#[65535, 90, 17, 90],
	#[1, -55, 4, -42, 19, -47, 30, -46, 12, -56, 27, -45, 18, -49, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 32, -50, 28, -33, 29, -34, 31, -54, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[5, -139],
	#[23, -145],
	#[5, -136],
	#[23, -133],
	#[1, -55, 18, -49, 37, -51, 4, -42, 19, -47, 29, -34, 12, -56, 36, -38, 20, -37, 21, 59, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 38, -41],
	#[23, -123],
	#[4, -116],
	#[23, -110],
	#[23, -103],
	#[1, 58, 2, 58, 5, 30, 6, 58, 7, 58, 20, 58, 21, 58, 22, 58, 24, 58, 40, 58, 42, 58, 47, 58],
	#[23, -60],
	#[1, -95, 20, -94, 7, 103, 22, -96, 47, 103],
	#[23, -92],
	#[23, -90],
	#[65535, 36, 1, 36, 2, 36, 20, 36, 6, 36, 7, 36, 24, 36, 42, 36, 21, 36, 22, 36, 40, 36, 47, 36],
	#[65535, 33, 1, 33, 2, 33, 6, 33, 7, 33, 47, 33, 20, 33, 21, 33, 22, 33, 24, 33, 40, 33, 42, 33],
	#[65535, 35, 1, 35, 2, 35, 6, 35, 7, 35, 20, 35, 21, 35, 22, 35, 24, 35, 40, 35, 42, 35, 47, 35],
	#[23, -88],
	#[23, -86],
	#[23, -84],
	#[5, -64],
	#[23, -62],
	#[1, -55, 2, 59, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[65535, 34, 1, 34, 2, 34, 6, 34, 7, 34, 20, 34, 21, 34, 22, 34, 24, 34, 40, 34, 42, 34, 47, 34],
	#[2, -176],
	#[1, -95, 2, 61, 6, 61, 7, 61, 20, -94, 21, 61, 22, -96],
	#[2, 60, 7, -131, 21, 60],
	#[1, -55, 4, -42, 30, -46, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -175],
	#[1, -55, 4, -42, 30, -46, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 31, -54, 32, -50, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -172, 20, -94, 22, -96],
	#[4, -70, 9, -66, 10, -67, 11, -71, 13, -74, 14, -73, 15, -72, 16, -69],
	#[65535, 29, 4, 29, 6, 29],
	#[65535, 22, 4, 22, 6, 22],
	#[65535, 23, 4, 23, 6, 23],
	#[6, -81],
	#[65535, 28, 4, 28, 6, 28],
	#[65535, 30, 1, 30, 4, 30, 5, 30, 6, 30, 7, 30, 42, 30],
	#[5, -77],
	#[65535, 27, 4, 27, 6, 27],
	#[5, -75],
	#[65535, 25, 4, 25, 6, 25],
	#[4, -70, 9, -66, 10, -67, 11, -71, 13, -74, 14, -73, 15, -72, 16, -69],
	#[6, -80],
	#[12, -78],
	#[6, -79],
	#[65535, 24, 6, 24, 4, 24],
	#[65535, 26, 4, 26, 6, 26],
	#[23, -82],
	#[1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -169, 20, -94, 22, -96],
	#[1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[7, -166, 1, -95, 20, -94, 22, -96],
	#[1, -55, 4, -42, 29, -34, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 7, -161],
	#[1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -158, 20, -94, 22, -96],
	#[12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41, 1, -55, 4, -42],
	#[1, -95, 7, -155, 20, -94, 22, -96],
	#[1, -55, 4, -42, 29, -34, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -97],
	#[12, -21],
	#[12, -21],
	#[4, -98],
	#[65535, 46, 1, 46, 2, 46, 6, 46, 7, 46, 20, 46, 21, 46, 22, 46, 24, 46, 40, 46, 42, 46, 47, 46],
	#[65535, 41, 1, 41, 2, 41, 6, 41, 7, 41, 20, 41, 21, 41, 22, 41, 24, 41, 40, 41, 42, 41, 47, 41],
	#[2, -100],
	#[65535, 39, 2, 39, 1, 39, 6, 39, 7, 39, 20, 39, 21, 39, 22, 39, 24, 39, 40, 39, 42, 39, 47, 39],
	#[21, -102],
	#[65535, 43, 7, 43, 2, 43, 6, 43, 1, 43, 20, 43, 21, 43, 22, 43, 24, 43, 40, 43, 42, 43, 47, 43],
	#[12, -56, 1, -55, 4, -42, 19, -47, 29, -34, 34, -52, 18, -49, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 33, -40, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -105, 20, -94, 22, -96],
	#[1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -107, 20, -94, 22, -96],
	#[1, -55, 4, -42, 29, -34, 28, -33, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -109],
	#[65535, 57, 2, 57, 7, 57, 6, 57, 1, 57, 20, 57, 21, 57, 22, 57, 24, 57, 40, 57, 42, 57, 47, 57],
	#[23, -39, 1, -55, 4, -42, 18, -49, 29, -34, 12, -56, 27, -45, 38, -41, 19, -47, 20, -37, 25, -36, 26, -43, 32, -50, 28, -33, 30, -46, 31, -54, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51],
	#[1, -95, 7, -112, 20, -94, 22, -96],
	#[1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -114],
	#[65535, 52, 1, 52, 2, 52, 6, 52, 7, 52, 20, 52, 21, 52, 22, 52, 24, 52, 40, 52, 42, 52, 47, 52],
	#[4, -42, 1, -55, 18, -49, 29, -34, 12, -56, 27, -45, 19, -47, 20, -37, 23, -39, 24, 70, 25, -36, 26, -43, 32, -50, 28, -33, 30, -46, 31, -54, 33, -40, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[65535, 58, 12, 58, 1, 58, 4, 58, 18, 58, 29, 58, 27, 58, 19, 58, 20, 58, 23, 58, 24, 58, 25, 58, 26, 58, 32, 58, 28, 58, 30, 58, 31, 58, 33, 58, 34, 58, 35, 58, 36, 58, 37, 58, 38, 58],
	#[7, -152, 24, 71],
	#[24, -151],
	#[1, -95, 7, 74, 20, -94, 22, -96, 24, 74, 40, -121],
	#[65535, 72, 7, 72, 24, 72],
	#[41, -122],
	#[65535, 75, 7, 75, 24, 75],
	#[4, -42, 34, -52, 1, -55, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 36, -38, 37, -51, 38, -41],
	#[1, -95, 7, -125, 20, -94, 22, -96],
	#[34, -52, 1, -55, 4, -42, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 29, -34, 30, -46, 31, -54, 32, -50, 33, -40, 36, -38, 37, -51, 38, -41],
	#[7, -127, 1, -95, 20, -94, 22, -96],
	#[1, -55, 4, -42, 29, -34, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 33, -40, 34, -52, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -129],
	#[65535, 55, 1, 55, 2, 55, 6, 55, 7, 55, 20, 55, 21, 55, 22, 55, 24, 55, 40, 55, 42, 55, 47, 55],
	#[21, -150],
	#[12, -56, 1, -55, 31, -54, 4, -42, 18, -49, 29, -34, 27, -45, 26, -43, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 33, -40, 32, -50, 28, -33, 30, -46, 34, -52, 36, -38, 37, -51, 38, -41],
	#[2, 62, 1, -95, 6, 62, 7, 62, 20, -94, 21, 62, 22, -96],
	#[30, -46, 4, -42, 23, -39, 12, -56, 1, -55, 18, -49, 26, -43, 19, -47, 20, -37, 25, -36, 33, -40, 27, -45, 28, -33, 29, -34, 31, -54, 32, -50, 34, -52, 35, -53, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -135],
	#[65535, 44, 7, 44, 1, 44, 2, 44, 6, 44, 20, 44, 21, 44, 22, 44, 24, 44, 40, 44, 42, 44, 47, 44],
	#[30, -46, 23, -39, 12, -56, 32, -50, 1, -55, 31, -54, 4, -42, 33, -40, 18, -49, 35, -53, 29, -34, 19, -47, 20, -37, 25, -36, 26, -43, 27, -45, 28, -33, 34, -52, 36, -38, 37, -51, 38, -41],
	#[6, -138, 7, -131],
	#[65535, 38, 6, 38, 47, 38, 1, 38, 2, 38, 7, 38, 20, 38, 21, 38, 22, 38, 24, 38, 40, 38, 42, 38],
	#[4, -70, 9, -66, 10, -67, 11, -71, 13, -74, 14, -73, 15, -72, 16, -69],
	#[6, -141],
	#[23, -142],
	#[30, -46, 4, -42, 23, -39, 37, -51, 1, -55, 25, -36, 12, -56, 38, -41, 18, -49, 19, -47, 20, -37, 32, -50, 26, -43, 27, -45, 28, -33, 29, -34, 31, -54, 33, -40, 34, -52, 35, -53, 36, -38],
	#[1, -95, 20, -94, 22, -96, 24, -144],
	#[65535, 47, 20, 47, 1, 47, 2, 47, 6, 47, 7, 47, 47, 47, 42, 47, 21, 47, 22, 47, 24, 47, 40, 47],
	#[4, -42, 12, -56, 32, -50, 1, -55, 31, -54, 33, -40, 18, -49, 35, -53, 29, -34, 27, -45, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 28, -33, 30, -46, 34, -52, 36, -38, 37, -51, 38, -41],
	#[20, -94, 1, -95, 7, -147, 22, -96],
	#[30, -46, 34, -52, 12, -56, 18, -49, 33, -40, 1, -55, 31, -54, 4, -42, 35, -53, 29, -34, 19, -47, 20, -37, 32, -50, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 36, -38, 37, -51, 38, -41],
	#[1, -95, 20, -94, 22, -96, 24, -149],
	#[65535, 48, 47, 48, 22, 48, 1, 48, 2, 48, 6, 48, 7, 48, 42, 48, 20, 48, 21, 48, 24, 48, 40, 48],
	#[65535, 40, 47, 40, 1, 40, 2, 40, 20, 40, 6, 40, 7, 40, 24, 40, 42, 40, 21, 40, 22, 40, 40, 40],
	#[65535, 42, 47, 42, 22, 42, 1, 42, 2, 42, 6, 42, 7, 42, 42, 42, 20, 42, 21, 42, 24, 42, 40, 42],
	#[30, -46, 34, -52, 12, -56, 37, -51, 1, -55, 31, -54, 4, -42, 33, -40, 18, -49, 29, -34, 27, -45, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 28, -33, 32, -50, 36, -38, 38, -41],
	#[22, -96, 1, -95, 20, -94, 7, 74, 40, -121, 24, 74],
	#[65535, 73, 7, 73, 24, 73],
	#[30, -46, 38, -41, 23, -39, 12, -56, 1, -55, 4, -42, 33, -40, 18, -49, 35, -53, 29, -34, 27, -45, 19, -47, 20, -37, 32, -50, 25, -36, 26, -43, 28, -33, 31, -54, 34, -52, 36, -38, 37, -51],
	#[1, -95, 20, -94, 22, -96, 24, -157],
	#[65535, 49, 1, 49, 2, 49, 6, 49, 7, 49, 20, 49, 21, 49, 22, 49, 24, 49, 40, 49, 42, 49, 47, 49],
	#[38, -41, 23, -39, 12, -56, 37, -51, 1, -55, 20, -37, 4, -42, 33, -40, 32, -50, 35, -53, 29, -34, 18, -49, 19, -47, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 34, -52, 36, -38],
	#[1, -95, 20, -94, 22, -96, 24, -160],
	#[65535, 51, 1, 51, 2, 51, 6, 51, 7, 51, 20, 51, 21, 51, 22, 51, 24, 51, 40, 51, 42, 51, 47, 51],
	#[23, -39, 38, -41, 37, -51, 1, -55, 31, -54, 20, -37, 4, -42, 33, -40, 32, -50, 35, -53, 29, -34, 12, -56, 27, -45, 18, -49, 19, -47, 25, -36, 26, -43, 28, -33, 30, -46, 34, -52, 36, -38],
	#[1, -95, 7, -163, 20, -94, 22, -96],
	#[34, -52, 1, -55, 4, -42, 30, -46, 29, -34, 12, -56, 18, -49, 19, -47, 20, -37, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 31, -54, 32, -50, 33, -40, 36, -38, 37, -51, 38, -41],
	#[22, -96, 1, -95, 20, -94, 24, -165],
	#[65535, 56, 1, 56, 2, 56, 6, 56, 7, 56, 20, 56, 21, 56, 22, 56, 24, 56, 40, 56, 42, 56, 47, 56],
	#[23, -39, 1, -55, 4, -42, 18, -49, 35, -53, 29, -34, 12, -56, 38, -41, 19, -47, 20, -37, 32, -50, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 33, -40, 34, -52, 36, -38, 37, -51],
	#[1, -95, 20, -94, 22, -96, 24, -168],
	#[65535, 53, 6, 53, 1, 53, 21, 53, 20, 53, 2, 53, 7, 53, 22, 53, 24, 53, 40, 53, 42, 53, 47, 53],
	#[38, -41, 37, -51, 1, -55, 4, -42, 32, -50, 35, -53, 29, -34, 12, -56, 18, -49, 19, -47, 20, -37, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 33, -40, 34, -52, 36, -38],
	#[20, -94, 1, -95, 22, -96, 24, -171],
	#[65535, 54, 20, 54, 22, 54, 1, 54, 2, 54, 6, 54, 7, 54, 47, 54, 42, 54, 21, 54, 24, 54, 40, 54],
	#[31, -54, 12, -56, 19, -47, 20, -37, 27, -45, 29, -34, 38, -41, 1, -55, 35, -53, 33, -40, 37, -51, 4, -42, 18, -49, 32, -50, 23, -39, 25, -36, 26, -43, 28, -33, 30, -46, 34, -52, 36, -38],
	#[22, -96, 1, -95, 20, -94, 24, -174],
	#[65535, 50, 47, 50, 6, 50, 21, 50, 20, 50, 2, 50, 1, 50, 42, 50, 7, 50, 22, 50, 24, 50, 40, 50],
	#[65535, 45, 20, 45, 2, 45, 6, 45, 42, 45, 1, 45, 22, 45, 40, 45, 21, 45, 24, 45, 7, 45, 47, 45],
	#[65535, 37, 42, 37, 7, 37, 22, 37, 24, 37, 40, 37, 47, 37, 6, 37, 21, 37, 1, 37, 20, 37, 2, 37],
	#[4, -17],
	#[65535, 99, 44, 99, 1, 99, 3, 99, 43, 99, 46, 99, 45, 99],
	#[65535, 102, 7, 102, 47, 102],
	#[65535, 2, 2, 2, 0, 2, 3, 2, 45, 2, 43, 2, 44, 2, 46, 2],
	#[#"eoi", #"accept"],
	#[65535, 10, 1, 10, 5, 10, 42, 10, 40, 10],
	#[65535, 13, 5, 13, 1, 13, 40, 13, 42, 13],
	#[5, -188, 1, 14, 42, 14, 40, 14],
	#[65535, 11, 1, 11, 5, 11, 42, 11, 40, 11],
	#[65535, 15, 1, 15, 40, 15, 42, 15],
	#[1, 77, 40, -203, 42, 77],
	#[4, 20, 13, 20, 8, -190, 9, 20, 10, 20, 11, 20, 14, 20, 15, 20, 16, 20],
	#[6, -199, 7, -198],
	#[65535, 21, 4, 21, 9, 21, 10, 21, 11, 21, 13, 21, 14, 21, 15, 21, 16, 21],
	#[13, -74, 14, -73, 11, -71, 15, -72, 9, -66, 16, -69, 10, -67, 4, -70],
	#[65535, 17, 6, 17, 7, 17],
	#[4, -194],
	#[7, 31, 6, 31, 17, -196, 42, 31],
	#[65535, 19, 7, 19, 6, 19, 42, 19],
	#[4, -42, 31, -54, 34, -52, 23, -39, 30, -46, 12, -56, 38, -41, 33, -40, 19, -47, 36, -38, 20, -37, 29, -34, 32, -50, 1, -55, 37, -51, 18, -49, 25, -36, 26, -43, 27, -45, 28, -33, 35, -53],
	#[6, 32, 20, -94, 1, -95, 7, 32, 42, 32, 22, -96],
	#[15, 20, 14, 20, 13, 20, 11, 20, 9, 20, 16, 20, 8, -190, 10, 20, 4, 20],
	#[65535, 16, 1, 16, 42, 16, 40, 16],
	#[65535, 18, 7, 18, 6, 18],
	#[1, -213, 42, -214],
	#[65535, 9, 3, 9, 2, 9, 0, 9, 46, 9, 43, 9, 44, 9, 45, 9],
	#[4, -70],
	#[7, 81, 1, 81, 5, -209, 42, 81],
	#[65535, 79, 7, 79, 1, 79, 42, 79],
	#[7, -207, 1, 78, 42, 78],
	#[4, -70],
	#[65535, 80, 7, 80, 1, 80, 42, 80],
	#[29, -34, 4, -42, 12, -56, 20, -37, 33, -40, 1, -55, 19, -47, 18, -49, 35, -53, 23, -39, 25, -36, 26, -43, 27, -45, 28, -33, 30, -46, 31, -54, 32, -50, 34, -52, 36, -38, 37, -51, 38, -41],
	#[6, -211, 7, -131],
	#[65535, 82, 1, 82, 7, 82, 42, 82],
	#[65535, 76, 3, 76, 46, 76, 2, 76, 43, 76, 0, 76, 45, 76, 44, 76],
	#[65535, 85, 14, 85, 4, 85, 11, 85, 9, 85, 15, 85, 2, 85, 13, 85, 16, 85, 8, 85, 10, 85, 43, 85],
	#[65535, 83, 3, 83, 0, 83, 46, 83, 2, 83, 43, 83, 45, 83, 44, 83],
	#[11, 20, 4, 20, 43, -219, 13, 20, 9, 20, 8, -190, 2, -217, 10, 20, 16, 20, 14, 20, 15, 20],
	#[42, -225],
	#[65535, 84, 3, 84, 2, 84, 0, 84, 43, 84, 45, 84, 44, 84, 46, 84],
	#[65535, 86, 2, 86, 10, 86, 15, 86, 4, 86, 43, 86, 13, 86, 8, 86, 9, 86, 11, 86, 14, 86, 16, 86],
	#[4, -220],
	#[1, -19, 17, 89],
	#[17, -222],
	#[4, -42, 23, -39, 30, -46, 38, -41, 19, -47, 12, -56, 36, -38, 31, -54, 34, -52, 26, -43, 20, -37, 29, -34, 1, -55, 37, -51, 25, -36, 18, -49, 32, -50, 27, -45, 28, -33, 33, -40, 35, -53],
	#[1, -95, 22, -96, 20, -94, 42, -224],
	#[65535, 88, 15, 88, 43, 88, 4, 88, 14, 88, 13, 88, 2, 88, 16, 88, 8, 88, 9, 88, 10, 88, 11, 88],
	#[65535, 87, 15, 87, 13, 87, 43, 87, 10, 87, 14, 87, 16, 87, 2, 87, 8, 87, 4, 87, 9, 87, 11, 87],
	#[40, -227],
	#[4, -70],
	#[42, -229],
	#[65535, 98, 3, 98, 0, 98, 2, 98, 46, 98, 45, 98, 43, 98, 44, 98],
	#[5, -188, 1, 14],
	#[65535, 94, 5, 94, 1, 94],
	#[1, -233],
	#[44, -8],
	#[65535, 97, 44, 97, 2, 97],
	#[65535, 95, 2, 95, 44, 95],
	#[2, -238, 44, -8],
	#[65535, 96, 2, 96, 44, 96],
	#[65535, 93, 2, 93, 46, 93, 0, 93, 3, 93, 43, 93, 45, 93, 44, 93],
	#[65535, 92, 1, 92, 42, 92, 40, 92],
	#[1, 77, 40, -203, 42, 77],
	#[65535, 91, 46, 91, 0, 91, 3, 91, 2, 91, 43, 91, 45, 91, 44, 91],
	#[65535, 8, 45, 8, 3, 8, 0, 8, 2, 8, 43, 8, 46, 8, 44, 8],
	#[46, -11, 3, -4, 45, -10, 43, -13, 44, -8],
	#[2, -245, 46, -11, 3, -4, 45, -10, 43, -13, 44, -8],
	#[65535, 7, 46, 7, 0, 7, 45, 7, 3, 7, 2, 7, 43, 7, 44, 7]],
  goto-table:
      #[#[87, 5, 86, 2, 82, 4, 80, 1, 51, 8, 45, 9, 44, 7, 43, 12, 49, 11, 50, 6, 3, 3, 46, 10],
	#[],
	#[],
	#[54, 184, 52, 183, 4, 182, 53, 181],
	#[],
	#[80, 1, 82, 4, 87, 5, 51, 8, 86, 2, 44, 7, 45, 9, 50, 241, 43, 12, 3, 3, 46, 10, 1, 242],
	#[],
	#[81, 239, 54, 184, 53, 238, 4, 182],
	#[],
	#[83, 229, 4, 230],
	#[4, 225],
	#[87, 5, 86, 2, 82, 4, 80, 1, 51, 8, 50, 179, 46, 10, 45, 9, 44, 7, 43, 12, 3, 3, 0, 180],
	#[88, 13],
	#[90, 14, 89, 15, 4, 16],
	#[],
	#[47, 177, 7, 176],
	#[79, 17, 1, 18],
	#[17, 31],
	#[68, 21, 67, 19, 12, 20],
	#[2, 30],
	#[12, 28, 39, 27],
	#[7, 22],
	#[12, 23],
	#[39, 24, 12, 25],
	#[12, 26],
	#[],
	#[],
	#[12, 29],
	#[],
	#[],
	#[],
	#[61, 34, 64, 47, 38, 40, 37, 50, 36, 37, 35, 52, 63, 43, 33, 39, 32, 49, 31, 53, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 34, 51, 20, 36, 19, 46, 18, 48, 12, 55, 30, 45, 4, 41, 1, 54, 23, 38],
	#[5, 138],
	#[23, 144],
	#[5, 135],
	#[23, 132],
	#[66, 58, 65, 129, 64, 47, 63, 57, 61, 34, 38, 40, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 35, 52, 20, 36, 36, 37, 12, 55, 29, 33, 18, 48, 19, 46, 4, 41, 1, 54, 37, 50],
	#[23, 122],
	#[64, 114, 4, 115],
	#[23, 109],
	#[23, 102],
	#[],
	#[23, 59],
	#[20, 93, 22, 95, 1, 94],
	#[23, 91],
	#[23, 89],
	#[],
	#[],
	#[],
	#[23, 87],
	#[23, 85],
	#[23, 83],
	#[5, 63],
	#[23, 61],
	#[66, 58, 65, 56, 64, 47, 63, 57, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 23, 38, 20, 36, 19, 46, 25, 35, 12, 55, 18, 48, 4, 41, 1, 54],
	#[],
	#[2, 175],
	#[22, 95, 20, 93, 1, 94],
	#[7, 130],
	#[64, 47, 63, 60, 61, 34, 38, 40, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 23, 38, 20, 36, 19, 46, 12, 55, 25, 35, 18, 48, 4, 41, 1, 54, 37, 50],
	#[24, 174, 22, 95, 20, 93, 1, 94],
	#[61, 34, 64, 47, 38, 40, 36, 37, 63, 62, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 34, 51, 35, 52, 20, 36, 19, 46, 12, 55, 18, 48, 4, 41, 1, 54, 37, 50, 23, 38],
	#[7, 171, 1, 94, 22, 95, 20, 93],
	#[61, 64, 60, 67, 16, 68, 15, 71, 14, 72, 13, 73, 10, 66, 4, 69, 9, 65, 11, 70],
	#[],
	#[],
	#[],
	#[6, 80],
	#[],
	#[],
	#[5, 76],
	#[],
	#[5, 74],
	#[],
	#[61, 64, 60, 75, 16, 68, 15, 71, 14, 72, 13, 73, 11, 70, 10, 66, 9, 65, 4, 69],
	#[6, 79],
	#[12, 77],
	#[6, 78],
	#[],
	#[],
	#[23, 81],
	#[64, 47, 63, 82, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 1, 94, 7, 168],
	#[64, 47, 63, 84, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 1, 94, 7, 165],
	#[64, 47, 63, 86, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[7, 160, 22, 95, 20, 93, 1, 94],
	#[64, 47, 63, 88, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 7, 157, 1, 94],
	#[4, 41, 1, 54, 64, 47, 63, 90, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55],
	#[22, 95, 20, 93, 7, 154, 1, 94],
	#[64, 47, 63, 92, 61, 34, 38, 40, 37, 50, 36, 37, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 35, 52, 20, 36, 19, 46, 18, 48, 12, 55, 29, 33, 4, 41, 1, 54],
	#[24, 96, 22, 95, 20, 93, 1, 94],
	#[67, 100, 68, 21, 12, 20],
	#[68, 21, 67, 98, 12, 20],
	#[4, 97],
	#[],
	#[],
	#[2, 99],
	#[],
	#[21, 101],
	#[],
	#[64, 47, 63, 103, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 7, 104, 1, 94],
	#[64, 47, 63, 105, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 1, 94, 7, 106],
	#[64, 47, 63, 107, 61, 34, 38, 40, 37, 50, 36, 37, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 27, 44, 26, 42, 25, 35, 23, 38, 35, 52, 20, 36, 19, 46, 18, 48, 12, 55, 28, 32, 29, 33, 4, 41, 1, 54],
	#[24, 108, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 61, 34, 37, 50, 36, 37, 35, 52, 63, 110, 33, 39, 31, 53, 30, 45, 28, 32, 32, 49, 26, 42, 25, 35, 20, 36, 19, 46, 18, 48, 38, 40, 27, 44, 34, 51, 29, 33, 4, 41, 1, 54, 12, 55, 23, 38],
	#[22, 95, 20, 93, 7, 111, 1, 94],
	#[64, 47, 63, 112, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[24, 113, 22, 95, 20, 93, 1, 94],
	#[],
	#[70, 116, 69, 117, 64, 47, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 63, 118, 31, 53, 30, 45, 28, 32, 32, 49, 33, 39, 25, 35, 34, 51, 20, 36, 19, 46, 18, 48, 26, 42, 27, 44, 12, 55, 29, 33, 4, 41, 1, 54, 23, 38],
	#[],
	#[7, 151],
	#[24, 150],
	#[71, 119, 40, 120, 22, 95, 20, 93, 1, 94],
	#[],
	#[41, 121],
	#[],
	#[64, 47, 63, 123, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 34, 51, 1, 54],
	#[22, 95, 20, 93, 7, 124, 1, 94],
	#[64, 47, 63, 125, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 34, 51, 4, 41, 1, 54, 12, 55],
	#[20, 93, 7, 126, 1, 94, 22, 95],
	#[64, 47, 63, 127, 61, 34, 38, 40, 37, 50, 36, 37, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 35, 52, 20, 36, 19, 46, 18, 48, 12, 55, 29, 33, 4, 41, 1, 54],
	#[24, 128, 22, 95, 20, 93, 1, 94],
	#[],
	#[21, 149],
	#[64, 47, 63, 131, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 29, 33, 4, 41, 1, 54],
	#[22, 95, 20, 93, 1, 94],
	#[64, 47, 63, 133, 61, 34, 37, 50, 34, 51, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 35, 52, 20, 36, 36, 37, 18, 48, 38, 40, 33, 39, 1, 54, 19, 46, 4, 41, 12, 55],
	#[24, 134, 22, 95, 20, 93, 1, 94],
	#[],
	#[66, 136, 64, 47, 63, 57, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[7, 130, 6, 137],
	#[],
	#[60, 139, 61, 64, 16, 68, 15, 71, 14, 72, 13, 73, 11, 70, 10, 66, 9, 65, 4, 69],
	#[6, 140],
	#[23, 141],
	#[64, 47, 63, 142, 61, 34, 34, 51, 31, 53, 30, 45, 29, 33, 28, 32, 32, 49, 33, 39, 25, 35, 23, 38, 35, 52, 20, 36, 36, 37, 18, 48, 38, 40, 27, 44, 12, 55, 1, 54, 19, 46, 4, 41, 26, 42, 37, 50],
	#[24, 143, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 63, 145, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[22, 95, 20, 93, 1, 94, 7, 146],
	#[64, 47, 63, 147, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 32, 49, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 33, 39, 18, 48, 19, 46, 4, 41, 31, 53, 1, 54, 12, 55],
	#[24, 148, 22, 95, 20, 93, 1, 94],
	#[],
	#[],
	#[],
	#[64, 47, 63, 152, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[71, 153, 40, 120, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 31, 53, 30, 45, 32, 49, 33, 39, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 26, 42, 27, 44, 12, 55, 28, 32, 29, 33, 63, 155, 1, 54, 4, 41],
	#[24, 156, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 37, 50, 36, 37, 35, 52, 34, 51, 32, 49, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 33, 39, 29, 33, 63, 158, 4, 41, 1, 54, 12, 55, 38, 40, 61, 34],
	#[24, 159, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 61, 34, 37, 50, 36, 37, 35, 52, 63, 161, 31, 53, 30, 45, 28, 32, 27, 44, 25, 35, 32, 49, 33, 39, 18, 48, 38, 40, 34, 51, 29, 33, 19, 46, 4, 41, 20, 36, 1, 54, 12, 55, 26, 42, 23, 38],
	#[22, 95, 20, 93, 7, 162, 1, 94],
	#[64, 47, 63, 163, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 34, 51, 33, 39, 32, 49, 31, 53, 30, 45, 29, 33, 28, 32, 27, 44, 26, 42, 25, 35, 23, 38, 20, 36, 19, 46, 18, 48, 12, 55, 4, 41, 1, 54],
	#[24, 164, 20, 93, 1, 94, 22, 95],
	#[],
	#[64, 47, 61, 34, 38, 40, 37, 50, 36, 37, 35, 52, 63, 166, 33, 39, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 32, 49, 20, 36, 19, 46, 18, 48, 34, 51, 29, 33, 4, 41, 1, 54, 12, 55, 23, 38],
	#[24, 167, 22, 95, 20, 93, 1, 94],
	#[],
	#[64, 47, 61, 34, 37, 50, 36, 37, 35, 52, 63, 169, 33, 39, 31, 53, 30, 45, 28, 32, 27, 44, 26, 42, 25, 35, 20, 36, 32, 49, 38, 40, 34, 51, 29, 33, 18, 48, 19, 46, 4, 41, 1, 54, 12, 55, 23, 38],
	#[22, 95, 24, 170, 1, 94, 20, 93],
	#[],
	#[64, 47, 61, 34, 37, 50, 36, 37, 30, 45, 28, 32, 27, 44, 33, 39, 25, 35, 18, 48, 4, 41, 32, 49, 35, 52, 23, 38, 26, 42, 38, 40, 20, 36, 19, 46, 34, 51, 12, 55, 63, 172, 31, 53, 29, 33, 1, 54],
	#[24, 173, 20, 93, 1, 94, 22, 95],
	#[],
	#[],
	#[],
	#[90, 178, 4, 16],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[56, 185, 55, 186, 5, 187],
	#[],
	#[],
	#[73, 200, 72, 201, 40, 202],
	#[59, 190, 58, 191, 57, 188, 8, 189],
	#[7, 197, 6, 198],
	#[],
	#[61, 64, 60, 192, 4, 69, 10, 66, 11, 70, 16, 68, 13, 73, 14, 72, 9, 65, 15, 71],
	#[],
	#[4, 193],
	#[17, 195, 62, 194],
	#[],
	#[64, 47, 63, 196, 35, 52, 28, 32, 27, 44, 33, 39, 30, 45, 37, 50, 18, 48, 23, 38, 38, 40, 29, 33, 25, 35, 20, 36, 61, 34, 26, 42, 19, 46, 31, 53, 32, 49, 36, 37, 1, 54, 34, 51, 4, 41, 12, 55],
	#[22, 95, 20, 93, 1, 94],
	#[58, 199, 59, 190, 8, 189],
	#[],
	#[],
	#[76, 211, 1, 212, 42, 213],
	#[],
	#[75, 204, 74, 205, 61, 203, 4, 69],
	#[5, 208],
	#[],
	#[7, 206],
	#[61, 203, 75, 207, 4, 69],
	#[],
	#[66, 209, 64, 47, 63, 57, 61, 34, 38, 40, 37, 50, 36, 37, 34, 51, 33, 39, 28, 32, 26, 42, 25, 35, 23, 38, 32, 49, 18, 48, 35, 52, 1, 54, 20, 36, 29, 33, 27, 44, 31, 53, 19, 46, 30, 45, 4, 41, 12, 55],
	#[7, 130, 6, 210],
	#[],
	#[],
	#[77, 214],
	#[],
	#[59, 190, 43, 218, 58, 215, 8, 189, 2, 216, 78, 217],
	#[42, 224],
	#[],
	#[],
	#[4, 219],
	#[79, 220, 1, 18],
	#[17, 221],
	#[64, 47, 63, 222, 61, 34, 35, 52, 28, 32, 27, 44, 32, 49, 30, 45, 37, 50, 18, 48, 23, 38, 26, 42, 38, 40, 20, 36, 36, 37, 25, 35, 19, 46, 34, 51, 12, 55, 33, 39, 31, 53, 1, 54, 4, 41, 29, 33],
	#[42, 223, 22, 95, 20, 93, 1, 94],
	#[],
	#[],
	#[40, 226],
	#[75, 227, 61, 203, 4, 69],
	#[42, 228],
	#[],
	#[56, 185, 5, 187, 55, 231],
	#[],
	#[1, 232],
	#[80, 233, 84, 235, 85, 234, 44, 7],
	#[],
	#[],
	#[85, 236, 80, 233, 44, 7, 2, 237],
	#[],
	#[],
	#[],
	#[73, 200, 72, 240, 40, 202],
	#[],
	#[],
	#[82, 4, 80, 1, 87, 5, 50, 6, 46, 10, 45, 9, 43, 12, 49, 243, 86, 2, 51, 8, 44, 7, 3, 3],
	#[87, 5, 80, 1, 86, 2, 50, 179, 82, 4, 45, 9, 46, 10, 44, 7, 2, 244, 3, 3, 43, 12, 51, 8],
	#[]],
  action-function-table:
	 vector(tablegen-parser-action0,
		tablegen-parser-action1,
		tablegen-parser-action0,
		tablegen-parser-action3,
		tablegen-parser-action3,
		tablegen-parser-action1,
		tablegen-parser-action1,
		tablegen-parser-action7,
		tablegen-parser-action8,
		tablegen-parser-action9,
		tablegen-parser-action10,
		tablegen-parser-action10,
		tablegen-parser-action12,
		tablegen-parser-action10,
		tablegen-parser-action14,
		tablegen-parser-action10,
		tablegen-parser-action16,
		tablegen-parser-action17,
		tablegen-parser-action18,
		tablegen-parser-action19,
		tablegen-parser-action20,
		tablegen-parser-action1,
		tablegen-parser-action22,
		tablegen-parser-action23,
		tablegen-parser-action24,
		tablegen-parser-action25,
		tablegen-parser-action26,
		tablegen-parser-action27,
		tablegen-parser-action28,
		tablegen-parser-action29,
		tablegen-parser-action30,
		tablegen-parser-action31,
		tablegen-parser-action32,
		tablegen-parser-action33,
		tablegen-parser-action34,
		tablegen-parser-action34,
		tablegen-parser-action36,
		tablegen-parser-action37,
		tablegen-parser-action38,
		tablegen-parser-action39,
		tablegen-parser-action40,
		tablegen-parser-action41,
		tablegen-parser-action42,
		tablegen-parser-action43,
		tablegen-parser-action44,
		tablegen-parser-action45,
		tablegen-parser-action46,
		tablegen-parser-action47,
		tablegen-parser-action48,
		tablegen-parser-action49,
		tablegen-parser-action50,
		tablegen-parser-action51,
		tablegen-parser-action52,
		tablegen-parser-action53,
		tablegen-parser-action54,
		tablegen-parser-action55,
		tablegen-parser-action56,
		tablegen-parser-action57,
		tablegen-parser-action10,
		tablegen-parser-action14,
		tablegen-parser-action60,
		tablegen-parser-action17,
		tablegen-parser-action18,
		tablegen-parser-action1,
		tablegen-parser-action1,
		tablegen-parser-action65,
		tablegen-parser-action0,
		tablegen-parser-action65,
		tablegen-parser-action68,
		tablegen-parser-action69,
		tablegen-parser-action20,
		tablegen-parser-action1,
		tablegen-parser-action0,
		tablegen-parser-action69,
		tablegen-parser-action20,
		tablegen-parser-action0,
		tablegen-parser-action76,
		tablegen-parser-action14,
		tablegen-parser-action78,
		tablegen-parser-action17,
		tablegen-parser-action18,
		tablegen-parser-action81,
		tablegen-parser-action82,
		tablegen-parser-action83,
		tablegen-parser-action16,
		tablegen-parser-action14,
		tablegen-parser-action86,
		tablegen-parser-action87,
		tablegen-parser-action88,
		tablegen-parser-action20,
		tablegen-parser-action90,
		tablegen-parser-action91,
		tablegen-parser-action10,
		tablegen-parser-action93,
		tablegen-parser-action10,
		tablegen-parser-action17,
		tablegen-parser-action86,
		tablegen-parser-action10,
		tablegen-parser-action68,
		tablegen-parser-action99,
		tablegen-parser-action100,
		tablegen-parser-action1,
		tablegen-parser-action65,
		tablegen-parser-action103),
  action-nargs-table: #[2, 1, 2, 1, 1, 1, 1, 4, 2, 4, 1, 1, 0, 1, 0, 1, 3, 1, 3, 4, 0, 1, 1, 1, 4, 1, 4, 1, 1, 1, 1, 0, 2, 1, 1, 1, 1, 3, 4, 4, 3, 3, 4, 4, 4, 4, 4, 7, 6, 6, 6, 6, 6, 6, 9, 8, 8, 8, 1, 0, 1, 1, 3, 1, 1, 3, 2, 3, 5, 4, 0, 1, 2, 4, 0, 2, 2, 0, 2, 1, 3, 1, 4, 1, 3, 0, 2, 2, 6, 0, 3, 3, 1, 6, 1, 1, 2, 1, 5, 4, 0, 1, 3, 4],
  action-nt-table: #[48, 49, 49, 50, 50, 50, 50, 50, 50, 51, 52, 53, 54, 54, 55, 55, 56, 57, 57, 58, 59, 59, 60, 60, 60, 60, 60, 60, 60, 60, 61, 62, 62, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 63, 64, 65, 65, 66, 66, 67, 68, 68, 68, 68, 68, 68, 69, 69, 70, 70, 71, 71, 72, 73, 73, 74, 74, 75, 75, 76, 76, 77, 77, 78, 78, 79, 79, 80, 81, 82, 83, 84, 84, 85, 86, 87, 88, 89, 89, 90]);
