Module:       llvm-asm-parser-internals
Author:       Peter S. Housel
Copyright:    Original Code is Copyright 2009-2010 Gwydion Dylan Maintainers
              All rights reserved.
License:      See License.txt in this distribution for details.
Warranty:     Distributed WITHOUT WARRANTY OF ANY KIND

define class <llvm-parse-state> (<object>)
  constant slot parse-module :: <llvm-module>,
    required-init-keyword: module:;

  slot parse-unnamed-type-index :: <integer> = 0;
  constant slot parse-unnamed-types :: <stretchy-object-vector>
    = make(<stretchy-object-vector>);

  slot parse-unnamed-global-index :: <integer> = 0;
  constant slot parse-unnamed-global-values :: <stretchy-object-vector>
    = make(<stretchy-object-vector>);

  slot parse-function :: false-or(<llvm-function>) = #f;
  slot parse-unnamed-local-index :: <integer> = 0;
  constant slot parse-unnamed-local-values :: <stretchy-object-vector>
    = make(<stretchy-object-vector>);

  constant slot parse-unnamed-metadata-values :: <stretchy-object-vector>
    = make(<stretchy-object-vector>);
end class;

define thread variable *llvm-parse-state* :: <llvm-parse-state>
  = make(<llvm-parse-state>, module: make(<llvm-module>, name: "<dummy>"));

define method do-define-type (name :: <string>, type :: <llvm-type>) => ();
  let definition
    = element(*llvm-parse-state*.parse-module.llvm-type-table, name,
              default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-type>))
      definition.llvm-placeholder-type-forward := type;
    else
      error("type %%%s is multiply defined", name);
    end if;
  end if;
  
  // Record this type 
  element(*llvm-parse-state*.parse-module.llvm-type-table, name) := type
end method;

define method do-define-type
    (supplied-index :: false-or(<integer>), type :: <llvm-type>) => ();
  let new-index = *llvm-parse-state*.parse-unnamed-type-index;
  if (supplied-index & supplied-index ~= new-index)
    error("Expected %%%d rather than %%%d for unnamed type",
          new-index, supplied-index);
  end;

  let definition
    = element(*llvm-parse-state*.parse-unnamed-types, new-index, default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-type>))
      definition.llvm-placeholder-type-forward := type;
    else
      error("type %%%d is multiply defined", new-index);
    end if;
  end if;

  // Record this type
  element(*llvm-parse-state*.parse-unnamed-types, new-index) := type;
  *llvm-parse-state*.parse-unnamed-type-index := new-index + 1;
  type
end method;

define method do-define-global-value
    (name :: <string>, value :: <llvm-constant-value>) => ();
  let definition
    = element(*llvm-parse-state*.parse-module.llvm-global-table, name,
              default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-constant>))
      definition.llvm-placeholder-value-forward := value;
      llvm-constrain-type(definition.llvm-value-type, llvm-value-type(value));
    else
      error("value @%s is multiply defined", name);
    end if;
  end if;
  
  // Record this value
  element(*llvm-parse-state*.parse-module.llvm-global-table, name) := value
end method;

define method do-define-global-value
    (supplied-index :: false-or(<integer>), value :: <llvm-constant-value>)
 => ();
  let new-index = *llvm-parse-state*.parse-unnamed-global-index;
  if (supplied-index & supplied-index ~= new-index)
    error("Expected @%d rather than @%d for unnamed global value",
          new-index, supplied-index);
  end;
  let definition
    = element(*llvm-parse-state*.parse-unnamed-global-values, new-index,
              default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-constant>))
      definition.llvm-placeholder-value-forward := value;
      llvm-constrain-type(definition.llvm-value-type, llvm-value-type(value));
    else
      error("value @%d is multiply defined", new-index);
    end if;
  end if;
  
  // Record this value
  element(*llvm-parse-state*.parse-unnamed-global-values, new-index) := value;
  *llvm-parse-state*.parse-unnamed-global-index := new-index + 1;
end method;

define method do-define-local-value
    (name :: <string>, value :: <llvm-value>) => (value :: <llvm-value>);
  let definition
    = element(*llvm-parse-state*.parse-function.llvm-function-value-table,
              name, default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-value>))
      definition.llvm-placeholder-value-forward := value;
      llvm-constrain-type(definition.llvm-value-type, llvm-value-type(value));
    else
      error("value %%%s is multiply defined", name);
    end if;
  end if;
  
  // Record this value
  element(*llvm-parse-state*.parse-function.llvm-function-value-table, name)
    := value
end method;

define method do-define-local-value
    (supplied-index :: false-or(<integer>), value :: <llvm-value>)
 => (value :: <llvm-value>);
  unless (llvm-void-type?(llvm-value-type(value)))
    let new-index = *llvm-parse-state*.parse-unnamed-local-index;
    if (supplied-index & supplied-index ~= new-index)
      error("Expected %%%d rather than %%%d for unnamed local value",
            new-index, supplied-index);
    end;
    let definition
      = element(*llvm-parse-state*.parse-unnamed-local-values, new-index,
                default: #f);
    if (definition)
      if (instance?(definition, <llvm-symbolic-value>))
        definition.llvm-placeholder-value-forward := value;
        llvm-constrain-type(definition.llvm-value-type, llvm-value-type(value));
      else
        error("value %%%d is multiply defined", new-index);
      end if;
    end if;
    
    // Record this value
    element(*llvm-parse-state*.parse-unnamed-local-values, new-index) := value;
    *llvm-parse-state*.parse-unnamed-local-index := new-index + 1;
  end unless;
  value
end method;

define method do-define-metadata-value
    (supplied-index :: <integer>, value :: <llvm-metadata-value>)
 => ();
  let definition
    = element(*llvm-parse-state*.parse-unnamed-metadata-values, supplied-index,
              default: #f);
  if (definition)
    if (instance?(definition, <llvm-symbolic-metadata>))
      definition.llvm-placeholder-value-forward := value;
      llvm-constrain-type(definition.llvm-value-type, llvm-value-type(value));
    else
      error("value !%d is multiply defined", supplied-index);
    end if;
  end if;

  // Record this value
  element(*llvm-parse-state*.parse-unnamed-metadata-values, supplied-index)
    := value;
end method;

define class <arg-entry> (<object>)
  constant slot arg-attributes :: <llvm-attributes>,
    init-value: $llvm-attribute-none, init-keyword: attributes:;
  constant slot arg-type :: <llvm-type>,
    init-keyword: type:;
end class;

define class <param-entry> (<object>)
  constant slot param-attributes :: <llvm-attributes>,
    init-value: $llvm-attribute-none, init-keyword: attributes:;
  constant slot param-value :: <llvm-value>,
    init-keyword: value:;
end class;

define function string-constant-contents
    (string-constant :: <string>)
 => (contents :: <sequence>);
  let contents = make(<stretchy-object-vector>);
  let string-constant-size = string-constant.size;

  iterate loop (i = 0)
    if (i < string-constant-size)
      let ch = string-constant[i];
      add!(contents, make(<llvm-integer-constant>,
                          type: $llvm-i8-type,
                          integer: as(<integer>, ch)));
      loop(i + 1);
    end if;
  end iterate;
  
  contents
end function;

// Change some attributes (that used to be function attributes) to
// return attributes
define constant $return-attributes-mask
  = %logior($llvm-attribute-zext, $llvm-attribute-sext, $llvm-attribute-inreg);
define function munge-attributes
    (return-attributes :: <llvm-attributes>,
     function-attributes :: <llvm-attributes>)
 => (return-attributes :: <llvm-attributes>,
     function-attributes :: <llvm-attributes>);
  values(%logior(return-attributes,
                 %logand(function-attributes, $return-attributes-mask)),
         %logand(function-attributes, %lognot($return-attributes-mask)))
end function;

define constant $EOF-token = 0;
define constant $%MODULE-token = 1;
define constant $%ASM-token = 2;
define constant $%ALIAS-token = 3;
define constant $%TARGET-token = 4;
define constant $%TRIPLE-token = 5;
define constant $%EQUALS-token = 6;
define constant $%STRINGCONSTANT-token = 7;
define constant $%DATALAYOUT-token = 8;
define constant $%DEPLIBS-token = 9;
define constant $%GLOBALVAL_ID-token = 10;
define constant $%LBRACK-token = 11;
define constant $%RBRACK-token = 12;
define constant $%COMMA-token = 13;
define constant $%PRIVATE-token = 14;
define constant $%LINKER_PRIVATE-token = 15;
define constant $%LINKER_PRIVATE_WEAK-token = 16;
define constant $%LINKER_PRIVATE_WEAK_DEF_AUTO-token = 17;
define constant $%INTERNAL-token = 18;
define constant $%WEAK-token = 19;
define constant $%WEAK_ODR-token = 20;
define constant $%LINKONCE-token = 21;
define constant $%LINKONCE_ODR-token = 22;
define constant $%AVAILABLE_EXTERNALLY-token = 23;
define constant $%APPENDING-token = 24;
define constant $%DLLEXPORT-token = 25;
define constant $%COMMON-token = 26;
define constant $%DLLIMPORT-token = 27;
define constant $%EXTERN_WEAK-token = 28;
define constant $%EXTERNAL-token = 29;
define constant $%TYPE-token = 30;
define constant $%LBRACE-token = 31;
define constant $%RBRACE-token = 32;
define constant $%LANGLE-token = 33;
define constant $%RANGLE-token = 34;
define constant $%OPAQUE-token = 35;
define constant $%INTTYPE-token = 36;
define constant $%STAR-token = 37;
define constant $%EUINT64VAL-token = 38;
define constant $%x-token = 39;
define constant $%DEFINE-token = 40;
define constant $%DECLARE-token = 41;
define constant $%LPAREN-token = 42;
define constant $%RPAREN-token = 43;
define constant $%CCC-token = 44;
define constant $%FASTCC-token = 45;
define constant $%COLDCC-token = 46;
define constant $%X86_STDCALLCC-token = 47;
define constant $%X86_FASTCALLCC-token = 48;
define constant $%ARM_APCSCC-token = 49;
define constant $%ARM_AAPCSCC-token = 50;
define constant $%ARM_AAPCS_VFPCC-token = 51;
define constant $%MSP430_INTRCC-token = 52;
define constant $%X86_THISCALLCC-token = 53;
define constant $%PTX_KERNELCC-token = 54;
define constant $%PTX_DEVICECC-token = 55;
define constant $%CC-token = 56;
define constant $%INREG-token = 57;
define constant $%ZEROEXT-token = 58;
define constant $%SIGNEXT-token = 59;
define constant $%NOALIAS-token = 60;
define constant $%BEGIN-token = 61;
define constant $%END-token = 62;
define constant $%LOCALVAL_ID-token = 63;
define constant $%FLOAT-token = 64;
define constant $%DOUBLE-token = 65;
define constant $%PPC_FP128-token = 66;
define constant $%FP128-token = 67;
define constant $%X86_FP80-token = 68;
define constant $%LABEL-token = 69;
define constant $%METADATA-token = 70;
define constant $%X86_MMX-token = 71;
define constant $%ADDRSPACE-token = 72;
define constant $%VOID-token = 73;
define constant $%DEFAULT-token = 74;
define constant $%HIDDEN-token = 75;
define constant $%PROTECTED-token = 76;
define constant $%THREAD_LOCAL-token = 77;
define constant $%LOCALDYNAMIC-token = 78;
define constant $%INITIALEXEC-token = 79;
define constant $%LOCALEXEC-token = 80;
define constant $%UNNAMED_ADDR-token = 81;
define constant $%GLOBAL-token = 82;
define constant $%CONSTANT-token = 83;
define constant $%C-token = 84;
define constant $%NULL-token = 85;
define constant $%UNDEF-token = 86;
define constant $%ZEROINITIALIZER-token = 87;
define constant $%ESINT64VAL-token = 88;
define constant $%FPVAL-token = 89;
define constant $%TRUE-token = 90;
define constant $%FALSE-token = 91;
define constant $%BLOCKADDRESS-token = 92;
define constant $%ALIGN-token = 93;
define constant $%SECTION-token = 94;
define constant $%BITCAST-token = 95;
define constant $%TO-token = 96;
define constant $%GETELEMENTPTR-token = 97;
define constant $%DOTDOTDOT-token = 98;
define constant $%SELECT-token = 99;
define constant $%ICMP-token = 100;
define constant $%FCMP-token = 101;
define constant $%EXTRACTELEMENT-token = 102;
define constant $%INSERTELEMENT-token = 103;
define constant $%SHUFFLEVECTOR-token = 104;
define constant $%EXTRACTVALUE-token = 105;
define constant $%INSERTVALUE-token = 106;
define constant $%TRUNC-token = 107;
define constant $%ZEXT-token = 108;
define constant $%SEXT-token = 109;
define constant $%FPTRUNC-token = 110;
define constant $%FPEXT-token = 111;
define constant $%UITOFP-token = 112;
define constant $%SITOFP-token = 113;
define constant $%FPTOUI-token = 114;
define constant $%FPTOSI-token = 115;
define constant $%INTTOPTR-token = 116;
define constant $%PTRTOINT-token = 117;
define constant $%LOCALVAR-token = 118;
define constant $%GLOBALVAR-token = 119;
define constant $%GC-token = 120;
define constant $%ADD-token = 121;
define constant $%FADD-token = 122;
define constant $%SUB-token = 123;
define constant $%FSUB-token = 124;
define constant $%MUL-token = 125;
define constant $%FMUL-token = 126;
define constant $%UDIV-token = 127;
define constant $%EXACT-token = 128;
define constant $%SDIV-token = 129;
define constant $%FDIV-token = 130;
define constant $%UREM-token = 131;
define constant $%SREM-token = 132;
define constant $%FREM-token = 133;
define constant $%NUW-token = 134;
define constant $%NSW-token = 135;
define constant $%SHL-token = 136;
define constant $%LSHR-token = 137;
define constant $%ASHR-token = 138;
define constant $%AND-token = 139;
define constant $%OR-token = 140;
define constant $%XOR-token = 141;
define constant $%EQ-token = 142;
define constant $%NE-token = 143;
define constant $%SLT-token = 144;
define constant $%SGT-token = 145;
define constant $%SLE-token = 146;
define constant $%SGE-token = 147;
define constant $%ULT-token = 148;
define constant $%UGT-token = 149;
define constant $%ULE-token = 150;
define constant $%UGE-token = 151;
define constant $%OEQ-token = 152;
define constant $%ONE-token = 153;
define constant $%OLT-token = 154;
define constant $%OGT-token = 155;
define constant $%OLE-token = 156;
define constant $%OGE-token = 157;
define constant $%ORD-token = 158;
define constant $%UNO-token = 159;
define constant $%UEQ-token = 160;
define constant $%UNE-token = 161;
define constant $%LABELSTR-token = 162;
define constant $%RET-token = 163;
define constant $%BR-token = 164;
define constant $%SWITCH-token = 165;
define constant $%INDIRECTBR-token = 166;
define constant $%INVOKE-token = 167;
define constant $%UNWIND-token = 168;
define constant $%UNREACHABLE-token = 169;
define constant $%RESUME-token = 170;
define constant $%NORETURN-token = 171;
define constant $%NOUNWIND-token = 172;
define constant $%READNONE-token = 173;
define constant $%READONLY-token = 174;
define constant $%NOINLINE-token = 175;
define constant $%ALWAYSINLINE-token = 176;
define constant $%INLINEHINT-token = 177;
define constant $%OPTSIZE-token = 178;
define constant $%SSP-token = 179;
define constant $%SSPREQ-token = 180;
define constant $%NOREDZONE-token = 181;
define constant $%NOIMPLICITFLOAT-token = 182;
define constant $%NAKED-token = 183;
define constant $%ALIGNSTACK-token = 184;
define constant $%UWTABLE-token = 185;
define constant $%RETURNS_TWICE-token = 186;
define constant $%NONLAZYBIND-token = 187;
define constant $%VAARG-token = 188;
define constant $%PHI-token = 189;
define constant $%LANDINGPAD-token = 190;
define constant $%PERSONALITY-token = 191;
define constant $%ALLOCA-token = 192;
define constant $%LOAD-token = 193;
define constant $%ATOMIC-token = 194;
define constant $%VOLATILE-token = 195;
define constant $%STORE-token = 196;
define constant $%CMPXCHG-token = 197;
define constant $%ATOMICRMW-token = 198;
define constant $%FENCE-token = 199;
define constant $%GETRESULT-token = 200;
define constant $%CLEANUP-token = 201;
define constant $%CATCH-token = 202;
define constant $%FILTER-token = 203;
define constant $%TAIL-token = 204;
define constant $%CALL-token = 205;
define constant $%SINGLETHREAD-token = 206;
define constant $%UNORDERED-token = 207;
define constant $%MONOTONIC-token = 208;
define constant $%ACQUIRE-token = 209;
define constant $%RELEASE-token = 210;
define constant $%ACQ_REL-token = 211;
define constant $%SEQ_CST-token = 212;
define constant $%XCHG-token = 213;
define constant $%NAND-token = 214;
define constant $%MAX-token = 215;
define constant $%MIN-token = 216;
define constant $%UMAX-token = 217;
define constant $%UMIN-token = 218;
define constant $%SIDEEFFECT-token = 219;
define constant $%SRET-token = 220;
define constant $%NOCAPTURE-token = 221;
define constant $%BYVAL-token = 222;
define constant $%NEST-token = 223;
define constant $%INBOUNDS-token = 224;
define constant $%METADATAVAR-token = 225;
define constant $%EXCLAIM-token = 226;

define function llvm-parser-action0 (arg$1, arg$2) => (value)
  #f
end llvm-parser-action0;

define function llvm-parser-action1 () => (value)
  #f
end llvm-parser-action1;

define function llvm-parser-action3 (arg$1, arg$2, arg$3) => (value)
  if (empty?(*llvm-parse-state*.parse-module.llvm-module-asm))
         *llvm-parse-state*.parse-module.llvm-module-asm := arg$3;
       else
         *llvm-parse-state*.parse-module.llvm-module-asm
           := concatenate(*llvm-parse-state*.parse-module.llvm-module-asm,
                          "\n", arg$3);
       end if
end llvm-parser-action3;

define function llvm-parser-action4 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  begin
         let pointer-type
           = make(<llvm-pointer-type>,
                  pointee: llvm-value-type(arg$8),
                  address-space: arg$5);
         let global
           = apply(make, <llvm-global-variable>,
                   name: arg$1,
                   type: pointer-type,
                   initializer: arg$8,
                   constant?: arg$7 == #"constant",
                   linkage: arg$2,
                   visibility: arg$3,
                   thread-local: arg$4,
		   unnamed-address?: arg$6,
                   arg$9);
         add!(*llvm-parse-state*.parse-module.llvm-module-globals, global);
         do-define-global-value(arg$1, global);
       end
end llvm-parser-action4;

define function llvm-parser-action5 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let pointer-type
           = make(<llvm-pointer-type>,
                  pointee: arg$5.llvm-value-type);
         let global
           = apply(make, <llvm-global-variable>,
                   type: pointer-type,
                   initializer: arg$5,
                   constant?: arg$4 == #"constant",
                   linkage: arg$1,
                   visibility: arg$2,
                   thread-local: arg$3,
                   arg$6);
         add!(*llvm-parse-state*.parse-module.llvm-module-globals, global);
         do-define-global-value(#f, global);
       end
end llvm-parser-action5;

define function llvm-parser-action6 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  begin
         let pointer-type
           = make(<llvm-pointer-type>,
                  pointee: arg$8,
                  address-space: arg$6);
         let global
           = apply(make, <llvm-global-variable>,
                   name: arg$1,
                   type: pointer-type,
                   constant?: arg$7 == #"constant",
                   linkage: arg$2,
                   visibility: arg$3,
                   thread-local: arg$4,
		   unnamed-address?: arg$5,
                   arg$9);
         add!(*llvm-parse-state*.parse-module.llvm-module-globals, global);
         do-define-global-value(arg$1, global);
       end
end llvm-parser-action6;

define function llvm-parser-action7 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let pointer-type = make(<llvm-pointer-type>, pointee: arg$5);
         let global
           = apply(make, <llvm-global-variable>,
                   type: pointer-type,
                   constant?: arg$4 == #"constant",
                   linkage: arg$1,
                   visibility: arg$2,
                   thread-local: arg$3,
                   arg$6);
         add!(*llvm-parse-state*.parse-module.llvm-module-globals, global);
         do-define-global-value(#f, global);
       end
end llvm-parser-action7;

define function llvm-parser-action8 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let alias
           = make(<llvm-global-alias>,
                  name: arg$1,
                  type: llvm-value-type(arg$6),
                  linkage: arg$5,
                  visibility: arg$3,
                  aliasee: arg$6);
         add!(*llvm-parse-state*.parse-module.llvm-module-aliases, alias);
         do-define-global-value(arg$1, alias);
       end
end llvm-parser-action8;

define function llvm-parser-action9 (arg$1, arg$2, arg$3, arg$4) => (value)
  *llvm-parse-state*.parse-module.llvm-module-target-triple := arg$4
end llvm-parser-action9;

define function llvm-parser-action10 (arg$1, arg$2, arg$3, arg$4) => (value)
  *llvm-parse-state*.parse-module.llvm-module-data-layout := arg$4
end llvm-parser-action10;

define function llvm-parser-action11 (arg$1, arg$2, arg$3) => (value)
  concatenate!(*llvm-parse-state*.parse-module.llvm-module-dependent-libraries, arg$3)
end llvm-parser-action11;

define function llvm-parser-action12 (arg$1) => (value)
  arg$1
end llvm-parser-action12;

define function llvm-parser-action13 (arg$1, arg$2) => (value)
  arg$1
end llvm-parser-action13;

define function llvm-parser-action15 (arg$1, arg$2, arg$3) => (value)
  reverse!(arg$2)
end llvm-parser-action15;

define function llvm-parser-action16 () => (value)
  #()
end llvm-parser-action16;

define function llvm-parser-action17 (arg$1) => (value)
  list(arg$1)
end llvm-parser-action17;

define function llvm-parser-action18 (arg$1, arg$2, arg$3) => (value)
  pair(arg$3, arg$1)
end llvm-parser-action18;

define function llvm-parser-action19 () => (value)
  #"external"
end llvm-parser-action19;

define function llvm-parser-action20 (arg$1) => (value)
  #"private"
end llvm-parser-action20;

define function llvm-parser-action21 (arg$1) => (value)
  #"linker-private"
end llvm-parser-action21;

define function llvm-parser-action22 (arg$1) => (value)
  #"linker-private-weak"
end llvm-parser-action22;

define function llvm-parser-action23 (arg$1) => (value)
  #"linker-private-weak-def-auto"
end llvm-parser-action23;

define function llvm-parser-action24 (arg$1) => (value)
  #"internal"
end llvm-parser-action24;

define function llvm-parser-action25 (arg$1) => (value)
  #"weak"
end llvm-parser-action25;

define function llvm-parser-action26 (arg$1) => (value)
  #"weak-odr"
end llvm-parser-action26;

define function llvm-parser-action27 (arg$1) => (value)
  #"linkonce"
end llvm-parser-action27;

define function llvm-parser-action28 (arg$1) => (value)
  #"linkonce-odr"
end llvm-parser-action28;

define function llvm-parser-action29 (arg$1) => (value)
  #"available-externally"
end llvm-parser-action29;

define function llvm-parser-action30 (arg$1) => (value)
  #"appending"
end llvm-parser-action30;

define function llvm-parser-action31 (arg$1) => (value)
  #"dllexport"
end llvm-parser-action31;

define function llvm-parser-action32 (arg$1) => (value)
  #"common"
end llvm-parser-action32;

define function llvm-parser-action33 (arg$1) => (value)
  #"dllimport"
end llvm-parser-action33;

define function llvm-parser-action34 (arg$1) => (value)
  #"extern-weak"
end llvm-parser-action34;

define function llvm-parser-action35 (arg$1) => (value)
  #"external"
end llvm-parser-action35;

define function llvm-parser-action36 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  begin
         let type
           = make(<llvm-struct-type>, name: arg$1, packed?: #f,
                  elements: reverse!(arg$4));
         do-define-type(arg$1, type);
       end
end llvm-parser-action36;

define function llvm-parser-action37 (arg$1, arg$2, arg$3, arg$4) => (value)
  begin
         let type
           = make(<llvm-struct-type>, name: arg$1, packed?: #f,
                  elements: #[]);
         do-define-type(arg$1, type);
       end
end llvm-parser-action37;

define function llvm-parser-action38 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         let type
           = make(<llvm-struct-type>, name: arg$1, packed?: #t,
                  elements: reverse!(arg$5));
         do-define-type(arg$1, type);
       end
end llvm-parser-action38;

define function llvm-parser-action39 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let type
           = make(<llvm-struct-type>, name: arg$1, packed?: #t,
                  elements: #[]);
         do-define-type(arg$1, type);
       end
end llvm-parser-action39;

define function llvm-parser-action40 (arg$1, arg$2, arg$3) => (value)
  begin
         let type = make(<llvm-opaque-type>, name: arg$1);
         do-define-type(arg$1, type);
       end
end llvm-parser-action40;

define function llvm-parser-action41 (arg$1, arg$2, arg$3) => (value)
  begin
         let type = make(<llvm-integer-type>, width: arg$3);
       end
end llvm-parser-action41;

define function llvm-parser-action42 (arg$1, arg$2, arg$3) => (value)
  begin
         let type = make(<llvm-primitive-type>, kind: arg$3);
         do-define-type(arg$1, type);
       end
end llvm-parser-action42;

define function llvm-parser-action43 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  begin
         let type
           = make(<llvm-pointer-type>, pointee: arg$3, address-space: arg$4);
         do-define-type(arg$1, type);
       end
end llvm-parser-action43;

define function llvm-parser-action44 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         let type = make(<llvm-array-type>, size: arg$4, element-type: arg$6);
         do-define-type(arg$1, type);
       end
end llvm-parser-action44;

define function llvm-parser-action45 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         let type = make(<llvm-vector-type>, size: arg$4, element-type: arg$6);
         do-define-type(arg$1, type);
       end
end llvm-parser-action45;

define function llvm-parser-action46 (arg$1, arg$2, arg$3) => (value)
  do-define-type(arg$1, arg$3)
end llvm-parser-action46;

define function llvm-parser-action51 (arg$1, arg$2) => (value)
  begin
         *llvm-parse-state*.parse-function := #f;
         *llvm-parse-state*.parse-unnamed-local-index := 0;
         *llvm-parse-state*.parse-unnamed-local-values.size := 0;
       end
end llvm-parser-action51;

define function llvm-parser-action52 (arg$1, arg$2, arg$3) => (value)
  begin
         let function
           = apply(make, <llvm-function>,
                   linkage: arg$1,
                   visibility: arg$2,
                   arg$3);
         add!(*llvm-parse-state*.parse-module.llvm-module-functions, function);
         do-define-global-value(function.llvm-global-name, function);

         // Install this function as the current one while parsing the
         // function-body
         *llvm-parse-state*.parse-function := function;

         // Install the function arguments as local values
         for (argument in function.llvm-function-arguments)
           do-define-local-value(argument.llvm-argument-name, argument);
         end for;
       end
end llvm-parser-action52;

define function llvm-parser-action53 (arg$1, arg$2, arg$3) => (value)
  begin
         let function
           = apply(make, <llvm-function>,
                   linkage: arg$1,
                   visibility: arg$2,
                   arg$3);
         add!(*llvm-parse-state*.parse-module.llvm-module-functions, function);
         do-define-global-value(function.llvm-global-name, function);
       end
end llvm-parser-action53;

define function llvm-parser-action54 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10, arg$11, arg$12) => (value)
  begin
         let function-type
           = make(<llvm-function-type>,
                  return-type: arg$3,
                  parameter-types: map(llvm-value-type, arg$6.head),
                  varargs?: arg$6.tail);
         let pointer-type
           = make(<llvm-pointer-type>, pointee: function-type);
         let (return-attributes, function-attributes)
           = munge-attributes(arg$2, arg$9);
         let attribute-list
           = make(<llvm-attribute-list>,
                  return-attributes: return-attributes,
                  function-attributes: function-attributes,
                  parameter-attributes:
                    map(llvm-argument-attributes, arg$6.head));
         vector(calling-convention: arg$1,
                type: pointer-type,
                name: arg$4,
                arguments: arg$6.head,
                attribute-list: attribute-list,
		unnamed-address?: arg$8,
                section: arg$10,
                alignment: arg$11,
                garbage-collector: arg$12)
       end
end llvm-parser-action54;

define function llvm-parser-action59 () => (value)
  $llvm-calling-convention-c
end llvm-parser-action59;

define function llvm-parser-action60 (arg$1) => (value)
  $llvm-calling-convention-c
end llvm-parser-action60;

define function llvm-parser-action61 (arg$1) => (value)
  $llvm-calling-convention-fast
end llvm-parser-action61;

define function llvm-parser-action62 (arg$1) => (value)
  $llvm-calling-convention-cold
end llvm-parser-action62;

define function llvm-parser-action63 (arg$1) => (value)
  $llvm-calling-convention-x86-stdcall
end llvm-parser-action63;

define function llvm-parser-action64 (arg$1) => (value)
  $llvm-calling-convention-x86-fastcall
end llvm-parser-action64;

define function llvm-parser-action65 (arg$1) => (value)
  $llvm-calling-convention-arm-apcs
end llvm-parser-action65;

define function llvm-parser-action66 (arg$1) => (value)
  $llvm-calling-convention-arm-aapcs
end llvm-parser-action66;

define function llvm-parser-action67 (arg$1) => (value)
  $llvm-calling-convention-arm-vfp
end llvm-parser-action67;

define function llvm-parser-action68 (arg$1) => (value)
  $llvm-calling-convention-msp430-intr
end llvm-parser-action68;

define function llvm-parser-action69 (arg$1) => (value)
  $llvm-calling-convention-x86-thiscall
end llvm-parser-action69;

define function llvm-parser-action70 (arg$1) => (value)
  $llvm-calling-convention-ptx-kernel
end llvm-parser-action70;

define function llvm-parser-action71 (arg$1) => (value)
  $llvm-calling-convention-ptx-device
end llvm-parser-action71;

define function llvm-parser-action72 (arg$1, arg$2) => (value)
  arg$2
end llvm-parser-action72;

define function llvm-parser-action73 () => (value)
  $llvm-attribute-none
end llvm-parser-action73;

define function llvm-parser-action74 (arg$1, arg$2) => (value)
  llvm-attribute-merge(arg$1, arg$2)
end llvm-parser-action74;

define function llvm-parser-action75 (arg$1) => (value)
  $llvm-attribute-inreg
end llvm-parser-action75;

define function llvm-parser-action76 (arg$1) => (value)
  $llvm-attribute-zext
end llvm-parser-action76;

define function llvm-parser-action77 (arg$1) => (value)
  $llvm-attribute-sext
end llvm-parser-action77;

define function llvm-parser-action78 (arg$1) => (value)
  $llvm-attribute-noalias
end llvm-parser-action78;

define function llvm-parser-action79 (arg$1, arg$2, arg$3) => (value)
  #f
end llvm-parser-action79;

define function llvm-parser-action81 (arg$1) => (value)
  add!(*llvm-parse-state*.parse-function.llvm-function-basic-blocks,
            arg$1)
end llvm-parser-action81;

define function llvm-parser-action82 (arg$1, arg$2) => (value)
  add!(*llvm-parse-state*.parse-function.llvm-function-basic-blocks,
            arg$2)
end llvm-parser-action82;

define function llvm-parser-action85 (arg$1) => (value)
  make(<llvm-opaque-type>)
end llvm-parser-action85;

define function llvm-parser-action86 (arg$1) => (value)
  make(<llvm-integer-type>, width: arg$1)
end llvm-parser-action86;

define function llvm-parser-action87 (arg$1) => (value)
  make(<llvm-primitive-type>, kind: arg$1)
end llvm-parser-action87;

define function llvm-parser-action88 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-pointer-type>, pointee: arg$1, address-space: arg$2)
end llvm-parser-action88;

define function llvm-parser-action89 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-unnamed-types, arg$1, default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-type>, name: arg$1);
               element(*llvm-parse-state*.parse-unnamed-types, arg$1)
                 := placeholder;
               placeholder;
             end
       end
end llvm-parser-action89;

define function llvm-parser-action90 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-module.llvm-type-table,
                 arg$1, default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-type>, name: arg$1);
               element(*llvm-parse-state*.parse-module.llvm-type-table, arg$1)
                 := placeholder;
               placeholder
             end
       end
end llvm-parser-action90;

define function llvm-parser-action91 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-array-type>, size: arg$2, element-type: arg$4)
end llvm-parser-action91;

define function llvm-parser-action92 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-vector-type>, size: arg$2, element-type: arg$4)
end llvm-parser-action92;

define function llvm-parser-action93 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-struct-type>, packed?: #f, elements: reverse!(arg$2))
end llvm-parser-action93;

define function llvm-parser-action94 (arg$1, arg$2) => (value)
  make(<llvm-struct-type>, packed?: #f, elements: #[])
end llvm-parser-action94;

define function llvm-parser-action95 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-struct-type>, packed?: #t, elements: reverse!(arg$3))
end llvm-parser-action95;

define function llvm-parser-action96 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-struct-type>, packed?: #t, elements: #[])
end llvm-parser-action96;

define function llvm-parser-action97 (arg$1) => (value)
  #"FLOAT"
end llvm-parser-action97;

define function llvm-parser-action98 (arg$1) => (value)
  #"DOUBLE"
end llvm-parser-action98;

define function llvm-parser-action99 (arg$1) => (value)
  #"PPC_FP128"
end llvm-parser-action99;

define function llvm-parser-action100 (arg$1) => (value)
  #"FP128"
end llvm-parser-action100;

define function llvm-parser-action101 (arg$1) => (value)
  #"X86_FP80"
end llvm-parser-action101;

define function llvm-parser-action102 (arg$1) => (value)
  #"LABEL"
end llvm-parser-action102;

define function llvm-parser-action103 (arg$1) => (value)
  #"METADATA"
end llvm-parser-action103;

define function llvm-parser-action104 (arg$1) => (value)
  #"X86_MMX"
end llvm-parser-action104;

define function llvm-parser-action105 () => (value)
  0
end llvm-parser-action105;

define function llvm-parser-action106 (arg$1, arg$2, arg$3, arg$4) => (value)
  arg$3
end llvm-parser-action106;

define function llvm-parser-action107 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-function-type>,
            return-type: arg$1,
            parameter-types: map(arg-type, arg$3.head),
            varargs?: arg$3.tail)
end llvm-parser-action107;

define function llvm-parser-action108 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-function-type>,
            return-type: $llvm-void-type,
            parameter-types: map(arg-type, arg$3.head),
            varargs?: arg$3.tail)
end llvm-parser-action108;

define function llvm-parser-action109 (arg$1) => (value)
  make(<arg-entry>, type: arg$1)
end llvm-parser-action109;

define function llvm-parser-action111 (arg$1, arg$2) => (value)
  make(<arg-entry>, type: arg$1, attributes: arg$2)
end llvm-parser-action111;

define function llvm-parser-action112 (arg$1, arg$2) => (value)
  make(<arg-entry>,
            type: arg$1.arg-type,
            attributes: llvm-attribute-merge(arg$1.arg-attributes, arg$2))
end llvm-parser-action112;

define function llvm-parser-action113 () => (value)
  #"default"
end llvm-parser-action113;

define function llvm-parser-action114 (arg$1) => (value)
  #"default"
end llvm-parser-action114;

define function llvm-parser-action115 (arg$1) => (value)
  #"hidden"
end llvm-parser-action115;

define function llvm-parser-action116 (arg$1) => (value)
  #"protected"
end llvm-parser-action116;

define function llvm-parser-action118 (arg$1) => (value)
  #t
end llvm-parser-action118;

define function llvm-parser-action120 (arg$1) => (value)
  #"localdynamic"
end llvm-parser-action120;

define function llvm-parser-action121 (arg$1) => (value)
  #"initialexec"
end llvm-parser-action121;

define function llvm-parser-action122 (arg$1) => (value)
  #"localexec"
end llvm-parser-action122;

define function llvm-parser-action125 (arg$1) => (value)
  #"global"
end llvm-parser-action125;

define function llvm-parser-action126 (arg$1) => (value)
  #"constant"
end llvm-parser-action126;

define function llvm-parser-action127 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-aggregate-constant>,
            type: arg$1,
            aggregate-values: reverse!(arg$3))
end llvm-parser-action127;

define function llvm-parser-action128 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-aggregate-constant>,
            type: arg$1,
            aggregate-values: #[])
end llvm-parser-action128;

define function llvm-parser-action129 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-aggregate-constant>,
            type: arg$1,
            aggregate-values: string-constant-contents(arg$3))
end llvm-parser-action129;

define function llvm-parser-action133 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<llvm-aggregate-constant>,
            type: arg$1,
            aggregate-values: reverse!(arg$4))
end llvm-parser-action133;

define function llvm-parser-action134 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-aggregate-constant>,
            type: arg$1,
            aggregate-values: #[])
end llvm-parser-action134;

define function llvm-parser-action135 (arg$1, arg$2) => (value)
  make(<llvm-null-constant>, type: arg$1)
end llvm-parser-action135;

define function llvm-parser-action136 (arg$1, arg$2) => (value)
  make(<llvm-undef-constant>, type: arg$1)
end llvm-parser-action136;

define function llvm-parser-action137 (arg$1, arg$2) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$2),
                             arg$1);
         arg$2
       end
end llvm-parser-action137;

define function llvm-parser-action140 (arg$1, arg$2) => (value)
  make(<llvm-integer-constant>,
            type: arg$1,
            integer: arg$2)
end llvm-parser-action140;

define function llvm-parser-action142 (arg$1, arg$2) => (value)
  make(<llvm-float-constant>,
            type: arg$1,
            float: arg$2)
end llvm-parser-action142;

define function llvm-parser-action143 (arg$1, arg$2) => (value)
  make(<llvm-integer-constant>,
            type: arg$1,
            integer: 1)
end llvm-parser-action143;

define function llvm-parser-action144 (arg$1, arg$2) => (value)
  make(<llvm-integer-constant>,
            type: arg$1,
            integer: 0)
end llvm-parser-action144;

define function llvm-parser-action145 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  error("%%BLOCKADDRESS")
end llvm-parser-action145;

define function llvm-parser-action147 () => (value)
  #[]
end llvm-parser-action147;

define function llvm-parser-action148 (arg$1, arg$2, arg$3) => (value)
  concatenate(arg$2, arg$3)
end llvm-parser-action148;

define function llvm-parser-action149 (arg$1) => (value)
  vector(section:, arg$1)
end llvm-parser-action149;

define function llvm-parser-action150 (arg$1, arg$2) => (value)
  vector(alignment:, arg$2)
end llvm-parser-action150;

define function llvm-parser-action153 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<llvm-cast-constant>,
            operator: #"BITCAST",
            type: arg$5,
            operands: vector(arg$3))
end llvm-parser-action153;

define function llvm-parser-action154 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<llvm-gep-constant>,
            in-bounds?: arg$2,
            operands: pair(arg$4, reverse!(arg$5)))
end llvm-parser-action154;

define function llvm-parser-action155 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-unnamed-local-values, arg$1,
                 default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-value>, name: arg$1);
               element(*llvm-parse-state*.parse-unnamed-local-values, arg$1)
                 := placeholder
             end
       end
end llvm-parser-action155;

define function llvm-parser-action156 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-unnamed-global-values, arg$1,
                 default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-constant>, name: arg$1);
               element(*llvm-parse-state*.parse-unnamed-global-values, arg$1)
                 := placeholder
             end
       end
end llvm-parser-action156;

define function llvm-parser-action157 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-function.llvm-function-value-table,
                 arg$1, default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-value>, name: arg$1);
               element(*llvm-parse-state*
                         .parse-function.llvm-function-value-table, arg$1)
                 := placeholder
             end
       end
end llvm-parser-action157;

define function llvm-parser-action158 (arg$1) => (value)
  begin
         element(*llvm-parse-state*.parse-module.llvm-global-table, arg$1,
                 default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-constant>, name: arg$1);
               element(*llvm-parse-state*
                         .parse-module.llvm-global-table, arg$1)
                 := placeholder
             end
       end
end llvm-parser-action158;

define function llvm-parser-action159 () => (value)
  pair(#(), #f)
end llvm-parser-action159;

define function llvm-parser-action160 (arg$1) => (value)
  pair(reverse!(arg$1), #f)
end llvm-parser-action160;

define function llvm-parser-action161 (arg$1, arg$2, arg$3) => (value)
  pair(reverse!(arg$1), #t)
end llvm-parser-action161;

define function llvm-parser-action162 (arg$1) => (value)
  pair(#(), #t)
end llvm-parser-action162;

define function llvm-parser-action169 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<llvm-cast-constant>,
            type: arg$5,
            operator: arg$1,
            operands: vector(arg$3))
end llvm-parser-action169;

define function llvm-parser-action171 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("const-expr %%SELECT")
end llvm-parser-action171;

define function llvm-parser-action172 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  apply(make, <llvm-binop-constant>,
             operator: arg$1.head,
             operands: vector(arg$3, arg$5),
             arg$1.tail)
end llvm-parser-action172;

define function llvm-parser-action173 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  apply(make, <llvm-binop-constant>,
            operator: arg$1.head,
            operands: vector(arg$3, arg$5),
	    arg$1.tail)
end llvm-parser-action173;

define function llvm-parser-action174 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         make(<llvm-icmp-constant>,
            predicate: arg$2,
            operands: vector(arg$4, arg$6))
       end
end llvm-parser-action174;

define function llvm-parser-action175 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         make(<llvm-fcmp-constant>,
            predicate: arg$2,
            operands: vector(arg$4, arg$6))
       end
end llvm-parser-action175;

define function llvm-parser-action176 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("const-expr %%EXTRACTELEMENT")
end llvm-parser-action176;

define function llvm-parser-action177 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("const-expr %%INSERTELEMENT")
end llvm-parser-action177;

define function llvm-parser-action178 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  error("const-expr %%SHUFFLEVECTOR")
end llvm-parser-action178;

define function llvm-parser-action179 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  error("const-expr %%EXTRACTVALUE")
end llvm-parser-action179;

define function llvm-parser-action180 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  error("const-expr %%INSERTVALUE")
end llvm-parser-action180;

define function llvm-parser-action181 (arg$1) => (value)
  #"TRUNC"
end llvm-parser-action181;

define function llvm-parser-action182 (arg$1) => (value)
  #"ZEXT"
end llvm-parser-action182;

define function llvm-parser-action183 (arg$1) => (value)
  #"SEXT"
end llvm-parser-action183;

define function llvm-parser-action184 (arg$1) => (value)
  #"FPTRUNC"
end llvm-parser-action184;

define function llvm-parser-action185 (arg$1) => (value)
  #"FPEXT"
end llvm-parser-action185;

define function llvm-parser-action186 (arg$1) => (value)
  #"BITCAST"
end llvm-parser-action186;

define function llvm-parser-action187 (arg$1) => (value)
  #"UITOFP"
end llvm-parser-action187;

define function llvm-parser-action188 (arg$1) => (value)
  #"SITOFP"
end llvm-parser-action188;

define function llvm-parser-action189 (arg$1) => (value)
  #"FPTOUI"
end llvm-parser-action189;

define function llvm-parser-action190 (arg$1) => (value)
  #"FPTOSI"
end llvm-parser-action190;

define function llvm-parser-action191 (arg$1) => (value)
  #"INTTOPTR"
end llvm-parser-action191;

define function llvm-parser-action192 (arg$1) => (value)
  #"PTRTOINT"
end llvm-parser-action192;

define function llvm-parser-action193 (arg$1, arg$2) => (value)
  list(arg$2)
end llvm-parser-action193;

define function llvm-parser-action196 (arg$1) => (value)
  $llvm-void-type
end llvm-parser-action196;

define function llvm-parser-action197 (arg$1, arg$2, arg$3) => (value)
  begin
         do-define-local-value(arg$2, arg$3);
         add!(arg$1.llvm-basic-block-instructions, arg$3);
         arg$1
       end
end llvm-parser-action197;

define function llvm-parser-action209 (arg$1, arg$2) => (value)
  list(make(<llvm-argument>,
                 type: arg$1.arg-type,
                 name: arg$2,
                 attributes: arg$1.arg-attributes,
                 index: 0))
end llvm-parser-action209;

define function llvm-parser-action210 (arg$1, arg$2, arg$3, arg$4) => (value)
  pair(make(<llvm-argument>,
                 type: arg$3.arg-type,
                 name: arg$4,
                 attributes: arg$3.arg-attributes,
                 index: arg$1.head.llvm-argument-index + 1),
            arg$1)
end llvm-parser-action210;

define function llvm-parser-action221 (arg$1, arg$2) => (value)
  pair(#"ADD", arg$2)
end llvm-parser-action221;

define function llvm-parser-action222 (arg$1) => (value)
  #(#"FADD")
end llvm-parser-action222;

define function llvm-parser-action223 (arg$1, arg$2) => (value)
  pair(#"SUB", arg$2)
end llvm-parser-action223;

define function llvm-parser-action224 (arg$1) => (value)
  #(#"FSUB")
end llvm-parser-action224;

define function llvm-parser-action225 (arg$1, arg$2) => (value)
  pair(#"MUL", arg$2)
end llvm-parser-action225;

define function llvm-parser-action226 (arg$1) => (value)
  #(#"FMUL")
end llvm-parser-action226;

define function llvm-parser-action227 (arg$1) => (value)
  #(#"UDIV")
end llvm-parser-action227;

define function llvm-parser-action228 (arg$1, arg$2) => (value)
  #(#"UDIV", exact?:, #t)
end llvm-parser-action228;

define function llvm-parser-action229 (arg$1) => (value)
  #(#"SDIV")
end llvm-parser-action229;

define function llvm-parser-action230 (arg$1, arg$2) => (value)
  #(#"SDIV", exact?:, #t)
end llvm-parser-action230;

define function llvm-parser-action231 (arg$1) => (value)
  #(#"FDIV")
end llvm-parser-action231;

define function llvm-parser-action232 (arg$1) => (value)
  #(#"UREM")
end llvm-parser-action232;

define function llvm-parser-action233 (arg$1) => (value)
  #(#"SREM")
end llvm-parser-action233;

define function llvm-parser-action234 (arg$1) => (value)
  #(#"FREM")
end llvm-parser-action234;

define function llvm-parser-action236 (arg$1) => (value)
  #(no-unsigned-wrap?:, #t)
end llvm-parser-action236;

define function llvm-parser-action237 (arg$1) => (value)
  #(no-signed-wrap?:, #t)
end llvm-parser-action237;

define function llvm-parser-action238 (arg$1, arg$2) => (value)
  #(no-unsigned-wrap?:, #t, no-signed-wrap?:, #t)
end llvm-parser-action238;

define function llvm-parser-action240 (arg$1, arg$2) => (value)
  pair(#"SHL", arg$2)
end llvm-parser-action240;

define function llvm-parser-action241 (arg$1) => (value)
  #(#"LSHR")
end llvm-parser-action241;

define function llvm-parser-action242 (arg$1, arg$2) => (value)
  #(#"LSHR", exact?:, #t)
end llvm-parser-action242;

define function llvm-parser-action243 (arg$1) => (value)
  #(#"ASHR")
end llvm-parser-action243;

define function llvm-parser-action244 (arg$1, arg$2) => (value)
  #(#"ASHR", exact?:, #t)
end llvm-parser-action244;

define function llvm-parser-action245 (arg$1) => (value)
  #(#"AND")
end llvm-parser-action245;

define function llvm-parser-action246 (arg$1) => (value)
  #(#"OR")
end llvm-parser-action246;

define function llvm-parser-action247 (arg$1) => (value)
  #(#"XOR")
end llvm-parser-action247;

define function llvm-parser-action248 (arg$1) => (value)
  #"EQ"
end llvm-parser-action248;

define function llvm-parser-action249 (arg$1) => (value)
  #"NE"
end llvm-parser-action249;

define function llvm-parser-action250 (arg$1) => (value)
  #"SLT"
end llvm-parser-action250;

define function llvm-parser-action251 (arg$1) => (value)
  #"SGT"
end llvm-parser-action251;

define function llvm-parser-action252 (arg$1) => (value)
  #"SLE"
end llvm-parser-action252;

define function llvm-parser-action253 (arg$1) => (value)
  #"SGE"
end llvm-parser-action253;

define function llvm-parser-action254 (arg$1) => (value)
  #"ULT"
end llvm-parser-action254;

define function llvm-parser-action255 (arg$1) => (value)
  #"UGT"
end llvm-parser-action255;

define function llvm-parser-action256 (arg$1) => (value)
  #"ULE"
end llvm-parser-action256;

define function llvm-parser-action257 (arg$1) => (value)
  #"UGE"
end llvm-parser-action257;

define function llvm-parser-action258 (arg$1) => (value)
  #"OEQ"
end llvm-parser-action258;

define function llvm-parser-action259 (arg$1) => (value)
  #"ONE"
end llvm-parser-action259;

define function llvm-parser-action260 (arg$1) => (value)
  #"OLT"
end llvm-parser-action260;

define function llvm-parser-action261 (arg$1) => (value)
  #"OGT"
end llvm-parser-action261;

define function llvm-parser-action262 (arg$1) => (value)
  #"OLE"
end llvm-parser-action262;

define function llvm-parser-action263 (arg$1) => (value)
  #"OGE"
end llvm-parser-action263;

define function llvm-parser-action264 (arg$1) => (value)
  #"ORD"
end llvm-parser-action264;

define function llvm-parser-action265 (arg$1) => (value)
  #"UNO"
end llvm-parser-action265;

define function llvm-parser-action266 (arg$1) => (value)
  #"UEQ"
end llvm-parser-action266;

define function llvm-parser-action267 (arg$1) => (value)
  #"UNE"
end llvm-parser-action267;

define function llvm-parser-action272 (arg$1) => (value)
  #"TRUE"
end llvm-parser-action272;

define function llvm-parser-action273 (arg$1) => (value)
  #"FALSE"
end llvm-parser-action273;

define function llvm-parser-action274 () => (value)
  begin
         let basic-block = make(<llvm-basic-block>);
         do-define-local-value(#f, basic-block);
         basic-block
       end
end llvm-parser-action274;

define function llvm-parser-action275 (arg$1) => (value)
  begin
         let basic-block = make(<llvm-basic-block>, name: arg$1);
         do-define-local-value(arg$1, basic-block);
         basic-block
       end
end llvm-parser-action275;

define function llvm-parser-action276 (arg$1, arg$2) => (value)
  begin
         add!(arg$1.llvm-basic-block-instructions, arg$2);
         arg$1
       end
end llvm-parser-action276;

define function llvm-parser-action277 (arg$1, arg$2) => (value)
  do-define-local-value(arg$1, arg$2)
end llvm-parser-action277;

define function llvm-parser-action279 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-return-instruction>,
            operands: reverse!(arg$2),
            metadata: arg$3)
end llvm-parser-action279;

define function llvm-parser-action280 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-return-instruction>, operands: #[], metadata: arg$3)
end llvm-parser-action280;

define function llvm-parser-action281 (arg$1, arg$2, arg$3, arg$4) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$3), $llvm-label-type);
         make(<llvm-branch-instruction>,
              operands: vector(arg$3),
              metadata: arg$4);
       end
end llvm-parser-action281;

define function llvm-parser-action282 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$3),
                             make(<llvm-integer-type>, width: arg$2));
         llvm-constrain-type(llvm-value-type(arg$6), $llvm-label-type);
         llvm-constrain-type(llvm-value-type(arg$9), $llvm-label-type);
         make(<llvm-branch-instruction>,
              operands: vector(arg$3, arg$6, arg$9),
              metadata: arg$10);
       end
end llvm-parser-action282;

define function llvm-parser-action283 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$3),
                             make(<llvm-integer-type>, width: arg$2));
         llvm-constrain-type(llvm-value-type(arg$6), $llvm-label-type);
         make(<llvm-switch-instruction>,
              operands: concatenate(vector(arg$3, arg$6), arg$8),
              metadata: arg$10)
       end
end llvm-parser-action283;

define function llvm-parser-action284 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$3),
                             make(<llvm-integer-type>, width: arg$2));
         llvm-constrain-type(llvm-value-type(arg$6), $llvm-label-type);
         make(<llvm-switch-instruction>,
              operands: vector(arg$3, arg$6),
              metadata: arg$9)
       end
end llvm-parser-action284;

define function llvm-parser-action285 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  error("%%INDIRECTBR")
end llvm-parser-action285;

define function llvm-parser-action286 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("%%INDIRECTBR")
end llvm-parser-action286;

define function llvm-parser-action287 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10, arg$11, arg$12, arg$13, arg$14, arg$15, arg$16) => (value)
  begin
         let type = arg$4;
         let return-type
           = if (instance?(type, <llvm-pointer-type>))
               let pointee
                 = llvm-type-forward(type.llvm-pointer-type-pointee);
               if (instance?(pointee, <llvm-function-type>))
                 llvm-constrain-type(llvm-value-type(arg$5), type);
                 llvm-type-forward(pointee.llvm-function-type-return-type)
               else
                 type
               end if
             else
               type
             end if;
         llvm-constrain-type(llvm-value-type(arg$12), $llvm-label-type);
         llvm-constrain-type(llvm-value-type(arg$15), $llvm-label-type);
         let param-list = reverse!(arg$7);
         let (return-attributes, function-attributes)
           = munge-attributes(arg$3, arg$9);
         let attribute-list
           = make(<llvm-attribute-list>,
                  return-attributes: return-attributes,
                  function-attributes: function-attributes,
                  parameter-attributes: map(param-attributes, param-list));
         make(<llvm-invoke-instruction>,
              type: return-type,
              attribute-list: attribute-list,
              calling-convention: arg$2,
              operands: concatenate(vector(arg$12, arg$15, arg$5),
                                    map(param-value, param-list)),
              metadata: arg$16)
       end
end llvm-parser-action287;

define function llvm-parser-action288 (arg$1, arg$2) => (value)
  make(<llvm-unreachable-instruction>, metadata: arg$2)
end llvm-parser-action288;

define function llvm-parser-action289 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-resume-instruction>,
            operands: vector(arg$2),
            metadata: arg$3)
end llvm-parser-action289;

define function llvm-parser-action290 (arg$1) => (value)
  $llvm-attribute-noreturn
end llvm-parser-action290;

define function llvm-parser-action291 (arg$1) => (value)
  $llvm-attribute-nounwind
end llvm-parser-action291;

define function llvm-parser-action295 (arg$1) => (value)
  $llvm-attribute-readnone
end llvm-parser-action295;

define function llvm-parser-action296 (arg$1) => (value)
  $llvm-attribute-readonly
end llvm-parser-action296;

define function llvm-parser-action297 (arg$1) => (value)
  $llvm-attribute-noinline
end llvm-parser-action297;

define function llvm-parser-action298 (arg$1) => (value)
  $llvm-attribute-alwaysinline
end llvm-parser-action298;

define function llvm-parser-action299 (arg$1) => (value)
  $llvm-attribute-inlinehint
end llvm-parser-action299;

define function llvm-parser-action300 (arg$1) => (value)
  $llvm-attribute-optsize
end llvm-parser-action300;

define function llvm-parser-action301 (arg$1) => (value)
  $llvm-attribute-ssp
end llvm-parser-action301;

define function llvm-parser-action302 (arg$1) => (value)
  $llvm-attribute-sspreq
end llvm-parser-action302;

define function llvm-parser-action303 (arg$1) => (value)
  $llvm-attribute-noredzone
end llvm-parser-action303;

define function llvm-parser-action304 (arg$1) => (value)
  $llvm-attribute-noimplicitfloat
end llvm-parser-action304;

define function llvm-parser-action305 (arg$1) => (value)
  $llvm-attribute-naked
end llvm-parser-action305;

define function llvm-parser-action306 (arg$1, arg$2, arg$3, arg$4) => (value)
  llvm-attribute-stack-alignment(arg$3)
end llvm-parser-action306;

define function llvm-parser-action307 (arg$1) => (value)
  $llvm-attribute-uwtable
end llvm-parser-action307;

define function llvm-parser-action308 (arg$1) => (value)
  $llvm-attribute-returns-twice
end llvm-parser-action308;

define function llvm-parser-action309 (arg$1) => (value)
  $llvm-attribute-non-lazy-bind
end llvm-parser-action309;

define function llvm-parser-action310 (arg$1, arg$2) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$2), arg$1);
         arg$2
       end
end llvm-parser-action310;

define function llvm-parser-action315 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$2),
                             make(<llvm-integer-type>, width: arg$1));
         llvm-constrain-type(llvm-value-type(arg$5), $llvm-label-type);
         vector(arg$2, arg$5)
       end
end llvm-parser-action315;

define function llvm-parser-action316 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$3),
                             make(<llvm-integer-type>, width: arg$2));
         llvm-constrain-type(llvm-value-type(arg$6), $llvm-label-type);
         concatenate(arg$1, vector(arg$3, arg$6))
       end
end llvm-parser-action316;

define function llvm-parser-action318 (arg$1, arg$2, arg$3, arg$4) => (value)
  #f
end llvm-parser-action318;

define function llvm-parser-action320 (arg$1, arg$2, arg$3) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$2), arg$1.arg-type);
         let entry
           = make(<param-entry>,
                  attributes: llvm-attribute-merge(arg$1.arg-attributes,
                                                   arg$3),
                  value: arg$2);
         list(entry)
       end
end llvm-parser-action320;

define function llvm-parser-action321 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  begin
         llvm-constrain-type(llvm-value-type(arg$4), arg$3.arg-type);
         let entry
           = make(<param-entry>,
                  attributes: llvm-attribute-merge(arg$3.arg-attributes,
                                                   arg$5),
                  value: arg$4);
         pair(entry, arg$1)
       end
end llvm-parser-action321;

define function llvm-parser-action326 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let type = arg$2;
         llvm-constrain-type(llvm-value-type(arg$3), type);
         llvm-constrain-type(llvm-value-type(arg$5), type);
         apply(make, <llvm-binop-instruction>,
               operator: arg$1.head,
               operands: vector(arg$3, arg$5),
               metadata: arg$6,
               arg$1.tail)
       end
end llvm-parser-action326;

define function llvm-parser-action327 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let type = arg$2;
         llvm-constrain-type(llvm-value-type(arg$3), type);
         llvm-constrain-type(llvm-value-type(arg$5), type);
         apply(make, <llvm-binop-instruction>,
              operator: arg$1.head,
              operands: vector(arg$3, arg$5),
              metadata: arg$6,
	      arg$1.tail)
       end
end llvm-parser-action327;

define function llvm-parser-action328 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         let type = arg$3;
         llvm-constrain-type(llvm-value-type(arg$4), type);
         llvm-constrain-type(llvm-value-type(arg$6), type);
         make(<llvm-icmp-instruction>,
              predicate: arg$2,
              operands: vector(arg$4, arg$6),
              metadata: arg$7)
       end
end llvm-parser-action328;

define function llvm-parser-action329 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  begin
         let type = arg$3;
         llvm-constrain-type(llvm-value-type(arg$4), type);
         llvm-constrain-type(llvm-value-type(arg$6), type);
         make(<llvm-fcmp-instruction>,
              predicate: arg$2,
              operands: vector(arg$4, arg$6),
              metadata: arg$7)
       end
end llvm-parser-action329;

define function llvm-parser-action330 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-cast-instruction>,
            type: arg$4,
            operator: arg$1,
            operands: vector(arg$2),
            metadata: arg$5)
end llvm-parser-action330;

define function llvm-parser-action331 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  make(<llvm-select-instruction>,
            operands: vector(arg$2, arg$4, arg$6),
            metadata: arg$7)
end llvm-parser-action331;

define function llvm-parser-action332 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-va-arg-instruction>,
            type: arg$4,
            operands: vector(arg$2),
            metadata: arg$5)
end llvm-parser-action332;

define function llvm-parser-action333 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-extractelement-instruction>,
            operands: vector(arg$2, arg$4),
            metadata: arg$5)
end llvm-parser-action333;

define function llvm-parser-action334 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  make(<llvm-insertelement-instruction>,
            operands: vector(arg$2, arg$4, arg$6),
            metadata: arg$7)
end llvm-parser-action334;

define function llvm-parser-action335 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  make(<llvm-shufflevector-instruction>,
            operands: vector(arg$2, arg$4, arg$6),
            metadata: arg$7)
end llvm-parser-action335;

define function llvm-parser-action336 (arg$1, arg$2, arg$3, arg$4) => (value)
  begin
         let type = arg$2;
         for (i from 0 below arg$3.size by 2)
           llvm-constrain-type(llvm-value-type(arg$3[i]), type);
           llvm-constrain-type(llvm-value-type(arg$3[i + 1]), $llvm-label-type)
         end for;
         make(<llvm-phi-node>, operands: arg$3, metadata: arg$4)
       end
end llvm-parser-action336;

define function llvm-parser-action337 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  make(<llvm-landingpad-instruction>,
            type: arg$2,
            operands: pair(arg$4, reverse!(arg$6)),
            cleanup?: arg$5,
            metadata: arg$7)
end llvm-parser-action337;

define function llvm-parser-action338 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10) => (value)
  begin
         let param-list = reverse!(arg$7);
         let type = arg$4;
         let return-type
           = if (instance?(type, <llvm-pointer-type>))
               let pointee
                 = llvm-type-forward(type.llvm-pointer-type-pointee);
               if (instance?(pointee, <llvm-function-type>))
                 llvm-constrain-type(llvm-value-type(arg$5), type);
                 llvm-type-forward(pointee.llvm-function-type-return-type)
               else
                 type
               end if
             else
               type
             end if;
         let (return-attributes, function-attributes)
           = munge-attributes(arg$3, arg$9);
         let attribute-list
           = make(<llvm-attribute-list>,
                  return-attributes: return-attributes,
                  function-attributes: function-attributes,
                  parameter-attributes: map(param-attributes, param-list));
         make(<llvm-call-instruction>,
              type: return-type,
              attribute-list: attribute-list,
              tail-call?: arg$1,
              calling-convention: arg$2,
              operands: pair(arg$5, map(param-value, param-list)),
              metadata: arg$10)
       end
end llvm-parser-action338;

define function llvm-parser-action339 (arg$1, arg$2, arg$3) => (value)
  begin
         let pointer-type
           = make(<llvm-pointer-type>, pointee: arg$2);
         let size-operand
           = make(<llvm-integer-constant>,
                  type: $llvm-i32-type,
                  integer: 1);
         apply(make, <llvm-alloca-instruction>,
               type: pointer-type,
               operands: vector(size-operand),
               arg$3)
       end
end llvm-parser-action339;

define function llvm-parser-action340 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  begin
         let pointer-type
           = make(<llvm-pointer-type>, pointee: arg$2);
         llvm-constrain-type(llvm-value-type(arg$5),
                             make(<llvm-integer-type>, width: arg$4));
         apply(make, <llvm-alloca-instruction>,
               type: pointer-type,
               operands: vector(arg$5),
               arg$6)
       end
end llvm-parser-action340;

define function llvm-parser-action341 (arg$1, arg$2, arg$3, arg$4) => (value)
  apply(make, <llvm-load-instruction>,
             volatile?: arg$2,
             operands: vector(arg$3),
             arg$4)
end llvm-parser-action341;

define function llvm-parser-action342 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  apply(make, <llvm-atomic-load-instruction>,
             volatile?: arg$3,
             operands: vector(arg$4),
             scope: arg$5,
             ordering: arg$6,
             arg$7)
end llvm-parser-action342;

define function llvm-parser-action343 (arg$1, arg$2, arg$3, arg$4) => (value)
  apply(make, <llvm-load-instruction>,
            volatile?: #t,
            operands: vector(arg$3),
            arg$4)
end llvm-parser-action343;

define function llvm-parser-action344 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  apply(make, <llvm-store-instruction>,
             volatile?: arg$2,
             operands: vector(arg$3, arg$5),
             arg$6)
end llvm-parser-action344;

define function llvm-parser-action345 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  apply(make, <llvm-atomic-store-instruction>,
             volatile?: arg$3,
             operands: vector(arg$4, arg$6),
             scope: arg$7,
             ordering: arg$8,
             arg$9)
end llvm-parser-action345;

define function llvm-parser-action346 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  apply(make, <llvm-store-instruction>,
             volatile?: #t,
             operands: vector(arg$3, arg$5),
             arg$6)
end llvm-parser-action346;

define function llvm-parser-action347 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9, arg$10) => (value)
  make(<llvm-cmpxchg-instruction>,
            volatile?: arg$2,
            operands: vector(arg$3, arg$5, arg$7),
            scope: arg$8,
            ordering: arg$9,
            metadata: arg$10)
end llvm-parser-action347;

define function llvm-parser-action348 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8, arg$9) => (value)
  make(<llvm-atomicrmw-instruction>,
            volatile?: arg$2,
            operation: arg$3,
            operands: vector(arg$4, arg$6),
            scope: arg$7,
            ordering: arg$8,
            metadata: arg$9)
end llvm-parser-action348;

define function llvm-parser-action349 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-fence-instruction>,
            scope: arg$2,
            ordering: arg$3,
            metadata: arg$4)
end llvm-parser-action349;

define function llvm-parser-action350 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-gep-instruction>,
            in-bounds?: arg$2,
            operands: pair(arg$3, reverse!(arg$4)),
            metadata: arg$5)
end llvm-parser-action350;

define function llvm-parser-action351 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  make(<llvm-insert-value-instruction>,
            operands: vector(arg$2, arg$4),
            indices: reverse!(arg$5),
            metadata: arg$6)
end llvm-parser-action351;

define function llvm-parser-action352 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-extract-value-instruction>,
            operands: vector(arg$2),
            indices: reverse!(arg$3),
            metadata: arg$4)
end llvm-parser-action352;

define function llvm-parser-action353 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-extract-value-instruction>,
            operands: vector(arg$2),
            indices: vector(arg$4),
            metadata: arg$5)
end llvm-parser-action353;

define function llvm-parser-action357 (arg$1, arg$2, arg$3) => (value)
  add(arg$1, arg$3)
end llvm-parser-action357;

define function llvm-parser-action359 (arg$1, arg$2) => (value)
  #t
end llvm-parser-action359;

define function llvm-parser-action360 (arg$1) => (value)
  #f
end llvm-parser-action360;

define function llvm-parser-action361 () => (value)
  #"cross-thread"
end llvm-parser-action361;

define function llvm-parser-action362 (arg$1) => (value)
  #"single-thread"
end llvm-parser-action362;

define function llvm-parser-action363 (arg$1) => (value)
  #"unordered"
end llvm-parser-action363;

define function llvm-parser-action364 (arg$1) => (value)
  #"monotonic"
end llvm-parser-action364;

define function llvm-parser-action365 (arg$1) => (value)
  #"acquire"
end llvm-parser-action365;

define function llvm-parser-action366 (arg$1) => (value)
  #"release"
end llvm-parser-action366;

define function llvm-parser-action367 (arg$1) => (value)
  #"acquire-release"
end llvm-parser-action367;

define function llvm-parser-action368 (arg$1) => (value)
  #"sequentially-consistent"
end llvm-parser-action368;

define function llvm-parser-action369 (arg$1) => (value)
  #"xchg"
end llvm-parser-action369;

define function llvm-parser-action370 (arg$1) => (value)
  #"add"
end llvm-parser-action370;

define function llvm-parser-action371 (arg$1) => (value)
  #"sub"
end llvm-parser-action371;

define function llvm-parser-action372 (arg$1) => (value)
  #"and"
end llvm-parser-action372;

define function llvm-parser-action373 (arg$1) => (value)
  #"nand"
end llvm-parser-action373;

define function llvm-parser-action374 (arg$1) => (value)
  #"or"
end llvm-parser-action374;

define function llvm-parser-action375 (arg$1) => (value)
  #"xor"
end llvm-parser-action375;

define function llvm-parser-action376 (arg$1) => (value)
  #"max"
end llvm-parser-action376;

define function llvm-parser-action377 (arg$1) => (value)
  #"min"
end llvm-parser-action377;

define function llvm-parser-action378 (arg$1) => (value)
  #"umax"
end llvm-parser-action378;

define function llvm-parser-action379 (arg$1) => (value)
  #"umin"
end llvm-parser-action379;

define function llvm-parser-action380 (arg$1) => (value)
  make(<llvm-integer-constant>, integer: arg$1)
end llvm-parser-action380;

define function llvm-parser-action382 (arg$1) => (value)
  make(<llvm-float-constant>, float: arg$1)
end llvm-parser-action382;

define function llvm-parser-action383 (arg$1) => (value)
  make(<llvm-integer-constant>, integer: 1)
end llvm-parser-action383;

define function llvm-parser-action384 (arg$1) => (value)
  make(<llvm-integer-constant>, integer: 0)
end llvm-parser-action384;

define function llvm-parser-action385 (arg$1) => (value)
  make(<llvm-null-constant>)
end llvm-parser-action385;

define function llvm-parser-action386 (arg$1) => (value)
  make(<llvm-undef-constant>)
end llvm-parser-action386;

define function llvm-parser-action388 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-aggregate-constant>, aggregate-values: reverse!(arg$2))
end llvm-parser-action388;

define function llvm-parser-action390 (arg$1, arg$2) => (value)
  make(<llvm-aggregate-constant>, aggregate-values: #[])
end llvm-parser-action390;

define function llvm-parser-action391 (arg$1, arg$2) => (value)
  make(<llvm-aggregate-constant>,
            aggregate-values: string-constant-contents(arg$2))
end llvm-parser-action391;

define function llvm-parser-action394 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  make(<llvm-aggregate-constant>, aggregate-values: reverse!(arg$3))
end llvm-parser-action394;

define function llvm-parser-action395 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-aggregate-constant>, aggregate-values: #[])
end llvm-parser-action395;

define function llvm-parser-action396 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  error("%%BLOCKADDRESS")
end llvm-parser-action396;

define function llvm-parser-action398 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  apply(make, <llvm-asm-constant>,
             asm-string: arg$3,
             constraint: arg$5,
             arg$2)
end llvm-parser-action398;

define function llvm-parser-action401 (arg$1) => (value)
  #[side-effect?:, #t]
end llvm-parser-action401;

define function llvm-parser-action402 (arg$1) => (value)
  #[align-stack?:, #t]
end llvm-parser-action402;

define function llvm-parser-action403 (arg$1, arg$2) => (value)
  #[side-effect?:, #t, align-stack?:, #t]
end llvm-parser-action403;

define function llvm-parser-action411 (arg$1) => (value)
  $llvm-attribute-sret
end llvm-parser-action411;

define function llvm-parser-action413 (arg$1) => (value)
  $llvm-attribute-nocapture
end llvm-parser-action413;

define function llvm-parser-action414 (arg$1) => (value)
  $llvm-attribute-byval
end llvm-parser-action414;

define function llvm-parser-action415 (arg$1) => (value)
  $llvm-attribute-nest
end llvm-parser-action415;

define function llvm-parser-action416 (arg$1, arg$2) => (value)
  llvm-attribute-alignment(arg$2)
end llvm-parser-action416;

define function llvm-parser-action417 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  vector(arg$2, arg$4)
end llvm-parser-action417;

define function llvm-parser-action418 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  concatenate(arg$1, vector(arg$4, arg$6))
end llvm-parser-action418;

define function llvm-parser-action423 (arg$1) => (value)
  vector(alignment: 0, metadata: arg$1)
end llvm-parser-action423;

define function llvm-parser-action424 (arg$1, arg$2, arg$3, arg$4) => (value)
  vector(alignment: arg$3, metadata: arg$4)
end llvm-parser-action424;

define function llvm-parser-action425 (arg$1, arg$2, arg$3, arg$4, arg$5) => (value)
  add!(*llvm-parse-state*.parse-module.llvm-module-named-metadata,
            make(<llvm-named-metadata>, name: arg$1, operands: #()))
end llvm-parser-action425;

define function llvm-parser-action426 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6) => (value)
  add!(*llvm-parse-state*.parse-module.llvm-module-named-metadata,
            make(<llvm-named-metadata>, name: arg$1,
	         operands: reverse!(arg$5)))
end llvm-parser-action426;

define function llvm-parser-action427 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7) => (value)
  do-define-metadata-value(arg$2,
                                make(<llvm-metadata-node>,
                                     node-values: #[],
                                     function-local?: #f))
end llvm-parser-action427;

define function llvm-parser-action428 (arg$1, arg$2, arg$3, arg$4, arg$5, arg$6, arg$7, arg$8) => (value)
  do-define-metadata-value(arg$2,
                                make(<llvm-metadata-node>,
                                     node-values: reverse!(arg$7),
                                     function-local?: #f))
end llvm-parser-action428;

define function llvm-parser-action432 (arg$1) => (value)
  list(#f)
end llvm-parser-action432;

define function llvm-parser-action434 (arg$1, arg$2, arg$3) => (value)
  pair(#f, arg$1)
end llvm-parser-action434;

define function llvm-parser-action436 (arg$1, arg$2, arg$3) => (value)
  make(<llvm-metadata-node>, node-values: #[], function-local?: #f)
end llvm-parser-action436;

define function llvm-parser-action437 (arg$1, arg$2, arg$3, arg$4) => (value)
  make(<llvm-metadata-node>, node-values: reverse!(arg$3))
end llvm-parser-action437;

define function llvm-parser-action438 (arg$1, arg$2) => (value)
  make(<llvm-metadata-string>, string: arg$2)
end llvm-parser-action438;

define function llvm-parser-action439 (arg$1, arg$2) => (value)
  begin
         element(*llvm-parse-state*.parse-unnamed-metadata-values, arg$2,
                 default: #f)
           | begin
               let placeholder = make(<llvm-symbolic-metadata>, name: arg$2);
               element(*llvm-parse-state*.parse-unnamed-metadata-values, arg$2)
                 := placeholder
             end
       end
end llvm-parser-action439;

define function llvm-parser-action442 (arg$1, arg$2) => (value)
  list(make(<llvm-named-metadata>, name: arg$1, operands: list(arg$2)))
end llvm-parser-action442;

define function llvm-parser-action443 (arg$1, arg$2, arg$3, arg$4) => (value)
  pair(make(<llvm-named-metadata>, name: arg$3, operands: list(arg$4)),
            arg$1)
end llvm-parser-action443;

define constant llvm-parser :: <parser>
  = make(<parser>,
  action-table:
      #[#[65535, 1, 0, 1, 1, 1, 4, 1, 19, 1, 7, 1, 9, 1, 10, 1, 14, 1, 16, 1, 17, 1, 25, 1, 20, 1, 21, 1, 22, 1, 23, 1, 24, 1, 27, 1, 28, 1, 29, 1, 41, 1, 40, 1, 76, 1, 119, 1, 63, 1, 83, 1, 82, 1, 74, 1, 75, 1, 77, 1, 118, 1, 225, 1, 226, 1, 18, 1, 15, 1, 26, 1],
	#[7, -13, 15, -30, 4, -12, 22, -31, 20, -23, 1, -22, 10, -25, 14, -9, 23, -16, 9, -35, 17, -29, 25, -26, 24, -3, 0, -17, 16, -14, 26, -28, 19, -38, 18, -15, 29, -6, 28, -18, 21, -10, 27, -34, 63, -8, 40, -39, 41, -19, 76, 19, 75, 19, 74, 19, 83, 19, 77, 19, 82, 19, 118, -36, 119, -33, 225, -32, 226, -21],
	#[65535, 30, 3, 30, 7, 30, 11, 30, 31, 30, 33, 30, 35, 30, 36, 30, 44, 30, 45, 30, 46, 30, 47, 30, 48, 30, 49, 30, 50, 30, 51, 30, 52, 30, 53, 30, 54, 30, 55, 30, 56, 30, 57, 30, 58, 30, 59, 30, 60, 30, 63, 30, 64, 30, 65, 30, 66, 30, 67, 30, 68, 30, 69, 30, 70, 30, 71, 30, 72, 30, 73, 30, 74, 30, 75, 30, 76, 30, 77, 30, 81, 30, 82, 30, 83, 30, 95, 30, 97, 30, 118, 30],
	#[6, -918],
	#[6, -917],
	#[65535, 35, 7, 35, 11, 35, 31, 35, 33, 35, 35, 35, 36, 35, 44, 35, 45, 35, 46, 35, 47, 35, 48, 35, 49, 35, 50, 35, 51, 35, 52, 35, 53, 35, 54, 35, 55, 35, 56, 35, 57, 35, 58, 35, 59, 35, 60, 35, 63, 35, 64, 35, 65, 35, 66, 35, 67, 35, 68, 35, 69, 35, 70, 35, 71, 35, 72, 35, 73, 35, 74, 35, 75, 35, 76, 35, 77, 35, 81, 35, 82, 35, 83, 35, 118, 35],
	#[30, -889],
	#[6, -748],
	#[65535, 20, 7, 20, 36, 20, 46, 20, 47, 20, 3, 20, 59, 20, 11, 20, 31, 20, 63, 20, 53, 20, 70, 20, 58, 20, 35, 20, 71, 20, 54, 20, 57, 20, 45, 20, 67, 20, 33, 20, 60, 20, 69, 20, 55, 20, 48, 20, 51, 20, 49, 20, 77, 20, 118, 20, 82, 20, 50, 20, 72, 20, 76, 20, 65, 20, 44, 20, 66, 20, 64, 20, 81, 20, 56, 20, 75, 20, 73, 20, 97, 20, 74, 20, 83, 20, 52, 20, 95, 20, 68, 20],
	#[65535, 27, 31, 27, 46, 27, 59, 27, 44, 27, 55, 27, 45, 27, 11, 27, 60, 27, 63, 27, 33, 27, 7, 27, 118, 27, 58, 27, 35, 27, 52, 27, 54, 27, 48, 27, 47, 27, 3, 27, 76, 27, 83, 27, 69, 27, 57, 27, 53, 27, 74, 27, 51, 27, 49, 27, 70, 27, 36, 27, 95, 27, 65, 27, 81, 27, 71, 27, 64, 27, 67, 27, 50, 27, 77, 27, 56, 27, 75, 27, 73, 27, 72, 27, 66, 27, 68, 27, 82, 27, 97, 27],
	#[74, -754, 75, -756, 76, -753, 77, 113, 82, 113, 83, 113],
	#[5, -879, 8, -878],
	#[65535, 202, 7, 202, 14, 202, 4, 202, 12, 202, 2, 202, 6, 202, 11, 202, 15, 202, 1, 202, 10, 202, 16, 202, 62, 202, 23, 202, 0, 202, 17, 202, 22, 202, 87, 202, 27, 202, 31, 202, 20, 202, 25, 202, 19, 202, 18, 202, 9, 202, 29, 202, 28, 202, 34, 202, 13, 202, 58, 202, 26, 202, 84, 202, 32, 202, 37, 202, 36, 202, 59, 202, 21, 202, 57, 202, 75, 202, 24, 202, 33, 202, 41, 202, 43, 202, 42, 202, 119, 202, 102, 202, 63, 202, 40, 202, 93, 202, 95, 202, 38, 202, 77, 202, 86, 202, 74, 202, 76, 202, 60, 202, 83, 202, 103, 202, 90, 202, 72, 202, 88, 202, 118, 202, 115, 202, 96, 202, 91, 202, 111, 202, 89, 202, 97, 202, 105, 202, 126, 202, 92, 202, 114, 202, 82, 202, 106, 202, 104, 202, 85, 202, 99, 202, 107, 202, 100, 202, 101, 202, 108, 202, 109, 202, 110, 202, 122, 202, 112, 202, 113, 202, 133, 202, 132, 202, 116, 202, 117, 202, 131, 202, 130, 202, 121, 202, 129, 202, 123, 202, 124, 202, 125, 202, 136, 202, 127, 202, 137, 202, 138, 202, 139, 202, 140, 202, 141, 202, 166, 202, 165, 202, 164, 202, 170, 202, 163, 202, 162, 202, 167, 202, 168, 202, 169, 202, 207, 202, 188, 202, 189, 202, 190, 202, 191, 202, 192, 202, 193, 202, 195, 202, 196, 202, 197, 202, 198, 202, 199, 202, 200, 202, 201, 202, 202, 202, 203, 202, 204, 202, 205, 202, 206, 202, 208, 202, 209, 202, 210, 202, 211, 202, 212, 202, 220, 202, 221, 202, 222, 202, 223, 202, 225, 202, 226, 202],
	#[65535, 22, 3, 22, 7, 22, 11, 22, 31, 22, 33, 22, 35, 22, 36, 22, 44, 22, 45, 22, 46, 22, 47, 22, 48, 22, 49, 22, 50, 22, 51, 22, 52, 22, 53, 22, 54, 22, 55, 22, 56, 22, 57, 22, 58, 22, 59, 22, 60, 22, 63, 22, 64, 22, 65, 22, 66, 22, 67, 22, 68, 22, 69, 22, 70, 22, 71, 22, 72, 22, 73, 22, 74, 22, 75, 22, 76, 22, 77, 22, 81, 22, 82, 22, 83, 22, 95, 22, 97, 22, 118, 22],
	#[65535, 24, 3, 24, 7, 24, 11, 24, 31, 24, 33, 24, 35, 24, 36, 24, 44, 24, 45, 24, 46, 24, 47, 24, 48, 24, 49, 24, 50, 24, 51, 24, 52, 24, 53, 24, 54, 24, 55, 24, 56, 24, 57, 24, 58, 24, 59, 24, 60, 24, 63, 24, 64, 24, 65, 24, 66, 24, 67, 24, 68, 24, 69, 24, 70, 24, 71, 24, 72, 24, 73, 24, 74, 24, 75, 24, 76, 24, 77, 24, 81, 24, 82, 24, 83, 24, 95, 24, 97, 24, 118, 24],
	#[65535, 29, 7, 29, 55, 29, 3, 29, 45, 29, 11, 29, 33, 29, 71, 29, 47, 29, 46, 29, 31, 29, 59, 29, 35, 29, 36, 29, 50, 29, 44, 29, 70, 29, 48, 29, 49, 29, 69, 29, 51, 29, 52, 29, 53, 29, 54, 29, 66, 29, 56, 29, 57, 29, 58, 29, 76, 29, 60, 29, 63, 29, 64, 29, 65, 29, 95, 29, 67, 29, 68, 29, 97, 29, 83, 29, 82, 29, 72, 29, 73, 29, 74, 29, 75, 29, 77, 29, 81, 29, 118, 29],
	#[#"eoi", #"accept"],
	#[65535, 34, 7, 34, 46, 34, 70, 34, 45, 34, 54, 34, 36, 34, 53, 34, 44, 34, 47, 34, 52, 34, 51, 34, 33, 34, 69, 34, 55, 34, 50, 34, 59, 34, 31, 34, 11, 34, 63, 34, 58, 34, 68, 34, 67, 34, 56, 34, 57, 34, 48, 34, 73, 34, 71, 34, 35, 34, 60, 34, 65, 34, 64, 34, 49, 34, 83, 34, 66, 34, 81, 34, 75, 34, 72, 34, 76, 34, 74, 34, 118, 34, 77, 34, 82, 34],
	#[14, -9, 46, 19, 17, -29, 23, -16, 55, 19, 11, 19, 60, 19, 33, 19, 7, 19, 19, -38, 26, -28, 18, -15, 15, -30, 45, 19, 47, 19, 76, 19, 24, -3, 16, -14, 25, -26, 20, -23, 21, -10, 22, -31, 27, -34, 28, -18, 29, -6, 31, 19, 59, 19, 35, 19, 36, 19, 65, 19, 50, 19, 44, 19, 58, 19, 48, 19, 49, 19, 57, 19, 51, 19, 52, 19, 53, 19, 54, 19, 66, 19, 56, 19, 63, 19, 64, 19, 67, 19, 68, 19, 69, 19, 70, 19, 71, 19, 73, 19, 74, 19, 75, 19, 118, 19],
	#[83, 113, 75, -756, 82, 113, 74, -754, 76, -753, 77, 113],
	#[38, -859],
	#[2, -856],
	#[65535, 26, 47, 26, 36, 26, 44, 26, 31, 26, 51, 26, 7, 26, 46, 26, 33, 26, 3, 26, 57, 26, 55, 26, 35, 26, 56, 26, 11, 26, 76, 26, 63, 26, 53, 26, 66, 26, 49, 26, 58, 26, 68, 26, 52, 26, 54, 26, 45, 26, 67, 26, 60, 26, 69, 26, 59, 26, 50, 26, 73, 26, 74, 26, 48, 26, 70, 26, 72, 26, 83, 26, 95, 26, 65, 26, 75, 26, 71, 26, 64, 26, 82, 26, 77, 26, 81, 26, 97, 26, 118, 26],
	#[65535, 2, 18, 2, 15, 2, 4, 2, 0, 2, 1, 2, 26, 2, 19, 2, 7, 2, 9, 2, 10, 2, 14, 2, 16, 2, 17, 2, 25, 2, 20, 2, 21, 2, 22, 2, 23, 2, 24, 2, 27, 2, 28, 2, 29, 2, 41, 2, 40, 2, 76, 2, 119, 2, 63, 2, 83, 2, 82, 2, 74, 2, 75, 2, 77, 2, 118, 2, 225, 2, 226, 2],
	#[6, -855],
	#[65535, 31, 55, 31, 54, 31, 44, 31, 31, 31, 63, 31, 7, 31, 46, 31, 70, 31, 11, 31, 47, 31, 3, 31, 68, 31, 59, 31, 45, 31, 35, 31, 49, 31, 60, 31, 67, 31, 53, 31, 48, 31, 58, 31, 77, 31, 82, 31, 33, 31, 36, 31, 74, 31, 57, 31, 66, 31, 76, 31, 83, 31, 69, 31, 73, 31, 72, 31, 51, 31, 52, 31, 118, 31, 81, 31, 50, 31, 95, 31, 64, 31, 71, 31, 97, 31, 56, 31, 65, 31, 75, 31],
	#[14, -9, 21, -10, 3, 19, 15, -30, 16, -14, 17, -29, 18, -15, 19, -38, 20, -23, 22, -31, 23, -16, 24, -3, 25, -26, 26, -28, 27, -34, 28, -18, 29, -6, 72, 19, 74, 19, 75, 19, 76, 19, 77, 19, 81, 19, 82, 19, 83, 19],
	#[65535, 32, 7, 32, 36, 32, 46, 32, 53, 32, 44, 32, 47, 32, 3, 32, 59, 32, 58, 32, 11, 32, 31, 32, 63, 32, 70, 32, 35, 32, 71, 32, 54, 32, 49, 32, 57, 32, 45, 32, 67, 32, 33, 32, 60, 32, 69, 32, 55, 32, 48, 32, 51, 32, 66, 32, 77, 32, 118, 32, 82, 32, 50, 32, 72, 32, 76, 32, 65, 32, 75, 32, 64, 32, 81, 32, 56, 32, 73, 32, 97, 32, 74, 32, 83, 32, 52, 32, 95, 32, 68, 32],
	#[65535, 23, 47, 23, 55, 23, 31, 23, 63, 23, 7, 23, 36, 23, 46, 23, 53, 23, 44, 23, 51, 23, 3, 23, 11, 23, 70, 23, 49, 23, 58, 23, 35, 23, 71, 23, 54, 23, 74, 23, 57, 23, 45, 23, 67, 23, 33, 23, 60, 23, 69, 23, 48, 23, 68, 23, 77, 23, 82, 23, 50, 23, 59, 23, 76, 23, 65, 23, 73, 23, 66, 23, 72, 23, 64, 23, 52, 23, 56, 23, 75, 23, 83, 23, 81, 23, 97, 23, 95, 23, 118, 23],
	#[65535, 21, 3, 21, 7, 21, 11, 21, 31, 21, 33, 21, 35, 21, 36, 21, 44, 21, 45, 21, 46, 21, 47, 21, 48, 21, 49, 21, 50, 21, 51, 21, 52, 21, 53, 21, 54, 21, 55, 21, 56, 21, 57, 21, 58, 21, 59, 21, 60, 21, 63, 21, 64, 21, 65, 21, 66, 21, 67, 21, 68, 21, 69, 21, 70, 21, 71, 21, 72, 21, 73, 21, 74, 21, 75, 21, 76, 21, 77, 21, 81, 21, 82, 21, 83, 21, 95, 21, 97, 21, 118, 21],
	#[65535, 28, 55, 28, 47, 28, 31, 28, 11, 28, 63, 28, 7, 28, 36, 28, 46, 28, 53, 28, 44, 28, 51, 28, 50, 28, 33, 28, 52, 28, 45, 28, 58, 28, 49, 28, 60, 28, 70, 28, 83, 28, 118, 28, 35, 28, 71, 28, 54, 28, 74, 28, 57, 28, 65, 28, 66, 28, 3, 28, 76, 28, 69, 28, 59, 28, 48, 28, 68, 28, 77, 28, 72, 28, 95, 28, 64, 28, 73, 28, 97, 28, 56, 28, 75, 28, 81, 28, 67, 28, 82, 28],
	#[6, -797],
	#[65535, 204, 7, 204, 4, 204, 23, 204, 6, 204, 36, 204, 18, 204, 11, 204, 14, 204, 27, 204, 60, 204, 0, 204, 21, 204, 15, 204, 103, 204, 22, 204, 1, 204, 9, 204, 10, 204, 29, 204, 20, 204, 41, 204, 13, 204, 26, 204, 43, 204, 76, 204, 93, 204, 28, 204, 12, 204, 75, 204, 34, 204, 62, 204, 58, 204, 24, 204, 95, 204, 16, 204, 57, 204, 127, 204, 42, 204, 59, 204, 118, 204, 105, 204, 115, 204, 17, 204, 25, 204, 77, 204, 102, 204, 126, 204, 111, 204, 19, 204, 122, 204, 99, 204, 40, 204, 114, 204, 110, 204, 119, 204, 74, 204, 63, 204, 32, 204, 97, 204, 133, 204, 82, 204, 96, 204, 101, 204, 108, 204, 83, 204, 167, 204, 116, 204, 100, 204, 188, 204, 129, 204, 137, 204, 124, 204, 190, 204, 106, 204, 141, 204, 130, 204, 164, 204, 136, 204, 121, 204, 131, 204, 132, 204, 112, 204, 109, 204, 104, 204, 189, 204, 113, 204, 204, 204, 107, 204, 123, 204, 163, 204, 117, 204, 162, 204, 125, 204, 170, 204, 165, 204, 138, 204, 139, 204, 140, 204, 169, 204, 166, 204, 193, 204, 192, 204, 206, 204, 205, 204, 168, 204, 195, 204, 203, 204, 196, 204, 197, 204, 198, 204, 199, 204, 200, 204, 201, 204, 202, 204, 207, 204, 208, 204, 209, 204, 210, 204, 211, 204, 212, 204, 220, 204, 221, 204, 222, 204, 223, 204, 225, 204, 226, 204],
	#[65535, 33, 7, 33, 55, 33, 53, 33, 71, 33, 70, 33, 36, 33, 54, 33, 60, 33, 44, 33, 31, 33, 63, 33, 33, 33, 45, 33, 50, 33, 48, 33, 46, 33, 59, 33, 35, 33, 75, 33, 11, 33, 47, 33, 77, 33, 74, 33, 49, 33, 57, 33, 52, 33, 64, 33, 58, 33, 82, 33, 69, 33, 73, 33, 67, 33, 118, 33, 51, 33, 65, 33, 66, 33, 76, 33, 81, 33, 56, 33, 83, 33, 72, 33, 68, 33],
	#[6, -789],
	#[65535, 201, 6, 201, 7, 201, 15, 201, 12, 201, 23, 201, 14, 201, 2, 201, 4, 201, 29, 201, 1, 201, 9, 201, 13, 201, 22, 201, 16, 201, 21, 201, 18, 201, 10, 201, 17, 201, 38, 201, 26, 201, 86, 201, 25, 201, 11, 201, 24, 201, 0, 201, 34, 201, 31, 201, 20, 201, 43, 201, 63, 201, 19, 201, 32, 201, 42, 201, 40, 201, 28, 201, 41, 201, 27, 201, 62, 201, 37, 201, 36, 201, 59, 201, 57, 201, 75, 201, 90, 201, 58, 201, 33, 201, 82, 201, 85, 201, 87, 201, 74, 201, 83, 201, 72, 201, 76, 201, 60, 201, 89, 201, 77, 201, 88, 201, 84, 201, 91, 201, 92, 201, 93, 201, 95, 201, 96, 201, 97, 201, 99, 201, 100, 201, 101, 201, 102, 201, 103, 201, 104, 201, 105, 201, 106, 201, 107, 201, 108, 201, 109, 201, 110, 201, 111, 201, 112, 201, 113, 201, 114, 201, 115, 201, 116, 201, 117, 201, 118, 201, 119, 201, 121, 201, 122, 201, 123, 201, 124, 201, 125, 201, 126, 201, 127, 201, 129, 201, 130, 201, 131, 201, 132, 201, 133, 201, 136, 201, 137, 201, 138, 201, 139, 201, 140, 201, 141, 201, 162, 201, 163, 201, 164, 201, 165, 201, 166, 201, 167, 201, 168, 201, 169, 201, 170, 201, 188, 201, 189, 201, 190, 201, 191, 201, 192, 201, 193, 201, 195, 201, 196, 201, 197, 201, 198, 201, 199, 201, 200, 201, 201, 201, 202, 201, 203, 201, 204, 201, 205, 201, 206, 201, 207, 201, 208, 201, 209, 201, 210, 201, 211, 201, 212, 201, 220, 201, 221, 201, 222, 201, 223, 201, 225, 201, 226, 201],
	#[65535, 48, 30, 48],
	#[65535, 25, 3, 25, 7, 25, 11, 25, 31, 25, 33, 25, 35, 25, 36, 25, 44, 25, 45, 25, 46, 25, 47, 25, 48, 25, 49, 25, 50, 25, 51, 25, 52, 25, 53, 25, 54, 25, 55, 25, 56, 25, 57, 25, 58, 25, 59, 25, 60, 25, 63, 25, 64, 25, 65, 25, 66, 25, 67, 25, 68, 25, 69, 25, 70, 25, 71, 25, 72, 25, 73, 25, 74, 25, 75, 25, 76, 25, 77, 25, 81, 25, 82, 25, 83, 25, 95, 25, 97, 25, 118, 25],
	#[55, 19, 21, -10, 54, 19, 18, -15, 31, 19, 14, -9, 63, 19, 7, 19, 36, 19, 20, -23, 22, -31, 46, 19, 71, 19, 15, -30, 53, 19, 44, 19, 47, 19, 17, -29, 24, -3, 23, -16, 68, 19, 52, 19, 59, 19, 25, -26, 16, -14, 69, 19, 74, 19, 67, 19, 70, 19, 49, 19, 19, -38, 65, 19, 75, 19, 51, 19, 57, 19, 45, 19, 66, 19, 58, 19, 35, 19, 11, 19, 73, 19, 26, -28, 33, 19, 50, 19, 76, 19, 56, 19, 48, 19, 64, 19, 118, 19, 60, 19],
	#[65535, 49, 4, 49, 1, 49, 7, 49, 19, 49, 9, 49, 26, 49, 0, 49, 15, 49, 10, 49, 17, 49, 23, 49, 14, 49, 16, 49, 18, 49, 24, 49, 20, 49, 21, 49, 22, 49, 25, 49, 27, 49, 28, 49, 29, 49, 40, 49, 41, 49, 119, 49, 63, 49, 83, 49, 82, 49, 74, 49, 75, 49, 76, 49, 77, 49, 118, 49, 225, 49, 226, 49],
	#[31, 113, 46, 113, 55, 113, 11, 113, 7, 113, 54, 113, 33, 113, 47, 113, 36, 113, 59, 113, 35, 113, 50, 113, 51, 113, 60, 113, 44, 113, 45, 113, 58, 113, 48, 113, 49, 113, 57, 113, 65, 113, 52, 113, 53, 113, 66, 113, 56, 113, 63, 113, 64, 113, 67, 113, 68, 113, 69, 113, 70, 113, 71, 113, 73, 113, 74, -754, 75, -756, 76, -753, 118, 113],
	#[31, -45, 61, -44],
	#[65535, 51, 7, 51, 15, 51, 14, 51, 0, 51, 1, 51, 4, 51, 19, 51, 18, 51, 9, 51, 10, 51, 16, 51, 17, 51, 20, 51, 21, 51, 22, 51, 23, 51, 24, 51, 25, 51, 26, 51, 27, 51, 28, 51, 29, 51, 40, 51, 41, 51, 63, 51, 74, 51, 75, 51, 76, 51, 77, 51, 82, 51, 83, 51, 118, 51, 119, 51, 225, 51, 226, 51],
	#[63, 274, 100, 274, 110, 274, 106, 274, 127, 274, 108, 274, 118, 274, 103, 274, 114, 274, 116, 274, 102, 274, 126, 274, 111, 274, 101, 274, 99, 274, 124, 274, 112, 274, 129, 274, 109, 274, 7, 274, 95, 274, 123, 274, 133, 274, 115, 274, 141, 274, 104, 274, 97, 274, 131, 274, 137, 274, 105, 274, 139, 274, 113, 274, 130, 274, 121, 274, 107, 274, 188, 274, 122, 274, 117, 274, 125, 274, 138, 274, 132, 274, 136, 274, 140, 274, 162, -47, 163, 274, 164, 274, 165, 274, 166, 274, 167, 274, 169, 274, 170, 274, 189, 274, 190, 274, 192, 274, 193, 274, 195, 274, 196, 274, 197, 274, 198, 274, 199, 274, 200, 274, 204, 274, 205, 274],
	#[111, 274, 7, 274, 95, 274, 108, 274, 99, 274, 102, 274, 113, 274, 101, 274, 104, 274, 114, 274, 110, 274, 63, 274, 97, 274, 122, 274, 118, 274, 100, 274, 127, 274, 106, 274, 126, 274, 129, 274, 164, 274, 116, 274, 103, 274, 188, 274, 123, 274, 105, 274, 124, 274, 137, 274, 112, 274, 132, 274, 130, 274, 109, 274, 121, 274, 133, 274, 115, 274, 136, 274, 107, 274, 163, 274, 117, 274, 131, 274, 125, 274, 170, 274, 165, 274, 138, 274, 139, 274, 140, 274, 141, 274, 166, 274, 162, -47, 167, 274, 169, 274, 189, 274, 190, 274, 192, 274, 193, 274, 195, 274, 196, 274, 197, 274, 198, 274, 199, 274, 200, 274, 204, 274, 205, 274],
	#[65535, 81, 7, 81, 32, 81, 62, 81, 63, 81, 95, 81, 97, 81, 99, 81, 100, 81, 101, 81, 102, 81, 103, 81, 104, 81, 105, 81, 106, 81, 107, 81, 108, 81, 109, 81, 110, 81, 111, 81, 112, 81, 113, 81, 114, 81, 115, 81, 116, 81, 117, 81, 118, 81, 121, 81, 122, 81, 123, 81, 124, 81, 125, 81, 126, 81, 127, 81, 129, 81, 130, 81, 131, 81, 132, 81, 133, 81, 136, 81, 137, 81, 138, 81, 139, 81, 140, 81, 141, 81, 162, 81, 163, 81, 164, 81, 165, 81, 166, 81, 167, 81, 169, 81, 170, 81, 188, 81, 189, 81, 190, 81, 192, 81, 193, 81, 195, 81, 196, 81, 197, 81, 198, 81, 199, 81, 200, 81, 204, 81, 205, 81],
	#[65535, 275, 118, 275, 102, 275, 110, 275, 95, 275, 108, 275, 122, 275, 103, 275, 100, 275, 126, 275, 111, 275, 101, 275, 99, 275, 114, 275, 115, 275, 63, 275, 7, 275, 116, 275, 121, 275, 127, 275, 106, 275, 129, 275, 164, 275, 105, 275, 107, 275, 188, 275, 123, 275, 124, 275, 137, 275, 132, 275, 130, 275, 109, 275, 104, 275, 112, 275, 133, 275, 113, 275, 136, 275, 97, 275, 163, 275, 117, 275, 131, 275, 125, 275, 170, 275, 165, 275, 138, 275, 139, 275, 140, 275, 141, 275, 166, 275, 167, 275, 169, 275, 189, 275, 190, 275, 192, 275, 193, 275, 195, 275, 196, 275, 197, 275, 198, 275, 199, 275, 200, 275, 204, 275, 205, 275],
	#[7, -13, 63, -8, 95, 199, 97, 199, 99, 199, 100, 199, 101, 199, 102, 199, 103, 199, 104, 199, 105, 199, 106, 199, 107, 199, 108, 199, 109, 199, 110, 199, 111, 199, 112, 199, 113, 199, 114, 199, 115, 199, 116, 199, 117, 199, 118, -36, 121, 199, 122, 199, 123, 199, 124, 199, 125, 199, 126, 199, 127, 199, 129, 199, 130, 199, 131, 199, 132, 199, 133, 199, 136, 199, 137, 199, 138, 199, 139, 199, 140, 199, 141, 199, 163, 199, 164, 199, 165, 199, 166, 199, 167, 199, 169, 199, 170, 199, 188, 199, 189, 199, 190, 199, 192, 199, 193, 199, 195, 199, 196, 199, 197, 199, 198, 199, 199, 199, 200, 199, 204, 199, 205, 199],
	#[63, 274, 7, 274, 95, 274, 127, 274, 100, 274, 103, 274, 110, 274, 99, 274, 32, -51, 111, 274, 101, 274, 107, 274, 141, 274, 104, 274, 113, 274, 102, 274, 109, 274, 115, 274, 108, 274, 97, 274, 105, 274, 133, 274, 121, 274, 129, 274, 112, 274, 123, 274, 116, 274, 106, 274, 136, 274, 118, 274, 122, 274, 124, 274, 114, 274, 170, 274, 138, 274, 137, 274, 126, 274, 164, 274, 139, 274, 132, 274, 130, 274, 188, 274, 131, 274, 163, 274, 117, 274, 125, 274, 165, 274, 140, 274, 166, 274, 162, -47, 167, 274, 169, 274, 189, 274, 190, 274, 192, 274, 193, 274, 195, 274, 196, 274, 197, 274, 198, 274, 199, 274, 200, 274, 204, 274, 205, 274],
	#[65535, 82, 102, 82, 100, 82, 107, 82, 111, 82, 32, 82, 101, 82, 99, 82, 62, 82, 110, 82, 63, 82, 164, 82, 7, 82, 95, 82, 118, 82, 133, 82, 127, 82, 103, 82, 126, 82, 112, 82, 105, 82, 106, 82, 116, 82, 115, 82, 131, 82, 108, 82, 113, 82, 104, 82, 129, 82, 132, 82, 114, 82, 97, 82, 121, 82, 138, 82, 139, 82, 130, 82, 109, 82, 123, 82, 122, 82, 117, 82, 136, 82, 141, 82, 137, 82, 124, 82, 125, 82, 170, 82, 165, 82, 140, 82, 166, 82, 163, 82, 162, 82, 167, 82, 169, 82, 188, 82, 189, 82, 190, 82, 192, 82, 193, 82, 195, 82, 196, 82, 197, 82, 198, 82, 199, 82, 200, 82, 204, 82, 205, 82],
	#[65535, 80, 4, 80, 1, 80, 10, 80, 7, 80, 19, 80, 26, 80, 0, 80, 15, 80, 9, 80, 17, 80, 23, 80, 14, 80, 16, 80, 18, 80, 24, 80, 20, 80, 21, 80, 22, 80, 25, 80, 27, 80, 28, 80, 29, 80, 40, 80, 41, 80, 119, 80, 63, 80, 83, 80, 82, 80, 74, 80, 75, 80, 76, 80, 77, 80, 118, 80, 225, 80, 226, 80],
	#[101, -109, 102, -78, 95, -100, 107, -118, 110, -65, 103, -106, 106, -69, 111, -56, 114, -72, 123, -105, 99, -92, 104, -79, 97, -66, 113, -94, 124, -76, 108, -89, 112, -97, 116, -117, 115, -70, 100, -71, 105, -95, 109, -80, 121, -58, 122, -107, 117, -119, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 188, -91, 163, -73, 164, -103, 165, -108, 166, -90, 167, -99, 169, -98, 170, -82, 192, -85, 189, -113, 190, -96, 193, -101, 195, -112, 196, -86, 197, -61, 198, -84, 199, -114, 200, -83, 204, -60, 205, -63],
	#[65535, 276, 7, 276, 63, 276, 101, 276, 95, 276, 97, 276, 99, 276, 100, 276, 102, 276, 103, 276, 104, 276, 105, 276, 106, 276, 107, 276, 108, 276, 109, 276, 110, 276, 111, 276, 112, 276, 113, 276, 114, 276, 115, 276, 116, 276, 117, 276, 118, 276, 121, 276, 122, 276, 123, 276, 124, 276, 125, 276, 126, 276, 127, 276, 129, 276, 130, 276, 131, 276, 132, 276, 133, 276, 136, 276, 137, 276, 138, 276, 139, 276, 140, 276, 141, 276, 163, 276, 164, 276, 165, 276, 166, 276, 167, 276, 169, 276, 170, 276, 188, 276, 189, 276, 190, 276, 192, 276, 193, 276, 195, 276, 196, 276, 197, 276, 198, 276, 199, 276, 200, 276, 204, 276, 205, 276],
	#[6, -747],
	#[101, -109, 111, -56, 123, -105, 99, -92, 116, -117, 114, -72, 95, -100, 127, -102, 124, -76, 108, -89, 97, -66, 103, -106, 106, -69, 102, -78, 126, -115, 113, -94, 188, -91, 122, -107, 104, -79, 110, -65, 129, -81, 130, -57, 100, -71, 164, -103, 105, -95, 137, -59, 112, -97, 132, -93, 109, -80, 121, -58, 133, -88, 115, -70, 136, -87, 107, -118, 163, -73, 117, -119, 131, -110, 125, -62, 170, -82, 165, -108, 138, -104, 139, -77, 140, -64, 141, -120, 166, -90, 167, -99, 169, -98, 189, -113, 190, -96, 192, -85, 193, -101, 195, -112, 196, -86, 197, -61, 198, -84, 199, -114, 200, -83, 204, -60, 205, -63],
	#[65535, 185, 7, 185, 11, 185, 31, 185, 33, 185, 35, 185, 36, 185, 42, 185, 63, 185, 64, 185, 65, 185, 66, 185, 67, 185, 68, 185, 69, 185, 70, 185, 71, 185, 73, 185, 118, 185],
	#[65535, 231, 31, 231, 63, 231, 71, 231, 68, 231, 35, 231, 7, 231, 118, 231, 11, 231, 66, 231, 69, 231, 33, 231, 70, 231, 36, 231, 65, 231, 67, 231, 64, 231, 42, 231, 73, 231],
	#[7, 235, 11, 235, 31, 235, 33, 235, 35, 235, 36, 235, 42, 235, 63, 235, 64, 235, 65, 235, 66, 235, 67, 235, 68, 235, 69, 235, 70, 235, 71, 235, 73, 235, 118, 235, 134, -332, 135, -333],
	#[7, 241, 11, 241, 31, 241, 33, 241, 35, 241, 36, 241, 42, 241, 65, 241, 66, 241, 63, 241, 64, 241, 67, 241, 68, 241, 69, 241, 70, 241, 71, 241, 73, 241, 118, 241, 128, -417],
	#[205, -746],
	#[7, 419, 11, 419, 31, 419, 33, 419, 35, 419, 36, 419, 63, 419, 64, 419, 65, 419, 66, 419, 67, 419, 68, 419, 69, 419, 70, 419, 71, 419, 73, 419, 118, 419, 195, -545],
	#[7, 235, 11, 235, 31, 235, 33, 235, 35, 235, 36, 235, 42, 235, 63, 235, 64, 235, 65, 235, 66, 235, 67, 235, 68, 235, 69, 235, 70, 235, 71, 235, 73, 235, 118, 235, 134, -332, 135, -333],
	#[65535, 360, 7, 360, 47, 360, 46, 360, 31, 360, 54, 360, 36, 360, 63, 360, 53, 360, 51, 360, 44, 360, 70, 360, 52, 360, 118, 360, 69, 360, 58, 360, 55, 360, 50, 360, 60, 360, 59, 360, 48, 360, 35, 360, 11, 360, 73, 360, 33, 360, 67, 360, 56, 360, 65, 360, 68, 360, 66, 360, 45, 360, 49, 360, 64, 360, 71, 360, 57, 360],
	#[65535, 246, 7, 246, 11, 246, 31, 246, 33, 246, 35, 246, 36, 246, 42, 246, 63, 246, 64, 246, 65, 246, 66, 246, 67, 246, 68, 246, 69, 246, 70, 246, 71, 246, 73, 246, 118, 246],
	#[65535, 184, 31, 184, 7, 184, 11, 184, 42, 184, 33, 184, 63, 184, 35, 184, 36, 184, 71, 184, 70, 184, 69, 184, 68, 184, 67, 184, 66, 184, 65, 184, 73, 184, 64, 184, 118, 184],
	#[31, 421, 63, 421, 7, 421, 11, 421, 70, 421, 69, 421, 33, 421, 68, 421, 35, 421, 36, 421, 67, 421, 71, 421, 66, 421, 65, 421, 73, 421, 64, 421, 118, 421, 224, -362],
	#[7, 59, 11, 59, 31, 59, 33, 59, 35, 59, 36, 59, 44, -558, 45, -561, 46, -563, 47, -565, 48, -567, 49, -555, 50, -559, 51, -560, 52, -562, 53, -564, 54, -566, 55, -554, 56, -556, 57, 59, 58, 59, 59, 59, 60, 59, 63, 59, 64, 59, 65, 59, 66, 59, 67, 59, 68, 59, 69, 59, 70, 59, 71, 59, 73, 59, 118, 59],
	#[31, -128, 63, -121, 7, -13, 11, -127, 70, -140, 69, -136, 33, -139, 68, -133, 35, -123, 36, -129, 67, -137, 71, -130, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[31, -128, 63, -121, 7, -13, 11, -127, 70, -140, 69, -136, 33, -139, 68, -133, 35, -123, 36, -129, 67, -137, 71, -130, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[65535, 190, 7, 190, 11, 190, 31, 190, 33, 190, 35, 190, 36, 190, 42, 190, 63, 190, 64, 190, 65, 190, 66, 190, 67, 190, 68, 190, 69, 190, 70, 190, 71, 190, 73, 190, 118, 190],
	#[142, -380, 143, -382, 144, -384, 145, -375, 146, -376, 147, -377, 148, -379, 149, -381, 150, -383, 151, -374],
	#[65535, 189, 7, 189, 11, 189, 31, 189, 33, 189, 35, 189, 36, 189, 42, 189, 63, 189, 64, 189, 65, 189, 66, 189, 67, 189, 68, 189, 69, 189, 70, 189, 71, 189, 73, 189, 118, 189],
	#[63, -121, 7, -13, 11, -127, 71, -130, 66, -126, 65, -122, 70, -140, 33, -139, 67, -137, 31, -128, 35, -123, 36, -129, 64, -141, 73, -702, 68, -133, 69, -136, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 277, 7, 277, 63, 277, 95, 277, 97, 277, 99, 277, 100, 277, 101, 277, 102, 277, 103, 277, 104, 277, 105, 277, 106, 277, 107, 277, 108, 277, 109, 277, 110, 277, 111, 277, 112, 277, 113, 277, 114, 277, 115, 277, 116, 277, 117, 277, 118, 277, 121, 277, 122, 277, 123, 277, 124, 277, 125, 277, 126, 277, 127, 277, 129, 277, 130, 277, 131, 277, 132, 277, 133, 277, 136, 277, 137, 277, 138, 277, 139, 277, 140, 277, 141, 277, 163, 277, 164, 277, 165, 277, 166, 277, 167, 277, 169, 277, 170, 277, 188, 277, 189, 277, 190, 277, 192, 277, 193, 277, 195, 277, 196, 277, 197, 277, 198, 277, 199, 277, 200, 277, 204, 277, 205, 277],
	#[65535, 224, 31, 224, 63, 224, 36, 224, 71, 224, 11, 224, 7, 224, 118, 224, 68, 224, 42, 224, 66, 224, 33, 224, 69, 224, 35, 224, 70, 224, 64, 224, 73, 224, 67, 224, 65, 224],
	#[65535, 245, 7, 245, 11, 245, 31, 245, 33, 245, 35, 245, 36, 245, 42, 245, 63, 245, 64, 245, 65, 245, 66, 245, 67, 245, 68, 245, 69, 245, 70, 245, 71, 245, 73, 245, 118, 245],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[31, -128, 7, -13, 36, -129, 70, -140, 35, -123, 11, -127, 63, -121, 64, -141, 71, -130, 66, -126, 65, -122, 33, -139, 67, -137, 73, -138, 68, -133, 69, -136, 118, -36],
	#[65535, 183, 7, 183, 11, 183, 31, 183, 33, 183, 35, 183, 36, 183, 42, 183, 63, 183, 64, 183, 65, 183, 66, 183, 67, 183, 68, 183, 69, 183, 70, 183, 71, 183, 73, 183, 118, 183],
	#[36, 229, 31, 229, 7, 229, 35, 229, 33, 229, 11, 229, 67, 229, 42, 229, 63, 229, 64, 229, 65, 229, 66, 229, 68, 229, 69, 229, 70, 229, 71, 229, 73, 229, 118, 229, 128, -360],
	#[7, -13, 11, -127, 31, -128, 33, -139, 63, -121, 35, -123, 36, -129, 71, -130, 70, -140, 69, -136, 68, -133, 67, -137, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[121, 419, 123, 419, 139, 419, 140, 419, 141, 419, 195, -545, 213, 419, 214, 419, 215, 419, 216, 419, 217, 419, 218, 419],
	#[70, -140, 69, -136, 31, -128, 63, -121, 7, -13, 36, -129, 71, -130, 35, -123, 67, -137, 68, -133, 66, -126, 11, -127, 64, -141, 33, -139, 73, -138, 65, -122, 118, -36],
	#[7, 419, 11, 419, 31, 419, 33, 419, 35, 419, 36, 419, 63, 419, 64, 419, 65, 419, 66, 419, 67, 419, 68, 419, 69, 419, 70, 419, 71, 419, 73, 419, 118, 419, 194, -641, 195, -545],
	#[7, 235, 135, -333, 31, 235, 11, 235, 42, 235, 33, 235, 63, 235, 35, 235, 36, 235, 71, 235, 70, 235, 69, 235, 68, 235, 67, 235, 66, 235, 65, 235, 73, 235, 64, 235, 118, 235, 134, -332],
	#[65535, 234, 7, 234, 11, 234, 31, 234, 33, 234, 35, 234, 36, 234, 42, 234, 63, 234, 64, 234, 65, 234, 66, 234, 67, 234, 68, 234, 69, 234, 70, 234, 71, 234, 73, 234, 118, 234],
	#[65535, 182, 7, 182, 11, 182, 35, 182, 31, 182, 33, 182, 36, 182, 42, 182, 63, 182, 64, 182, 65, 182, 66, 182, 67, 182, 68, 182, 69, 182, 70, 182, 71, 182, 73, 182, 118, 182],
	#[31, -128, 63, -121, 7, -13, 11, -127, 68, -133, 65, -122, 70, -140, 33, -139, 69, -136, 35, -123, 36, -129, 67, -137, 71, -130, 66, -126, 73, -138, 64, -141, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, -13, 35, -123, 11, -127, 31, -128, 33, -139, 36, -129, 65, -122, 67, -137, 66, -126, 63, -121, 64, -141, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 233, 7, 233, 70, 233, 69, 233, 31, 233, 63, 233, 68, 233, 36, 233, 71, 233, 67, 233, 42, 233, 66, 233, 35, 233, 11, 233, 64, 233, 33, 233, 73, 233, 65, 233, 118, 233],
	#[65535, 188, 7, 188, 11, 188, 31, 188, 33, 188, 35, 188, 36, 188, 42, 188, 63, 188, 64, 188, 65, 188, 66, 188, 67, 188, 68, 188, 69, 188, 70, 188, 71, 188, 73, 188, 118, 188],
	#[31, -128, 7, -13, 36, -129, 35, -123, 11, -127, 63, -121, 70, -140, 67, -137, 33, -139, 71, -130, 64, -141, 73, -138, 66, -126, 65, -122, 69, -136, 68, -133, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 65, -122, 66, -126, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 187, 31, 187, 36, 187, 7, 187, 35, 187, 11, 187, 33, 187, 42, 187, 67, 187, 63, 187, 64, 187, 65, 187, 66, 187, 68, 187, 69, 187, 70, 187, 71, 187, 73, 187, 118, 187],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[7, 59, 11, 59, 31, 59, 33, 59, 35, 59, 36, 59, 44, -558, 45, -561, 46, -563, 47, -565, 48, -567, 49, -555, 50, -559, 51, -560, 52, -562, 53, -564, 54, -566, 55, -554, 56, -556, 57, 59, 58, 59, 59, 59, 60, 59, 63, 59, 64, 59, 65, 59, 66, 59, 67, 59, 68, 59, 69, 59, 70, 59, 71, 59, 73, 59, 118, 59],
	#[65535, 186, 31, 186, 7, 186, 71, 186, 42, 186, 36, 186, 35, 186, 11, 186, 63, 186, 70, 186, 67, 186, 33, 186, 64, 186, 73, 186, 65, 186, 69, 186, 68, 186, 66, 186, 118, 186],
	#[31, 419, 36, 419, 7, 419, 11, 419, 68, 419, 35, 419, 33, 419, 63, 419, 64, 419, 65, 419, 66, 419, 67, 419, 69, 419, 70, 419, 71, 419, 73, 419, 118, 419, 194, -544, 195, -545],
	#[7, 227, 11, 227, 31, 227, 33, 227, 35, 227, 36, 227, 42, 227, 63, 227, 64, 227, 65, 227, 66, 227, 67, 227, 68, 227, 69, 227, 70, 227, 71, 227, 73, 227, 118, 227, 128, -337],
	#[36, -532, 69, -533],
	#[7, 243, 11, 243, 31, 243, 33, 243, 35, 243, 36, 243, 42, 243, 63, 243, 64, 243, 65, 243, 66, 243, 67, 243, 68, 243, 69, 243, 70, 243, 71, 243, 73, 243, 118, 243, 128, -336],
	#[31, 235, 135, -333, 70, 235, 68, 235, 7, 235, 67, 235, 71, 235, 64, 235, 33, 235, 42, 235, 11, 235, 63, 235, 35, 235, 65, 235, 73, 235, 36, 235, 66, 235, 69, 235, 134, -332, 118, 235],
	#[31, -128, 7, -13, 11, -127, 33, -139, 36, -129, 35, -123, 65, -122, 66, -126, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 222, 7, 222, 11, 222, 31, 222, 33, 222, 35, 222, 36, 222, 42, 222, 63, 222, 64, 222, 65, 222, 66, 222, 67, 222, 68, 222, 69, 222, 70, 222, 71, 222, 73, 222, 118, 222],
	#[36, -505],
	#[90, -301, 91, -299, 148, -296, 149, -300, 150, -303, 151, -288, 152, -290, 153, -293, 154, -297, 155, -291, 156, -302, 157, -304, 158, -295, 159, -292, 160, -294, 161, -298],
	#[65535, 232, 7, 232, 11, 232, 31, 232, 33, 232, 35, 232, 36, 232, 42, 232, 63, 232, 64, 232, 65, 232, 66, 232, 67, 232, 68, 232, 69, 232, 70, 232, 71, 232, 73, 232, 118, 232],
	#[65535, 197, 7, 197, 32, 197, 62, 197, 63, 197, 101, 197, 95, 197, 97, 197, 99, 197, 100, 197, 102, 197, 103, 197, 104, 197, 105, 197, 106, 197, 107, 197, 108, 197, 109, 197, 110, 197, 111, 197, 112, 197, 113, 197, 114, 197, 115, 197, 116, 197, 117, 197, 118, 197, 121, 197, 122, 197, 123, 197, 124, 197, 125, 197, 126, 197, 127, 197, 129, 197, 130, 197, 131, 197, 132, 197, 133, 197, 136, 197, 137, 197, 138, 197, 139, 197, 140, 197, 141, 197, 162, 197, 163, 197, 164, 197, 165, 197, 166, 197, 167, 197, 169, 197, 170, 197, 188, 197, 189, 197, 190, 197, 192, 197, 193, 197, 195, 197, 196, 197, 197, 197, 198, 197, 199, 197, 200, 197, 204, 197, 205, 197],
	#[193, -486, 196, -487],
	#[31, -128, 36, -129, 7, -13, 11, -127, 33, -139, 35, -123, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[209, 361, 208, 361, 206, -462, 207, 361, 210, 361, 211, 361, 212, 361],
	#[65535, 226, 7, 226, 11, 226, 31, 226, 33, 226, 35, 226, 36, 226, 42, 226, 63, 226, 64, 226, 65, 226, 66, 226, 67, 226, 68, 226, 69, 226, 70, 226, 71, 226, 73, 226, 118, 226],
	#[31, -128, 7, -13, 11, -127, 33, -139, 36, -129, 35, -123, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 191, 31, 191, 7, 191, 11, 191, 33, 191, 36, 191, 35, 191, 42, 191, 63, 191, 64, 191, 65, 191, 66, 191, 67, 191, 68, 191, 69, 191, 70, 191, 71, 191, 73, 191, 118, 191],
	#[65535, 181, 7, 181, 11, 181, 31, 181, 33, 181, 35, 181, 36, 181, 42, 181, 63, 181, 64, 181, 65, 181, 66, 181, 67, 181, 68, 181, 69, 181, 70, 181, 71, 181, 73, 181, 118, 181],
	#[65535, 192, 31, 192, 11, 192, 7, 192, 36, 192, 33, 192, 42, 192, 35, 192, 65, 192, 66, 192, 63, 192, 64, 192, 67, 192, 68, 192, 69, 192, 70, 192, 71, 192, 73, 192, 118, 192],
	#[65535, 247, 7, 247, 11, 247, 31, 247, 33, 247, 35, 247, 36, 247, 42, 247, 63, 247, 64, 247, 65, 247, 66, 247, 67, 247, 68, 247, 69, 247, 70, 247, 71, 247, 73, 247, 118, 247],
	#[65535, 89, 31, 89, 1, 89, 10, 89, 14, 89, 0, 89, 87, 89, 11, 89, 13, 89, 22, 89, 7, 89, 2, 89, 9, 89, 18, 89, 15, 89, 4, 89, 27, 89, 17, 89, 29, 89, 23, 89, 12, 89, 33, 89, 91, 89, 20, 89, 26, 89, 63, 89, 19, 89, 59, 89, 95, 89, 24, 89, 75, 89, 25, 89, 34, 89, 58, 89, 40, 89, 16, 89, 43, 89, 37, 89, 74, 89, 77, 89, 21, 89, 32, 89, 41, 89, 28, 89, 38, 89, 86, 89, 110, 89, 42, 89, 60, 89, 103, 89, 85, 89, 116, 89, 82, 89, 57, 89, 109, 89, 111, 89, 89, 89, 93, 89, 102, 89, 83, 89, 88, 89, 76, 89, 99, 89, 119, 89, 106, 89, 101, 89, 113, 89, 100, 89, 84, 89, 90, 89, 104, 89, 105, 89, 123, 89, 72, 89, 124, 89, 108, 89, 92, 89, 97, 89, 112, 89, 121, 89, 107, 89, 115, 89, 114, 89, 122, 89, 117, 89, 118, 89, 125, 89, 126, 89, 127, 89, 129, 89, 130, 89, 131, 89, 132, 89, 133, 89, 136, 89, 137, 89, 138, 89, 139, 89, 140, 89, 141, 89, 188, 89, 163, 89, 164, 89, 165, 89, 166, 89, 167, 89, 169, 89, 170, 89, 189, 89, 190, 89, 191, 89, 192, 89, 193, 89, 195, 89, 196, 89, 197, 89, 198, 89, 199, 89, 200, 89, 204, 89, 205, 89, 220, 89, 221, 89, 222, 89, 223, 89, 225, 89, 226, 89],
	#[65535, 98, 4, 98, 14, 98, 12, 98, 15, 98, 22, 98, 23, 98, 13, 98, 17, 98, 19, 98, 31, 98, 21, 98, 9, 98, 63, 98, 7, 98, 0, 98, 1, 98, 29, 98, 42, 98, 10, 98, 92, 98, 34, 98, 2, 98, 11, 98, 93, 98, 37, 98, 25, 98, 77, 98, 107, 98, 84, 98, 16, 98, 32, 98, 38, 98, 33, 98, 87, 98, 72, 98, 18, 98, 60, 98, 43, 98, 95, 98, 57, 98, 59, 98, 99, 98, 82, 98, 75, 98, 40, 98, 20, 98, 103, 98, 27, 98, 24, 98, 76, 98, 83, 98, 26, 98, 86, 98, 111, 98, 100, 98, 91, 98, 113, 98, 90, 98, 119, 98, 89, 98, 41, 98, 106, 98, 85, 98, 102, 98, 28, 98, 112, 98, 97, 98, 58, 98, 114, 98, 110, 98, 104, 98, 105, 98, 124, 98, 108, 98, 101, 98, 74, 98, 116, 98, 123, 98, 109, 98, 115, 98, 88, 98, 121, 98, 122, 98, 117, 98, 118, 98, 125, 98, 126, 98, 127, 98, 129, 98, 130, 98, 131, 98, 132, 98, 133, 98, 136, 98, 137, 98, 138, 98, 139, 98, 140, 98, 141, 98, 188, 98, 163, 98, 164, 98, 165, 98, 166, 98, 167, 98, 169, 98, 170, 98, 192, 98, 189, 98, 190, 98, 191, 98, 193, 98, 195, 98, 196, 98, 197, 98, 198, 98, 199, 98, 200, 98, 204, 98, 205, 98, 220, 98, 221, 98, 222, 98, 223, 98, 225, 98, 226, 98],
	#[65535, 85, 0, 85, 1, 85, 2, 85, 4, 85, 7, 85, 9, 85, 10, 85, 11, 85, 12, 85, 13, 85, 14, 85, 15, 85, 16, 85, 17, 85, 18, 85, 19, 85, 20, 85, 21, 85, 22, 85, 23, 85, 24, 85, 25, 85, 26, 85, 27, 85, 28, 85, 29, 85, 31, 85, 32, 85, 33, 85, 34, 85, 37, 85, 38, 85, 40, 85, 41, 85, 42, 85, 43, 85, 57, 85, 58, 85, 59, 85, 60, 85, 63, 85, 72, 85, 74, 85, 75, 85, 76, 85, 77, 85, 82, 85, 83, 85, 84, 85, 85, 85, 86, 85, 87, 85, 88, 85, 89, 85, 90, 85, 91, 85, 92, 85, 93, 85, 95, 85, 97, 85, 99, 85, 100, 85, 101, 85, 102, 85, 103, 85, 104, 85, 105, 85, 106, 85, 107, 85, 108, 85, 109, 85, 110, 85, 111, 85, 112, 85, 113, 85, 114, 85, 115, 85, 116, 85, 117, 85, 118, 85, 119, 85, 121, 85, 122, 85, 123, 85, 124, 85, 125, 85, 126, 85, 127, 85, 129, 85, 130, 85, 131, 85, 132, 85, 133, 85, 136, 85, 137, 85, 138, 85, 139, 85, 140, 85, 141, 85, 163, 85, 164, 85, 165, 85, 166, 85, 167, 85, 169, 85, 170, 85, 188, 85, 189, 85, 190, 85, 191, 85, 192, 85, 193, 85, 195, 85, 196, 85, 197, 85, 198, 85, 199, 85, 200, 85, 204, 85, 205, 85, 220, 85, 221, 85, 222, 85, 223, 85, 225, 85, 226, 85],
	#[65535, 83, 14, 83, 7, 83, 15, 83, 4, 83, 10, 83, 17, 83, 2, 83, 23, 83, 0, 83, 11, 83, 87, 83, 1, 83, 31, 83, 20, 83, 16, 83, 18, 83, 9, 83, 29, 83, 28, 83, 12, 83, 13, 83, 27, 83, 26, 83, 84, 83, 43, 83, 37, 83, 19, 83, 77, 83, 21, 83, 22, 83, 34, 83, 24, 83, 25, 83, 33, 83, 41, 83, 85, 83, 74, 83, 42, 83, 32, 83, 102, 83, 63, 83, 40, 83, 75, 83, 95, 83, 38, 83, 91, 83, 86, 83, 92, 83, 72, 83, 88, 83, 83, 83, 103, 83, 90, 83, 89, 83, 97, 83, 105, 83, 76, 83, 82, 83, 106, 83, 104, 83, 99, 83, 107, 83, 100, 83, 101, 83, 108, 83, 109, 83, 110, 83, 111, 83, 112, 83, 113, 83, 114, 83, 115, 83, 116, 83, 117, 83, 118, 83, 119, 83, 121, 83, 122, 83, 123, 83, 124, 83, 125, 83, 126, 83, 127, 83, 129, 83, 130, 83, 131, 83, 132, 83, 133, 83, 136, 83, 137, 83, 138, 83, 139, 83, 140, 83, 141, 83, 163, 83, 164, 83, 165, 83, 166, 83, 167, 83, 169, 83, 170, 83, 192, 83, 188, 83, 189, 83, 190, 83, 191, 83, 193, 83, 195, 83, 196, 83, 197, 83, 198, 83, 199, 83, 200, 83, 204, 83, 205, 83, 225, 83, 226, 83],
	#[65535, 84, 31, 84, 23, 84, 10, 84, 14, 84, 7, 84, 34, 84, 37, 84, 9, 84, 15, 84, 1, 84, 27, 84, 2, 84, 12, 84, 16, 84, 11, 84, 0, 84, 4, 84, 19, 84, 18, 84, 24, 84, 29, 84, 25, 84, 33, 84, 13, 84, 26, 84, 17, 84, 20, 84, 21, 84, 22, 84, 32, 84, 28, 84, 38, 84, 40, 84, 41, 84, 42, 84, 43, 84, 63, 84, 84, 84, 82, 84, 72, 84, 74, 84, 75, 84, 76, 84, 77, 84, 101, 84, 83, 84, 85, 84, 86, 84, 87, 84, 88, 84, 89, 84, 90, 84, 91, 84, 92, 84, 95, 84, 97, 84, 99, 84, 100, 84, 102, 84, 103, 84, 104, 84, 105, 84, 106, 84, 107, 84, 108, 84, 109, 84, 110, 84, 111, 84, 112, 84, 113, 84, 114, 84, 115, 84, 116, 84, 117, 84, 118, 84, 119, 84, 121, 84, 122, 84, 123, 84, 124, 84, 125, 84, 126, 84, 127, 84, 129, 84, 130, 84, 131, 84, 132, 84, 133, 84, 136, 84, 137, 84, 138, 84, 139, 84, 140, 84, 141, 84, 163, 84, 164, 84, 165, 84, 166, 84, 167, 84, 169, 84, 170, 84, 188, 84, 189, 84, 190, 84, 191, 84, 192, 84, 193, 84, 195, 84, 196, 84, 197, 84, 198, 84, 199, 84, 200, 84, 204, 84, 205, 84, 225, 84, 226, 84],
	#[65535, 99, 0, 99, 1, 99, 2, 99, 4, 99, 7, 99, 9, 99, 10, 99, 11, 99, 12, 99, 13, 99, 14, 99, 15, 99, 16, 99, 17, 99, 18, 99, 19, 99, 20, 99, 21, 99, 22, 99, 23, 99, 24, 99, 25, 99, 26, 99, 27, 99, 28, 99, 29, 99, 31, 99, 32, 99, 33, 99, 34, 99, 37, 99, 38, 99, 40, 99, 41, 99, 42, 99, 43, 99, 57, 99, 58, 99, 59, 99, 60, 99, 63, 99, 72, 99, 74, 99, 75, 99, 76, 99, 77, 99, 82, 99, 83, 99, 84, 99, 85, 99, 86, 99, 87, 99, 88, 99, 89, 99, 90, 99, 91, 99, 92, 99, 93, 99, 95, 99, 97, 99, 99, 99, 100, 99, 101, 99, 102, 99, 103, 99, 104, 99, 105, 99, 106, 99, 107, 99, 108, 99, 109, 99, 110, 99, 111, 99, 112, 99, 113, 99, 114, 99, 115, 99, 116, 99, 117, 99, 118, 99, 119, 99, 121, 99, 122, 99, 123, 99, 124, 99, 125, 99, 126, 99, 127, 99, 129, 99, 130, 99, 131, 99, 132, 99, 133, 99, 136, 99, 137, 99, 138, 99, 139, 99, 140, 99, 141, 99, 163, 99, 164, 99, 165, 99, 166, 99, 167, 99, 169, 99, 170, 99, 188, 99, 189, 99, 190, 99, 191, 99, 192, 99, 193, 99, 195, 99, 196, 99, 197, 99, 198, 99, 199, 99, 200, 99, 204, 99, 205, 99, 220, 99, 221, 99, 222, 99, 223, 99, 225, 99, 226, 99],
	#[38, -200],
	#[7, -13, 11, -127, 31, -128, 32, -198, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 86, 7, 86, 12, 86, 10, 86, 37, 86, 31, 86, 14, 86, 9, 86, 63, 86, 26, 86, 0, 86, 1, 86, 22, 86, 25, 86, 15, 86, 2, 86, 11, 86, 28, 86, 18, 86, 23, 86, 16, 86, 19, 86, 17, 86, 38, 86, 4, 86, 86, 86, 29, 86, 92, 86, 24, 86, 60, 86, 42, 86, 33, 86, 43, 86, 34, 86, 118, 86, 127, 86, 40, 86, 20, 86, 27, 86, 108, 86, 76, 86, 13, 86, 58, 86, 103, 86, 110, 86, 77, 86, 21, 86, 32, 86, 72, 86, 85, 86, 87, 86, 74, 86, 83, 86, 119, 86, 59, 86, 41, 86, 90, 86, 75, 86, 95, 86, 82, 86, 91, 86, 101, 86, 99, 86, 116, 86, 100, 86, 57, 86, 109, 86, 111, 86, 93, 86, 102, 86, 138, 86, 106, 86, 115, 86, 113, 86, 84, 86, 88, 86, 97, 86, 105, 86, 123, 86, 121, 86, 124, 86, 104, 86, 89, 86, 112, 86, 107, 86, 114, 86, 122, 86, 117, 86, 131, 86, 125, 86, 126, 86, 165, 86, 129, 86, 130, 86, 136, 86, 132, 86, 133, 86, 204, 86, 137, 86, 167, 86, 139, 86, 140, 86, 141, 86, 190, 86, 188, 86, 169, 86, 170, 86, 191, 86, 163, 86, 164, 86, 193, 86, 166, 86, 205, 86, 195, 86, 189, 86, 192, 86, 196, 86, 197, 86, 198, 86, 199, 86, 200, 86, 220, 86, 221, 86, 222, 86, 223, 86, 225, 86, 226, 86],
	#[65535, 104, 31, 104, 23, 104, 14, 104, 29, 104, 7, 104, 15, 104, 2, 104, 1, 104, 9, 104, 0, 104, 4, 104, 18, 104, 10, 104, 11, 104, 12, 104, 13, 104, 27, 104, 26, 104, 16, 104, 17, 104, 37, 104, 19, 104, 20, 104, 21, 104, 22, 104, 34, 104, 24, 104, 25, 104, 33, 104, 32, 104, 28, 104, 38, 104, 40, 104, 41, 104, 42, 104, 43, 104, 57, 104, 58, 104, 59, 104, 60, 104, 63, 104, 84, 104, 72, 104, 74, 104, 75, 104, 76, 104, 77, 104, 102, 104, 82, 104, 83, 104, 85, 104, 86, 104, 87, 104, 88, 104, 89, 104, 90, 104, 91, 104, 92, 104, 93, 104, 95, 104, 97, 104, 99, 104, 100, 104, 101, 104, 103, 104, 104, 104, 105, 104, 106, 104, 107, 104, 108, 104, 109, 104, 110, 104, 111, 104, 112, 104, 113, 104, 114, 104, 115, 104, 116, 104, 117, 104, 118, 104, 119, 104, 121, 104, 122, 104, 123, 104, 124, 104, 125, 104, 126, 104, 127, 104, 129, 104, 130, 104, 131, 104, 132, 104, 133, 104, 136, 104, 137, 104, 138, 104, 139, 104, 140, 104, 141, 104, 163, 104, 164, 104, 165, 104, 166, 104, 167, 104, 169, 104, 170, 104, 188, 104, 189, 104, 190, 104, 191, 104, 192, 104, 193, 104, 195, 104, 196, 104, 197, 104, 198, 104, 199, 104, 200, 104, 204, 104, 205, 104, 220, 104, 221, 104, 222, 104, 223, 104, 225, 104, 226, 104],
	#[96, -452],
	#[65535, 87, 31, 87, 7, 87, 2, 87, 15, 87, 22, 87, 23, 87, 13, 87, 14, 87, 17, 87, 1, 87, 10, 87, 0, 87, 37, 87, 9, 87, 84, 87, 25, 87, 24, 87, 11, 87, 16, 87, 4, 87, 27, 87, 29, 87, 60, 87, 59, 87, 12, 87, 18, 87, 42, 87, 87, 87, 34, 87, 32, 87, 20, 87, 19, 87, 92, 87, 75, 87, 28, 87, 41, 87, 58, 87, 26, 87, 103, 87, 43, 87, 83, 87, 33, 87, 77, 87, 21, 87, 57, 87, 72, 87, 85, 87, 74, 87, 102, 87, 63, 87, 40, 87, 93, 87, 95, 87, 38, 87, 91, 87, 86, 87, 97, 87, 76, 87, 88, 87, 101, 87, 90, 87, 89, 87, 105, 87, 82, 87, 106, 87, 104, 87, 99, 87, 107, 87, 100, 87, 108, 87, 109, 87, 110, 87, 111, 87, 112, 87, 113, 87, 114, 87, 115, 87, 116, 87, 117, 87, 118, 87, 119, 87, 121, 87, 122, 87, 123, 87, 124, 87, 125, 87, 126, 87, 127, 87, 129, 87, 130, 87, 131, 87, 132, 87, 133, 87, 136, 87, 137, 87, 138, 87, 139, 87, 140, 87, 141, 87, 170, 87, 163, 87, 164, 87, 165, 87, 166, 87, 167, 87, 169, 87, 189, 87, 188, 87, 192, 87, 190, 87, 191, 87, 193, 87, 195, 87, 196, 87, 197, 87, 198, 87, 199, 87, 200, 87, 204, 87, 205, 87, 220, 87, 221, 87, 222, 87, 223, 87, 225, 87, 226, 87],
	#[65535, 101, 31, 101, 22, 101, 1, 101, 13, 101, 7, 101, 2, 101, 15, 101, 4, 101, 10, 101, 17, 101, 25, 101, 14, 101, 23, 101, 12, 101, 18, 101, 11, 101, 0, 101, 16, 101, 32, 101, 20, 101, 19, 101, 59, 101, 9, 101, 29, 101, 28, 101, 34, 101, 27, 101, 26, 101, 84, 101, 43, 101, 37, 101, 33, 101, 77, 101, 21, 101, 57, 101, 75, 101, 24, 101, 58, 101, 41, 101, 85, 101, 74, 101, 42, 101, 72, 101, 102, 101, 63, 101, 40, 101, 93, 101, 95, 101, 38, 101, 91, 101, 86, 101, 92, 101, 76, 101, 60, 101, 83, 101, 103, 101, 90, 101, 89, 101, 88, 101, 105, 101, 82, 101, 106, 101, 104, 101, 99, 101, 87, 101, 107, 101, 97, 101, 100, 101, 101, 101, 108, 101, 109, 101, 110, 101, 111, 101, 112, 101, 113, 101, 114, 101, 115, 101, 116, 101, 117, 101, 118, 101, 119, 101, 121, 101, 122, 101, 123, 101, 124, 101, 125, 101, 126, 101, 127, 101, 129, 101, 130, 101, 131, 101, 132, 101, 133, 101, 136, 101, 137, 101, 138, 101, 139, 101, 140, 101, 141, 101, 163, 101, 164, 101, 165, 101, 166, 101, 167, 101, 169, 101, 170, 101, 188, 101, 189, 101, 190, 101, 191, 101, 192, 101, 193, 101, 195, 101, 196, 101, 197, 101, 198, 101, 199, 101, 200, 101, 204, 101, 205, 101, 220, 101, 221, 101, 222, 101, 223, 101, 225, 101, 226, 101],
	#[65535, 90, 0, 90, 1, 90, 2, 90, 4, 90, 7, 90, 9, 90, 10, 90, 11, 90, 12, 90, 13, 90, 14, 90, 15, 90, 16, 90, 17, 90, 18, 90, 19, 90, 20, 90, 21, 90, 22, 90, 23, 90, 24, 90, 25, 90, 26, 90, 27, 90, 28, 90, 29, 90, 31, 90, 32, 90, 33, 90, 34, 90, 37, 90, 38, 90, 40, 90, 41, 90, 42, 90, 43, 90, 57, 90, 58, 90, 59, 90, 60, 90, 63, 90, 72, 90, 74, 90, 75, 90, 76, 90, 77, 90, 82, 90, 83, 90, 84, 90, 85, 90, 86, 90, 87, 90, 88, 90, 89, 90, 90, 90, 91, 90, 92, 90, 93, 90, 95, 90, 97, 90, 99, 90, 100, 90, 101, 90, 102, 90, 103, 90, 104, 90, 105, 90, 106, 90, 107, 90, 108, 90, 109, 90, 110, 90, 111, 90, 112, 90, 113, 90, 114, 90, 115, 90, 116, 90, 117, 90, 118, 90, 119, 90, 121, 90, 122, 90, 123, 90, 124, 90, 125, 90, 126, 90, 127, 90, 129, 90, 130, 90, 131, 90, 132, 90, 133, 90, 136, 90, 137, 90, 138, 90, 139, 90, 140, 90, 141, 90, 163, 90, 164, 90, 165, 90, 166, 90, 167, 90, 169, 90, 170, 90, 188, 90, 189, 90, 190, 90, 191, 90, 192, 90, 193, 90, 195, 90, 196, 90, 197, 90, 198, 90, 199, 90, 200, 90, 204, 90, 205, 90, 220, 90, 221, 90, 222, 90, 223, 90, 225, 90, 226, 90],
	#[31, -228, 2, -227, 7, -13, 10, -223, 11, -250, 33, -256, 37, 105, 38, -253, 42, -159, 63, -232, 72, -161, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 102, 7, 102, 1, 102, 14, 102, 11, 102, 0, 102, 16, 102, 2, 102, 4, 102, 19, 102, 18, 102, 9, 102, 10, 102, 25, 102, 12, 102, 13, 102, 15, 102, 17, 102, 20, 102, 21, 102, 22, 102, 23, 102, 24, 102, 26, 102, 27, 102, 28, 102, 29, 102, 31, 102, 32, 102, 33, 102, 34, 102, 37, 102, 38, 102, 40, 102, 41, 102, 42, 102, 43, 102, 57, 102, 58, 102, 59, 102, 60, 102, 63, 102, 82, 102, 72, 102, 74, 102, 75, 102, 76, 102, 77, 102, 101, 102, 83, 102, 84, 102, 85, 102, 86, 102, 87, 102, 88, 102, 89, 102, 90, 102, 91, 102, 92, 102, 93, 102, 95, 102, 97, 102, 99, 102, 100, 102, 102, 102, 103, 102, 104, 102, 105, 102, 106, 102, 107, 102, 108, 102, 109, 102, 110, 102, 111, 102, 112, 102, 113, 102, 114, 102, 115, 102, 116, 102, 117, 102, 118, 102, 119, 102, 121, 102, 122, 102, 123, 102, 124, 102, 125, 102, 126, 102, 127, 102, 129, 102, 130, 102, 131, 102, 132, 102, 133, 102, 136, 102, 137, 102, 138, 102, 139, 102, 140, 102, 141, 102, 163, 102, 164, 102, 165, 102, 166, 102, 167, 102, 169, 102, 170, 102, 188, 102, 189, 102, 190, 102, 191, 102, 192, 102, 193, 102, 195, 102, 196, 102, 197, 102, 198, 102, 199, 102, 200, 102, 204, 102, 205, 102, 220, 102, 221, 102, 222, 102, 223, 102, 225, 102, 226, 102],
	#[65535, 100, 14, 100, 4, 100, 21, 100, 15, 100, 1, 100, 9, 100, 2, 100, 23, 100, 12, 100, 11, 100, 0, 100, 27, 100, 31, 100, 24, 100, 7, 100, 10, 100, 25, 100, 34, 100, 13, 100, 26, 100, 16, 100, 17, 100, 18, 100, 19, 100, 20, 100, 22, 100, 32, 100, 28, 100, 29, 100, 33, 100, 63, 100, 37, 100, 38, 100, 40, 100, 41, 100, 42, 100, 43, 100, 59, 100, 57, 100, 58, 100, 60, 100, 75, 100, 74, 100, 72, 100, 76, 100, 77, 100, 82, 100, 83, 100, 84, 100, 85, 100, 86, 100, 87, 100, 88, 100, 89, 100, 90, 100, 91, 100, 92, 100, 93, 100, 95, 100, 97, 100, 99, 100, 100, 100, 101, 100, 102, 100, 103, 100, 104, 100, 105, 100, 106, 100, 107, 100, 108, 100, 109, 100, 110, 100, 111, 100, 112, 100, 113, 100, 114, 100, 115, 100, 116, 100, 117, 100, 118, 100, 119, 100, 121, 100, 122, 100, 123, 100, 124, 100, 125, 100, 126, 100, 127, 100, 129, 100, 130, 100, 131, 100, 132, 100, 133, 100, 136, 100, 137, 100, 138, 100, 139, 100, 140, 100, 141, 100, 170, 100, 163, 100, 164, 100, 165, 100, 166, 100, 167, 100, 169, 100, 189, 100, 188, 100, 190, 100, 191, 100, 192, 100, 193, 100, 195, 100, 196, 100, 197, 100, 198, 100, 199, 100, 200, 100, 204, 100, 205, 100, 220, 100, 221, 100, 222, 100, 223, 100, 225, 100, 226, 100],
	#[42, -151],
	#[31, -143, 38, -142],
	#[65535, 103, 14, 103, 7, 103, 2, 103, 21, 103, 15, 103, 23, 103, 12, 103, 0, 103, 1, 103, 31, 103, 4, 103, 19, 103, 18, 103, 9, 103, 10, 103, 11, 103, 34, 103, 13, 103, 27, 103, 26, 103, 16, 103, 17, 103, 25, 103, 24, 103, 20, 103, 22, 103, 32, 103, 28, 103, 29, 103, 33, 103, 63, 103, 37, 103, 38, 103, 40, 103, 41, 103, 42, 103, 43, 103, 58, 103, 57, 103, 87, 103, 59, 103, 60, 103, 74, 103, 82, 103, 72, 103, 103, 103, 75, 103, 76, 103, 77, 103, 90, 103, 111, 103, 83, 103, 84, 103, 85, 103, 86, 103, 110, 103, 88, 103, 89, 103, 119, 103, 91, 103, 92, 103, 93, 103, 95, 103, 97, 103, 127, 103, 99, 103, 100, 103, 101, 103, 102, 103, 126, 103, 104, 103, 105, 103, 106, 103, 107, 103, 108, 103, 109, 103, 123, 103, 122, 103, 112, 103, 113, 103, 114, 103, 115, 103, 116, 103, 117, 103, 118, 103, 130, 103, 121, 103, 141, 103, 140, 103, 124, 103, 125, 103, 139, 103, 138, 103, 129, 103, 137, 103, 131, 103, 132, 103, 133, 103, 136, 103, 163, 103, 164, 103, 165, 103, 166, 103, 167, 103, 169, 103, 170, 103, 188, 103, 189, 103, 190, 103, 191, 103, 192, 103, 193, 103, 195, 103, 196, 103, 197, 103, 198, 103, 199, 103, 200, 103, 204, 103, 205, 103, 220, 103, 221, 103, 222, 103, 223, 103, 225, 103, 226, 103],
	#[65535, 97, 15, 97, 0, 97, 1, 97, 2, 97, 4, 97, 7, 97, 9, 97, 10, 97, 11, 97, 12, 97, 13, 97, 14, 97, 26, 97, 16, 97, 17, 97, 18, 97, 19, 97, 20, 97, 21, 97, 22, 97, 23, 97, 24, 97, 25, 97, 27, 97, 28, 97, 29, 97, 31, 97, 32, 97, 33, 97, 34, 97, 37, 97, 38, 97, 40, 97, 41, 97, 42, 97, 43, 97, 57, 97, 58, 97, 59, 97, 60, 97, 63, 97, 72, 97, 74, 97, 75, 97, 76, 97, 77, 97, 82, 97, 83, 97, 84, 97, 85, 97, 86, 97, 87, 97, 88, 97, 89, 97, 90, 97, 91, 97, 92, 97, 93, 97, 95, 97, 97, 97, 99, 97, 100, 97, 101, 97, 102, 97, 103, 97, 104, 97, 105, 97, 106, 97, 107, 97, 108, 97, 109, 97, 110, 97, 111, 97, 112, 97, 113, 97, 114, 97, 115, 97, 116, 97, 117, 97, 118, 97, 119, 97, 121, 97, 122, 97, 123, 97, 124, 97, 125, 97, 126, 97, 127, 97, 129, 97, 130, 97, 131, 97, 132, 97, 133, 97, 136, 97, 137, 97, 138, 97, 139, 97, 140, 97, 141, 97, 163, 97, 164, 97, 165, 97, 166, 97, 167, 97, 169, 97, 170, 97, 188, 97, 189, 97, 190, 97, 191, 97, 192, 97, 193, 97, 195, 97, 196, 97, 197, 97, 198, 97, 199, 97, 200, 97, 204, 97, 205, 97, 220, 97, 221, 97, 222, 97, 223, 97, 225, 97, 226, 97],
	#[39, -220],
	#[7, -13, 11, -127, 31, -128, 32, -146, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, 165, 32, 165, 37, 105, 42, -159, 72, -161],
	#[13, -149, 32, -148],
	#[34, -147],
	#[65535, 96, 0, 96, 1, 96, 2, 96, 4, 96, 7, 96, 9, 96, 10, 96, 11, 96, 12, 96, 13, 96, 14, 96, 15, 96, 16, 96, 17, 96, 18, 96, 19, 96, 20, 96, 21, 96, 22, 96, 23, 96, 24, 96, 25, 96, 26, 96, 27, 96, 28, 96, 29, 96, 31, 96, 32, 96, 33, 96, 34, 96, 37, 96, 38, 96, 40, 96, 41, 96, 42, 96, 43, 96, 57, 96, 58, 96, 59, 96, 60, 96, 63, 96, 72, 96, 74, 96, 75, 96, 76, 96, 77, 96, 82, 96, 83, 96, 84, 96, 85, 96, 86, 96, 87, 96, 88, 96, 89, 96, 90, 96, 91, 96, 92, 96, 93, 96, 95, 96, 97, 96, 99, 96, 100, 96, 101, 96, 102, 96, 103, 96, 104, 96, 105, 96, 106, 96, 107, 96, 108, 96, 109, 96, 110, 96, 111, 96, 112, 96, 113, 96, 114, 96, 115, 96, 116, 96, 117, 96, 118, 96, 119, 96, 121, 96, 122, 96, 123, 96, 124, 96, 125, 96, 126, 96, 127, 96, 129, 96, 130, 96, 131, 96, 132, 96, 133, 96, 136, 96, 137, 96, 138, 96, 139, 96, 140, 96, 141, 96, 163, 96, 164, 96, 165, 96, 166, 96, 167, 96, 169, 96, 170, 96, 188, 96, 189, 96, 190, 96, 191, 96, 192, 96, 193, 96, 195, 96, 196, 96, 197, 96, 198, 96, 199, 96, 200, 96, 204, 96, 205, 96, 220, 96, 221, 96, 222, 96, 223, 96, 225, 96, 226, 96],
	#[34, -219],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 65, -122, 66, -126, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[37, 105, 13, 166, 42, -159, 32, 166, 72, -161],
	#[31, -128, 63, -121, 67, -137, 7, -13, 11, -127, 70, -140, 43, 159, 69, -136, 33, -139, 68, -133, 35, -123, 36, -129, 71, -130, 98, -156, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[13, 163, 43, 163, 60, -209, 93, -211, 108, -208, 109, -213, 220, -207, 221, -210, 222, -212, 223, -214],
	#[31, 109, 37, 84, 2, 109, 7, 109, 10, 109, 11, 109, 33, 109, 13, 109, 43, 109, 42, 84, 38, 109, 60, 109, 63, 109, 84, 109, 72, 84, 101, 109, 85, 109, 86, 109, 87, 109, 88, 109, 89, 109, 90, 109, 91, 109, 92, 109, 93, 109, 95, 109, 97, 109, 99, 109, 100, 109, 102, 109, 103, 109, 104, 109, 105, 109, 106, 109, 107, 109, 108, 109, 109, 109, 110, 109, 111, 109, 112, 109, 113, 109, 114, 109, 115, 109, 116, 109, 117, 109, 118, 109, 119, 109, 121, 109, 122, 109, 123, 109, 124, 109, 125, 109, 126, 109, 127, 109, 129, 109, 130, 109, 131, 109, 132, 109, 133, 109, 136, 109, 137, 109, 138, 109, 139, 109, 140, 109, 141, 109, 220, 109, 221, 109, 222, 109, 223, 109, 226, 109],
	#[13, -204, 43, 160],
	#[43, -217],
	#[65535, 162, 43, 162],
	#[2, 110, 7, 110, 10, 110, 11, 110, 13, 110, 31, 110, 33, 110, 37, 83, 38, 110, 42, 83, 43, 110, 57, -170, 58, -167, 59, -169, 60, 110, 63, 110, 72, 83, 101, 110, 84, 110, 85, 110, 86, 110, 87, 110, 88, 110, 89, 110, 90, 110, 91, 110, 92, 110, 93, 110, 95, 110, 97, 110, 99, 110, 100, 110, 102, 110, 103, 110, 104, 110, 105, 110, 106, 110, 107, 110, 108, 110, 109, 110, 110, 110, 111, 110, 112, 110, 113, 110, 114, 110, 115, 110, 116, 110, 117, 110, 118, 110, 119, 110, 121, 110, 122, 110, 123, 110, 124, 110, 125, 110, 126, 110, 127, 110, 129, 110, 130, 110, 131, 110, 132, 110, 133, 110, 136, 110, 137, 110, 138, 110, 139, 110, 140, 110, 141, 110, 220, 110, 221, 110, 222, 110, 223, 110, 226, 110],
	#[37, 105, 42, -159, 72, -161],
	#[69, -136, 70, -140, 31, -128, 63, -121, 7, -13, 11, -127, 67, -137, 43, 159, 33, -139, 68, -133, 35, -123, 36, -129, 71, -130, 98, -156, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[37, -165],
	#[42, -162],
	#[38, -163],
	#[43, -164],
	#[65535, 106, 37, 106, 81, 106, 83, 106, 82, 106],
	#[65535, 88, 14, 88, 9, 88, 2, 88, 7, 88, 0, 88, 1, 88, 22, 88, 10, 88, 13, 88, 15, 88, 75, 88, 11, 88, 28, 88, 12, 88, 25, 88, 23, 88, 16, 88, 19, 88, 17, 88, 38, 88, 4, 88, 86, 88, 29, 88, 31, 88, 18, 88, 60, 88, 42, 88, 33, 88, 43, 88, 34, 88, 37, 88, 59, 88, 26, 88, 40, 88, 20, 88, 27, 88, 24, 88, 76, 88, 41, 88, 58, 88, 103, 88, 110, 88, 77, 88, 21, 88, 32, 88, 82, 88, 85, 88, 87, 88, 74, 88, 83, 88, 72, 88, 89, 88, 63, 88, 90, 88, 93, 88, 95, 88, 88, 88, 91, 88, 101, 88, 99, 88, 116, 88, 100, 88, 57, 88, 121, 88, 109, 88, 111, 88, 104, 88, 97, 88, 102, 88, 119, 88, 106, 88, 115, 88, 113, 88, 84, 88, 107, 88, 127, 88, 105, 88, 123, 88, 138, 88, 124, 88, 108, 88, 92, 88, 112, 88, 132, 88, 114, 88, 122, 88, 117, 88, 118, 88, 125, 88, 126, 88, 165, 88, 129, 88, 130, 88, 131, 88, 189, 88, 133, 88, 136, 88, 137, 88, 167, 88, 139, 88, 140, 88, 141, 88, 190, 88, 188, 88, 169, 88, 170, 88, 191, 88, 163, 88, 164, 88, 193, 88, 166, 88, 205, 88, 195, 88, 204, 88, 192, 88, 196, 88, 197, 88, 198, 88, 199, 88, 200, 88, 220, 88, 221, 88, 222, 88, 223, 88, 225, 88, 226, 88],
	#[43, -171],
	#[65535, 406, 13, 406, 86, 406, 87, 406, 2, 406, 7, 406, 10, 406, 11, 406, 118, 406, 43, 406, 31, 406, 119, 406, 33, 406, 63, 406, 90, 406, 93, 406, 95, 406, 38, 406, 108, 406, 92, 406, 57, 406, 103, 406, 59, 406, 58, 406, 105, 406, 84, 406, 106, 406, 91, 406, 89, 406, 107, 406, 102, 406, 126, 406, 88, 406, 114, 406, 60, 406, 97, 406, 132, 406, 85, 406, 99, 406, 104, 406, 100, 406, 101, 406, 129, 406, 109, 406, 110, 406, 111, 406, 112, 406, 113, 406, 133, 406, 115, 406, 116, 406, 117, 406, 131, 406, 130, 406, 121, 406, 122, 406, 123, 406, 124, 406, 125, 406, 136, 406, 127, 406, 137, 406, 138, 406, 139, 406, 140, 406, 141, 406, 220, 406, 221, 406, 222, 406, 223, 406, 226, 406],
	#[65535, 111, 2, 111, 7, 111, 10, 111, 11, 111, 13, 111, 31, 111, 33, 111, 38, 111, 43, 111, 60, 111, 63, 111, 101, 111, 84, 111, 85, 111, 86, 111, 87, 111, 88, 111, 89, 111, 90, 111, 91, 111, 92, 111, 93, 111, 95, 111, 97, 111, 99, 111, 100, 111, 102, 111, 103, 111, 104, 111, 105, 111, 106, 111, 107, 111, 108, 111, 109, 111, 110, 111, 111, 111, 112, 111, 113, 111, 114, 111, 115, 111, 116, 111, 117, 111, 118, 111, 119, 111, 121, 111, 122, 111, 123, 111, 124, 111, 125, 111, 126, 111, 127, 111, 129, 111, 130, 111, 131, 111, 132, 111, 133, 111, 136, 111, 137, 111, 138, 111, 139, 111, 140, 111, 141, 111, 220, 111, 221, 111, 222, 111, 223, 111, 226, 111],
	#[65535, 407, 2, 407, 7, 407, 10, 407, 11, 407, 13, 407, 31, 407, 33, 407, 38, 407, 43, 407, 57, 407, 58, 407, 59, 407, 60, 407, 63, 407, 84, 407, 85, 407, 86, 407, 87, 407, 88, 407, 89, 407, 90, 407, 91, 407, 92, 407, 93, 407, 95, 407, 97, 407, 99, 407, 100, 407, 101, 407, 102, 407, 103, 407, 104, 407, 105, 407, 106, 407, 107, 407, 108, 407, 109, 407, 110, 407, 111, 407, 112, 407, 113, 407, 114, 407, 115, 407, 116, 407, 117, 407, 118, 407, 119, 407, 121, 407, 122, 407, 123, 407, 124, 407, 125, 407, 126, 407, 127, 407, 129, 407, 130, 407, 131, 407, 132, 407, 133, 407, 136, 407, 137, 407, 138, 407, 139, 407, 140, 407, 141, 407, 220, 407, 221, 407, 222, 407, 223, 407, 226, 407],
	#[65535, 408, 38, 408, 86, 408, 60, 408, 7, 408, 2, 408, 11, 408, 10, 408, 33, 408, 13, 408, 58, 408, 103, 408, 110, 408, 85, 408, 87, 408, 43, 408, 31, 408, 59, 408, 63, 408, 93, 408, 95, 408, 88, 408, 91, 408, 101, 408, 99, 408, 116, 408, 100, 408, 57, 408, 109, 408, 111, 408, 89, 408, 97, 408, 102, 408, 119, 408, 106, 408, 115, 408, 113, 408, 84, 408, 90, 408, 104, 408, 105, 408, 123, 408, 121, 408, 124, 408, 108, 408, 92, 408, 112, 408, 107, 408, 114, 408, 122, 408, 117, 408, 118, 408, 125, 408, 126, 408, 127, 408, 129, 408, 130, 408, 131, 408, 132, 408, 133, 408, 136, 408, 137, 408, 138, 408, 139, 408, 140, 408, 141, 408, 220, 408, 221, 408, 222, 408, 223, 408, 226, 408],
	#[65535, 211, 0, 211, 1, 211, 2, 211, 4, 211, 7, 211, 9, 211, 10, 211, 11, 211, 12, 211, 13, 211, 14, 211, 15, 211, 16, 211, 17, 211, 18, 211, 19, 211, 20, 211, 21, 211, 22, 211, 23, 211, 24, 211, 25, 211, 26, 211, 27, 211, 28, 211, 29, 211, 31, 211, 32, 211, 33, 211, 34, 211, 37, 211, 38, 211, 40, 211, 41, 211, 42, 211, 43, 211, 57, 211, 58, 211, 59, 211, 60, 211, 63, 211, 72, 211, 74, 211, 75, 211, 76, 211, 77, 211, 82, 211, 83, 211, 84, 211, 85, 211, 86, 211, 87, 211, 88, 211, 89, 211, 90, 211, 91, 211, 92, 211, 93, 211, 95, 211, 97, 211, 99, 211, 100, 211, 101, 211, 102, 211, 103, 211, 104, 211, 105, 211, 106, 211, 107, 211, 108, 211, 109, 211, 110, 211, 111, 211, 112, 211, 113, 211, 114, 211, 115, 211, 116, 211, 117, 211, 118, 211, 119, 211, 121, 211, 122, 211, 123, 211, 124, 211, 125, 211, 126, 211, 127, 211, 129, 211, 130, 211, 131, 211, 132, 211, 133, 211, 136, 211, 137, 211, 138, 211, 139, 211, 140, 211, 141, 211, 163, 211, 164, 211, 165, 211, 166, 211, 167, 211, 169, 211, 170, 211, 171, 211, 172, 211, 173, 211, 174, 211, 175, 211, 176, 211, 177, 211, 178, 211, 179, 211, 180, 211, 181, 211, 182, 211, 183, 211, 184, 211, 185, 211, 186, 211, 187, 211, 188, 211, 189, 211, 190, 211, 191, 211, 192, 211, 193, 211, 195, 211, 196, 211, 197, 211, 198, 211, 199, 211, 200, 211, 204, 211, 205, 211, 220, 211, 221, 211, 222, 211, 223, 211, 225, 211, 226, 211],
	#[0, 107, 1, 107, 2, 107, 4, 107, 7, 107, 9, 107, 10, 107, 11, 107, 12, 107, 13, 107, 14, 107, 15, 107, 16, 107, 17, 107, 18, 107, 19, 107, 20, 107, 21, 107, 22, 107, 23, 107, 24, 107, 25, 107, 26, 107, 27, 107, 28, 107, 29, 107, 31, 107, 32, 107, 33, 107, 34, 107, 37, 107, 38, 107, 40, 107, 41, 107, 42, 107, 43, 107, 57, -192, 58, -175, 59, -179, 60, 107, 63, 107, 72, 107, 74, 107, 75, 107, 76, 107, 77, 107, 82, 107, 83, 107, 84, 107, 85, 107, 86, 107, 87, 107, 88, 107, 89, 107, 90, 107, 91, 107, 92, 107, 93, 107, 95, 107, 97, 107, 99, 107, 100, 107, 101, 107, 102, 107, 103, 107, 104, 107, 105, 107, 106, 107, 107, 107, 108, 107, 109, 107, 110, 107, 111, 107, 112, 107, 113, 107, 114, 107, 115, 107, 116, 107, 117, 107, 118, 107, 119, 107, 121, 107, 122, 107, 123, 107, 124, 107, 125, 107, 126, 107, 127, 107, 129, 107, 130, 107, 131, 107, 132, 107, 133, 107, 136, 107, 137, 107, 138, 107, 139, 107, 140, 107, 141, 107, 163, 107, 164, 107, 165, 107, 166, 107, 167, 107, 169, 107, 170, 107, 171, -180, 172, -191, 173, -182, 174, -184, 175, -178, 176, -185, 177, -193, 178, -176, 179, -181, 180, -183, 181, -186, 182, -189, 183, -174, 184, -177, 185, -187, 186, -190, 187, -173, 188, 107, 189, 107, 190, 107, 191, 107, 192, 107, 193, 107, 195, 107, 196, 107, 197, 107, 198, 107, 199, 107, 200, 107, 204, 107, 205, 107, 220, 107, 221, 107, 222, 107, 223, 107, 225, 107, 226, 107],
	#[65535, 309, 7, 309, 14, 309, 11, 309, 0, 309, 1, 309, 2, 309, 4, 309, 19, 309, 18, 309, 9, 309, 10, 309, 16, 309, 12, 309, 13, 309, 15, 309, 17, 309, 20, 309, 21, 309, 22, 309, 23, 309, 24, 309, 25, 309, 26, 309, 27, 309, 28, 309, 29, 309, 31, 309, 32, 309, 33, 309, 34, 309, 37, 309, 38, 309, 40, 309, 41, 309, 42, 309, 43, 309, 57, 309, 58, 309, 59, 309, 60, 309, 61, 309, 63, 309, 82, 309, 72, 309, 74, 309, 75, 309, 76, 309, 77, 309, 101, 309, 83, 309, 84, 309, 85, 309, 86, 309, 87, 309, 88, 309, 89, 309, 90, 309, 91, 309, 92, 309, 93, 309, 94, 309, 95, 309, 96, 309, 97, 309, 99, 309, 100, 309, 102, 309, 103, 309, 104, 309, 105, 309, 106, 309, 107, 309, 108, 309, 109, 309, 110, 309, 111, 309, 112, 309, 113, 309, 114, 309, 115, 309, 116, 309, 117, 309, 118, 309, 119, 309, 120, 309, 121, 309, 122, 309, 123, 309, 124, 309, 125, 309, 126, 309, 127, 309, 129, 309, 130, 309, 131, 309, 132, 309, 133, 309, 136, 309, 137, 309, 138, 309, 139, 309, 140, 309, 141, 309, 163, 309, 164, 309, 165, 309, 166, 309, 167, 309, 169, 309, 170, 309, 171, 309, 172, 309, 173, 309, 174, 309, 175, 309, 176, 309, 177, 309, 178, 309, 179, 309, 180, 309, 181, 309, 182, 309, 183, 309, 184, 309, 185, 309, 186, 309, 187, 309, 188, 309, 189, 309, 190, 309, 191, 309, 192, 309, 193, 309, 195, 309, 196, 309, 197, 309, 198, 309, 199, 309, 200, 309, 204, 309, 205, 309, 220, 309, 221, 309, 222, 309, 223, 309, 225, 309, 226, 309],
	#[65535, 305, 15, 305, 1, 305, 14, 305, 7, 305, 2, 305, 21, 305, 4, 305, 10, 305, 17, 305, 0, 305, 23, 305, 12, 305, 11, 305, 22, 305, 87, 305, 16, 305, 31, 305, 20, 305, 61, 305, 19, 305, 18, 305, 9, 305, 29, 305, 25, 305, 34, 305, 13, 305, 27, 305, 26, 305, 84, 305, 32, 305, 37, 305, 33, 305, 59, 305, 57, 305, 75, 305, 24, 305, 58, 305, 60, 305, 41, 305, 28, 305, 43, 305, 42, 305, 119, 305, 102, 305, 63, 305, 40, 305, 93, 305, 95, 305, 38, 305, 77, 305, 86, 305, 74, 305, 76, 305, 88, 305, 83, 305, 103, 305, 90, 305, 72, 305, 94, 305, 118, 305, 115, 305, 96, 305, 91, 305, 111, 305, 89, 305, 97, 305, 105, 305, 126, 305, 92, 305, 114, 305, 82, 305, 106, 305, 104, 305, 85, 305, 99, 305, 107, 305, 100, 305, 101, 305, 108, 305, 109, 305, 110, 305, 122, 305, 112, 305, 113, 305, 133, 305, 132, 305, 116, 305, 117, 305, 131, 305, 130, 305, 120, 305, 121, 305, 129, 305, 123, 305, 124, 305, 125, 305, 136, 305, 127, 305, 137, 305, 138, 305, 139, 305, 140, 305, 141, 305, 166, 305, 171, 305, 165, 305, 164, 305, 170, 305, 163, 305, 167, 305, 169, 305, 172, 305, 173, 305, 174, 305, 175, 305, 176, 305, 177, 305, 178, 305, 179, 305, 180, 305, 181, 305, 182, 305, 183, 305, 184, 305, 185, 305, 186, 305, 187, 305, 188, 305, 189, 305, 190, 305, 191, 305, 192, 305, 193, 305, 195, 305, 196, 305, 197, 305, 198, 305, 199, 305, 200, 305, 204, 305, 205, 305, 220, 305, 221, 305, 222, 305, 223, 305, 225, 305, 226, 305],
	#[65535, 293, 23, 293, 31, 293, 7, 293, 22, 293, 11, 293, 29, 293, 4, 293, 13, 293, 14, 293, 38, 293, 34, 293, 2, 293, 59, 293, 17, 293, 25, 293, 0, 293, 15, 293, 102, 293, 27, 293, 76, 293, 60, 293, 86, 293, 12, 293, 74, 293, 42, 293, 10, 293, 87, 293, 1, 293, 94, 293, 119, 293, 61, 293, 63, 293, 19, 293, 18, 293, 95, 293, 9, 293, 75, 293, 16, 293, 58, 293, 26, 293, 103, 293, 32, 293, 37, 293, 24, 293, 20, 293, 21, 293, 72, 293, 28, 293, 43, 293, 33, 293, 41, 293, 93, 293, 115, 293, 82, 293, 77, 293, 40, 293, 92, 293, 107, 293, 57, 293, 83, 293, 90, 293, 99, 293, 118, 293, 84, 293, 91, 293, 111, 293, 89, 293, 97, 293, 105, 293, 126, 293, 88, 293, 114, 293, 117, 293, 106, 293, 85, 293, 96, 293, 104, 293, 100, 293, 101, 293, 108, 293, 109, 293, 110, 293, 122, 293, 112, 293, 113, 293, 133, 293, 132, 293, 116, 293, 175, 293, 131, 293, 130, 293, 120, 293, 121, 293, 129, 293, 123, 293, 124, 293, 125, 293, 136, 293, 127, 293, 137, 293, 138, 293, 139, 293, 140, 293, 141, 293, 166, 293, 171, 293, 165, 293, 164, 293, 170, 293, 163, 293, 174, 293, 167, 293, 169, 293, 172, 293, 173, 293, 176, 293, 177, 293, 178, 293, 179, 293, 180, 293, 181, 293, 182, 293, 183, 293, 184, 293, 185, 293, 186, 293, 187, 293, 188, 293, 189, 293, 190, 293, 191, 293, 192, 293, 193, 293, 195, 293, 196, 293, 197, 293, 198, 293, 199, 293, 200, 293, 204, 293, 205, 293, 220, 293, 221, 293, 222, 293, 223, 293, 225, 293, 226, 293],
	#[65535, 300, 23, 300, 7, 300, 27, 300, 15, 300, 103, 300, 22, 300, 29, 300, 20, 300, 4, 300, 28, 300, 31, 300, 14, 300, 11, 300, 2, 300, 86, 300, 26, 300, 95, 300, 43, 300, 83, 300, 42, 300, 10, 300, 13, 300, 34, 300, 75, 300, 1, 300, 93, 300, 37, 300, 25, 300, 21, 300, 0, 300, 111, 300, 19, 300, 17, 300, 38, 300, 40, 300, 85, 300, 82, 300, 92, 300, 18, 300, 60, 300, 63, 300, 33, 300, 12, 300, 94, 300, 57, 300, 59, 300, 127, 300, 61, 300, 16, 300, 9, 300, 24, 300, 76, 300, 41, 300, 58, 300, 84, 300, 110, 300, 77, 300, 96, 300, 117, 300, 97, 300, 88, 300, 87, 300, 74, 300, 116, 300, 32, 300, 102, 300, 126, 300, 90, 300, 113, 300, 72, 300, 133, 300, 91, 300, 101, 300, 99, 300, 114, 300, 100, 300, 138, 300, 121, 300, 109, 300, 130, 300, 89, 300, 107, 300, 105, 300, 119, 300, 106, 300, 115, 300, 132, 300, 122, 300, 125, 300, 104, 300, 120, 300, 123, 300, 140, 300, 124, 300, 108, 300, 190, 300, 112, 300, 137, 300, 141, 300, 163, 300, 175, 300, 118, 300, 183, 300, 170, 300, 165, 300, 174, 300, 129, 300, 164, 300, 131, 300, 189, 300, 191, 300, 178, 300, 136, 300, 188, 300, 167, 300, 139, 300, 179, 300, 169, 300, 186, 300, 166, 300, 184, 300, 173, 300, 172, 300, 177, 300, 171, 300, 182, 300, 181, 300, 180, 300, 185, 300, 193, 300, 176, 300, 205, 300, 195, 300, 204, 300, 187, 300, 192, 300, 196, 300, 197, 300, 198, 300, 199, 300, 200, 300, 220, 300, 221, 300, 222, 300, 223, 300, 225, 300, 226, 300],
	#[42, -194],
	#[65535, 297, 15, 297, 1, 297, 7, 297, 2, 297, 20, 297, 4, 297, 10, 297, 17, 297, 0, 297, 14, 297, 26, 297, 23, 297, 12, 297, 11, 297, 22, 297, 87, 297, 16, 297, 31, 297, 61, 297, 19, 297, 18, 297, 9, 297, 29, 297, 25, 297, 34, 297, 13, 297, 27, 297, 38, 297, 84, 297, 32, 297, 37, 297, 33, 297, 59, 297, 21, 297, 57, 297, 75, 297, 24, 297, 58, 297, 60, 297, 41, 297, 28, 297, 43, 297, 42, 297, 119, 297, 102, 297, 63, 297, 40, 297, 93, 297, 95, 297, 77, 297, 86, 297, 74, 297, 76, 297, 88, 297, 83, 297, 103, 297, 90, 297, 72, 297, 94, 297, 118, 297, 115, 297, 96, 297, 91, 297, 111, 297, 89, 297, 97, 297, 105, 297, 126, 297, 92, 297, 114, 297, 82, 297, 106, 297, 104, 297, 85, 297, 99, 297, 107, 297, 100, 297, 101, 297, 108, 297, 109, 297, 110, 297, 122, 297, 112, 297, 113, 297, 133, 297, 132, 297, 116, 297, 117, 297, 131, 297, 130, 297, 120, 297, 121, 297, 129, 297, 123, 297, 124, 297, 125, 297, 136, 297, 127, 297, 137, 297, 138, 297, 139, 297, 140, 297, 141, 297, 166, 297, 171, 297, 165, 297, 164, 297, 170, 297, 163, 297, 167, 297, 169, 297, 172, 297, 173, 297, 174, 297, 175, 297, 176, 297, 177, 297, 178, 297, 179, 297, 180, 297, 181, 297, 182, 297, 183, 297, 184, 297, 185, 297, 186, 297, 187, 297, 188, 297, 189, 297, 190, 297, 191, 297, 192, 297, 193, 297, 195, 297, 196, 297, 197, 297, 198, 297, 199, 297, 200, 297, 204, 297, 205, 297, 220, 297, 221, 297, 222, 297, 223, 297, 225, 297, 226, 297],
	#[65535, 294, 0, 294, 1, 294, 2, 294, 4, 294, 7, 294, 9, 294, 10, 294, 11, 294, 12, 294, 13, 294, 14, 294, 15, 294, 16, 294, 17, 294, 18, 294, 19, 294, 20, 294, 21, 294, 22, 294, 23, 294, 24, 294, 25, 294, 26, 294, 27, 294, 28, 294, 29, 294, 31, 294, 32, 294, 33, 294, 34, 294, 37, 294, 38, 294, 40, 294, 41, 294, 42, 294, 43, 294, 57, 294, 58, 294, 59, 294, 60, 294, 61, 294, 63, 294, 72, 294, 74, 294, 75, 294, 76, 294, 77, 294, 82, 294, 83, 294, 84, 294, 85, 294, 86, 294, 87, 294, 88, 294, 89, 294, 90, 294, 91, 294, 92, 294, 93, 294, 94, 294, 95, 294, 96, 294, 97, 294, 99, 294, 100, 294, 101, 294, 102, 294, 103, 294, 104, 294, 105, 294, 106, 294, 107, 294, 108, 294, 109, 294, 110, 294, 111, 294, 112, 294, 113, 294, 114, 294, 115, 294, 116, 294, 117, 294, 118, 294, 119, 294, 120, 294, 121, 294, 122, 294, 123, 294, 124, 294, 125, 294, 126, 294, 127, 294, 129, 294, 130, 294, 131, 294, 132, 294, 133, 294, 136, 294, 137, 294, 138, 294, 139, 294, 140, 294, 141, 294, 163, 294, 164, 294, 165, 294, 166, 294, 167, 294, 169, 294, 170, 294, 171, 294, 172, 294, 173, 294, 174, 294, 175, 294, 176, 294, 177, 294, 178, 294, 179, 294, 180, 294, 181, 294, 182, 294, 183, 294, 184, 294, 185, 294, 186, 294, 187, 294, 188, 294, 189, 294, 190, 294, 191, 294, 192, 294, 193, 294, 195, 294, 196, 294, 197, 294, 198, 294, 199, 294, 200, 294, 204, 294, 205, 294, 220, 294, 221, 294, 222, 294, 223, 294, 225, 294, 226, 294],
	#[65535, 290, 7, 290, 15, 290, 4, 290, 10, 290, 14, 290, 13, 290, 12, 290, 11, 290, 22, 290, 0, 290, 1, 290, 2, 290, 20, 290, 26, 290, 19, 290, 18, 290, 9, 290, 17, 290, 16, 290, 21, 290, 23, 290, 24, 290, 25, 290, 27, 290, 28, 290, 29, 290, 31, 290, 32, 290, 33, 290, 34, 290, 37, 290, 38, 290, 40, 290, 41, 290, 42, 290, 43, 290, 57, 290, 58, 290, 59, 290, 60, 290, 61, 290, 63, 290, 72, 290, 74, 290, 75, 290, 76, 290, 77, 290, 82, 290, 83, 290, 84, 290, 85, 290, 86, 290, 87, 290, 88, 290, 89, 290, 90, 290, 91, 290, 92, 290, 93, 290, 94, 290, 95, 290, 96, 290, 97, 290, 99, 290, 100, 290, 101, 290, 102, 290, 103, 290, 104, 290, 105, 290, 106, 290, 107, 290, 108, 290, 109, 290, 110, 290, 111, 290, 112, 290, 113, 290, 114, 290, 115, 290, 116, 290, 117, 290, 118, 290, 119, 290, 120, 290, 121, 290, 122, 290, 123, 290, 124, 290, 125, 290, 126, 290, 127, 290, 129, 290, 130, 290, 131, 290, 132, 290, 133, 290, 136, 290, 137, 290, 138, 290, 139, 290, 140, 290, 141, 290, 163, 290, 164, 290, 165, 290, 166, 290, 167, 290, 169, 290, 170, 290, 171, 290, 172, 290, 173, 290, 174, 290, 175, 290, 176, 290, 177, 290, 178, 290, 179, 290, 180, 290, 181, 290, 182, 290, 183, 290, 184, 290, 185, 290, 186, 290, 187, 290, 188, 290, 189, 290, 190, 290, 191, 290, 192, 290, 193, 290, 195, 290, 196, 290, 197, 290, 198, 290, 199, 290, 200, 290, 204, 290, 205, 290, 220, 290, 221, 290, 222, 290, 223, 290, 225, 290, 226, 290],
	#[65535, 301, 4, 301, 7, 301, 12, 301, 22, 301, 23, 301, 13, 301, 1, 301, 28, 301, 31, 301, 14, 301, 9, 301, 2, 301, 26, 301, 0, 301, 18, 301, 29, 301, 17, 301, 10, 301, 94, 301, 15, 301, 75, 301, 11, 301, 93, 301, 37, 301, 25, 301, 33, 301, 24, 301, 42, 301, 16, 301, 19, 301, 34, 301, 38, 301, 40, 301, 86, 301, 32, 301, 92, 301, 76, 301, 60, 301, 63, 301, 72, 301, 43, 301, 95, 301, 57, 301, 59, 301, 84, 301, 61, 301, 20, 301, 27, 301, 108, 301, 89, 301, 41, 301, 58, 301, 103, 301, 110, 301, 77, 301, 21, 301, 117, 301, 82, 301, 85, 301, 87, 301, 74, 301, 83, 301, 100, 301, 102, 301, 97, 301, 90, 301, 96, 301, 114, 301, 88, 301, 91, 301, 101, 301, 99, 301, 116, 301, 118, 301, 113, 301, 121, 301, 109, 301, 111, 301, 104, 301, 107, 301, 105, 301, 119, 301, 106, 301, 115, 301, 132, 301, 122, 301, 125, 301, 127, 301, 120, 301, 123, 301, 138, 301, 124, 301, 166, 301, 130, 301, 112, 301, 137, 301, 141, 301, 163, 301, 175, 301, 131, 301, 183, 301, 126, 301, 165, 301, 174, 301, 129, 301, 164, 301, 136, 301, 189, 301, 133, 301, 173, 301, 182, 301, 170, 301, 167, 301, 139, 301, 140, 301, 169, 301, 177, 301, 181, 301, 190, 301, 188, 301, 172, 301, 178, 301, 185, 301, 171, 301, 184, 301, 176, 301, 191, 301, 180, 301, 186, 301, 193, 301, 179, 301, 205, 301, 195, 301, 204, 301, 187, 301, 192, 301, 196, 301, 197, 301, 198, 301, 199, 301, 200, 301, 220, 301, 221, 301, 222, 301, 223, 301, 225, 301, 226, 301],
	#[65535, 295, 7, 295, 23, 295, 2, 295, 15, 295, 1, 295, 9, 295, 14, 295, 12, 295, 16, 295, 11, 295, 0, 295, 27, 295, 31, 295, 4, 295, 19, 295, 18, 295, 24, 295, 10, 295, 25, 295, 13, 295, 26, 295, 17, 295, 20, 295, 21, 295, 22, 295, 32, 295, 28, 295, 29, 295, 33, 295, 34, 295, 37, 295, 38, 295, 40, 295, 41, 295, 42, 295, 43, 295, 57, 295, 58, 295, 59, 295, 60, 295, 61, 295, 63, 295, 82, 295, 72, 295, 74, 295, 75, 295, 76, 295, 77, 295, 101, 295, 83, 295, 84, 295, 85, 295, 86, 295, 87, 295, 88, 295, 89, 295, 90, 295, 91, 295, 92, 295, 93, 295, 94, 295, 95, 295, 96, 295, 97, 295, 99, 295, 100, 295, 102, 295, 103, 295, 104, 295, 105, 295, 106, 295, 107, 295, 108, 295, 109, 295, 110, 295, 111, 295, 112, 295, 113, 295, 114, 295, 115, 295, 116, 295, 117, 295, 118, 295, 119, 295, 120, 295, 121, 295, 122, 295, 123, 295, 124, 295, 125, 295, 126, 295, 127, 295, 129, 295, 130, 295, 131, 295, 132, 295, 133, 295, 136, 295, 137, 295, 138, 295, 139, 295, 140, 295, 141, 295, 175, 295, 174, 295, 163, 295, 164, 295, 165, 295, 166, 295, 167, 295, 169, 295, 170, 295, 171, 295, 172, 295, 173, 295, 176, 295, 177, 295, 178, 295, 179, 295, 180, 295, 181, 295, 182, 295, 183, 295, 184, 295, 185, 295, 186, 295, 187, 295, 188, 295, 189, 295, 190, 295, 191, 295, 192, 295, 193, 295, 195, 295, 196, 295, 197, 295, 198, 295, 199, 295, 200, 295, 204, 295, 205, 295, 220, 295, 221, 295, 222, 295, 223, 295, 225, 295, 226, 295],
	#[65535, 302, 31, 302, 14, 302, 7, 302, 15, 302, 4, 302, 2, 302, 23, 302, 0, 302, 1, 302, 21, 302, 24, 302, 18, 302, 9, 302, 10, 302, 11, 302, 12, 302, 13, 302, 27, 302, 26, 302, 16, 302, 17, 302, 25, 302, 19, 302, 20, 302, 22, 302, 32, 302, 28, 302, 29, 302, 33, 302, 34, 302, 37, 302, 38, 302, 40, 302, 41, 302, 42, 302, 43, 302, 57, 302, 58, 302, 59, 302, 60, 302, 61, 302, 63, 302, 72, 302, 74, 302, 75, 302, 76, 302, 77, 302, 82, 302, 83, 302, 84, 302, 85, 302, 86, 302, 87, 302, 88, 302, 89, 302, 90, 302, 91, 302, 92, 302, 93, 302, 94, 302, 95, 302, 96, 302, 97, 302, 99, 302, 100, 302, 101, 302, 102, 302, 103, 302, 104, 302, 105, 302, 106, 302, 107, 302, 108, 302, 109, 302, 110, 302, 111, 302, 112, 302, 113, 302, 114, 302, 115, 302, 116, 302, 117, 302, 118, 302, 119, 302, 120, 302, 121, 302, 122, 302, 123, 302, 124, 302, 125, 302, 126, 302, 127, 302, 129, 302, 130, 302, 131, 302, 132, 302, 133, 302, 136, 302, 137, 302, 138, 302, 139, 302, 140, 302, 141, 302, 163, 302, 164, 302, 165, 302, 166, 302, 167, 302, 169, 302, 170, 302, 171, 302, 172, 302, 173, 302, 174, 302, 175, 302, 176, 302, 177, 302, 178, 302, 179, 302, 180, 302, 181, 302, 182, 302, 183, 302, 184, 302, 185, 302, 186, 302, 187, 302, 188, 302, 189, 302, 190, 302, 191, 302, 192, 302, 193, 302, 195, 302, 196, 302, 197, 302, 198, 302, 199, 302, 200, 302, 204, 302, 205, 302, 220, 302, 221, 302, 222, 302, 223, 302, 225, 302, 226, 302],
	#[65535, 296, 7, 296, 14, 296, 11, 296, 0, 296, 1, 296, 2, 296, 4, 296, 19, 296, 18, 296, 9, 296, 10, 296, 16, 296, 12, 296, 13, 296, 15, 296, 17, 296, 20, 296, 21, 296, 22, 296, 23, 296, 24, 296, 25, 296, 26, 296, 27, 296, 28, 296, 29, 296, 31, 296, 32, 296, 33, 296, 34, 296, 37, 296, 38, 296, 40, 296, 41, 296, 42, 296, 43, 296, 57, 296, 58, 296, 59, 296, 60, 296, 61, 296, 63, 296, 82, 296, 72, 296, 74, 296, 75, 296, 76, 296, 77, 296, 101, 296, 83, 296, 84, 296, 85, 296, 86, 296, 87, 296, 88, 296, 89, 296, 90, 296, 91, 296, 92, 296, 93, 296, 94, 296, 95, 296, 96, 296, 97, 296, 99, 296, 100, 296, 102, 296, 103, 296, 104, 296, 105, 296, 106, 296, 107, 296, 108, 296, 109, 296, 110, 296, 111, 296, 112, 296, 113, 296, 114, 296, 115, 296, 116, 296, 117, 296, 118, 296, 119, 296, 120, 296, 121, 296, 122, 296, 123, 296, 124, 296, 125, 296, 126, 296, 127, 296, 129, 296, 130, 296, 131, 296, 132, 296, 133, 296, 136, 296, 137, 296, 138, 296, 139, 296, 140, 296, 141, 296, 163, 296, 164, 296, 165, 296, 166, 296, 167, 296, 169, 296, 170, 296, 171, 296, 172, 296, 173, 296, 174, 296, 175, 296, 176, 296, 177, 296, 178, 296, 179, 296, 180, 296, 181, 296, 182, 296, 183, 296, 184, 296, 185, 296, 186, 296, 187, 296, 188, 296, 189, 296, 190, 296, 191, 296, 192, 296, 193, 296, 195, 296, 196, 296, 197, 296, 198, 296, 199, 296, 200, 296, 204, 296, 205, 296, 220, 296, 221, 296, 222, 296, 223, 296, 225, 296, 226, 296],
	#[65535, 298, 23, 298, 31, 298, 7, 298, 29, 298, 4, 298, 37, 298, 42, 298, 14, 298, 20, 298, 10, 298, 28, 298, 12, 298, 2, 298, 59, 298, 32, 298, 61, 298, 15, 298, 102, 298, 9, 298, 43, 298, 60, 298, 26, 298, 1, 298, 0, 298, 11, 298, 87, 298, 34, 298, 94, 298, 17, 298, 63, 298, 19, 298, 18, 298, 57, 298, 24, 298, 75, 298, 16, 298, 13, 298, 58, 298, 40, 298, 103, 298, 86, 298, 25, 298, 33, 298, 77, 298, 21, 298, 22, 298, 27, 298, 85, 298, 74, 298, 84, 298, 41, 298, 90, 298, 93, 298, 95, 298, 38, 298, 91, 298, 106, 298, 92, 298, 72, 298, 88, 298, 83, 298, 96, 298, 89, 298, 105, 298, 76, 298, 82, 298, 97, 298, 104, 298, 99, 298, 107, 298, 100, 298, 101, 298, 108, 298, 109, 298, 110, 298, 111, 298, 112, 298, 113, 298, 114, 298, 115, 298, 116, 298, 117, 298, 118, 298, 119, 298, 120, 298, 121, 298, 122, 298, 123, 298, 124, 298, 125, 298, 126, 298, 127, 298, 129, 298, 130, 298, 131, 298, 132, 298, 133, 298, 136, 298, 137, 298, 138, 298, 139, 298, 140, 298, 141, 298, 175, 298, 174, 298, 187, 298, 163, 298, 164, 298, 165, 298, 166, 298, 167, 298, 169, 298, 170, 298, 171, 298, 172, 298, 173, 298, 176, 298, 177, 298, 178, 298, 179, 298, 180, 298, 181, 298, 182, 298, 183, 298, 184, 298, 185, 298, 186, 298, 192, 298, 188, 298, 189, 298, 190, 298, 191, 298, 193, 298, 195, 298, 196, 298, 197, 298, 198, 298, 199, 298, 200, 298, 204, 298, 205, 298, 220, 298, 221, 298, 222, 298, 223, 298, 225, 298, 226, 298],
	#[65535, 303, 7, 303, 14, 303, 11, 303, 0, 303, 1, 303, 2, 303, 4, 303, 19, 303, 18, 303, 9, 303, 10, 303, 16, 303, 12, 303, 13, 303, 15, 303, 17, 303, 20, 303, 21, 303, 22, 303, 23, 303, 24, 303, 25, 303, 26, 303, 27, 303, 28, 303, 29, 303, 31, 303, 32, 303, 33, 303, 34, 303, 37, 303, 38, 303, 40, 303, 41, 303, 42, 303, 43, 303, 57, 303, 58, 303, 59, 303, 60, 303, 61, 303, 63, 303, 82, 303, 72, 303, 74, 303, 75, 303, 76, 303, 77, 303, 101, 303, 83, 303, 84, 303, 85, 303, 86, 303, 87, 303, 88, 303, 89, 303, 90, 303, 91, 303, 92, 303, 93, 303, 94, 303, 95, 303, 96, 303, 97, 303, 99, 303, 100, 303, 102, 303, 103, 303, 104, 303, 105, 303, 106, 303, 107, 303, 108, 303, 109, 303, 110, 303, 111, 303, 112, 303, 113, 303, 114, 303, 115, 303, 116, 303, 117, 303, 118, 303, 119, 303, 120, 303, 121, 303, 122, 303, 123, 303, 124, 303, 125, 303, 126, 303, 127, 303, 129, 303, 130, 303, 131, 303, 132, 303, 133, 303, 136, 303, 137, 303, 138, 303, 139, 303, 140, 303, 141, 303, 163, 303, 164, 303, 165, 303, 166, 303, 167, 303, 169, 303, 170, 303, 171, 303, 172, 303, 173, 303, 174, 303, 175, 303, 176, 303, 177, 303, 178, 303, 179, 303, 180, 303, 181, 303, 182, 303, 183, 303, 184, 303, 185, 303, 186, 303, 187, 303, 188, 303, 189, 303, 190, 303, 191, 303, 192, 303, 193, 303, 195, 303, 196, 303, 197, 303, 198, 303, 199, 303, 200, 303, 204, 303, 205, 303, 220, 303, 221, 303, 222, 303, 223, 303, 225, 303, 226, 303],
	#[65535, 307, 0, 307, 1, 307, 2, 307, 4, 307, 7, 307, 9, 307, 10, 307, 11, 307, 12, 307, 13, 307, 14, 307, 15, 307, 16, 307, 17, 307, 18, 307, 19, 307, 20, 307, 21, 307, 22, 307, 23, 307, 24, 307, 25, 307, 26, 307, 27, 307, 28, 307, 29, 307, 31, 307, 32, 307, 33, 307, 34, 307, 37, 307, 38, 307, 40, 307, 41, 307, 42, 307, 43, 307, 57, 307, 58, 307, 59, 307, 60, 307, 61, 307, 63, 307, 72, 307, 74, 307, 75, 307, 76, 307, 77, 307, 82, 307, 83, 307, 84, 307, 85, 307, 86, 307, 87, 307, 88, 307, 89, 307, 90, 307, 91, 307, 92, 307, 93, 307, 94, 307, 95, 307, 96, 307, 97, 307, 99, 307, 100, 307, 101, 307, 102, 307, 103, 307, 104, 307, 105, 307, 106, 307, 107, 307, 108, 307, 109, 307, 110, 307, 111, 307, 112, 307, 113, 307, 114, 307, 115, 307, 116, 307, 117, 307, 118, 307, 119, 307, 120, 307, 121, 307, 122, 307, 123, 307, 124, 307, 125, 307, 126, 307, 127, 307, 129, 307, 130, 307, 131, 307, 132, 307, 133, 307, 136, 307, 137, 307, 138, 307, 139, 307, 140, 307, 141, 307, 163, 307, 164, 307, 165, 307, 166, 307, 167, 307, 169, 307, 170, 307, 171, 307, 172, 307, 173, 307, 174, 307, 175, 307, 176, 307, 177, 307, 178, 307, 179, 307, 180, 307, 181, 307, 182, 307, 183, 307, 184, 307, 185, 307, 186, 307, 187, 307, 188, 307, 189, 307, 190, 307, 191, 307, 192, 307, 193, 307, 195, 307, 196, 307, 197, 307, 198, 307, 199, 307, 200, 307, 204, 307, 205, 307, 220, 307, 221, 307, 222, 307, 223, 307, 225, 307, 226, 307],
	#[65535, 212, 4, 212, 12, 212, 10, 212, 18, 212, 7, 212, 14, 212, 1, 212, 2, 212, 43, 212, 26, 212, 19, 212, 24, 212, 22, 212, 84, 212, 9, 212, 94, 212, 25, 212, 23, 212, 13, 212, 57, 212, 92, 212, 0, 212, 93, 212, 28, 212, 31, 212, 75, 212, 11, 212, 17, 212, 86, 212, 33, 212, 40, 212, 16, 212, 85, 212, 29, 212, 42, 212, 108, 212, 15, 212, 103, 212, 58, 212, 37, 212, 32, 212, 77, 212, 41, 212, 117, 212, 61, 212, 72, 212, 34, 212, 38, 212, 89, 212, 110, 212, 76, 212, 60, 212, 63, 212, 95, 212, 59, 212, 20, 212, 27, 212, 90, 212, 21, 212, 82, 212, 105, 212, 87, 212, 74, 212, 83, 212, 100, 212, 102, 212, 97, 212, 96, 212, 114, 212, 88, 212, 91, 212, 101, 212, 99, 212, 116, 212, 118, 212, 113, 212, 121, 212, 109, 212, 111, 212, 104, 212, 107, 212, 122, 212, 119, 212, 106, 212, 115, 212, 132, 212, 125, 212, 127, 212, 120, 212, 123, 212, 138, 212, 124, 212, 166, 212, 130, 212, 112, 212, 137, 212, 141, 212, 163, 212, 175, 212, 131, 212, 183, 212, 126, 212, 165, 212, 174, 212, 129, 212, 164, 212, 136, 212, 189, 212, 133, 212, 173, 212, 182, 212, 170, 212, 167, 212, 139, 212, 140, 212, 169, 212, 177, 212, 181, 212, 190, 212, 188, 212, 172, 212, 178, 212, 185, 212, 171, 212, 184, 212, 176, 212, 191, 212, 180, 212, 186, 212, 193, 212, 179, 212, 205, 212, 195, 212, 204, 212, 187, 212, 192, 212, 196, 212, 197, 212, 198, 212, 199, 212, 200, 212, 220, 212, 221, 212, 222, 212, 223, 212, 225, 212, 226, 212],
	#[65535, 304, 15, 304, 0, 304, 1, 304, 2, 304, 4, 304, 7, 304, 9, 304, 10, 304, 11, 304, 12, 304, 13, 304, 14, 304, 16, 304, 17, 304, 18, 304, 19, 304, 20, 304, 21, 304, 22, 304, 23, 304, 24, 304, 25, 304, 26, 304, 27, 304, 28, 304, 29, 304, 31, 304, 32, 304, 33, 304, 34, 304, 37, 304, 38, 304, 40, 304, 41, 304, 42, 304, 43, 304, 57, 304, 58, 304, 59, 304, 60, 304, 61, 304, 63, 304, 72, 304, 74, 304, 75, 304, 76, 304, 77, 304, 82, 304, 83, 304, 84, 304, 85, 304, 86, 304, 87, 304, 88, 304, 89, 304, 90, 304, 91, 304, 92, 304, 93, 304, 94, 304, 95, 304, 96, 304, 97, 304, 99, 304, 100, 304, 101, 304, 102, 304, 103, 304, 104, 304, 105, 304, 106, 304, 107, 304, 108, 304, 109, 304, 110, 304, 111, 304, 112, 304, 113, 304, 114, 304, 115, 304, 116, 304, 117, 304, 118, 304, 119, 304, 120, 304, 121, 304, 122, 304, 123, 304, 124, 304, 125, 304, 126, 304, 127, 304, 129, 304, 130, 304, 131, 304, 132, 304, 133, 304, 136, 304, 137, 304, 138, 304, 139, 304, 140, 304, 141, 304, 163, 304, 164, 304, 165, 304, 166, 304, 167, 304, 169, 304, 170, 304, 171, 304, 172, 304, 173, 304, 174, 304, 175, 304, 176, 304, 177, 304, 178, 304, 179, 304, 180, 304, 181, 304, 182, 304, 183, 304, 184, 304, 185, 304, 186, 304, 187, 304, 188, 304, 189, 304, 190, 304, 191, 304, 192, 304, 193, 304, 195, 304, 196, 304, 197, 304, 198, 304, 199, 304, 200, 304, 204, 304, 205, 304, 220, 304, 221, 304, 222, 304, 223, 304, 225, 304, 226, 304],
	#[65535, 308, 14, 308, 13, 308, 4, 308, 21, 308, 19, 308, 7, 308, 28, 308, 12, 308, 11, 308, 1, 308, 9, 308, 15, 308, 20, 308, 34, 308, 2, 308, 37, 308, 0, 308, 23, 308, 18, 308, 61, 308, 27, 308, 10, 308, 17, 308, 43, 308, 33, 308, 31, 308, 57, 308, 63, 308, 24, 308, 25, 308, 95, 308, 93, 308, 76, 308, 60, 308, 86, 308, 58, 308, 26, 308, 82, 308, 59, 308, 22, 308, 84, 308, 32, 308, 94, 308, 29, 308, 40, 308, 41, 308, 111, 308, 109, 308, 92, 308, 83, 308, 102, 308, 87, 308, 42, 308, 77, 308, 75, 308, 38, 308, 90, 308, 88, 308, 89, 308, 97, 308, 16, 308, 85, 308, 127, 308, 125, 308, 108, 308, 99, 308, 118, 308, 103, 308, 104, 308, 114, 308, 91, 308, 72, 308, 106, 308, 173, 308, 117, 308, 113, 308, 174, 308, 101, 308, 115, 308, 119, 308, 74, 308, 130, 308, 112, 308, 122, 308, 100, 308, 166, 308, 120, 308, 105, 308, 124, 308, 129, 308, 96, 308, 164, 308, 116, 308, 107, 308, 110, 308, 123, 308, 182, 308, 137, 308, 180, 308, 132, 308, 136, 308, 126, 308, 121, 308, 133, 308, 138, 308, 167, 308, 188, 308, 140, 308, 163, 308, 175, 308, 131, 308, 183, 308, 170, 308, 165, 308, 172, 308, 139, 308, 179, 308, 141, 308, 186, 308, 181, 308, 171, 308, 176, 308, 185, 308, 178, 308, 169, 308, 187, 308, 177, 308, 184, 308, 189, 308, 190, 308, 191, 308, 192, 308, 193, 308, 195, 308, 196, 308, 197, 308, 198, 308, 199, 308, 200, 308, 204, 308, 205, 308, 220, 308, 221, 308, 222, 308, 223, 308, 225, 308, 226, 308],
	#[65535, 291, 18, 291, 20, 291, 15, 291, 28, 291, 14, 291, 38, 291, 23, 291, 13, 291, 11, 291, 9, 291, 22, 291, 4, 291, 31, 291, 21, 291, 19, 291, 61, 291, 7, 291, 95, 291, 12, 291, 26, 291, 0, 291, 1, 291, 27, 291, 17, 291, 42, 291, 43, 291, 41, 291, 34, 291, 2, 291, 37, 291, 40, 291, 25, 291, 33, 291, 111, 291, 91, 291, 57, 291, 10, 291, 32, 291, 87, 291, 85, 291, 58, 291, 59, 291, 82, 291, 63, 291, 24, 291, 74, 291, 108, 291, 93, 291, 76, 291, 60, 291, 86, 291, 110, 291, 72, 291, 88, 291, 116, 291, 101, 291, 84, 291, 75, 291, 94, 291, 29, 291, 125, 291, 90, 291, 124, 291, 114, 291, 92, 291, 83, 291, 102, 291, 89, 291, 99, 291, 77, 291, 119, 291, 117, 291, 100, 291, 97, 291, 16, 291, 106, 291, 127, 291, 130, 291, 122, 291, 103, 291, 96, 291, 112, 291, 107, 291, 123, 291, 104, 291, 115, 291, 120, 291, 105, 291, 137, 291, 109, 291, 113, 291, 121, 291, 163, 291, 175, 291, 118, 291, 183, 291, 126, 291, 138, 291, 129, 291, 164, 291, 131, 291, 132, 291, 133, 291, 173, 291, 136, 291, 170, 291, 167, 291, 139, 291, 140, 291, 141, 291, 166, 291, 174, 291, 188, 291, 172, 291, 185, 291, 171, 291, 182, 291, 176, 291, 181, 291, 180, 291, 186, 291, 165, 291, 179, 291, 178, 291, 169, 291, 177, 291, 187, 291, 192, 291, 184, 291, 189, 291, 190, 291, 191, 291, 193, 291, 195, 291, 196, 291, 197, 291, 198, 291, 199, 291, 200, 291, 204, 291, 205, 291, 220, 291, 221, 291, 222, 291, 223, 291, 225, 291, 226, 291],
	#[65535, 292, 7, 292, 1, 292, 14, 292, 11, 292, 0, 292, 16, 292, 2, 292, 4, 292, 19, 292, 18, 292, 9, 292, 10, 292, 25, 292, 12, 292, 13, 292, 15, 292, 17, 292, 20, 292, 21, 292, 22, 292, 23, 292, 24, 292, 26, 292, 27, 292, 28, 292, 29, 292, 31, 292, 32, 292, 33, 292, 34, 292, 37, 292, 38, 292, 40, 292, 41, 292, 42, 292, 43, 292, 57, 292, 58, 292, 59, 292, 60, 292, 61, 292, 63, 292, 82, 292, 72, 292, 74, 292, 75, 292, 76, 292, 77, 292, 101, 292, 83, 292, 84, 292, 85, 292, 86, 292, 87, 292, 88, 292, 89, 292, 90, 292, 91, 292, 92, 292, 93, 292, 94, 292, 95, 292, 96, 292, 97, 292, 99, 292, 100, 292, 102, 292, 103, 292, 104, 292, 105, 292, 106, 292, 107, 292, 108, 292, 109, 292, 110, 292, 111, 292, 112, 292, 113, 292, 114, 292, 115, 292, 116, 292, 117, 292, 118, 292, 119, 292, 120, 292, 121, 292, 122, 292, 123, 292, 124, 292, 125, 292, 126, 292, 127, 292, 129, 292, 130, 292, 131, 292, 132, 292, 133, 292, 136, 292, 137, 292, 138, 292, 139, 292, 140, 292, 141, 292, 163, 292, 164, 292, 165, 292, 166, 292, 167, 292, 169, 292, 170, 292, 171, 292, 172, 292, 173, 292, 174, 292, 175, 292, 176, 292, 177, 292, 178, 292, 179, 292, 180, 292, 181, 292, 182, 292, 183, 292, 184, 292, 185, 292, 186, 292, 187, 292, 188, 292, 189, 292, 190, 292, 191, 292, 192, 292, 193, 292, 195, 292, 196, 292, 197, 292, 198, 292, 199, 292, 200, 292, 204, 292, 205, 292, 220, 292, 221, 292, 222, 292, 223, 292, 225, 292, 226, 292],
	#[65535, 299, 31, 299, 23, 299, 10, 299, 2, 299, 7, 299, 37, 299, 15, 299, 1, 299, 27, 299, 14, 299, 19, 299, 11, 299, 0, 299, 4, 299, 16, 299, 18, 299, 9, 299, 29, 299, 25, 299, 12, 299, 13, 299, 26, 299, 17, 299, 20, 299, 21, 299, 22, 299, 34, 299, 24, 299, 33, 299, 32, 299, 28, 299, 38, 299, 40, 299, 41, 299, 42, 299, 43, 299, 57, 299, 58, 299, 59, 299, 60, 299, 61, 299, 63, 299, 84, 299, 82, 299, 72, 299, 74, 299, 75, 299, 76, 299, 77, 299, 101, 299, 83, 299, 85, 299, 86, 299, 87, 299, 88, 299, 89, 299, 90, 299, 91, 299, 92, 299, 93, 299, 94, 299, 95, 299, 96, 299, 97, 299, 99, 299, 100, 299, 102, 299, 103, 299, 104, 299, 105, 299, 106, 299, 107, 299, 108, 299, 109, 299, 110, 299, 111, 299, 112, 299, 113, 299, 114, 299, 115, 299, 116, 299, 117, 299, 118, 299, 119, 299, 120, 299, 121, 299, 122, 299, 123, 299, 124, 299, 125, 299, 126, 299, 127, 299, 129, 299, 130, 299, 131, 299, 132, 299, 133, 299, 136, 299, 137, 299, 138, 299, 139, 299, 140, 299, 141, 299, 163, 299, 164, 299, 165, 299, 166, 299, 167, 299, 169, 299, 170, 299, 171, 299, 172, 299, 173, 299, 174, 299, 175, 299, 176, 299, 177, 299, 178, 299, 179, 299, 180, 299, 181, 299, 182, 299, 183, 299, 184, 299, 185, 299, 186, 299, 187, 299, 188, 299, 189, 299, 190, 299, 191, 299, 192, 299, 193, 299, 195, 299, 196, 299, 197, 299, 198, 299, 199, 299, 200, 299, 204, 299, 205, 299, 220, 299, 221, 299, 222, 299, 223, 299, 225, 299, 226, 299],
	#[38, -195],
	#[43, -196],
	#[65535, 306, 31, 306, 23, 306, 10, 306, 2, 306, 7, 306, 37, 306, 21, 306, 15, 306, 1, 306, 27, 306, 14, 306, 9, 306, 11, 306, 0, 306, 4, 306, 19, 306, 18, 306, 24, 306, 29, 306, 16, 306, 12, 306, 13, 306, 26, 306, 17, 306, 20, 306, 22, 306, 34, 306, 25, 306, 33, 306, 32, 306, 28, 306, 38, 306, 40, 306, 41, 306, 42, 306, 43, 306, 57, 306, 58, 306, 59, 306, 60, 306, 61, 306, 63, 306, 84, 306, 82, 306, 72, 306, 74, 306, 75, 306, 76, 306, 77, 306, 101, 306, 83, 306, 85, 306, 86, 306, 87, 306, 88, 306, 89, 306, 90, 306, 91, 306, 92, 306, 93, 306, 94, 306, 95, 306, 96, 306, 97, 306, 99, 306, 100, 306, 102, 306, 103, 306, 104, 306, 105, 306, 106, 306, 107, 306, 108, 306, 109, 306, 110, 306, 111, 306, 112, 306, 113, 306, 114, 306, 115, 306, 116, 306, 117, 306, 118, 306, 119, 306, 120, 306, 121, 306, 122, 306, 123, 306, 124, 306, 125, 306, 126, 306, 127, 306, 129, 306, 130, 306, 131, 306, 132, 306, 133, 306, 136, 306, 137, 306, 138, 306, 139, 306, 140, 306, 141, 306, 163, 306, 164, 306, 165, 306, 166, 306, 167, 306, 169, 306, 170, 306, 171, 306, 172, 306, 173, 306, 174, 306, 175, 306, 176, 306, 177, 306, 178, 306, 179, 306, 180, 306, 181, 306, 182, 306, 183, 306, 184, 306, 185, 306, 186, 306, 187, 306, 188, 306, 189, 306, 190, 306, 191, 306, 192, 306, 193, 306, 195, 306, 196, 306, 197, 306, 198, 306, 199, 306, 200, 306, 204, 306, 205, 306, 220, 306, 221, 306, 222, 306, 223, 306, 225, 306, 226, 306],
	#[13, -149, 32, -199],
	#[65535, 94, 15, 94, 13, 94, 20, 94, 14, 94, 1, 94, 31, 94, 12, 94, 58, 94, 22, 94, 19, 94, 10, 94, 4, 94, 9, 94, 2, 94, 27, 94, 76, 94, 43, 94, 18, 94, 85, 94, 38, 94, 23, 94, 84, 94, 11, 94, 32, 94, 26, 94, 87, 94, 25, 94, 33, 94, 77, 94, 21, 94, 63, 94, 0, 94, 7, 94, 95, 94, 75, 94, 24, 94, 127, 94, 29, 94, 106, 94, 17, 94, 42, 94, 103, 94, 16, 94, 34, 94, 59, 94, 37, 94, 92, 94, 82, 94, 60, 94, 111, 94, 28, 94, 57, 94, 101, 94, 72, 94, 100, 94, 104, 94, 41, 94, 119, 94, 99, 94, 102, 94, 109, 94, 40, 94, 108, 94, 93, 94, 88, 94, 133, 94, 86, 94, 110, 94, 112, 94, 89, 94, 116, 94, 191, 94, 74, 94, 188, 94, 118, 94, 122, 94, 97, 94, 124, 94, 114, 94, 121, 94, 83, 94, 126, 94, 164, 94, 132, 94, 129, 94, 91, 94, 141, 94, 113, 94, 130, 94, 137, 94, 190, 94, 123, 94, 115, 94, 105, 94, 90, 94, 136, 94, 189, 94, 138, 94, 204, 94, 107, 94, 163, 94, 117, 94, 131, 94, 125, 94, 170, 94, 165, 94, 167, 94, 139, 94, 140, 94, 169, 94, 166, 94, 193, 94, 205, 94, 195, 94, 192, 94, 196, 94, 197, 94, 198, 94, 199, 94, 200, 94, 220, 94, 221, 94, 222, 94, 223, 94, 225, 94, 226, 94],
	#[65535, 93, 31, 93, 7, 93, 23, 93, 2, 93, 10, 93, 34, 93, 37, 93, 9, 93, 84, 93, 21, 93, 15, 93, 1, 93, 27, 93, 14, 93, 19, 93, 12, 93, 17, 93, 11, 93, 0, 93, 4, 93, 33, 93, 16, 93, 18, 93, 24, 93, 29, 93, 25, 93, 13, 93, 26, 93, 20, 93, 22, 93, 32, 93, 28, 93, 38, 93, 40, 93, 41, 93, 42, 93, 43, 93, 57, 93, 58, 93, 59, 93, 60, 93, 63, 93, 82, 93, 72, 93, 74, 93, 75, 93, 76, 93, 77, 93, 101, 93, 83, 93, 85, 93, 86, 93, 87, 93, 88, 93, 89, 93, 90, 93, 91, 93, 92, 93, 93, 93, 95, 93, 97, 93, 99, 93, 100, 93, 102, 93, 103, 93, 104, 93, 105, 93, 106, 93, 107, 93, 108, 93, 109, 93, 110, 93, 111, 93, 112, 93, 113, 93, 114, 93, 115, 93, 116, 93, 117, 93, 118, 93, 119, 93, 121, 93, 122, 93, 123, 93, 124, 93, 125, 93, 126, 93, 127, 93, 129, 93, 130, 93, 131, 93, 132, 93, 133, 93, 136, 93, 137, 93, 138, 93, 139, 93, 140, 93, 141, 93, 163, 93, 164, 93, 165, 93, 166, 93, 167, 93, 169, 93, 170, 93, 188, 93, 189, 93, 190, 93, 191, 93, 192, 93, 193, 93, 195, 93, 196, 93, 197, 93, 198, 93, 199, 93, 200, 93, 204, 93, 205, 93, 220, 93, 221, 93, 222, 93, 223, 93, 225, 93, 226, 93],
	#[39, -201],
	#[31, -128, 33, -139, 7, -13, 11, -127, 36, -129, 35, -123, 68, -133, 67, -137, 63, -121, 64, -141, 65, -122, 66, -126, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[12, -203, 37, 105, 42, -159, 72, -161],
	#[65535, 91, 0, 91, 1, 91, 2, 91, 4, 91, 7, 91, 9, 91, 10, 91, 11, 91, 12, 91, 13, 91, 14, 91, 15, 91, 16, 91, 17, 91, 18, 91, 19, 91, 20, 91, 21, 91, 22, 91, 23, 91, 24, 91, 25, 91, 26, 91, 27, 91, 28, 91, 29, 91, 31, 91, 32, 91, 33, 91, 34, 91, 37, 91, 38, 91, 40, 91, 41, 91, 42, 91, 43, 91, 57, 91, 58, 91, 59, 91, 60, 91, 63, 91, 72, 91, 74, 91, 75, 91, 76, 91, 77, 91, 82, 91, 83, 91, 84, 91, 85, 91, 86, 91, 87, 91, 88, 91, 89, 91, 90, 91, 91, 91, 92, 91, 93, 91, 95, 91, 97, 91, 99, 91, 100, 91, 101, 91, 102, 91, 103, 91, 104, 91, 105, 91, 106, 91, 107, 91, 108, 91, 109, 91, 110, 91, 111, 91, 112, 91, 113, 91, 114, 91, 115, 91, 116, 91, 117, 91, 118, 91, 119, 91, 121, 91, 122, 91, 123, 91, 124, 91, 125, 91, 126, 91, 127, 91, 129, 91, 130, 91, 131, 91, 132, 91, 133, 91, 136, 91, 137, 91, 138, 91, 139, 91, 140, 91, 141, 91, 163, 91, 164, 91, 165, 91, 166, 91, 167, 91, 169, 91, 170, 91, 188, 91, 189, 91, 190, 91, 191, 91, 192, 91, 193, 91, 195, 91, 196, 91, 197, 91, 198, 91, 199, 91, 200, 91, 204, 91, 205, 91, 220, 91, 221, 91, 222, 91, 223, 91, 225, 91, 226, 91],
	#[7, -13, 11, -127, 31, -128, 33, -139, 63, -121, 35, -123, 36, -129, 67, -137, 71, -130, 70, -140, 69, -136, 68, -133, 98, -206, 66, -126, 65, -122, 73, -138, 64, -141, 118, -36],
	#[13, 164, 93, -211, 43, 164, 109, -213, 108, -208, 60, -209, 220, -207, 221, -210, 222, -212, 223, -214],
	#[65535, 161, 43, 161],
	#[65535, 411, 87, 411, 2, 411, 7, 411, 10, 411, 11, 411, 13, 411, 84, 411, 59, 411, 33, 411, 43, 411, 31, 411, 119, 411, 102, 411, 63, 411, 93, 411, 95, 411, 38, 411, 86, 411, 57, 411, 103, 411, 90, 411, 58, 411, 88, 411, 118, 411, 115, 411, 91, 411, 111, 411, 89, 411, 97, 411, 105, 411, 126, 411, 92, 411, 114, 411, 60, 411, 106, 411, 104, 411, 85, 411, 99, 411, 107, 411, 100, 411, 101, 411, 108, 411, 109, 411, 110, 411, 122, 411, 112, 411, 113, 411, 133, 411, 132, 411, 116, 411, 117, 411, 131, 411, 130, 411, 121, 411, 129, 411, 123, 411, 124, 411, 125, 411, 136, 411, 127, 411, 137, 411, 138, 411, 139, 411, 140, 411, 141, 411, 220, 411, 221, 411, 222, 411, 223, 411, 226, 411],
	#[65535, 409, 7, 409, 13, 409, 43, 409, 93, 409, 57, 409, 58, 409, 59, 409, 60, 409, 109, 409, 108, 409, 118, 409, 220, 409, 221, 409, 222, 409, 223, 409],
	#[65535, 412, 7, 412, 84, 412, 2, 412, 87, 412, 31, 412, 92, 412, 10, 412, 11, 412, 13, 412, 58, 412, 103, 412, 33, 412, 57, 412, 85, 412, 43, 412, 102, 412, 63, 412, 93, 412, 95, 412, 38, 412, 91, 412, 86, 412, 97, 412, 88, 412, 90, 412, 89, 412, 105, 412, 59, 412, 60, 412, 106, 412, 104, 412, 99, 412, 107, 412, 100, 412, 101, 412, 108, 412, 109, 412, 110, 412, 111, 412, 112, 412, 113, 412, 114, 412, 115, 412, 116, 412, 117, 412, 118, 412, 119, 412, 121, 412, 122, 412, 123, 412, 124, 412, 125, 412, 126, 412, 127, 412, 129, 412, 130, 412, 131, 412, 132, 412, 133, 412, 136, 412, 137, 412, 138, 412, 139, 412, 140, 412, 141, 412, 220, 412, 221, 412, 222, 412, 223, 412, 226, 412],
	#[65535, 413, 31, 413, 86, 413, 95, 413, 85, 413, 92, 413, 2, 413, 93, 413, 84, 413, 111, 413, 38, 413, 58, 413, 87, 413, 91, 413, 13, 413, 63, 413, 7, 413, 118, 413, 59, 413, 99, 413, 102, 413, 10, 413, 33, 413, 60, 413, 88, 413, 100, 413, 97, 413, 133, 413, 57, 413, 90, 413, 119, 413, 43, 413, 124, 413, 101, 413, 106, 413, 113, 413, 104, 413, 129, 413, 11, 413, 89, 413, 116, 413, 103, 413, 110, 413, 105, 413, 108, 413, 112, 413, 132, 413, 130, 413, 137, 413, 109, 413, 126, 413, 114, 413, 115, 413, 121, 413, 107, 413, 123, 413, 122, 413, 117, 413, 131, 413, 125, 413, 136, 413, 127, 413, 138, 413, 139, 413, 140, 413, 141, 413, 220, 413, 221, 413, 222, 413, 223, 413, 226, 413],
	#[38, -216],
	#[65535, 414, 13, 414, 2, 414, 31, 414, 7, 414, 60, 414, 10, 414, 11, 414, 33, 414, 58, 414, 103, 414, 86, 414, 110, 414, 87, 414, 43, 414, 59, 414, 63, 414, 93, 414, 95, 414, 38, 414, 91, 414, 101, 414, 85, 414, 116, 414, 84, 414, 106, 414, 109, 414, 111, 414, 89, 414, 97, 414, 102, 414, 57, 414, 88, 414, 104, 414, 99, 414, 119, 414, 112, 414, 105, 414, 113, 414, 100, 414, 90, 414, 123, 414, 121, 414, 124, 414, 108, 414, 92, 414, 107, 414, 115, 414, 114, 414, 122, 414, 117, 414, 118, 414, 125, 414, 126, 414, 127, 414, 129, 414, 130, 414, 131, 414, 132, 414, 133, 414, 136, 414, 137, 414, 138, 414, 139, 414, 140, 414, 141, 414, 220, 414, 221, 414, 222, 414, 223, 414, 226, 414],
	#[65535, 410, 7, 410, 13, 410, 43, 410, 57, 410, 58, 410, 59, 410, 60, 410, 93, 410, 108, 410, 109, 410, 118, 410, 220, 410, 221, 410, 222, 410, 223, 410],
	#[65535, 415, 31, 415, 2, 415, 13, 415, 7, 415, 10, 415, 11, 415, 33, 415, 38, 415, 43, 415, 84, 415, 57, 415, 58, 415, 59, 415, 60, 415, 63, 415, 85, 415, 86, 415, 87, 415, 88, 415, 89, 415, 90, 415, 91, 415, 92, 415, 93, 415, 95, 415, 97, 415, 99, 415, 100, 415, 101, 415, 102, 415, 103, 415, 104, 415, 105, 415, 106, 415, 107, 415, 108, 415, 109, 415, 110, 415, 111, 415, 112, 415, 113, 415, 114, 415, 115, 415, 116, 415, 117, 415, 118, 415, 119, 415, 121, 415, 122, 415, 123, 415, 124, 415, 125, 415, 126, 415, 127, 415, 129, 415, 130, 415, 131, 415, 132, 415, 133, 415, 136, 415, 137, 415, 138, 415, 139, 415, 140, 415, 141, 415, 220, 415, 221, 415, 222, 415, 223, 415, 226, 415],
	#[65535, 112, 31, 112, 2, 112, 7, 112, 10, 112, 11, 112, 13, 112, 33, 112, 38, 112, 43, 112, 60, 112, 63, 112, 84, 112, 101, 112, 85, 112, 86, 112, 87, 112, 88, 112, 89, 112, 90, 112, 91, 112, 92, 112, 93, 112, 95, 112, 97, 112, 99, 112, 100, 112, 102, 112, 103, 112, 104, 112, 105, 112, 106, 112, 107, 112, 108, 112, 109, 112, 110, 112, 111, 112, 112, 112, 113, 112, 114, 112, 115, 112, 116, 112, 117, 112, 118, 112, 119, 112, 121, 112, 122, 112, 123, 112, 124, 112, 125, 112, 126, 112, 127, 112, 129, 112, 130, 112, 131, 112, 132, 112, 133, 112, 136, 112, 137, 112, 138, 112, 139, 112, 140, 112, 141, 112, 220, 112, 221, 112, 222, 112, 223, 112, 226, 112],
	#[65535, 416, 2, 416, 7, 416, 10, 416, 11, 416, 13, 416, 31, 416, 33, 416, 38, 416, 43, 416, 57, 416, 58, 416, 59, 416, 60, 416, 63, 416, 84, 416, 85, 416, 86, 416, 87, 416, 88, 416, 89, 416, 90, 416, 91, 416, 92, 416, 93, 416, 95, 416, 97, 416, 99, 416, 100, 416, 101, 416, 102, 416, 103, 416, 104, 416, 105, 416, 106, 416, 107, 416, 108, 416, 109, 416, 110, 416, 111, 416, 112, 416, 113, 416, 114, 416, 115, 416, 116, 416, 117, 416, 118, 416, 119, 416, 121, 416, 122, 416, 123, 416, 124, 416, 125, 416, 126, 416, 127, 416, 129, 416, 130, 416, 131, 416, 132, 416, 133, 416, 136, 416, 137, 416, 138, 416, 139, 416, 140, 416, 141, 416, 220, 416, 221, 416, 222, 416, 223, 416, 226, 416],
	#[65535, 211, 4, 211, 2, 211, 22, 211, 23, 211, 13, 211, 1, 211, 28, 211, 31, 211, 14, 211, 9, 211, 17, 211, 86, 211, 7, 211, 0, 211, 16, 211, 20, 211, 29, 211, 19, 211, 10, 211, 15, 211, 75, 211, 11, 211, 37, 211, 25, 211, 77, 211, 102, 211, 42, 211, 57, 211, 32, 211, 38, 211, 33, 211, 87, 211, 110, 211, 18, 211, 40, 211, 63, 211, 12, 211, 34, 211, 59, 211, 82, 211, 74, 211, 103, 211, 27, 211, 24, 211, 90, 211, 60, 211, 26, 211, 93, 211, 91, 211, 58, 211, 43, 211, 83, 211, 41, 211, 72, 211, 95, 211, 85, 211, 127, 211, 21, 211, 108, 211, 92, 211, 97, 211, 105, 211, 88, 211, 101, 211, 99, 211, 116, 211, 118, 211, 84, 211, 109, 211, 111, 211, 89, 211, 107, 211, 122, 211, 126, 211, 121, 211, 76, 211, 117, 211, 119, 211, 106, 211, 115, 211, 113, 211, 100, 211, 125, 211, 104, 211, 114, 211, 123, 211, 138, 211, 124, 211, 166, 211, 130, 211, 112, 211, 132, 211, 137, 211, 141, 211, 163, 211, 175, 211, 131, 211, 183, 211, 136, 211, 165, 211, 174, 211, 129, 211, 164, 211, 186, 211, 189, 211, 133, 211, 173, 211, 182, 211, 170, 211, 167, 211, 139, 211, 140, 211, 169, 211, 177, 211, 181, 211, 190, 211, 188, 211, 172, 211, 178, 211, 171, 211, 184, 211, 176, 211, 191, 211, 180, 211, 185, 211, 193, 211, 179, 211, 205, 211, 195, 211, 204, 211, 187, 211, 192, 211, 196, 211, 197, 211, 198, 211, 199, 211, 200, 211, 220, 211, 221, 211, 222, 211, 223, 211, 225, 211, 226, 211],
	#[31, 108, 1, 108, 7, 108, 2, 108, 15, 108, 4, 108, 10, 108, 14, 108, 23, 108, 0, 108, 26, 108, 18, 108, 9, 108, 29, 108, 11, 108, 12, 108, 13, 108, 27, 108, 16, 108, 17, 108, 37, 108, 19, 108, 20, 108, 21, 108, 22, 108, 34, 108, 24, 108, 25, 108, 33, 108, 32, 108, 28, 108, 38, 108, 40, 108, 41, 108, 42, 108, 43, 108, 57, -192, 58, -175, 59, -179, 60, 108, 63, 108, 84, 108, 72, 108, 74, 108, 75, 108, 76, 108, 77, 108, 82, 108, 83, 108, 85, 108, 86, 108, 87, 108, 88, 108, 89, 108, 90, 108, 91, 108, 92, 108, 93, 108, 95, 108, 97, 108, 99, 108, 100, 108, 101, 108, 102, 108, 103, 108, 104, 108, 105, 108, 106, 108, 107, 108, 108, 108, 109, 108, 110, 108, 111, 108, 112, 108, 113, 108, 114, 108, 115, 108, 116, 108, 117, 108, 118, 108, 119, 108, 121, 108, 122, 108, 123, 108, 124, 108, 125, 108, 126, 108, 127, 108, 129, 108, 130, 108, 131, 108, 132, 108, 133, 108, 136, 108, 137, 108, 138, 108, 139, 108, 140, 108, 141, 108, 163, 108, 164, 108, 165, 108, 166, 108, 167, 108, 169, 108, 170, 108, 171, -180, 172, -191, 173, -182, 174, -184, 175, -178, 176, -185, 177, -193, 178, -176, 179, -181, 180, -183, 181, -186, 182, -189, 183, -174, 184, -177, 185, -187, 186, -190, 187, -173, 188, 108, 189, 108, 190, 108, 191, 108, 192, 108, 193, 108, 195, 108, 196, 108, 197, 108, 198, 108, 199, 108, 200, 108, 204, 108, 205, 108, 220, 108, 221, 108, 222, 108, 223, 108, 225, 108, 226, 108],
	#[65535, 95, 0, 95, 1, 95, 2, 95, 4, 95, 7, 95, 9, 95, 10, 95, 11, 95, 12, 95, 13, 95, 14, 95, 15, 95, 16, 95, 17, 95, 18, 95, 19, 95, 20, 95, 21, 95, 22, 95, 23, 95, 24, 95, 25, 95, 26, 95, 27, 95, 28, 95, 29, 95, 31, 95, 32, 95, 33, 95, 34, 95, 37, 95, 38, 95, 40, 95, 41, 95, 42, 95, 43, 95, 57, 95, 58, 95, 59, 95, 60, 95, 63, 95, 72, 95, 74, 95, 75, 95, 76, 95, 77, 95, 82, 95, 83, 95, 84, 95, 85, 95, 86, 95, 87, 95, 88, 95, 89, 95, 90, 95, 91, 95, 92, 95, 93, 95, 95, 95, 97, 95, 99, 95, 100, 95, 101, 95, 102, 95, 103, 95, 104, 95, 105, 95, 106, 95, 107, 95, 108, 95, 109, 95, 110, 95, 111, 95, 112, 95, 113, 95, 114, 95, 115, 95, 116, 95, 117, 95, 118, 95, 119, 95, 121, 95, 122, 95, 123, 95, 124, 95, 125, 95, 126, 95, 127, 95, 129, 95, 130, 95, 131, 95, 132, 95, 133, 95, 136, 95, 137, 95, 138, 95, 139, 95, 140, 95, 141, 95, 163, 95, 164, 95, 165, 95, 166, 95, 167, 95, 169, 95, 170, 95, 188, 95, 189, 95, 190, 95, 191, 95, 192, 95, 193, 95, 195, 95, 196, 95, 197, 95, 198, 95, 199, 95, 200, 95, 204, 95, 205, 95, 220, 95, 221, 95, 222, 95, 223, 95, 225, 95, 226, 95],
	#[31, -128, 36, -129, 33, -139, 7, -13, 35, -123, 11, -127, 67, -137, 65, -122, 63, -121, 64, -141, 66, -126, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[34, -222, 37, 105, 42, -159, 72, -161],
	#[65535, 92, 31, 92, 7, 92, 29, 92, 14, 92, 15, 92, 34, 92, 2, 92, 9, 92, 21, 92, 1, 92, 10, 92, 23, 92, 12, 92, 11, 92, 0, 92, 17, 92, 4, 92, 19, 92, 18, 92, 24, 92, 25, 92, 13, 92, 27, 92, 26, 92, 16, 92, 37, 92, 33, 92, 20, 92, 22, 92, 32, 92, 28, 92, 38, 92, 40, 92, 41, 92, 42, 92, 43, 92, 84, 92, 57, 92, 58, 92, 59, 92, 60, 92, 63, 92, 72, 92, 74, 92, 75, 92, 76, 92, 77, 92, 82, 92, 83, 92, 85, 92, 86, 92, 87, 92, 88, 92, 89, 92, 90, 92, 91, 92, 92, 92, 93, 92, 95, 92, 97, 92, 127, 92, 99, 92, 100, 92, 101, 92, 102, 92, 103, 92, 104, 92, 105, 92, 106, 92, 107, 92, 108, 92, 109, 92, 110, 92, 111, 92, 112, 92, 113, 92, 114, 92, 115, 92, 116, 92, 117, 92, 118, 92, 119, 92, 121, 92, 122, 92, 123, 92, 124, 92, 125, 92, 126, 92, 138, 92, 129, 92, 130, 92, 131, 92, 132, 92, 133, 92, 136, 92, 137, 92, 167, 92, 139, 92, 140, 92, 141, 92, 163, 92, 164, 92, 165, 92, 166, 92, 169, 92, 170, 92, 188, 92, 189, 92, 190, 92, 191, 92, 192, 92, 193, 92, 195, 92, 196, 92, 197, 92, 198, 92, 199, 92, 200, 92, 204, 92, 205, 92, 220, 92, 221, 92, 222, 92, 223, 92, 225, 92, 226, 92],
	#[65535, 156, 1, 156, 10, 156, 14, 156, 23, 156, 0, 156, 4, 156, 19, 156, 7, 156, 9, 156, 29, 156, 11, 156, 12, 156, 13, 156, 27, 156, 15, 156, 16, 156, 17, 156, 18, 156, 36, 156, 20, 156, 21, 156, 22, 156, 34, 156, 24, 156, 25, 156, 26, 156, 32, 156, 28, 156, 40, 156, 41, 156, 42, 156, 43, 156, 57, 156, 58, 156, 59, 156, 60, 156, 62, 156, 63, 156, 74, 156, 75, 156, 76, 156, 77, 156, 82, 156, 83, 156, 93, 156, 95, 156, 96, 156, 97, 156, 99, 156, 100, 156, 101, 156, 102, 156, 103, 156, 104, 156, 105, 156, 106, 156, 107, 156, 108, 156, 109, 156, 110, 156, 111, 156, 112, 156, 113, 156, 114, 156, 115, 156, 116, 156, 117, 156, 118, 156, 119, 156, 121, 156, 122, 156, 123, 156, 124, 156, 125, 156, 126, 156, 127, 156, 129, 156, 130, 156, 131, 156, 132, 156, 133, 156, 136, 156, 137, 156, 138, 156, 139, 156, 140, 156, 141, 156, 162, 156, 163, 156, 164, 156, 165, 156, 166, 156, 167, 156, 168, 156, 169, 156, 170, 156, 188, 156, 189, 156, 190, 156, 192, 156, 193, 156, 195, 156, 196, 156, 197, 156, 198, 156, 199, 156, 200, 156, 201, 156, 202, 156, 203, 156, 204, 156, 205, 156, 206, 156, 207, 156, 208, 156, 209, 156, 210, 156, 211, 156, 212, 156, 220, 156, 221, 156, 222, 156, 223, 156, 225, 156, 226, 156],
	#[65535, 397, 7, 397, 11, 397, 12, 397, 13, 397, 32, 397, 36, 397, 42, 397, 43, 397, 57, 397, 58, 397, 59, 397, 60, 397, 62, 397, 63, 397, 93, 397, 95, 397, 96, 397, 97, 397, 99, 397, 100, 397, 101, 397, 102, 397, 103, 397, 104, 397, 105, 397, 106, 397, 107, 397, 108, 397, 109, 397, 110, 397, 111, 397, 112, 397, 113, 397, 114, 397, 115, 397, 116, 397, 117, 397, 118, 397, 121, 397, 122, 397, 123, 397, 124, 397, 125, 397, 126, 397, 127, 397, 129, 397, 130, 397, 131, 397, 132, 397, 133, 397, 136, 397, 137, 397, 138, 397, 139, 397, 140, 397, 141, 397, 162, 397, 163, 397, 164, 397, 165, 397, 166, 397, 167, 397, 168, 397, 169, 397, 170, 397, 188, 397, 189, 397, 190, 397, 192, 397, 193, 397, 195, 397, 196, 397, 197, 397, 198, 397, 199, 397, 200, 397, 201, 397, 202, 397, 203, 397, 204, 397, 205, 397, 206, 397, 207, 397, 208, 397, 209, 397, 210, 397, 211, 397, 212, 397, 220, 397, 221, 397, 222, 397, 223, 397],
	#[65535, 383, 32, 383, 36, 383, 42, 383, 43, 383, 57, 383, 58, 383, 59, 383, 60, 383, 62, 383, 63, 383, 93, 383, 95, 383, 96, 383, 97, 383, 99, 383, 100, 383, 101, 383, 102, 383, 103, 383, 104, 383, 105, 383, 106, 383, 107, 383, 108, 383, 109, 383, 110, 383, 111, 383, 112, 383, 113, 383, 114, 383, 115, 383, 116, 383, 117, 383, 118, 383, 121, 383, 122, 383, 123, 383, 124, 383, 125, 383, 126, 383, 127, 383, 129, 383, 130, 383, 131, 383, 132, 383, 133, 383, 136, 383, 137, 383, 138, 383, 139, 383, 140, 383, 141, 383, 162, 383, 163, 383, 164, 383, 165, 383, 166, 383, 167, 383, 168, 383, 169, 383, 170, 383, 188, 383, 189, 383, 190, 383, 192, 383, 193, 383, 195, 383, 196, 383, 197, 383, 198, 383, 199, 383, 200, 383, 201, 383, 202, 383, 203, 383, 204, 383, 205, 383, 206, 383, 207, 383, 208, 383, 209, 383, 210, 383, 211, 383, 212, 383, 220, 383, 221, 383, 222, 383, 223, 383, 7, 383, 11, 383, 12, 383, 13, 383],
	#[42, -418],
	#[7, 400, 184, -445, 219, -446],
	#[7, -13, 11, -127, 31, -128, 32, -443, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[42, -406],
	#[65535, 399, 63, 399, 103, 399, 93, 399, 59, 399, 11, 399, 100, 399, 7, 399, 95, 399, 36, 399, 57, 399, 42, 399, 108, 399, 13, 399, 118, 399, 105, 399, 58, 399, 114, 399, 12, 399, 116, 399, 102, 399, 126, 399, 96, 399, 32, 399, 101, 399, 99, 399, 62, 399, 124, 399, 110, 399, 104, 399, 60, 399, 109, 399, 107, 399, 43, 399, 122, 399, 123, 399, 133, 399, 127, 399, 188, 399, 106, 399, 113, 399, 129, 399, 97, 399, 164, 399, 111, 399, 136, 399, 137, 399, 115, 399, 112, 399, 130, 399, 132, 399, 121, 399, 163, 399, 117, 399, 131, 399, 125, 399, 170, 399, 165, 399, 138, 399, 139, 399, 140, 399, 141, 399, 166, 399, 162, 399, 167, 399, 168, 399, 169, 399, 189, 399, 190, 399, 192, 399, 193, 399, 195, 399, 196, 399, 197, 399, 198, 399, 199, 399, 200, 399, 201, 399, 202, 399, 203, 399, 204, 399, 205, 399, 206, 399, 207, 399, 208, 399, 209, 399, 210, 399, 211, 399, 212, 399, 220, 399, 221, 399, 222, 399, 223, 399],
	#[42, -395],
	#[65535, 155, 0, 155, 1, 155, 4, 155, 7, 155, 9, 155, 10, 155, 11, 155, 12, 155, 13, 155, 14, 155, 15, 155, 16, 155, 17, 155, 18, 155, 19, 155, 20, 155, 21, 155, 22, 155, 23, 155, 24, 155, 25, 155, 26, 155, 27, 155, 28, 155, 29, 155, 32, 155, 34, 155, 36, 155, 40, 155, 41, 155, 42, 155, 43, 155, 57, 155, 58, 155, 59, 155, 60, 155, 62, 155, 63, 155, 74, 155, 75, 155, 76, 155, 77, 155, 82, 155, 83, 155, 93, 155, 95, 155, 96, 155, 97, 155, 99, 155, 100, 155, 101, 155, 102, 155, 103, 155, 104, 155, 105, 155, 106, 155, 107, 155, 108, 155, 109, 155, 110, 155, 111, 155, 112, 155, 113, 155, 114, 155, 115, 155, 116, 155, 117, 155, 118, 155, 119, 155, 121, 155, 122, 155, 123, 155, 124, 155, 125, 155, 126, 155, 127, 155, 129, 155, 130, 155, 131, 155, 132, 155, 133, 155, 136, 155, 137, 155, 138, 155, 139, 155, 140, 155, 141, 155, 162, 155, 163, 155, 164, 155, 165, 155, 166, 155, 167, 155, 168, 155, 169, 155, 170, 155, 188, 155, 189, 155, 190, 155, 192, 155, 193, 155, 195, 155, 196, 155, 197, 155, 198, 155, 199, 155, 200, 155, 201, 155, 202, 155, 203, 155, 204, 155, 205, 155, 206, 155, 207, 155, 208, 155, 209, 155, 210, 155, 211, 155, 212, 155, 220, 155, 221, 155, 222, 155, 223, 155, 225, 155, 226, 155],
	#[65535, 382, 7, 382, 11, 382, 12, 382, 13, 382, 36, 382, 43, 382, 32, 382, 102, 382, 63, 382, 93, 382, 95, 382, 62, 382, 42, 382, 57, 382, 103, 382, 59, 382, 118, 382, 115, 382, 96, 382, 111, 382, 97, 382, 105, 382, 126, 382, 58, 382, 114, 382, 60, 382, 106, 382, 104, 382, 99, 382, 107, 382, 100, 382, 101, 382, 108, 382, 109, 382, 110, 382, 122, 382, 112, 382, 113, 382, 133, 382, 132, 382, 116, 382, 117, 382, 131, 382, 130, 382, 121, 382, 129, 382, 123, 382, 124, 382, 125, 382, 136, 382, 127, 382, 137, 382, 138, 382, 139, 382, 140, 382, 141, 382, 166, 382, 165, 382, 164, 382, 170, 382, 163, 382, 162, 382, 167, 382, 168, 382, 169, 382, 188, 382, 189, 382, 190, 382, 202, 382, 192, 382, 193, 382, 195, 382, 196, 382, 197, 382, 198, 382, 199, 382, 200, 382, 201, 382, 209, 382, 203, 382, 204, 382, 205, 382, 206, 382, 207, 382, 208, 382, 210, 382, 211, 382, 212, 382, 220, 382, 221, 382, 222, 382, 223, 382],
	#[42, -390],
	#[142, -380, 143, -382, 144, -384, 145, -375, 146, -376, 147, -377, 148, -379, 149, -381, 150, -383, 151, -374],
	#[42, -369],
	#[42, 421, 224, -362],
	#[65535, 158, 14, 158, 13, 158, 22, 158, 7, 158, 1, 158, 9, 158, 4, 158, 10, 158, 17, 158, 16, 158, 23, 158, 12, 158, 18, 158, 11, 158, 0, 158, 20, 158, 43, 158, 63, 158, 19, 158, 32, 158, 24, 158, 29, 158, 28, 158, 34, 158, 41, 158, 27, 158, 15, 158, 62, 158, 93, 158, 25, 158, 36, 158, 21, 158, 57, 158, 75, 158, 40, 158, 26, 158, 82, 158, 127, 158, 74, 158, 42, 158, 60, 158, 103, 158, 59, 158, 58, 158, 100, 158, 107, 158, 83, 158, 77, 158, 76, 158, 99, 158, 95, 158, 96, 158, 97, 158, 101, 158, 102, 158, 114, 158, 104, 158, 105, 158, 106, 158, 112, 158, 108, 158, 109, 158, 110, 158, 111, 158, 113, 158, 115, 158, 116, 158, 117, 158, 118, 158, 119, 158, 121, 158, 122, 158, 123, 158, 124, 158, 125, 158, 126, 158, 138, 158, 129, 158, 130, 158, 131, 158, 132, 158, 133, 158, 136, 158, 137, 158, 167, 158, 139, 158, 140, 158, 141, 158, 162, 158, 163, 158, 164, 158, 165, 158, 166, 158, 190, 158, 168, 158, 169, 158, 170, 158, 195, 158, 193, 158, 192, 158, 188, 158, 189, 158, 196, 158, 197, 158, 198, 158, 199, 158, 200, 158, 201, 158, 202, 158, 203, 158, 204, 158, 205, 158, 206, 158, 207, 158, 208, 158, 209, 158, 210, 158, 211, 158, 212, 158, 220, 158, 221, 158, 222, 158, 223, 158, 225, 158, 226, 158],
	#[42, -401],
	#[65535, 313, 7, 313, 11, 313, 12, 313, 13, 313, 32, 313, 36, 313, 42, 313, 43, 313, 57, 313, 58, 313, 59, 313, 60, 313, 62, 313, 63, 313, 93, 313, 95, 313, 96, 313, 97, 313, 99, 313, 100, 313, 101, 313, 102, 313, 103, 313, 104, 313, 105, 313, 106, 313, 107, 313, 108, 313, 109, 313, 110, 313, 111, 313, 112, 313, 113, 313, 114, 313, 115, 313, 116, 313, 117, 313, 118, 313, 121, 313, 122, 313, 123, 313, 124, 313, 125, 313, 126, 313, 127, 313, 129, 313, 130, 313, 131, 313, 132, 313, 133, 313, 136, 313, 137, 313, 138, 313, 139, 313, 140, 313, 141, 313, 162, 313, 163, 313, 164, 313, 165, 313, 166, 313, 167, 313, 168, 313, 169, 313, 170, 313, 188, 313, 189, 313, 190, 313, 192, 313, 193, 313, 195, 313, 196, 313, 197, 313, 198, 313, 199, 313, 200, 313, 201, 313, 202, 313, 203, 313, 204, 313, 205, 313, 206, 313, 207, 313, 208, 313, 209, 313, 210, 313, 211, 313, 212, 313, 220, 313, 221, 313, 222, 313, 223, 313],
	#[42, -437],
	#[65535, 386, 7, 386, 93, 386, 42, 386, 12, 386, 95, 386, 57, 386, 11, 386, 103, 386, 62, 386, 59, 386, 111, 386, 100, 386, 96, 386, 36, 386, 58, 386, 43, 386, 63, 386, 108, 386, 106, 386, 99, 386, 60, 386, 102, 386, 164, 386, 32, 386, 13, 386, 116, 386, 114, 386, 107, 386, 110, 386, 104, 386, 105, 386, 124, 386, 122, 386, 129, 386, 101, 386, 118, 386, 97, 386, 132, 386, 112, 386, 109, 386, 126, 386, 113, 386, 115, 386, 133, 386, 127, 386, 130, 386, 121, 386, 136, 386, 188, 386, 137, 386, 123, 386, 163, 386, 117, 386, 131, 386, 125, 386, 170, 386, 165, 386, 138, 386, 139, 386, 140, 386, 141, 386, 166, 386, 162, 386, 167, 386, 168, 386, 169, 386, 189, 386, 190, 386, 192, 386, 193, 386, 195, 386, 196, 386, 197, 386, 198, 386, 199, 386, 200, 386, 201, 386, 202, 386, 203, 386, 204, 386, 205, 386, 206, 386, 207, 386, 208, 386, 209, 386, 210, 386, 211, 386, 212, 386, 220, 386, 221, 386, 222, 386, 223, 386],
	#[65535, 157, 0, 157, 1, 157, 4, 157, 7, 157, 9, 157, 10, 157, 11, 157, 12, 157, 13, 157, 14, 157, 15, 157, 16, 157, 17, 157, 18, 157, 19, 157, 20, 157, 21, 157, 22, 157, 23, 157, 24, 157, 25, 157, 26, 157, 27, 157, 28, 157, 29, 157, 32, 157, 34, 157, 36, 157, 40, 157, 41, 157, 42, 157, 43, 157, 57, 157, 58, 157, 59, 157, 60, 157, 62, 157, 63, 157, 74, 157, 75, 157, 76, 157, 77, 157, 82, 157, 83, 157, 93, 157, 95, 157, 96, 157, 97, 157, 99, 157, 100, 157, 101, 157, 102, 157, 103, 157, 104, 157, 105, 157, 106, 157, 107, 157, 108, 157, 109, 157, 110, 157, 111, 157, 112, 157, 113, 157, 114, 157, 115, 157, 116, 157, 117, 157, 118, 157, 119, 157, 121, 157, 122, 157, 123, 157, 124, 157, 125, 157, 126, 157, 127, 157, 129, 157, 130, 157, 131, 157, 132, 157, 133, 157, 136, 157, 137, 157, 138, 157, 139, 157, 140, 157, 141, 157, 162, 157, 163, 157, 164, 157, 165, 157, 166, 157, 167, 157, 168, 157, 169, 157, 170, 157, 188, 157, 189, 157, 190, 157, 192, 157, 193, 157, 195, 157, 196, 157, 197, 157, 198, 157, 199, 157, 200, 157, 201, 157, 202, 157, 203, 157, 204, 157, 205, 157, 206, 157, 207, 157, 208, 157, 209, 157, 210, 157, 211, 157, 212, 157, 220, 157, 221, 157, 222, 157, 223, 157, 225, 157, 226, 157],
	#[65535, 310, 63, 310, 7, 310, 95, 310, 13, 310, 62, 310, 32, 310, 110, 310, 43, 310, 103, 310, 116, 310, 109, 310, 111, 310, 102, 310, 99, 310, 96, 310, 106, 310, 101, 310, 113, 310, 100, 310, 104, 310, 105, 310, 123, 310, 124, 310, 108, 310, 97, 310, 112, 310, 121, 310, 107, 310, 115, 310, 114, 310, 122, 310, 117, 310, 118, 310, 125, 310, 126, 310, 127, 310, 129, 310, 130, 310, 131, 310, 132, 310, 133, 310, 136, 310, 137, 310, 138, 310, 139, 310, 140, 310, 141, 310, 188, 310, 162, 310, 163, 310, 164, 310, 165, 310, 166, 310, 167, 310, 169, 310, 170, 310, 189, 310, 190, 310, 192, 310, 193, 310, 195, 310, 196, 310, 197, 310, 198, 310, 199, 310, 200, 310, 201, 310, 202, 310, 203, 310, 204, 310, 205, 310, 206, 310, 207, 310, 208, 310, 209, 310, 210, 310, 211, 310, 212, 310],
	#[42, -347],
	#[42, -339],
	#[7, -436],
	#[65535, 380, 7, 380, 11, 380, 12, 380, 13, 380, 32, 380, 36, 380, 42, 380, 43, 380, 57, 380, 58, 380, 59, 380, 60, 380, 62, 380, 63, 380, 93, 380, 95, 380, 96, 380, 97, 380, 99, 380, 100, 380, 101, 380, 102, 380, 103, 380, 104, 380, 105, 380, 106, 380, 107, 380, 108, 380, 109, 380, 110, 380, 111, 380, 112, 380, 113, 380, 114, 380, 115, 380, 116, 380, 117, 380, 118, 380, 121, 380, 122, 380, 123, 380, 124, 380, 125, 380, 126, 380, 127, 380, 129, 380, 130, 380, 131, 380, 132, 380, 133, 380, 136, 380, 137, 380, 138, 380, 139, 380, 140, 380, 141, 380, 162, 380, 163, 380, 164, 380, 165, 380, 166, 380, 167, 380, 168, 380, 169, 380, 170, 380, 188, 380, 189, 380, 190, 380, 192, 380, 193, 380, 195, 380, 196, 380, 197, 380, 198, 380, 199, 380, 200, 380, 201, 380, 202, 380, 203, 380, 204, 380, 205, 380, 206, 380, 207, 380, 208, 380, 209, 380, 210, 380, 211, 380, 212, 380, 220, 380, 221, 380, 222, 380, 223, 380],
	#[65535, 314, 11, 314, 63, 314, 7, 314, 36, 314, 109, 314, 58, 314, 93, 314, 42, 314, 57, 314, 32, 314, 99, 314, 62, 314, 110, 314, 13, 314, 12, 314, 95, 314, 116, 314, 59, 314, 103, 314, 60, 314, 111, 314, 106, 314, 100, 314, 105, 314, 96, 314, 112, 314, 107, 314, 124, 314, 108, 314, 113, 314, 102, 314, 115, 314, 97, 314, 123, 314, 101, 314, 114, 314, 43, 314, 104, 314, 121, 314, 122, 314, 117, 314, 118, 314, 125, 314, 126, 314, 127, 314, 129, 314, 130, 314, 131, 314, 132, 314, 133, 314, 136, 314, 137, 314, 138, 314, 139, 314, 140, 314, 141, 314, 162, 314, 163, 314, 164, 314, 165, 314, 166, 314, 167, 314, 168, 314, 169, 314, 170, 314, 192, 314, 188, 314, 189, 314, 190, 314, 202, 314, 193, 314, 195, 314, 196, 314, 197, 314, 198, 314, 199, 314, 200, 314, 201, 314, 209, 314, 203, 314, 204, 314, 205, 314, 206, 314, 207, 314, 208, 314, 210, 314, 211, 314, 212, 314, 220, 314, 221, 314, 222, 314, 223, 314],
	#[63, -121, 7, -13, 11, -127, 12, -434, 67, -137, 36, -129, 35, -123, 70, -140, 31, -128, 33, -139, 64, -141, 71, -130, 66, -126, 69, -136, 68, -133, 73, -138, 65, -122, 118, -36],
	#[7, -317, 31, -318, 38, -319],
	#[65535, 384, 7, 384, 11, 384, 12, 384, 13, 384, 36, 384, 43, 384, 32, 384, 102, 384, 63, 384, 93, 384, 95, 384, 62, 384, 127, 384, 42, 384, 57, 384, 103, 384, 59, 384, 118, 384, 115, 384, 96, 384, 111, 384, 97, 384, 105, 384, 126, 384, 58, 384, 114, 384, 60, 384, 106, 384, 104, 384, 99, 384, 107, 384, 100, 384, 101, 384, 108, 384, 109, 384, 110, 384, 122, 384, 112, 384, 113, 384, 133, 384, 132, 384, 116, 384, 117, 384, 131, 384, 130, 384, 121, 384, 129, 384, 123, 384, 124, 384, 125, 384, 136, 384, 138, 384, 137, 384, 167, 384, 139, 384, 140, 384, 141, 384, 166, 384, 165, 384, 164, 384, 170, 384, 163, 384, 162, 384, 190, 384, 168, 384, 169, 384, 188, 384, 189, 384, 203, 384, 192, 384, 193, 384, 195, 384, 196, 384, 197, 384, 198, 384, 199, 384, 200, 384, 201, 384, 202, 384, 208, 384, 204, 384, 205, 384, 206, 384, 207, 384, 209, 384, 210, 384, 211, 384, 212, 384, 220, 384, 221, 384, 222, 384, 223, 384],
	#[65535, 381, 7, 381, 11, 381, 12, 381, 13, 381, 36, 381, 59, 381, 43, 381, 32, 381, 102, 381, 63, 381, 93, 381, 95, 381, 62, 381, 110, 381, 42, 381, 60, 381, 101, 381, 103, 381, 58, 381, 57, 381, 109, 381, 111, 381, 97, 381, 124, 381, 108, 381, 123, 381, 104, 381, 99, 381, 96, 381, 106, 381, 105, 381, 113, 381, 116, 381, 100, 381, 115, 381, 112, 381, 121, 381, 107, 381, 114, 381, 122, 381, 117, 381, 118, 381, 125, 381, 126, 381, 127, 381, 129, 381, 130, 381, 131, 381, 132, 381, 133, 381, 136, 381, 137, 381, 138, 381, 139, 381, 140, 381, 141, 381, 162, 381, 163, 381, 164, 381, 165, 381, 166, 381, 167, 381, 168, 381, 169, 381, 170, 381, 188, 381, 189, 381, 190, 381, 202, 381, 192, 381, 193, 381, 195, 381, 196, 381, 197, 381, 198, 381, 199, 381, 200, 381, 201, 381, 209, 381, 203, 381, 204, 381, 205, 381, 206, 381, 207, 381, 208, 381, 210, 381, 211, 381, 212, 381, 220, 381, 221, 381, 222, 381, 223, 381],
	#[42, -310],
	#[159, -292, 90, -301, 158, -295, 91, -299, 156, -302, 155, -291, 154, -297, 161, -298, 148, -296, 149, -300, 150, -303, 151, -288, 152, -290, 153, -293, 157, -304, 160, -294],
	#[7, -13, 11, -127, 31, -261, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 387, 63, 387, 7, 387, 95, 387, 11, 387, 12, 387, 13, 387, 62, 387, 43, 387, 36, 387, 32, 387, 110, 387, 42, 387, 60, 387, 103, 387, 116, 387, 109, 387, 111, 387, 93, 387, 102, 387, 57, 387, 58, 387, 59, 387, 99, 387, 96, 387, 106, 387, 101, 387, 113, 387, 100, 387, 104, 387, 105, 387, 123, 387, 124, 387, 108, 387, 97, 387, 112, 387, 121, 387, 107, 387, 115, 387, 114, 387, 122, 387, 117, 387, 118, 387, 125, 387, 126, 387, 127, 387, 129, 387, 130, 387, 131, 387, 132, 387, 133, 387, 136, 387, 137, 387, 138, 387, 139, 387, 140, 387, 141, 387, 188, 387, 162, 387, 163, 387, 164, 387, 165, 387, 166, 387, 167, 387, 168, 387, 169, 387, 170, 387, 189, 387, 190, 387, 192, 387, 193, 387, 195, 387, 196, 387, 197, 387, 198, 387, 199, 387, 200, 387, 201, 387, 202, 387, 203, 387, 204, 387, 205, 387, 206, 387, 207, 387, 208, 387, 209, 387, 210, 387, 211, 387, 212, 387, 220, 387, 221, 387, 222, 387, 223, 387],
	#[65535, 435, 11, 435, 10, 435, 34, 435, 23, 435, 12, 435, 22, 435, 0, 435, 1, 435, 36, 435, 4, 435, 7, 435, 9, 435, 29, 435, 28, 435, 32, 435, 13, 435, 14, 435, 15, 435, 16, 435, 17, 435, 18, 435, 19, 435, 20, 435, 21, 435, 57, 435, 75, 435, 24, 435, 25, 435, 26, 435, 27, 435, 43, 435, 42, 435, 119, 435, 102, 435, 63, 435, 93, 435, 95, 435, 62, 435, 40, 435, 41, 435, 76, 435, 60, 435, 83, 435, 103, 435, 59, 435, 58, 435, 118, 435, 115, 435, 96, 435, 74, 435, 111, 435, 97, 435, 105, 435, 126, 435, 77, 435, 114, 435, 82, 435, 106, 435, 104, 435, 99, 435, 107, 435, 100, 435, 101, 435, 108, 435, 109, 435, 110, 435, 122, 435, 112, 435, 113, 435, 133, 435, 132, 435, 116, 435, 117, 435, 131, 435, 130, 435, 121, 435, 129, 435, 123, 435, 124, 435, 125, 435, 136, 435, 127, 435, 137, 435, 138, 435, 139, 435, 140, 435, 141, 435, 166, 435, 165, 435, 164, 435, 170, 435, 163, 435, 162, 435, 167, 435, 168, 435, 169, 435, 188, 435, 189, 435, 190, 435, 192, 435, 193, 435, 195, 435, 196, 435, 197, 435, 198, 435, 199, 435, 200, 435, 201, 435, 202, 435, 203, 435, 204, 435, 205, 435, 206, 435, 207, 435, 208, 435, 209, 435, 210, 435, 211, 435, 212, 435, 220, 435, 221, 435, 222, 435, 223, 435, 225, 435, 226, 435],
	#[65535, 385, 7, 385, 11, 385, 12, 385, 13, 385, 32, 385, 36, 385, 42, 385, 43, 385, 57, 385, 58, 385, 59, 385, 60, 385, 62, 385, 63, 385, 93, 385, 95, 385, 96, 385, 97, 385, 99, 385, 100, 385, 101, 385, 102, 385, 103, 385, 104, 385, 105, 385, 106, 385, 107, 385, 108, 385, 109, 385, 110, 385, 111, 385, 112, 385, 113, 385, 114, 385, 115, 385, 116, 385, 117, 385, 118, 385, 121, 385, 122, 385, 123, 385, 124, 385, 125, 385, 126, 385, 127, 385, 129, 385, 130, 385, 131, 385, 132, 385, 133, 385, 136, 385, 137, 385, 138, 385, 139, 385, 140, 385, 141, 385, 162, 385, 163, 385, 164, 385, 165, 385, 166, 385, 167, 385, 168, 385, 169, 385, 170, 385, 188, 385, 189, 385, 190, 385, 192, 385, 193, 385, 195, 385, 196, 385, 197, 385, 198, 385, 199, 385, 200, 385, 201, 385, 202, 385, 203, 385, 204, 385, 205, 385, 206, 385, 207, 385, 208, 385, 209, 385, 210, 385, 211, 385, 212, 385, 220, 385, 221, 385, 222, 385, 223, 385],
	#[65535, 167, 12, 167, 13, 167, 32, 167, 34, 167],
	#[7, -13, 11, -127, 31, -128, 32, -429, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -282, 34, -427],
	#[37, 105, 84, -272, 87, -278, 31, -267, 7, -13, 95, -100, 10, -223, 11, -274, 33, -277, 86, -271, 111, -56, 118, -36, 72, -161, 42, -159, 119, -33, 102, -236, 63, -232, 90, -266, 115, -70, 38, -264, 91, -275, 89, -269, 92, -270, 105, -246, 101, -255, 103, -254, 97, -237, 126, -115, 88, -273, 114, -72, 106, -231, 104, -229, 85, -276, 99, -245, 107, -118, 100, -235, 108, -89, 109, -80, 110, -65, 122, -107, 112, -97, 113, -94, 133, -88, 132, -93, 116, -117, 117, -119, 131, -110, 130, -57, 121, -58, 129, -81, 123, -105, 124, -76, 125, -62, 136, -87, 127, -102, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 141, 0, 141, 1, 141, 4, 141, 7, 141, 9, 141, 10, 141, 12, 141, 13, 141, 14, 141, 15, 141, 16, 141, 17, 141, 18, 141, 19, 141, 20, 141, 21, 141, 22, 141, 23, 141, 24, 141, 25, 141, 26, 141, 27, 141, 28, 141, 29, 141, 32, 141, 34, 141, 40, 141, 41, 141, 43, 141, 63, 141, 74, 141, 75, 141, 76, 141, 77, 141, 82, 141, 83, 141, 95, 141, 96, 141, 97, 141, 99, 141, 100, 141, 101, 141, 102, 141, 103, 141, 104, 141, 105, 141, 106, 141, 107, 141, 108, 141, 109, 141, 110, 141, 111, 141, 112, 141, 113, 141, 114, 141, 115, 141, 116, 141, 117, 141, 118, 141, 119, 141, 121, 141, 122, 141, 123, 141, 124, 141, 125, 141, 126, 141, 127, 141, 129, 141, 130, 141, 131, 141, 132, 141, 133, 141, 136, 141, 137, 141, 138, 141, 139, 141, 140, 141, 141, 141, 163, 141, 164, 141, 165, 141, 166, 141, 167, 141, 169, 141, 170, 141, 188, 141, 189, 141, 190, 141, 192, 141, 193, 141, 195, 141, 196, 141, 197, 141, 198, 141, 199, 141, 200, 141, 202, 141, 203, 141, 204, 141, 205, 141, 225, 141, 226, 141],
	#[65535, 146, 7, 146, 14, 146, 0, 146, 1, 146, 4, 146, 19, 146, 18, 146, 9, 146, 10, 146, 16, 146, 12, 146, 13, 146, 15, 146, 17, 146, 20, 146, 21, 146, 22, 146, 23, 146, 24, 146, 25, 146, 26, 146, 27, 146, 28, 146, 29, 146, 32, 146, 34, 146, 40, 146, 41, 146, 43, 146, 63, 146, 82, 146, 74, 146, 75, 146, 76, 146, 77, 146, 101, 146, 83, 146, 95, 146, 96, 146, 97, 146, 99, 146, 100, 146, 102, 146, 103, 146, 104, 146, 105, 146, 106, 146, 107, 146, 108, 146, 109, 146, 110, 146, 111, 146, 112, 146, 113, 146, 114, 146, 115, 146, 116, 146, 117, 146, 118, 146, 119, 146, 121, 146, 122, 146, 123, 146, 124, 146, 125, 146, 126, 146, 127, 146, 129, 146, 130, 146, 131, 146, 132, 146, 133, 146, 136, 146, 137, 146, 138, 146, 139, 146, 140, 146, 141, 146, 163, 146, 164, 146, 165, 146, 166, 146, 167, 146, 169, 146, 170, 146, 188, 146, 189, 146, 190, 146, 192, 146, 193, 146, 195, 146, 196, 146, 197, 146, 198, 146, 199, 146, 200, 146, 202, 146, 203, 146, 204, 146, 205, 146, 225, 146, 226, 146],
	#[65535, 143, 22, 143, 1, 143, 10, 143, 14, 143, 23, 143, 0, 143, 4, 143, 63, 143, 19, 143, 7, 143, 95, 143, 9, 143, 29, 143, 12, 143, 13, 143, 27, 143, 15, 143, 16, 143, 17, 143, 18, 143, 74, 143, 20, 143, 21, 143, 32, 143, 34, 143, 24, 143, 25, 143, 26, 143, 41, 143, 28, 143, 40, 143, 110, 143, 43, 143, 103, 143, 116, 143, 82, 143, 96, 143, 109, 143, 111, 143, 102, 143, 83, 143, 77, 143, 76, 143, 99, 143, 119, 143, 106, 143, 101, 143, 113, 143, 100, 143, 104, 143, 105, 143, 123, 143, 124, 143, 108, 143, 75, 143, 97, 143, 112, 143, 121, 143, 107, 143, 115, 143, 114, 143, 122, 143, 117, 143, 118, 143, 125, 143, 126, 143, 127, 143, 129, 143, 130, 143, 131, 143, 132, 143, 133, 143, 136, 143, 137, 143, 138, 143, 139, 143, 140, 143, 141, 143, 188, 143, 163, 143, 164, 143, 165, 143, 166, 143, 167, 143, 169, 143, 170, 143, 189, 143, 190, 143, 192, 143, 193, 143, 195, 143, 196, 143, 197, 143, 198, 143, 199, 143, 200, 143, 202, 143, 203, 143, 204, 143, 205, 143, 225, 143, 226, 143],
	#[7, -13, 11, -127, 31, -128, 32, -414, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 137, 14, 137, 23, 137, 10, 137, 0, 137, 1, 137, 4, 137, 63, 137, 7, 137, 9, 137, 29, 137, 12, 137, 13, 137, 27, 137, 15, 137, 16, 137, 17, 137, 18, 137, 19, 137, 20, 137, 21, 137, 22, 137, 34, 137, 24, 137, 25, 137, 26, 137, 32, 137, 28, 137, 119, 137, 102, 137, 126, 137, 75, 137, 95, 137, 40, 137, 41, 137, 76, 137, 43, 137, 83, 137, 103, 137, 118, 137, 115, 137, 96, 137, 74, 137, 111, 137, 97, 137, 105, 137, 77, 137, 114, 137, 82, 137, 106, 137, 104, 137, 99, 137, 107, 137, 100, 137, 101, 137, 108, 137, 109, 137, 110, 137, 122, 137, 112, 137, 113, 137, 133, 137, 132, 137, 116, 137, 117, 137, 131, 137, 130, 137, 121, 137, 129, 137, 123, 137, 124, 137, 125, 137, 136, 137, 127, 137, 137, 137, 138, 137, 139, 137, 140, 137, 141, 137, 166, 137, 165, 137, 164, 137, 170, 137, 163, 137, 167, 137, 169, 137, 188, 137, 189, 137, 190, 137, 192, 137, 193, 137, 195, 137, 196, 137, 197, 137, 198, 137, 199, 137, 200, 137, 202, 137, 203, 137, 204, 137, 205, 137, 225, 137, 226, 137],
	#[65535, 142, 14, 142, 23, 142, 0, 142, 1, 142, 4, 142, 7, 142, 9, 142, 10, 142, 12, 142, 13, 142, 27, 142, 15, 142, 16, 142, 17, 142, 18, 142, 19, 142, 20, 142, 21, 142, 22, 142, 34, 142, 24, 142, 25, 142, 26, 142, 41, 142, 28, 142, 29, 142, 74, 142, 32, 142, 63, 142, 95, 142, 40, 142, 110, 142, 43, 142, 103, 142, 82, 142, 96, 142, 109, 142, 111, 142, 102, 142, 83, 142, 77, 142, 76, 142, 99, 142, 119, 142, 106, 142, 101, 142, 113, 142, 116, 142, 100, 142, 104, 142, 105, 142, 123, 142, 124, 142, 108, 142, 75, 142, 97, 142, 112, 142, 121, 142, 107, 142, 115, 142, 114, 142, 122, 142, 117, 142, 118, 142, 125, 142, 126, 142, 127, 142, 129, 142, 130, 142, 131, 142, 132, 142, 133, 142, 136, 142, 137, 142, 138, 142, 139, 142, 140, 142, 141, 142, 188, 142, 163, 142, 164, 142, 165, 142, 166, 142, 167, 142, 169, 142, 170, 142, 189, 142, 190, 142, 192, 142, 193, 142, 195, 142, 196, 142, 197, 142, 198, 142, 199, 142, 200, 142, 202, 142, 203, 142, 204, 142, 205, 142, 225, 142, 226, 142],
	#[42, -354],
	#[65535, 136, 23, 136, 0, 136, 1, 136, 4, 136, 63, 136, 7, 136, 9, 136, 10, 136, 12, 136, 13, 136, 14, 136, 15, 136, 16, 136, 17, 136, 18, 136, 19, 136, 20, 136, 21, 136, 22, 136, 34, 136, 24, 136, 25, 136, 26, 136, 27, 136, 28, 136, 29, 136, 74, 136, 32, 136, 41, 136, 95, 136, 82, 136, 40, 136, 110, 136, 43, 136, 103, 136, 100, 136, 109, 136, 111, 136, 102, 136, 83, 136, 77, 136, 76, 136, 99, 136, 119, 136, 106, 136, 101, 136, 113, 136, 116, 136, 96, 136, 104, 136, 105, 136, 123, 136, 124, 136, 108, 136, 75, 136, 97, 136, 112, 136, 121, 136, 107, 136, 115, 136, 114, 136, 122, 136, 117, 136, 118, 136, 125, 136, 126, 136, 127, 136, 129, 136, 130, 136, 131, 136, 132, 136, 133, 136, 136, 136, 137, 136, 138, 136, 139, 136, 140, 136, 141, 136, 188, 136, 163, 136, 164, 136, 165, 136, 166, 136, 167, 136, 169, 136, 170, 136, 189, 136, 190, 136, 192, 136, 193, 136, 195, 136, 196, 136, 197, 136, 198, 136, 199, 136, 200, 136, 202, 136, 203, 136, 204, 136, 205, 136, 225, 136, 226, 136],
	#[7, -338],
	#[65535, 140, 7, 140, 95, 140, 22, 140, 18, 140, 21, 140, 23, 140, 10, 140, 19, 140, 14, 140, 20, 140, 4, 140, 29, 140, 9, 140, 13, 140, 63, 140, 28, 140, 12, 140, 1, 140, 16, 140, 77, 140, 25, 140, 106, 140, 15, 140, 103, 140, 27, 140, 17, 140, 76, 140, 26, 140, 82, 140, 0, 140, 75, 140, 100, 140, 41, 140, 40, 140, 96, 140, 32, 140, 119, 140, 43, 140, 83, 140, 108, 140, 105, 140, 74, 140, 24, 140, 102, 140, 164, 140, 34, 140, 97, 140, 116, 140, 114, 140, 107, 140, 110, 140, 104, 140, 99, 140, 124, 140, 122, 140, 111, 140, 101, 140, 118, 140, 112, 140, 132, 140, 129, 140, 137, 140, 109, 140, 126, 140, 113, 140, 115, 140, 133, 140, 127, 140, 130, 140, 121, 140, 136, 140, 188, 140, 123, 140, 163, 140, 117, 140, 131, 140, 125, 140, 170, 140, 165, 140, 138, 140, 139, 140, 140, 140, 141, 140, 166, 140, 167, 140, 169, 140, 189, 140, 190, 140, 192, 140, 193, 140, 195, 140, 196, 140, 197, 140, 198, 140, 199, 140, 200, 140, 202, 140, 203, 140, 204, 140, 205, 140, 225, 140, 226, 140],
	#[70, -140, 7, -13, 11, -127, 12, -329, 36, -129, 35, -123, 31, -128, 33, -139, 63, -121, 73, -138, 67, -137, 71, -130, 64, -141, 69, -136, 68, -133, 66, -126, 65, -122, 118, -36],
	#[65535, 144, 0, 144, 1, 144, 4, 144, 7, 144, 9, 144, 10, 144, 12, 144, 13, 144, 14, 144, 15, 144, 16, 144, 17, 144, 18, 144, 19, 144, 20, 144, 21, 144, 22, 144, 23, 144, 24, 144, 25, 144, 26, 144, 27, 144, 28, 144, 29, 144, 32, 144, 34, 144, 40, 144, 41, 144, 43, 144, 63, 144, 74, 144, 75, 144, 76, 144, 77, 144, 82, 144, 83, 144, 95, 144, 96, 144, 97, 144, 99, 144, 100, 144, 101, 144, 102, 144, 103, 144, 104, 144, 105, 144, 106, 144, 107, 144, 108, 144, 109, 144, 110, 144, 111, 144, 112, 144, 113, 144, 114, 144, 115, 144, 116, 144, 117, 144, 118, 144, 119, 144, 121, 144, 122, 144, 123, 144, 124, 144, 125, 144, 126, 144, 127, 144, 129, 144, 130, 144, 131, 144, 132, 144, 133, 144, 136, 144, 137, 144, 138, 144, 139, 144, 140, 144, 141, 144, 163, 144, 164, 144, 165, 144, 166, 144, 167, 144, 169, 144, 170, 144, 188, 144, 189, 144, 190, 144, 192, 144, 193, 144, 195, 144, 196, 144, 197, 144, 198, 144, 199, 144, 200, 144, 202, 144, 203, 144, 204, 144, 205, 144, 225, 144, 226, 144],
	#[65535, 135, 23, 135, 10, 135, 14, 135, 13, 135, 22, 135, 7, 135, 9, 135, 77, 135, 21, 135, 75, 135, 15, 135, 1, 135, 27, 135, 29, 135, 12, 135, 16, 135, 0, 135, 20, 135, 4, 135, 63, 135, 19, 135, 18, 135, 95, 135, 24, 135, 102, 135, 25, 135, 34, 135, 41, 135, 26, 135, 17, 135, 83, 135, 74, 135, 32, 135, 97, 135, 28, 135, 40, 135, 110, 135, 76, 135, 43, 135, 101, 135, 103, 135, 116, 135, 82, 135, 96, 135, 109, 135, 111, 135, 124, 135, 108, 135, 123, 135, 104, 135, 99, 135, 119, 135, 106, 135, 105, 135, 113, 135, 100, 135, 115, 135, 112, 135, 121, 135, 107, 135, 114, 135, 122, 135, 117, 135, 118, 135, 125, 135, 126, 135, 127, 135, 129, 135, 130, 135, 131, 135, 132, 135, 133, 135, 136, 135, 137, 135, 138, 135, 139, 135, 140, 135, 141, 135, 163, 135, 164, 135, 165, 135, 166, 135, 167, 135, 169, 135, 170, 135, 188, 135, 189, 135, 190, 135, 192, 135, 193, 135, 195, 135, 196, 135, 197, 135, 198, 135, 199, 135, 200, 135, 202, 135, 203, 135, 204, 135, 205, 135, 225, 135, 226, 135],
	#[7, -13, 11, -127, 31, -280, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 139, 23, 139, 1, 139, 14, 139, 12, 139, 0, 139, 27, 139, 21, 139, 4, 139, 7, 139, 9, 139, 10, 139, 16, 139, 34, 139, 13, 139, 15, 139, 17, 139, 18, 139, 19, 139, 20, 139, 22, 139, 24, 139, 25, 139, 26, 139, 41, 139, 28, 139, 29, 139, 74, 139, 83, 139, 32, 139, 102, 139, 63, 139, 75, 139, 95, 139, 77, 139, 40, 139, 110, 139, 76, 139, 43, 139, 101, 139, 103, 139, 82, 139, 96, 139, 109, 139, 111, 139, 97, 139, 124, 139, 108, 139, 123, 139, 104, 139, 99, 139, 119, 139, 106, 139, 105, 139, 113, 139, 116, 139, 100, 139, 115, 139, 112, 139, 121, 139, 107, 139, 114, 139, 122, 139, 117, 139, 118, 139, 125, 139, 126, 139, 127, 139, 129, 139, 130, 139, 131, 139, 132, 139, 133, 139, 136, 139, 137, 139, 138, 139, 139, 139, 140, 139, 141, 139, 163, 139, 164, 139, 165, 139, 166, 139, 167, 139, 169, 139, 170, 139, 188, 139, 189, 139, 190, 139, 192, 139, 193, 139, 195, 139, 196, 139, 197, 139, 198, 139, 199, 139, 200, 139, 202, 139, 203, 139, 204, 139, 205, 139, 225, 139, 226, 139],
	#[65535, 138, 22, 138, 1, 138, 10, 138, 14, 138, 23, 138, 9, 138, 0, 138, 4, 138, 63, 138, 19, 138, 7, 138, 95, 138, 29, 138, 12, 138, 13, 138, 27, 138, 15, 138, 16, 138, 17, 138, 18, 138, 74, 138, 20, 138, 21, 138, 32, 138, 34, 138, 24, 138, 25, 138, 26, 138, 41, 138, 28, 138, 40, 138, 110, 138, 43, 138, 103, 138, 116, 138, 82, 138, 96, 138, 109, 138, 111, 138, 102, 138, 83, 138, 77, 138, 76, 138, 99, 138, 119, 138, 106, 138, 101, 138, 113, 138, 100, 138, 104, 138, 105, 138, 123, 138, 124, 138, 108, 138, 75, 138, 97, 138, 112, 138, 121, 138, 107, 138, 115, 138, 114, 138, 122, 138, 117, 138, 118, 138, 125, 138, 126, 138, 127, 138, 129, 138, 130, 138, 131, 138, 132, 138, 133, 138, 136, 138, 137, 138, 138, 138, 139, 138, 140, 138, 141, 138, 188, 138, 163, 138, 164, 138, 165, 138, 166, 138, 167, 138, 169, 138, 170, 138, 189, 138, 190, 138, 192, 138, 193, 138, 195, 138, 196, 138, 197, 138, 198, 138, 199, 138, 200, 138, 202, 138, 203, 138, 204, 138, 205, 138, 225, 138, 226, 138],
	#[7, -13, 11, -127, 31, -128, 32, -286, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -282, 34, -283],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 130, 0, 130, 1, 130, 4, 130, 7, 130, 9, 130, 10, 130, 12, 130, 13, 130, 14, 130, 15, 130, 16, 130, 17, 130, 18, 130, 19, 130, 20, 130, 21, 130, 22, 130, 23, 130, 24, 130, 25, 130, 26, 130, 27, 130, 28, 130, 29, 130, 32, 130, 34, 130, 40, 130, 41, 130, 43, 130, 63, 130, 74, 130, 75, 130, 76, 130, 77, 130, 82, 130, 83, 130, 95, 130, 96, 130, 97, 130, 99, 130, 100, 130, 101, 130, 102, 130, 103, 130, 104, 130, 105, 130, 106, 130, 107, 130, 108, 130, 109, 130, 110, 130, 111, 130, 112, 130, 113, 130, 114, 130, 115, 130, 116, 130, 117, 130, 118, 130, 119, 130, 121, 130, 122, 130, 123, 130, 124, 130, 125, 130, 126, 130, 127, 130, 129, 130, 130, 130, 131, 130, 132, 130, 133, 130, 136, 130, 137, 130, 138, 130, 139, 130, 140, 130, 141, 130, 163, 130, 164, 130, 165, 130, 166, 130, 167, 130, 169, 130, 170, 130, 188, 130, 189, 130, 190, 130, 192, 130, 193, 130, 195, 130, 196, 130, 197, 130, 198, 130, 199, 130, 200, 130, 202, 130, 203, 130, 204, 130, 205, 130, 225, 130, 226, 130],
	#[65535, 168, 12, 168, 13, 168, 34, 168, 32, 168],
	#[13, -282, 32, -425],
	#[31, 94, 111, 94, 10, 94, 84, 94, 87, 94, 119, 94, 7, 94, 95, 94, 85, 94, 102, 94, 11, 94, 33, 94, 91, 94, 114, 94, 103, 94, 86, 94, 37, 94, 34, -424, 118, 94, 42, 94, 89, 94, 63, 94, 115, 94, 38, 94, 72, 94, 88, 94, 90, 94, 99, 94, 106, 94, 104, 94, 105, 94, 126, 94, 92, 94, 97, 94, 107, 94, 100, 94, 101, 94, 108, 94, 109, 94, 110, 94, 122, 94, 112, 94, 113, 94, 133, 94, 132, 94, 116, 94, 117, 94, 131, 94, 130, 94, 121, 94, 129, 94, 123, 94, 124, 94, 125, 94, 136, 94, 127, 94, 137, 94, 138, 94, 139, 94, 140, 94, 141, 94, 226, 94],
	#[31, -267, 86, -271, 42, -159, 87, -278, 11, -274, 91, -275, 13, 165, 63, -232, 32, 165, 7, -13, 95, -100, 37, 105, 99, -245, 103, -254, 10, -223, 92, -270, 33, -277, 38, -264, 88, -273, 111, -56, 72, -161, 100, -235, 84, -272, 105, -246, 90, -266, 119, -33, 124, -76, 97, -237, 106, -231, 113, -94, 85, -276, 102, -236, 115, -70, 89, -269, 132, -93, 114, -72, 112, -97, 110, -65, 104, -229, 101, -255, 118, -36, 116, -117, 123, -105, 109, -80, 126, -115, 121, -58, 133, -88, 108, -89, 130, -57, 107, -118, 129, -81, 122, -107, 117, -119, 131, -110, 125, -62, 136, -87, 127, -102, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 271, 7, 271, 11, 271, 31, 271, 33, 271, 35, 271, 36, 271, 42, 271, 63, 271, 64, 271, 65, 271, 66, 271, 67, 271, 68, 271, 69, 271, 70, 271, 71, 271, 73, 271, 118, 271],
	#[42, -305],
	#[65535, 258, 7, 258, 11, 258, 31, 258, 33, 258, 35, 258, 36, 258, 42, 258, 63, 258, 64, 258, 65, 258, 66, 258, 67, 258, 68, 258, 69, 258, 70, 258, 71, 258, 73, 258, 118, 258],
	#[65535, 261, 69, 261, 71, 261, 68, 261, 63, 261, 7, 261, 11, 261, 64, 261, 31, 261, 33, 261, 35, 261, 36, 261, 42, 261, 65, 261, 73, 261, 70, 261, 67, 261, 66, 261, 118, 261],
	#[65535, 265, 7, 265, 11, 265, 31, 265, 33, 265, 35, 265, 36, 265, 42, 265, 63, 265, 64, 265, 65, 265, 66, 265, 67, 265, 68, 265, 69, 265, 70, 265, 71, 265, 73, 265, 118, 265],
	#[65535, 259, 7, 259, 11, 259, 31, 259, 33, 259, 35, 259, 36, 259, 42, 259, 63, 259, 64, 259, 65, 259, 66, 259, 67, 259, 68, 259, 69, 259, 70, 259, 71, 259, 73, 259, 118, 259],
	#[65535, 266, 7, 266, 11, 266, 31, 266, 33, 266, 35, 266, 36, 266, 42, 266, 63, 266, 64, 266, 65, 266, 66, 266, 67, 266, 68, 266, 69, 266, 70, 266, 71, 266, 73, 266, 118, 266],
	#[65535, 264, 7, 264, 11, 264, 31, 264, 33, 264, 35, 264, 36, 264, 42, 264, 63, 264, 64, 264, 65, 264, 66, 264, 67, 264, 68, 264, 69, 264, 70, 264, 71, 264, 73, 264, 118, 264],
	#[65535, 268, 7, 268, 11, 268, 31, 268, 33, 268, 35, 268, 36, 268, 42, 268, 63, 268, 64, 268, 65, 268, 66, 268, 67, 268, 68, 268, 69, 268, 70, 268, 71, 268, 73, 268, 118, 268],
	#[65535, 260, 31, 260, 7, 260, 11, 260, 42, 260, 33, 260, 63, 260, 35, 260, 36, 260, 71, 260, 70, 260, 69, 260, 68, 260, 67, 260, 66, 260, 65, 260, 73, 260, 64, 260, 118, 260],
	#[65535, 267, 7, 267, 11, 267, 31, 267, 33, 267, 35, 267, 36, 267, 42, 267, 63, 267, 64, 267, 65, 267, 66, 267, 67, 267, 68, 267, 69, 267, 70, 267, 71, 267, 73, 267, 118, 267],
	#[65535, 273, 31, 273, 7, 273, 11, 273, 42, 273, 69, 273, 33, 273, 63, 273, 35, 273, 36, 273, 71, 273, 64, 273, 70, 273, 118, 273, 68, 273, 67, 273, 66, 273, 65, 273, 73, 273],
	#[65535, 269, 35, 269, 42, 269, 70, 269, 31, 269, 63, 269, 7, 269, 65, 269, 11, 269, 67, 269, 36, 269, 64, 269, 69, 269, 33, 269, 68, 269, 73, 269, 71, 269, 66, 269, 118, 269],
	#[65535, 272, 7, 272, 11, 272, 31, 272, 33, 272, 35, 272, 36, 272, 42, 272, 63, 272, 64, 272, 65, 272, 66, 272, 67, 272, 68, 272, 69, 272, 70, 272, 71, 272, 73, 272, 118, 272],
	#[65535, 262, 31, 262, 7, 262, 11, 262, 69, 262, 42, 262, 73, 262, 33, 262, 63, 262, 35, 262, 36, 262, 71, 262, 66, 262, 70, 262, 64, 262, 68, 262, 67, 262, 65, 262, 118, 262],
	#[65535, 270, 31, 270, 63, 270, 7, 270, 11, 270, 42, 270, 70, 270, 69, 270, 33, 270, 68, 270, 35, 270, 36, 270, 67, 270, 71, 270, 66, 270, 65, 270, 73, 270, 64, 270, 118, 270],
	#[65535, 263, 7, 263, 11, 263, 31, 263, 33, 263, 35, 263, 36, 263, 42, 263, 63, 263, 64, 263, 65, 263, 66, 263, 67, 263, 68, 263, 69, 263, 70, 263, 71, 263, 73, 263, 118, 263],
	#[31, -128, 36, -129, 11, -127, 35, -123, 33, -139, 7, -13, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -307],
	#[31, -128, 36, -129, 69, -136, 11, -127, 33, -139, 7, -13, 35, -123, 66, -126, 65, -122, 73, -138, 63, -121, 64, -141, 67, -137, 68, -133, 70, -140, 71, -130, 118, -36],
	#[43, -309],
	#[65535, 175, 23, 175, 62, 175, 1, 175, 9, 175, 14, 175, 12, 175, 11, 175, 10, 175, 0, 175, 27, 175, 21, 175, 20, 175, 4, 175, 19, 175, 7, 175, 42, 175, 17, 175, 16, 175, 34, 175, 13, 175, 15, 175, 43, 175, 18, 175, 36, 175, 59, 175, 22, 175, 75, 175, 24, 175, 25, 175, 26, 175, 41, 175, 28, 175, 29, 175, 74, 175, 83, 175, 32, 175, 102, 175, 63, 175, 93, 175, 95, 175, 77, 175, 40, 175, 110, 175, 76, 175, 60, 175, 101, 175, 103, 175, 58, 175, 82, 175, 57, 175, 96, 175, 109, 175, 111, 175, 97, 175, 124, 175, 108, 175, 123, 175, 104, 175, 99, 175, 119, 175, 106, 175, 105, 175, 113, 175, 116, 175, 100, 175, 115, 175, 112, 175, 121, 175, 107, 175, 114, 175, 122, 175, 117, 175, 118, 175, 125, 175, 126, 175, 127, 175, 129, 175, 130, 175, 131, 175, 132, 175, 133, 175, 136, 175, 137, 175, 138, 175, 139, 175, 140, 175, 141, 175, 162, 175, 163, 175, 164, 175, 165, 175, 166, 175, 167, 175, 168, 175, 169, 175, 170, 175, 188, 175, 189, 175, 190, 175, 202, 175, 192, 175, 193, 175, 195, 175, 196, 175, 197, 175, 198, 175, 199, 175, 200, 175, 201, 175, 209, 175, 203, 175, 204, 175, 205, 175, 206, 175, 207, 175, 208, 175, 210, 175, 211, 175, 212, 175, 220, 175, 221, 175, 222, 175, 223, 175, 225, 175, 226, 175],
	#[31, -128, 33, -139, 7, -13, 11, -127, 69, -136, 36, -129, 35, -123, 66, -126, 65, -122, 73, -138, 63, -121, 64, -141, 67, -137, 68, -133, 70, -140, 71, -130, 118, -36],
	#[13, -312],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -314],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 71, -130, 70, -140, 69, -136, 68, -133, 67, -137, 66, -126, 63, -121, 64, -141, 65, -122, 73, -138, 118, -36],
	#[43, -316],
	#[65535, 177, 1, 177, 10, 177, 14, 177, 23, 177, 9, 177, 0, 177, 4, 177, 19, 177, 7, 177, 29, 177, 11, 177, 12, 177, 13, 177, 27, 177, 15, 177, 16, 177, 17, 177, 18, 177, 36, 177, 20, 177, 21, 177, 22, 177, 34, 177, 24, 177, 25, 177, 26, 177, 32, 177, 28, 177, 40, 177, 41, 177, 42, 177, 43, 177, 57, 177, 58, 177, 59, 177, 60, 177, 62, 177, 63, 177, 74, 177, 75, 177, 76, 177, 77, 177, 82, 177, 83, 177, 93, 177, 95, 177, 96, 177, 97, 177, 99, 177, 100, 177, 101, 177, 102, 177, 103, 177, 104, 177, 105, 177, 106, 177, 107, 177, 108, 177, 109, 177, 110, 177, 111, 177, 112, 177, 113, 177, 114, 177, 115, 177, 116, 177, 117, 177, 118, 177, 119, 177, 121, 177, 122, 177, 123, 177, 124, 177, 125, 177, 126, 177, 127, 177, 129, 177, 130, 177, 131, 177, 132, 177, 133, 177, 136, 177, 137, 177, 138, 177, 139, 177, 140, 177, 141, 177, 162, 177, 163, 177, 164, 177, 165, 177, 166, 177, 167, 177, 168, 177, 169, 177, 170, 177, 188, 177, 189, 177, 190, 177, 192, 177, 193, 177, 195, 177, 196, 177, 197, 177, 198, 177, 199, 177, 200, 177, 201, 177, 202, 177, 203, 177, 204, 177, 205, 177, 206, 177, 207, 177, 208, 177, 209, 177, 210, 177, 211, 177, 212, 177, 220, 177, 221, 177, 222, 177, 223, 177, 225, 177, 226, 177],
	#[65535, 438, 62, 438, 21, 438, 14, 438, 12, 438, 11, 438, 10, 438, 0, 438, 1, 438, 9, 438, 20, 438, 4, 438, 63, 438, 19, 438, 7, 438, 42, 438, 17, 438, 16, 438, 34, 438, 13, 438, 15, 438, 43, 438, 18, 438, 36, 438, 59, 438, 22, 438, 23, 438, 24, 438, 25, 438, 26, 438, 27, 438, 28, 438, 29, 438, 74, 438, 83, 438, 32, 438, 102, 438, 41, 438, 75, 438, 95, 438, 82, 438, 77, 438, 40, 438, 110, 438, 76, 438, 60, 438, 101, 438, 103, 438, 58, 438, 100, 438, 57, 438, 106, 438, 109, 438, 111, 438, 93, 438, 124, 438, 108, 438, 123, 438, 104, 438, 99, 438, 119, 438, 97, 438, 105, 438, 113, 438, 116, 438, 96, 438, 115, 438, 112, 438, 121, 438, 107, 438, 114, 438, 122, 438, 117, 438, 118, 438, 125, 438, 126, 438, 127, 438, 129, 438, 130, 438, 131, 438, 132, 438, 133, 438, 136, 438, 137, 438, 138, 438, 139, 438, 140, 438, 141, 438, 162, 438, 163, 438, 164, 438, 165, 438, 166, 438, 167, 438, 168, 438, 169, 438, 170, 438, 188, 438, 189, 438, 190, 438, 202, 438, 192, 438, 193, 438, 195, 438, 196, 438, 197, 438, 198, 438, 199, 438, 200, 438, 201, 438, 209, 438, 203, 438, 204, 438, 205, 438, 206, 438, 207, 438, 208, 438, 210, 438, 211, 438, 212, 438, 220, 438, 221, 438, 222, 438, 223, 438, 225, 438, 226, 438],
	#[31, -128, 7, -13, 11, -127, 69, -136, 32, -320, 33, -139, 35, -123, 36, -129, 73, -138, 85, -323, 66, -126, 65, -122, 63, -121, 64, -141, 67, -137, 68, -133, 70, -140, 71, -130, 118, -36],
	#[65535, 439, 15, 439, 14, 439, 60, 439, 106, 439, 11, 439, 0, 439, 1, 439, 9, 439, 20, 439, 4, 439, 7, 439, 10, 439, 16, 439, 12, 439, 13, 439, 109, 439, 103, 439, 17, 439, 18, 439, 19, 439, 59, 439, 21, 439, 22, 439, 23, 439, 24, 439, 25, 439, 26, 439, 27, 439, 28, 439, 29, 439, 32, 439, 102, 439, 34, 439, 36, 439, 95, 439, 77, 439, 40, 439, 41, 439, 42, 439, 43, 439, 101, 439, 58, 439, 96, 439, 74, 439, 111, 439, 93, 439, 124, 439, 57, 439, 123, 439, 76, 439, 82, 439, 119, 439, 62, 439, 63, 439, 110, 439, 116, 439, 100, 439, 104, 439, 83, 439, 113, 439, 112, 439, 99, 439, 108, 439, 75, 439, 97, 439, 105, 439, 121, 439, 107, 439, 115, 439, 114, 439, 122, 439, 117, 439, 118, 439, 125, 439, 126, 439, 127, 439, 129, 439, 130, 439, 131, 439, 132, 439, 133, 439, 136, 439, 137, 439, 138, 439, 139, 439, 140, 439, 141, 439, 188, 439, 162, 439, 163, 439, 164, 439, 165, 439, 166, 439, 167, 439, 168, 439, 169, 439, 170, 439, 189, 439, 190, 439, 192, 439, 193, 439, 195, 439, 196, 439, 197, 439, 198, 439, 199, 439, 200, 439, 201, 439, 202, 439, 203, 439, 204, 439, 205, 439, 206, 439, 207, 439, 208, 439, 209, 439, 210, 439, 211, 439, 212, 439, 220, 439, 221, 439, 222, 439, 223, 439, 225, 439, 226, 439],
	#[65535, 436, 0, 436, 1, 436, 4, 436, 7, 436, 9, 436, 10, 436, 11, 436, 12, 436, 13, 436, 14, 436, 15, 436, 16, 436, 17, 436, 18, 436, 19, 436, 20, 436, 21, 436, 22, 436, 23, 436, 24, 436, 25, 436, 26, 436, 27, 436, 28, 436, 29, 436, 32, 436, 34, 436, 36, 436, 40, 436, 41, 436, 42, 436, 43, 436, 57, 436, 58, 436, 59, 436, 60, 436, 62, 436, 63, 436, 74, 436, 75, 436, 76, 436, 77, 436, 82, 436, 83, 436, 93, 436, 95, 436, 96, 436, 97, 436, 99, 436, 100, 436, 101, 436, 102, 436, 103, 436, 104, 436, 105, 436, 106, 436, 107, 436, 108, 436, 109, 436, 110, 436, 111, 436, 112, 436, 113, 436, 114, 436, 115, 436, 116, 436, 117, 436, 118, 436, 119, 436, 121, 436, 122, 436, 123, 436, 124, 436, 125, 436, 126, 436, 127, 436, 129, 436, 130, 436, 131, 436, 132, 436, 133, 436, 136, 436, 137, 436, 138, 436, 139, 436, 140, 436, 141, 436, 162, 436, 163, 436, 164, 436, 165, 436, 166, 436, 167, 436, 168, 436, 169, 436, 170, 436, 188, 436, 189, 436, 190, 436, 192, 436, 193, 436, 195, 436, 196, 436, 197, 436, 198, 436, 199, 436, 200, 436, 201, 436, 202, 436, 203, 436, 204, 436, 205, 436, 206, 436, 207, 436, 208, 436, 209, 436, 210, 436, 211, 436, 212, 436, 220, 436, 221, 436, 222, 436, 223, 436, 225, 436, 226, 436],
	#[65535, 431, 13, 431, 32, 431],
	#[13, -325, 32, -324],
	#[65535, 432, 13, 432, 32, 432],
	#[65535, 437, 14, 437, 23, 437, 10, 437, 0, 437, 1, 437, 9, 437, 4, 437, 7, 437, 17, 437, 11, 437, 12, 437, 13, 437, 27, 437, 15, 437, 16, 437, 43, 437, 18, 437, 19, 437, 20, 437, 21, 437, 22, 437, 34, 437, 24, 437, 25, 437, 26, 437, 41, 437, 28, 437, 29, 437, 74, 437, 32, 437, 59, 437, 63, 437, 36, 437, 95, 437, 62, 437, 40, 437, 110, 437, 42, 437, 60, 437, 103, 437, 58, 437, 82, 437, 57, 437, 96, 437, 109, 437, 111, 437, 93, 437, 102, 437, 83, 437, 77, 437, 76, 437, 99, 437, 119, 437, 106, 437, 101, 437, 113, 437, 116, 437, 100, 437, 104, 437, 105, 437, 123, 437, 124, 437, 108, 437, 75, 437, 97, 437, 112, 437, 121, 437, 107, 437, 115, 437, 114, 437, 122, 437, 117, 437, 118, 437, 125, 437, 126, 437, 127, 437, 129, 437, 130, 437, 131, 437, 132, 437, 133, 437, 136, 437, 137, 437, 138, 437, 139, 437, 140, 437, 141, 437, 188, 437, 162, 437, 163, 437, 164, 437, 165, 437, 166, 437, 167, 437, 168, 437, 169, 437, 170, 437, 189, 437, 190, 437, 192, 437, 193, 437, 195, 437, 196, 437, 197, 437, 198, 437, 199, 437, 200, 437, 201, 437, 202, 437, 203, 437, 204, 437, 205, 437, 206, 437, 207, 437, 208, 437, 209, 437, 210, 437, 211, 437, 212, 437, 220, 437, 221, 437, 222, 437, 223, 437, 225, 437, 226, 437],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 85, -327, 118, -36],
	#[65535, 433, 13, 433, 32, 433],
	#[65535, 434, 13, 434, 32, 434],
	#[12, -330, 13, -282],
	#[65535, 128, 29, 128, 12, 128, 63, 128, 7, 128, 4, 128, 19, 128, 75, 128, 23, 128, 14, 128, 26, 128, 108, 128, 43, 128, 18, 128, 0, 128, 1, 128, 20, 128, 25, 128, 83, 128, 76, 128, 9, 128, 10, 128, 126, 128, 34, 128, 13, 128, 27, 128, 15, 128, 16, 128, 17, 128, 110, 128, 24, 128, 41, 128, 21, 128, 22, 128, 118, 128, 28, 128, 74, 128, 116, 128, 32, 128, 102, 128, 40, 128, 96, 128, 95, 128, 82, 128, 101, 128, 103, 128, 99, 128, 115, 128, 106, 128, 109, 128, 111, 128, 104, 128, 97, 128, 124, 128, 77, 128, 119, 128, 113, 128, 100, 128, 107, 128, 127, 128, 105, 128, 123, 128, 140, 128, 129, 128, 139, 128, 130, 128, 133, 128, 112, 128, 121, 128, 114, 128, 137, 128, 141, 128, 122, 128, 117, 128, 131, 128, 125, 128, 138, 128, 132, 128, 136, 128, 188, 128, 163, 128, 164, 128, 165, 128, 166, 128, 167, 128, 169, 128, 170, 128, 192, 128, 189, 128, 190, 128, 193, 128, 195, 128, 196, 128, 197, 128, 198, 128, 199, 128, 200, 128, 202, 128, 203, 128, 204, 128, 205, 128, 225, 128, 226, 128],
	#[65535, 127, 14, 127, 18, 127, 0, 127, 1, 127, 20, 127, 4, 127, 19, 127, 7, 127, 9, 127, 10, 127, 16, 127, 12, 127, 13, 127, 15, 127, 17, 127, 41, 127, 21, 127, 22, 127, 23, 127, 24, 127, 25, 127, 26, 127, 27, 127, 28, 127, 29, 127, 43, 127, 32, 127, 34, 127, 40, 127, 63, 127, 82, 127, 74, 127, 75, 127, 76, 127, 77, 127, 101, 127, 83, 127, 95, 127, 96, 127, 97, 127, 99, 127, 100, 127, 102, 127, 103, 127, 104, 127, 105, 127, 106, 127, 107, 127, 108, 127, 109, 127, 110, 127, 111, 127, 112, 127, 113, 127, 114, 127, 115, 127, 116, 127, 117, 127, 118, 127, 119, 127, 121, 127, 122, 127, 123, 127, 124, 127, 125, 127, 126, 127, 127, 127, 129, 127, 130, 127, 131, 127, 132, 127, 133, 127, 136, 127, 137, 127, 138, 127, 139, 127, 140, 127, 141, 127, 163, 127, 164, 127, 165, 127, 166, 127, 167, 127, 169, 127, 170, 127, 188, 127, 189, 127, 190, 127, 192, 127, 193, 127, 195, 127, 196, 127, 197, 127, 198, 127, 199, 127, 200, 127, 202, 127, 203, 127, 204, 127, 205, 127, 225, 127, 226, 127],
	#[65535, 223, 7, 223, 11, 223, 33, 223, 42, 223, 35, 223, 31, 223, 36, 223, 63, 223, 64, 223, 65, 223, 66, 223, 67, 223, 68, 223, 69, 223, 70, 223, 71, 223, 73, 223, 118, 223],
	#[7, 236, 11, 236, 31, 236, 33, 236, 35, 236, 36, 236, 42, 236, 63, 236, 64, 236, 65, 236, 66, 236, 67, 236, 68, 236, 69, 236, 70, 236, 71, 236, 73, 236, 118, 236, 135, -335],
	#[7, 237, 11, 237, 42, 237, 31, 237, 33, 237, 35, 237, 36, 237, 63, 237, 64, 237, 65, 237, 66, 237, 67, 237, 68, 237, 69, 237, 70, 237, 71, 237, 73, 237, 118, 237, 134, -334],
	#[65535, 239, 31, 239, 7, 239, 11, 239, 33, 239, 35, 239, 42, 239, 36, 239, 63, 239, 64, 239, 65, 239, 66, 239, 67, 239, 68, 239, 69, 239, 70, 239, 71, 239, 73, 239, 118, 239],
	#[65535, 238, 31, 238, 7, 238, 11, 238, 42, 238, 33, 238, 35, 238, 36, 238, 63, 238, 64, 238, 65, 238, 66, 238, 67, 238, 68, 238, 69, 238, 70, 238, 71, 238, 73, 238, 118, 238],
	#[65535, 244, 7, 244, 11, 244, 31, 244, 33, 244, 35, 244, 36, 244, 42, 244, 63, 244, 64, 244, 65, 244, 66, 244, 67, 244, 68, 244, 69, 244, 70, 244, 71, 244, 73, 244, 118, 244],
	#[65535, 228, 7, 228, 11, 228, 42, 228, 31, 228, 33, 228, 35, 228, 36, 228, 63, 228, 64, 228, 65, 228, 66, 228, 67, 228, 68, 228, 69, 228, 70, 228, 71, 228, 73, 228, 118, 228],
	#[65535, 129, 14, 129, 12, 129, 18, 129, 0, 129, 1, 129, 4, 129, 19, 129, 7, 129, 9, 129, 10, 129, 16, 129, 34, 129, 13, 129, 15, 129, 17, 129, 20, 129, 21, 129, 22, 129, 23, 129, 24, 129, 25, 129, 26, 129, 27, 129, 28, 129, 29, 129, 43, 129, 32, 129, 41, 129, 40, 129, 63, 129, 82, 129, 74, 129, 75, 129, 76, 129, 77, 129, 101, 129, 83, 129, 95, 129, 96, 129, 97, 129, 99, 129, 100, 129, 102, 129, 103, 129, 104, 129, 105, 129, 106, 129, 107, 129, 108, 129, 109, 129, 110, 129, 111, 129, 112, 129, 113, 129, 114, 129, 115, 129, 116, 129, 117, 129, 118, 129, 119, 129, 121, 129, 122, 129, 123, 129, 124, 129, 125, 129, 126, 129, 127, 129, 129, 129, 130, 129, 131, 129, 132, 129, 133, 129, 136, 129, 137, 129, 138, 129, 139, 129, 140, 129, 141, 129, 163, 129, 164, 129, 165, 129, 166, 129, 167, 129, 169, 129, 170, 129, 188, 129, 189, 129, 190, 129, 192, 129, 193, 129, 195, 129, 196, 129, 197, 129, 198, 129, 199, 129, 200, 129, 202, 129, 203, 129, 204, 129, 205, 129, 225, 129, 226, 129],
	#[7, -13, 11, -127, 33, -139, 35, -123, 31, -128, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -341],
	#[38, -346],
	#[13, -344, 43, -343],
	#[65535, 179, 7, 179, 23, 179, 14, 179, 11, 179, 0, 179, 1, 179, 4, 179, 19, 179, 18, 179, 9, 179, 10, 179, 16, 179, 12, 179, 13, 179, 15, 179, 17, 179, 20, 179, 21, 179, 22, 179, 34, 179, 24, 179, 25, 179, 26, 179, 27, 179, 28, 179, 29, 179, 43, 179, 42, 179, 32, 179, 41, 179, 40, 179, 36, 179, 57, 179, 58, 179, 59, 179, 60, 179, 62, 179, 63, 179, 82, 179, 74, 179, 75, 179, 76, 179, 77, 179, 101, 179, 83, 179, 93, 179, 95, 179, 96, 179, 97, 179, 99, 179, 100, 179, 102, 179, 103, 179, 104, 179, 105, 179, 106, 179, 107, 179, 108, 179, 109, 179, 110, 179, 111, 179, 112, 179, 113, 179, 114, 179, 115, 179, 116, 179, 117, 179, 118, 179, 119, 179, 121, 179, 122, 179, 123, 179, 124, 179, 125, 179, 126, 179, 127, 179, 129, 179, 130, 179, 131, 179, 132, 179, 133, 179, 136, 179, 137, 179, 138, 179, 139, 179, 140, 179, 141, 179, 162, 179, 163, 179, 164, 179, 165, 179, 166, 179, 167, 179, 168, 179, 169, 179, 170, 179, 188, 179, 189, 179, 190, 179, 192, 179, 193, 179, 195, 179, 196, 179, 197, 179, 198, 179, 199, 179, 200, 179, 201, 179, 202, 179, 203, 179, 204, 179, 205, 179, 206, 179, 207, 179, 208, 179, 209, 179, 210, 179, 211, 179, 212, 179, 220, 179, 221, 179, 222, 179, 223, 179, 225, 179, 226, 179],
	#[38, -345],
	#[65535, 194, 7, 194, 13, 194, 43, 194, 63, 194, 101, 194, 95, 194, 97, 194, 99, 194, 100, 194, 102, 194, 103, 194, 104, 194, 105, 194, 106, 194, 107, 194, 108, 194, 109, 194, 110, 194, 111, 194, 112, 194, 113, 194, 114, 194, 115, 194, 116, 194, 117, 194, 118, 194, 121, 194, 122, 194, 123, 194, 124, 194, 125, 194, 126, 194, 127, 194, 129, 194, 130, 194, 131, 194, 132, 194, 133, 194, 136, 194, 137, 194, 138, 194, 139, 194, 140, 194, 141, 194, 163, 194, 164, 194, 165, 194, 166, 194, 167, 194, 169, 194, 170, 194, 188, 194, 189, 194, 190, 194, 192, 194, 193, 194, 195, 194, 196, 194, 197, 194, 198, 194, 199, 194, 200, 194, 204, 194, 205, 194],
	#[65535, 193, 7, 193, 13, 193, 102, 193, 63, 193, 95, 193, 110, 193, 43, 193, 101, 193, 103, 193, 106, 193, 109, 193, 111, 193, 124, 193, 108, 193, 123, 193, 104, 193, 99, 193, 97, 193, 105, 193, 113, 193, 116, 193, 100, 193, 115, 193, 112, 193, 121, 193, 107, 193, 114, 193, 122, 193, 117, 193, 118, 193, 125, 193, 126, 193, 127, 193, 129, 193, 130, 193, 131, 193, 132, 193, 133, 193, 136, 193, 137, 193, 138, 193, 139, 193, 140, 193, 141, 193, 163, 193, 164, 193, 165, 193, 166, 193, 167, 193, 169, 193, 170, 193, 188, 193, 189, 193, 190, 193, 192, 193, 193, 193, 195, 193, 196, 193, 197, 193, 198, 193, 199, 193, 200, 193, 204, 193, 205, 193],
	#[31, -128, 11, -127, 36, -129, 7, -13, 33, -139, 35, -123, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -349],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -351],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[43, -353],
	#[65535, 171, 7, 171, 0, 171, 1, 171, 4, 171, 18, 171, 9, 171, 10, 171, 11, 171, 12, 171, 13, 171, 14, 171, 15, 171, 16, 171, 17, 171, 25, 171, 19, 171, 20, 171, 21, 171, 22, 171, 23, 171, 24, 171, 26, 171, 27, 171, 28, 171, 29, 171, 32, 171, 34, 171, 36, 171, 40, 171, 41, 171, 42, 171, 43, 171, 57, 171, 58, 171, 59, 171, 60, 171, 62, 171, 63, 171, 74, 171, 75, 171, 76, 171, 77, 171, 82, 171, 83, 171, 93, 171, 95, 171, 96, 171, 97, 171, 99, 171, 100, 171, 101, 171, 102, 171, 103, 171, 104, 171, 105, 171, 106, 171, 107, 171, 108, 171, 109, 171, 110, 171, 111, 171, 112, 171, 113, 171, 114, 171, 115, 171, 116, 171, 117, 171, 118, 171, 119, 171, 121, 171, 122, 171, 123, 171, 124, 171, 125, 171, 126, 171, 127, 171, 129, 171, 130, 171, 131, 171, 132, 171, 133, 171, 136, 171, 137, 171, 138, 171, 139, 171, 140, 171, 141, 171, 162, 171, 163, 171, 164, 171, 165, 171, 166, 171, 167, 171, 168, 171, 169, 171, 170, 171, 188, 171, 189, 171, 190, 171, 192, 171, 193, 171, 195, 171, 196, 171, 197, 171, 198, 171, 199, 171, 200, 171, 201, 171, 202, 171, 203, 171, 204, 171, 205, 171, 206, 171, 207, 171, 208, 171, 209, 171, 210, 171, 211, 171, 212, 171, 220, 171, 221, 171, 222, 171, 223, 171, 225, 171, 226, 171],
	#[119, -33],
	#[13, -356],
	#[7, -13, 118, -36],
	#[43, -358],
	#[65535, 145, 21, 145, 20, 145, 4, 145, 23, 145, 28, 145, 10, 145, 19, 145, 119, 145, 13, 145, 18, 145, 7, 145, 95, 145, 14, 145, 1, 145, 110, 145, 12, 145, 15, 145, 103, 145, 22, 145, 16, 145, 17, 145, 29, 145, 0, 145, 27, 145, 111, 145, 26, 145, 9, 145, 76, 145, 102, 145, 74, 145, 41, 145, 77, 145, 132, 145, 34, 145, 63, 145, 101, 145, 82, 145, 107, 145, 43, 145, 127, 145, 24, 145, 25, 145, 99, 145, 139, 145, 75, 145, 115, 145, 40, 145, 129, 145, 133, 145, 106, 145, 100, 145, 126, 145, 83, 145, 32, 145, 122, 145, 104, 145, 124, 145, 108, 145, 125, 145, 121, 145, 109, 145, 97, 145, 123, 145, 114, 145, 117, 145, 105, 145, 112, 145, 141, 145, 96, 145, 205, 145, 116, 145, 113, 145, 118, 145, 140, 145, 138, 145, 137, 145, 130, 145, 131, 145, 136, 145, 170, 145, 203, 145, 163, 145, 164, 145, 165, 145, 166, 145, 167, 145, 169, 145, 204, 145, 198, 145, 188, 145, 189, 145, 190, 145, 192, 145, 193, 145, 195, 145, 196, 145, 197, 145, 199, 145, 200, 145, 202, 145, 225, 145, 226, 145],
	#[65535, 240, 31, 240, 7, 240, 11, 240, 33, 240, 42, 240, 35, 240, 36, 240, 63, 240, 64, 240, 65, 240, 66, 240, 67, 240, 68, 240, 69, 240, 70, 240, 71, 240, 73, 240, 118, 240],
	#[65535, 230, 7, 230, 11, 230, 31, 230, 33, 230, 35, 230, 36, 230, 42, 230, 63, 230, 64, 230, 65, 230, 66, 230, 67, 230, 68, 230, 69, 230, 70, 230, 71, 230, 73, 230, 118, 230],
	#[42, -363],
	#[65535, 422, 31, 422, 7, 422, 11, 422, 42, 422, 33, 422, 35, 422, 36, 422, 63, 422, 64, 422, 65, 422, 66, 422, 67, 422, 68, 422, 69, 422, 70, 422, 71, 422, 73, 422, 118, 422],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 219, 13, 219, 43, 219],
	#[13, -367, 43, -366],
	#[65535, 170, 62, 170, 110, 170, 106, 170, 109, 170, 34, 170, 21, 170, 102, 170, 108, 170, 14, 170, 60, 170, 12, 170, 18, 170, 11, 170, 0, 170, 1, 170, 9, 170, 20, 170, 4, 170, 19, 170, 7, 170, 57, 170, 10, 170, 16, 170, 32, 170, 13, 170, 15, 170, 17, 170, 25, 170, 36, 170, 77, 170, 22, 170, 23, 170, 24, 170, 76, 170, 26, 170, 27, 170, 28, 170, 29, 170, 43, 170, 42, 170, 100, 170, 59, 170, 63, 170, 75, 170, 124, 170, 40, 170, 41, 170, 107, 170, 101, 170, 103, 170, 58, 170, 82, 170, 115, 170, 167, 170, 74, 170, 111, 170, 93, 170, 83, 170, 123, 170, 104, 170, 99, 170, 119, 170, 97, 170, 105, 170, 113, 170, 116, 170, 95, 170, 112, 170, 121, 170, 96, 170, 114, 170, 122, 170, 117, 170, 118, 170, 125, 170, 126, 170, 127, 170, 129, 170, 130, 170, 131, 170, 132, 170, 133, 170, 136, 170, 137, 170, 138, 170, 139, 170, 140, 170, 141, 170, 162, 170, 163, 170, 164, 170, 165, 170, 166, 170, 190, 170, 168, 170, 169, 170, 170, 170, 206, 170, 205, 170, 204, 170, 188, 170, 189, 170, 203, 170, 202, 170, 192, 170, 193, 170, 195, 170, 196, 170, 197, 170, 198, 170, 199, 170, 200, 170, 201, 170, 209, 170, 207, 170, 208, 170, 210, 170, 211, 170, 212, 170, 220, 170, 221, 170, 222, 170, 223, 170, 225, 170, 226, 170],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 220, 7, 220, 13, 220, 43, 220, 63, 220, 95, 220, 97, 220, 99, 220, 100, 220, 101, 220, 102, 220, 103, 220, 104, 220, 105, 220, 106, 220, 107, 220, 108, 220, 109, 220, 110, 220, 111, 220, 112, 220, 113, 220, 114, 220, 115, 220, 116, 220, 117, 220, 118, 220, 121, 220, 122, 220, 123, 220, 124, 220, 125, 220, 126, 220, 127, 220, 129, 220, 130, 220, 131, 220, 132, 220, 133, 220, 136, 220, 137, 220, 138, 220, 139, 220, 140, 220, 141, 220, 163, 220, 164, 220, 165, 220, 166, 220, 167, 220, 169, 220, 170, 220, 188, 220, 189, 220, 190, 220, 192, 220, 193, 220, 195, 220, 196, 220, 197, 220, 198, 220, 199, 220, 200, 220, 204, 220, 205, 220],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -371],
	#[7, -13, 69, -136, 65, -122, 36, -129, 35, -123, 11, -127, 31, -128, 73, -138, 66, -126, 68, -133, 71, -130, 67, -137, 33, -139, 70, -140, 118, -36, 64, -141, 63, -121],
	#[43, -373],
	#[65535, 176, 0, 176, 1, 176, 4, 176, 7, 176, 9, 176, 10, 176, 11, 176, 12, 176, 13, 176, 14, 176, 15, 176, 16, 176, 17, 176, 18, 176, 19, 176, 20, 176, 21, 176, 22, 176, 23, 176, 24, 176, 25, 176, 26, 176, 27, 176, 28, 176, 29, 176, 32, 176, 34, 176, 36, 176, 40, 176, 41, 176, 42, 176, 43, 176, 57, 176, 58, 176, 59, 176, 60, 176, 62, 176, 63, 176, 74, 176, 75, 176, 76, 176, 77, 176, 82, 176, 83, 176, 93, 176, 95, 176, 96, 176, 97, 176, 99, 176, 100, 176, 101, 176, 102, 176, 103, 176, 104, 176, 105, 176, 106, 176, 107, 176, 108, 176, 109, 176, 110, 176, 111, 176, 112, 176, 113, 176, 114, 176, 115, 176, 116, 176, 117, 176, 118, 176, 119, 176, 121, 176, 122, 176, 123, 176, 124, 176, 125, 176, 126, 176, 127, 176, 129, 176, 130, 176, 131, 176, 132, 176, 133, 176, 136, 176, 137, 176, 138, 176, 139, 176, 140, 176, 141, 176, 162, 176, 163, 176, 164, 176, 165, 176, 166, 176, 167, 176, 168, 176, 169, 176, 170, 176, 188, 176, 189, 176, 190, 176, 192, 176, 193, 176, 195, 176, 196, 176, 197, 176, 198, 176, 199, 176, 200, 176, 201, 176, 202, 176, 203, 176, 204, 176, 205, 176, 206, 176, 207, 176, 208, 176, 209, 176, 210, 176, 211, 176, 212, 176, 220, 176, 221, 176, 222, 176, 223, 176, 225, 176, 226, 176],
	#[65535, 257, 31, 257, 7, 257, 11, 257, 36, 257, 42, 257, 35, 257, 33, 257, 71, 257, 70, 257, 69, 257, 68, 257, 67, 257, 65, 257, 73, 257, 63, 257, 64, 257, 66, 257, 118, 257],
	#[65535, 251, 31, 251, 7, 251, 11, 251, 36, 251, 35, 251, 33, 251, 42, 251, 65, 251, 63, 251, 64, 251, 66, 251, 67, 251, 68, 251, 69, 251, 70, 251, 71, 251, 73, 251, 118, 251],
	#[65535, 252, 7, 252, 11, 252, 31, 252, 33, 252, 35, 252, 36, 252, 42, 252, 63, 252, 64, 252, 65, 252, 66, 252, 67, 252, 68, 252, 69, 252, 70, 252, 71, 252, 73, 252, 118, 252],
	#[65535, 253, 31, 253, 7, 253, 11, 253, 42, 253, 33, 253, 35, 253, 36, 253, 63, 253, 64, 253, 65, 253, 66, 253, 67, 253, 68, 253, 69, 253, 70, 253, 71, 253, 73, 253, 118, 253],
	#[42, -385],
	#[65535, 254, 7, 254, 31, 254, 11, 254, 42, 254, 33, 254, 35, 254, 36, 254, 63, 254, 64, 254, 65, 254, 66, 254, 67, 254, 68, 254, 69, 254, 70, 254, 71, 254, 73, 254, 118, 254],
	#[65535, 248, 7, 248, 11, 248, 31, 248, 33, 248, 35, 248, 36, 248, 42, 248, 63, 248, 64, 248, 65, 248, 66, 248, 67, 248, 68, 248, 69, 248, 70, 248, 71, 248, 73, 248, 118, 248],
	#[65535, 255, 7, 255, 11, 255, 31, 255, 33, 255, 35, 255, 36, 255, 42, 255, 63, 255, 64, 255, 65, 255, 66, 255, 67, 255, 68, 255, 69, 255, 70, 255, 71, 255, 73, 255, 118, 255],
	#[65535, 249, 7, 249, 31, 249, 11, 249, 42, 249, 33, 249, 35, 249, 36, 249, 65, 249, 63, 249, 64, 249, 66, 249, 67, 249, 68, 249, 69, 249, 70, 249, 71, 249, 73, 249, 118, 249],
	#[65535, 256, 7, 256, 31, 256, 11, 256, 42, 256, 33, 256, 35, 256, 36, 256, 65, 256, 63, 256, 64, 256, 66, 256, 67, 256, 68, 256, 69, 256, 70, 256, 71, 256, 73, 256, 118, 256],
	#[65535, 250, 7, 250, 11, 250, 31, 250, 33, 250, 35, 250, 36, 250, 42, 250, 63, 250, 64, 250, 65, 250, 66, 250, 67, 250, 68, 250, 69, 250, 70, 250, 71, 250, 73, 250, 118, 250],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -387],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 65, -122, 63, -121, 64, -141, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[43, -389],
	#[65535, 174, 0, 174, 1, 174, 4, 174, 7, 174, 9, 174, 10, 174, 11, 174, 12, 174, 13, 174, 14, 174, 15, 174, 16, 174, 17, 174, 18, 174, 19, 174, 20, 174, 21, 174, 22, 174, 23, 174, 24, 174, 25, 174, 26, 174, 27, 174, 28, 174, 29, 174, 32, 174, 34, 174, 36, 174, 40, 174, 41, 174, 42, 174, 43, 174, 57, 174, 58, 174, 59, 174, 60, 174, 62, 174, 63, 174, 74, 174, 75, 174, 76, 174, 77, 174, 82, 174, 83, 174, 93, 174, 95, 174, 96, 174, 97, 174, 99, 174, 100, 174, 101, 174, 102, 174, 103, 174, 104, 174, 105, 174, 106, 174, 107, 174, 108, 174, 109, 174, 110, 174, 111, 174, 112, 174, 113, 174, 114, 174, 115, 174, 116, 174, 117, 174, 118, 174, 119, 174, 121, 174, 122, 174, 123, 174, 124, 174, 125, 174, 126, 174, 127, 174, 129, 174, 130, 174, 131, 174, 132, 174, 133, 174, 136, 174, 137, 174, 138, 174, 139, 174, 140, 174, 141, 174, 162, 174, 163, 174, 164, 174, 165, 174, 166, 174, 167, 174, 168, 174, 169, 174, 170, 174, 188, 174, 189, 174, 190, 174, 192, 174, 193, 174, 195, 174, 196, 174, 197, 174, 198, 174, 199, 174, 200, 174, 201, 174, 202, 174, 203, 174, 204, 174, 205, 174, 206, 174, 207, 174, 208, 174, 209, 174, 210, 174, 211, 174, 212, 174, 220, 174, 221, 174, 222, 174, 223, 174, 225, 174, 226, 174],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -392],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[43, -394],
	#[65535, 172, 0, 172, 1, 172, 4, 172, 7, 172, 9, 172, 10, 172, 11, 172, 12, 172, 13, 172, 14, 172, 15, 172, 16, 172, 17, 172, 18, 172, 19, 172, 20, 172, 21, 172, 22, 172, 23, 172, 24, 172, 25, 172, 26, 172, 27, 172, 28, 172, 29, 172, 32, 172, 34, 172, 36, 172, 40, 172, 41, 172, 42, 172, 43, 172, 57, 172, 58, 172, 59, 172, 60, 172, 62, 172, 63, 172, 74, 172, 75, 172, 76, 172, 77, 172, 82, 172, 83, 172, 93, 172, 95, 172, 96, 172, 97, 172, 99, 172, 100, 172, 101, 172, 102, 172, 103, 172, 104, 172, 105, 172, 106, 172, 107, 172, 108, 172, 109, 172, 110, 172, 111, 172, 112, 172, 113, 172, 114, 172, 115, 172, 116, 172, 117, 172, 118, 172, 119, 172, 121, 172, 122, 172, 123, 172, 124, 172, 125, 172, 126, 172, 127, 172, 129, 172, 130, 172, 131, 172, 132, 172, 133, 172, 136, 172, 137, 172, 138, 172, 139, 172, 140, 172, 141, 172, 162, 172, 163, 172, 164, 172, 165, 172, 166, 172, 167, 172, 168, 172, 169, 172, 170, 172, 188, 172, 189, 172, 190, 172, 192, 172, 193, 172, 195, 172, 196, 172, 197, 172, 198, 172, 199, 172, 200, 172, 201, 172, 202, 172, 203, 172, 204, 172, 205, 172, 206, 172, 207, 172, 208, 172, 209, 172, 210, 172, 211, 172, 212, 172, 220, 172, 221, 172, 222, 172, 223, 172, 225, 172, 226, 172],
	#[31, -128, 36, -129, 11, -127, 63, -121, 7, -13, 33, -139, 35, -123, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -397],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -341],
	#[13, -344, 43, -400],
	#[65535, 180, 0, 180, 1, 180, 4, 180, 7, 180, 9, 180, 10, 180, 11, 180, 12, 180, 13, 180, 14, 180, 15, 180, 16, 180, 17, 180, 18, 180, 19, 180, 20, 180, 21, 180, 22, 180, 23, 180, 24, 180, 25, 180, 26, 180, 27, 180, 28, 180, 29, 180, 32, 180, 34, 180, 36, 180, 40, 180, 41, 180, 42, 180, 43, 180, 57, 180, 58, 180, 59, 180, 60, 180, 62, 180, 63, 180, 74, 180, 75, 180, 76, 180, 77, 180, 82, 180, 83, 180, 93, 180, 95, 180, 96, 180, 97, 180, 99, 180, 100, 180, 101, 180, 102, 180, 103, 180, 104, 180, 105, 180, 106, 180, 107, 180, 108, 180, 109, 180, 110, 180, 111, 180, 112, 180, 113, 180, 114, 180, 115, 180, 116, 180, 117, 180, 118, 180, 119, 180, 121, 180, 122, 180, 123, 180, 124, 180, 125, 180, 126, 180, 127, 180, 129, 180, 130, 180, 131, 180, 132, 180, 133, 180, 136, 180, 137, 180, 138, 180, 139, 180, 140, 180, 141, 180, 162, 180, 163, 180, 164, 180, 165, 180, 166, 180, 167, 180, 168, 180, 169, 180, 170, 180, 188, 180, 189, 180, 190, 180, 192, 180, 193, 180, 195, 180, 196, 180, 197, 180, 198, 180, 199, 180, 200, 180, 201, 180, 202, 180, 203, 180, 204, 180, 205, 180, 206, 180, 207, 180, 208, 180, 209, 180, 210, 180, 211, 180, 212, 180, 220, 180, 221, 180, 222, 180, 223, 180, 225, 180, 226, 180],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -403],
	#[11, -127, 31, -128, 63, -121, 33, -139, 35, -123, 64, -141, 36, -129, 67, -137, 65, -122, 73, -138, 66, -126, 70, -140, 118, -36, 7, -13, 69, -136, 68, -133, 71, -130],
	#[43, -405],
	#[65535, 173, 0, 173, 1, 173, 4, 173, 7, 173, 9, 173, 10, 173, 11, 173, 12, 173, 13, 173, 14, 173, 15, 173, 16, 173, 17, 173, 18, 173, 19, 173, 20, 173, 21, 173, 22, 173, 23, 173, 24, 173, 25, 173, 26, 173, 27, 173, 28, 173, 29, 173, 32, 173, 34, 173, 36, 173, 40, 173, 41, 173, 42, 173, 43, 173, 57, 173, 58, 173, 59, 173, 60, 173, 62, 173, 63, 173, 74, 173, 75, 173, 76, 173, 77, 173, 82, 173, 83, 173, 93, 173, 95, 173, 96, 173, 97, 173, 99, 173, 100, 173, 101, 173, 102, 173, 103, 173, 104, 173, 105, 173, 106, 173, 107, 173, 108, 173, 109, 173, 110, 173, 111, 173, 112, 173, 113, 173, 114, 173, 115, 173, 116, 173, 117, 173, 118, 173, 119, 173, 121, 173, 122, 173, 123, 173, 124, 173, 125, 173, 126, 173, 127, 173, 129, 173, 130, 173, 131, 173, 132, 173, 133, 173, 136, 173, 137, 173, 138, 173, 139, 173, 140, 173, 141, 173, 162, 173, 163, 173, 164, 173, 165, 173, 166, 173, 167, 173, 168, 173, 169, 173, 170, 173, 188, 173, 189, 173, 190, 173, 192, 173, 193, 173, 195, 173, 196, 173, 197, 173, 198, 173, 199, 173, 200, 173, 201, 173, 202, 173, 203, 173, 204, 173, 205, 173, 206, 173, 207, 173, 208, 173, 209, 173, 210, 173, 211, 173, 212, 173, 220, 173, 221, 173, 222, 173, 223, 173, 225, 173, 226, 173],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -408],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -410],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[43, -412],
	#[65535, 178, 10, 178, 34, 178, 1, 178, 12, 178, 0, 178, 4, 178, 19, 178, 7, 178, 9, 178, 29, 178, 11, 178, 32, 178, 13, 178, 14, 178, 15, 178, 16, 178, 17, 178, 18, 178, 36, 178, 20, 178, 21, 178, 22, 178, 23, 178, 24, 178, 25, 178, 26, 178, 27, 178, 28, 178, 40, 178, 41, 178, 42, 178, 43, 178, 57, 178, 58, 178, 59, 178, 60, 178, 62, 178, 63, 178, 74, 178, 75, 178, 76, 178, 77, 178, 82, 178, 83, 178, 93, 178, 95, 178, 96, 178, 97, 178, 99, 178, 100, 178, 101, 178, 102, 178, 103, 178, 104, 178, 105, 178, 106, 178, 107, 178, 108, 178, 109, 178, 110, 178, 111, 178, 112, 178, 113, 178, 114, 178, 115, 178, 116, 178, 117, 178, 118, 178, 119, 178, 121, 178, 122, 178, 123, 178, 124, 178, 125, 178, 126, 178, 127, 178, 129, 178, 130, 178, 131, 178, 132, 178, 133, 178, 136, 178, 137, 178, 138, 178, 139, 178, 140, 178, 141, 178, 162, 178, 163, 178, 164, 178, 165, 178, 166, 178, 167, 178, 168, 178, 169, 178, 170, 178, 188, 178, 189, 178, 190, 178, 192, 178, 193, 178, 195, 178, 196, 178, 197, 178, 198, 178, 199, 178, 200, 178, 201, 178, 202, 178, 203, 178, 204, 178, 205, 178, 206, 178, 207, 178, 208, 178, 209, 178, 210, 178, 211, 178, 212, 178, 220, 178, 221, 178, 222, 178, 223, 178, 225, 178, 226, 178],
	#[13, -282, 32, -415],
	#[65535, 132, 15, 132, 103, 132, 27, 132, 12, 132, 0, 132, 1, 132, 4, 132, 7, 132, 9, 132, 10, 132, 34, 132, 13, 132, 14, 132, 16, 132, 17, 132, 18, 132, 19, 132, 20, 132, 21, 132, 22, 132, 23, 132, 24, 132, 25, 132, 26, 132, 32, 132, 28, 132, 29, 132, 74, 132, 119, 132, 63, 132, 95, 132, 77, 132, 40, 132, 41, 132, 43, 132, 82, 132, 115, 132, 109, 132, 111, 132, 102, 132, 108, 132, 123, 132, 76, 132, 99, 132, 96, 132, 106, 132, 101, 132, 110, 132, 116, 132, 100, 132, 83, 132, 113, 132, 124, 132, 104, 132, 75, 132, 97, 132, 105, 132, 121, 132, 107, 132, 112, 132, 114, 132, 122, 132, 117, 132, 118, 132, 125, 132, 126, 132, 127, 132, 129, 132, 130, 132, 131, 132, 132, 132, 133, 132, 136, 132, 137, 132, 138, 132, 139, 132, 140, 132, 141, 132, 188, 132, 163, 132, 164, 132, 165, 132, 166, 132, 167, 132, 169, 132, 170, 132, 189, 132, 190, 132, 192, 132, 193, 132, 195, 132, 196, 132, 197, 132, 198, 132, 199, 132, 200, 132, 202, 132, 203, 132, 204, 132, 205, 132, 225, 132, 226, 132],
	#[65535, 131, 23, 131, 10, 131, 29, 131, 22, 131, 19, 131, 77, 131, 21, 131, 1, 131, 27, 131, 14, 131, 12, 131, 96, 131, 0, 131, 4, 131, 63, 131, 16, 131, 7, 131, 95, 131, 9, 131, 102, 131, 25, 131, 34, 131, 13, 131, 15, 131, 17, 131, 18, 131, 123, 131, 20, 131, 32, 131, 122, 131, 24, 131, 101, 131, 26, 131, 41, 131, 28, 131, 82, 131, 40, 131, 110, 131, 43, 131, 83, 131, 103, 131, 112, 131, 116, 131, 75, 131, 109, 131, 97, 131, 124, 131, 108, 131, 121, 131, 76, 131, 99, 131, 119, 131, 106, 131, 74, 131, 113, 131, 100, 131, 115, 131, 105, 131, 104, 131, 107, 131, 114, 131, 111, 131, 117, 131, 118, 131, 125, 131, 126, 131, 127, 131, 129, 131, 130, 131, 131, 131, 132, 131, 133, 131, 136, 131, 137, 131, 138, 131, 139, 131, 140, 131, 141, 131, 163, 131, 164, 131, 165, 131, 166, 131, 167, 131, 169, 131, 170, 131, 188, 131, 189, 131, 190, 131, 192, 131, 193, 131, 195, 131, 196, 131, 197, 131, 198, 131, 199, 131, 200, 131, 202, 131, 203, 131, 204, 131, 205, 131, 225, 131, 226, 131],
	#[65535, 225, 7, 225, 11, 225, 31, 225, 33, 225, 35, 225, 36, 225, 42, 225, 63, 225, 64, 225, 65, 225, 66, 225, 67, 225, 68, 225, 69, 225, 70, 225, 71, 225, 73, 225, 118, 225],
	#[65535, 242, 7, 242, 11, 242, 31, 242, 33, 242, 35, 242, 36, 242, 42, 242, 63, 242, 64, 242, 65, 242, 66, 242, 67, 242, 68, 242, 69, 242, 70, 242, 71, 242, 73, 242, 118, 242],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[96, -420],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[37, 105, 42, -159, 43, -422, 72, -161],
	#[65535, 169, 7, 169, 60, 169, 77, 169, 106, 169, 12, 169, 18, 169, 42, 169, 10, 169, 0, 169, 1, 169, 4, 169, 83, 169, 9, 169, 29, 169, 11, 169, 34, 169, 13, 169, 14, 169, 15, 169, 16, 169, 17, 169, 110, 169, 19, 169, 20, 169, 21, 169, 22, 169, 23, 169, 24, 169, 25, 169, 26, 169, 27, 169, 28, 169, 43, 169, 116, 169, 32, 169, 109, 169, 41, 169, 40, 169, 36, 169, 95, 169, 62, 169, 101, 169, 103, 169, 99, 169, 57, 169, 96, 169, 74, 169, 111, 169, 93, 169, 102, 169, 108, 169, 58, 169, 59, 169, 82, 169, 119, 169, 63, 169, 113, 169, 100, 169, 97, 169, 105, 169, 123, 169, 124, 169, 104, 169, 75, 169, 76, 169, 112, 169, 121, 169, 107, 169, 115, 169, 114, 169, 122, 169, 117, 169, 118, 169, 125, 169, 126, 169, 127, 169, 129, 169, 130, 169, 131, 169, 132, 169, 133, 169, 136, 169, 137, 169, 138, 169, 139, 169, 140, 169, 141, 169, 188, 169, 162, 169, 163, 169, 164, 169, 165, 169, 166, 169, 167, 169, 168, 169, 169, 169, 170, 169, 192, 169, 189, 169, 190, 169, 193, 169, 195, 169, 196, 169, 197, 169, 198, 169, 199, 169, 200, 169, 201, 169, 202, 169, 203, 169, 204, 169, 205, 169, 206, 169, 207, 169, 208, 169, 209, 169, 210, 169, 211, 169, 212, 169, 220, 169, 221, 169, 222, 169, 223, 169, 225, 169, 226, 169],
	#[65535, 221, 7, 221, 11, 221, 42, 221, 31, 221, 33, 221, 35, 221, 36, 221, 63, 221, 64, 221, 65, 221, 66, 221, 67, 221, 68, 221, 69, 221, 70, 221, 71, 221, 73, 221, 118, 221],
	#[65535, 134, 0, 134, 1, 134, 4, 134, 7, 134, 9, 134, 10, 134, 12, 134, 13, 134, 14, 134, 15, 134, 16, 134, 17, 134, 18, 134, 19, 134, 20, 134, 21, 134, 22, 134, 23, 134, 24, 134, 25, 134, 26, 134, 27, 134, 28, 134, 29, 134, 32, 134, 34, 134, 40, 134, 41, 134, 43, 134, 63, 134, 74, 134, 75, 134, 76, 134, 77, 134, 82, 134, 83, 134, 95, 134, 96, 134, 97, 134, 99, 134, 100, 134, 101, 134, 102, 134, 103, 134, 104, 134, 105, 134, 106, 134, 107, 134, 108, 134, 109, 134, 110, 134, 111, 134, 112, 134, 113, 134, 114, 134, 115, 134, 116, 134, 117, 134, 118, 134, 119, 134, 121, 134, 122, 134, 123, 134, 124, 134, 125, 134, 126, 134, 127, 134, 129, 134, 130, 134, 131, 134, 132, 134, 133, 134, 136, 134, 137, 134, 138, 134, 139, 134, 140, 134, 141, 134, 163, 134, 164, 134, 165, 134, 166, 134, 167, 134, 169, 134, 170, 134, 188, 134, 189, 134, 190, 134, 192, 134, 193, 134, 195, 134, 196, 134, 197, 134, 198, 134, 199, 134, 200, 134, 202, 134, 203, 134, 204, 134, 205, 134, 225, 134, 226, 134],
	#[34, -426],
	#[65535, 133, 0, 133, 1, 133, 4, 133, 7, 133, 9, 133, 10, 133, 12, 133, 13, 133, 14, 133, 15, 133, 16, 133, 17, 133, 18, 133, 19, 133, 20, 133, 21, 133, 22, 133, 23, 133, 24, 133, 25, 133, 26, 133, 27, 133, 28, 133, 29, 133, 32, 133, 34, 133, 40, 133, 41, 133, 43, 133, 63, 133, 74, 133, 75, 133, 76, 133, 77, 133, 167, 133, 82, 133, 83, 133, 95, 133, 96, 133, 97, 133, 99, 133, 100, 133, 101, 133, 102, 133, 103, 133, 104, 133, 105, 133, 106, 133, 107, 133, 108, 133, 109, 133, 110, 133, 111, 133, 112, 133, 113, 133, 114, 133, 115, 133, 116, 133, 117, 133, 118, 133, 119, 133, 121, 133, 122, 133, 123, 133, 124, 133, 125, 133, 126, 133, 127, 133, 129, 133, 130, 133, 131, 133, 132, 133, 133, 133, 136, 133, 137, 133, 138, 133, 139, 133, 140, 133, 141, 133, 170, 133, 163, 133, 164, 133, 165, 133, 166, 133, 190, 133, 169, 133, 189, 133, 195, 133, 193, 133, 192, 133, 188, 133, 196, 133, 197, 133, 198, 133, 199, 133, 200, 133, 202, 133, 203, 133, 204, 133, 205, 133, 225, 133, 226, 133],
	#[65535, 388, 109, 388, 126, 388, 7, 388, 95, 388, 141, 388, 125, 388, 13, 388, 210, 388, 60, 388, 57, 388, 133, 388, 167, 388, 42, 388, 59, 388, 58, 388, 11, 388, 63, 388, 118, 388, 32, 388, 36, 388, 103, 388, 43, 388, 12, 388, 132, 388, 93, 388, 117, 388, 124, 388, 108, 388, 99, 388, 190, 388, 131, 388, 102, 388, 62, 388, 101, 388, 130, 388, 116, 388, 100, 388, 122, 388, 107, 388, 127, 388, 123, 388, 140, 388, 129, 388, 195, 388, 104, 388, 114, 388, 111, 388, 138, 388, 113, 388, 121, 388, 110, 388, 97, 388, 106, 388, 96, 388, 112, 388, 105, 388, 136, 388, 139, 388, 115, 388, 137, 388, 221, 388, 162, 388, 206, 388, 212, 388, 169, 388, 163, 388, 164, 388, 165, 388, 166, 388, 205, 388, 168, 388, 202, 388, 170, 388, 192, 388, 222, 388, 203, 388, 207, 388, 193, 388, 188, 388, 189, 388, 196, 388, 197, 388, 198, 388, 199, 388, 200, 388, 201, 388, 209, 388, 208, 388, 204, 388, 211, 388, 220, 388, 223, 388],
	#[13, -282, 32, -431],
	#[7, 94, 10, 94, 11, 94, 31, 94, 33, 94, 34, -430, 37, 94, 38, 94, 42, 94, 63, 94, 72, 94, 84, 94, 85, 94, 86, 94, 87, 94, 88, 94, 89, 94, 90, 94, 91, 94, 92, 94, 95, 94, 97, 94, 99, 94, 100, 94, 101, 94, 102, 94, 103, 94, 104, 94, 105, 94, 106, 94, 107, 94, 108, 94, 109, 94, 110, 94, 111, 94, 112, 94, 113, 94, 114, 94, 115, 94, 116, 94, 117, 94, 118, 94, 119, 94, 121, 94, 122, 94, 123, 94, 124, 94, 125, 94, 126, 94, 127, 94, 129, 94, 130, 94, 131, 94, 132, 94, 133, 94, 136, 94, 137, 94, 138, 94, 139, 94, 140, 94, 141, 94, 226, 94],
	#[65535, 395, 7, 395, 11, 395, 12, 395, 13, 395, 36, 395, 43, 395, 32, 395, 63, 395, 93, 395, 95, 395, 62, 395, 110, 395, 42, 395, 60, 395, 103, 395, 57, 395, 109, 395, 111, 395, 102, 395, 108, 395, 58, 395, 59, 395, 99, 395, 96, 395, 106, 395, 101, 395, 113, 395, 116, 395, 100, 395, 105, 395, 123, 395, 124, 395, 104, 395, 97, 395, 112, 395, 121, 395, 107, 395, 115, 395, 114, 395, 122, 395, 117, 395, 118, 395, 125, 395, 126, 395, 127, 395, 129, 395, 130, 395, 131, 395, 132, 395, 133, 395, 136, 395, 137, 395, 138, 395, 139, 395, 140, 395, 141, 395, 188, 395, 162, 395, 163, 395, 164, 395, 165, 395, 166, 395, 167, 395, 168, 395, 169, 395, 170, 395, 189, 395, 190, 395, 192, 395, 193, 395, 195, 395, 196, 395, 197, 395, 198, 395, 199, 395, 200, 395, 201, 395, 202, 395, 203, 395, 204, 395, 205, 395, 206, 395, 207, 395, 208, 395, 209, 395, 210, 395, 211, 395, 212, 395, 220, 395, 221, 395, 222, 395, 223, 395],
	#[34, -432],
	#[65535, 394, 7, 394, 11, 394, 12, 394, 13, 394, 32, 394, 36, 394, 42, 394, 43, 394, 57, 394, 58, 394, 59, 394, 60, 394, 62, 394, 63, 394, 93, 394, 95, 394, 96, 394, 97, 394, 99, 394, 100, 394, 101, 394, 102, 394, 103, 394, 104, 394, 105, 394, 106, 394, 107, 394, 108, 394, 109, 394, 110, 394, 111, 394, 112, 394, 113, 394, 114, 394, 115, 394, 116, 394, 117, 394, 118, 394, 121, 394, 122, 394, 123, 394, 124, 394, 125, 394, 126, 394, 127, 394, 129, 394, 130, 394, 131, 394, 132, 394, 133, 394, 136, 394, 137, 394, 138, 394, 139, 394, 140, 394, 141, 394, 162, 394, 163, 394, 164, 394, 165, 394, 166, 394, 167, 394, 168, 394, 169, 394, 170, 394, 188, 394, 189, 394, 190, 394, 192, 394, 193, 394, 195, 394, 196, 394, 197, 394, 198, 394, 199, 394, 200, 394, 201, 394, 202, 394, 203, 394, 204, 394, 205, 394, 206, 394, 207, 394, 208, 394, 209, 394, 210, 394, 211, 394, 212, 394, 220, 394, 221, 394, 222, 394, 223, 394],
	#[12, -435, 13, -282],
	#[65535, 390, 7, 390, 11, 390, 12, 390, 13, 390, 43, 390, 32, 390, 63, 390, 36, 390, 95, 390, 62, 390, 110, 390, 42, 390, 60, 390, 103, 390, 109, 390, 111, 390, 93, 390, 102, 390, 57, 390, 58, 390, 59, 390, 99, 390, 96, 390, 106, 390, 101, 390, 113, 390, 116, 390, 100, 390, 104, 390, 105, 390, 123, 390, 124, 390, 108, 390, 97, 390, 112, 390, 121, 390, 107, 390, 115, 390, 114, 390, 122, 390, 117, 390, 118, 390, 125, 390, 126, 390, 127, 390, 129, 390, 130, 390, 131, 390, 132, 390, 133, 390, 136, 390, 137, 390, 138, 390, 139, 390, 140, 390, 141, 390, 188, 390, 162, 390, 163, 390, 164, 390, 165, 390, 166, 390, 167, 390, 168, 390, 169, 390, 170, 390, 189, 390, 190, 390, 192, 390, 193, 390, 195, 390, 196, 390, 197, 390, 198, 390, 199, 390, 200, 390, 201, 390, 202, 390, 203, 390, 204, 390, 205, 390, 206, 390, 207, 390, 208, 390, 209, 390, 210, 390, 211, 390, 212, 390, 220, 390, 221, 390, 222, 390, 223, 390],
	#[65535, 389, 7, 389, 11, 389, 12, 389, 13, 389, 110, 389, 36, 389, 42, 389, 60, 389, 43, 389, 116, 389, 32, 389, 63, 389, 93, 389, 95, 389, 62, 389, 103, 389, 106, 389, 109, 389, 111, 389, 102, 389, 57, 389, 58, 389, 59, 389, 99, 389, 96, 389, 101, 389, 113, 389, 100, 389, 104, 389, 105, 389, 123, 389, 124, 389, 108, 389, 97, 389, 112, 389, 121, 389, 107, 389, 115, 389, 114, 389, 122, 389, 117, 389, 118, 389, 125, 389, 126, 389, 127, 389, 129, 389, 130, 389, 131, 389, 132, 389, 133, 389, 136, 389, 137, 389, 138, 389, 139, 389, 140, 389, 141, 389, 188, 389, 162, 389, 163, 389, 164, 389, 165, 389, 166, 389, 167, 389, 168, 389, 169, 389, 170, 389, 192, 389, 189, 389, 190, 389, 193, 389, 195, 389, 196, 389, 197, 389, 198, 389, 199, 389, 200, 389, 201, 389, 202, 389, 203, 389, 204, 389, 205, 389, 206, 389, 207, 389, 208, 389, 209, 389, 210, 389, 211, 389, 212, 389, 220, 389, 221, 389, 222, 389, 223, 389],
	#[65535, 391, 7, 391, 11, 391, 12, 391, 13, 391, 43, 391, 32, 391, 36, 391, 42, 391, 57, 391, 58, 391, 59, 391, 60, 391, 62, 391, 63, 391, 111, 391, 106, 391, 99, 391, 110, 391, 93, 391, 107, 391, 95, 391, 96, 391, 97, 391, 104, 391, 100, 391, 101, 391, 102, 391, 103, 391, 105, 391, 124, 391, 108, 391, 109, 391, 123, 391, 122, 391, 112, 391, 113, 391, 114, 391, 115, 391, 116, 391, 117, 391, 118, 391, 121, 391, 125, 391, 126, 391, 127, 391, 129, 391, 130, 391, 131, 391, 132, 391, 133, 391, 136, 391, 137, 391, 138, 391, 139, 391, 140, 391, 141, 391, 170, 391, 162, 391, 163, 391, 164, 391, 165, 391, 166, 391, 167, 391, 168, 391, 169, 391, 199, 391, 198, 391, 197, 391, 196, 391, 195, 391, 188, 391, 189, 391, 190, 391, 192, 391, 193, 391, 201, 391, 200, 391, 202, 391, 203, 391, 204, 391, 205, 391, 206, 391, 207, 391, 208, 391, 209, 391, 210, 391, 211, 391, 212, 391, 220, 391, 221, 391, 222, 391, 223, 391],
	#[119, -33],
	#[13, -439],
	#[7, -13, 118, -36],
	#[43, -441],
	#[65535, 396, 36, 396, 7, 396, 11, 396, 12, 396, 13, 396, 32, 396, 42, 396, 43, 396, 57, 396, 58, 396, 59, 396, 60, 396, 62, 396, 63, 396, 101, 396, 93, 396, 95, 396, 96, 396, 97, 396, 99, 396, 100, 396, 102, 396, 103, 396, 104, 396, 105, 396, 106, 396, 107, 396, 108, 396, 109, 396, 110, 396, 111, 396, 112, 396, 113, 396, 114, 396, 115, 396, 116, 396, 117, 396, 118, 396, 121, 396, 122, 396, 123, 396, 124, 396, 125, 396, 126, 396, 127, 396, 129, 396, 130, 396, 131, 396, 132, 396, 133, 396, 136, 396, 137, 396, 138, 396, 139, 396, 140, 396, 141, 396, 162, 396, 163, 396, 164, 396, 165, 396, 166, 396, 167, 396, 168, 396, 169, 396, 170, 396, 188, 396, 189, 396, 190, 396, 192, 396, 193, 396, 195, 396, 196, 396, 197, 396, 198, 396, 199, 396, 200, 396, 201, 396, 202, 396, 203, 396, 204, 396, 205, 396, 206, 396, 207, 396, 208, 396, 209, 396, 210, 396, 211, 396, 212, 396, 220, 396, 221, 396, 222, 396, 223, 396],
	#[13, -282, 32, -444],
	#[65535, 393, 7, 393, 42, 393, 13, 393, 126, 393, 111, 393, 101, 393, 11, 393, 60, 393, 63, 393, 164, 393, 12, 393, 95, 393, 118, 393, 58, 393, 36, 393, 103, 393, 108, 393, 125, 393, 62, 393, 109, 393, 199, 393, 32, 393, 93, 393, 57, 393, 43, 393, 190, 393, 59, 393, 102, 393, 122, 393, 97, 393, 100, 393, 167, 393, 106, 393, 115, 393, 110, 393, 131, 393, 112, 393, 107, 393, 104, 393, 105, 393, 123, 393, 140, 393, 99, 393, 117, 393, 114, 393, 116, 393, 138, 393, 141, 393, 113, 393, 121, 393, 129, 393, 124, 393, 133, 393, 96, 393, 130, 393, 127, 393, 136, 393, 139, 393, 132, 393, 137, 393, 206, 393, 212, 393, 162, 393, 163, 393, 221, 393, 165, 393, 166, 393, 205, 393, 168, 393, 169, 393, 170, 393, 210, 393, 207, 393, 188, 393, 189, 393, 203, 393, 202, 393, 192, 393, 193, 393, 195, 393, 196, 393, 197, 393, 198, 393, 222, 393, 200, 393, 201, 393, 209, 393, 208, 393, 204, 393, 211, 393, 220, 393, 223, 393],
	#[65535, 392, 36, 392, 7, 392, 11, 392, 12, 392, 13, 392, 32, 392, 42, 392, 43, 392, 57, 392, 58, 392, 59, 392, 60, 392, 62, 392, 63, 392, 101, 392, 93, 392, 95, 392, 96, 392, 97, 392, 99, 392, 100, 392, 102, 392, 103, 392, 104, 392, 105, 392, 106, 392, 107, 392, 108, 392, 109, 392, 110, 392, 111, 392, 112, 392, 113, 392, 114, 392, 115, 392, 116, 392, 117, 392, 118, 392, 121, 392, 122, 392, 123, 392, 124, 392, 125, 392, 126, 392, 127, 392, 129, 392, 130, 392, 131, 392, 132, 392, 133, 392, 136, 392, 137, 392, 138, 392, 139, 392, 140, 392, 141, 392, 162, 392, 163, 392, 164, 392, 165, 392, 166, 392, 167, 392, 168, 392, 169, 392, 170, 392, 188, 392, 189, 392, 190, 392, 192, 392, 193, 392, 195, 392, 196, 392, 197, 392, 198, 392, 199, 392, 200, 392, 201, 392, 202, 392, 203, 392, 204, 392, 205, 392, 206, 392, 207, 392, 208, 392, 209, 392, 210, 392, 211, 392, 212, 392, 220, 392, 221, 392, 222, 392, 223, 392],
	#[65535, 402, 7, 402],
	#[7, 401, 184, -451],
	#[7, -448],
	#[13, -449],
	#[7, -450],
	#[65535, 398, 7, 398, 11, 398, 12, 398, 13, 398, 32, 398, 36, 398, 42, 398, 43, 398, 57, 398, 58, 398, 59, 398, 60, 398, 62, 398, 63, 398, 101, 398, 93, 398, 95, 398, 96, 398, 97, 398, 99, 398, 100, 398, 102, 398, 103, 398, 104, 398, 105, 398, 106, 398, 107, 398, 108, 398, 109, 398, 110, 398, 111, 398, 112, 398, 113, 398, 114, 398, 115, 398, 116, 398, 117, 398, 118, 398, 121, 398, 122, 398, 123, 398, 124, 398, 125, 398, 126, 398, 127, 398, 129, 398, 130, 398, 131, 398, 132, 398, 133, 398, 136, 398, 137, 398, 138, 398, 139, 398, 140, 398, 141, 398, 162, 398, 163, 398, 164, 398, 165, 398, 166, 398, 167, 398, 168, 398, 169, 398, 170, 398, 188, 398, 189, 398, 190, 398, 192, 398, 193, 398, 195, 398, 196, 398, 197, 398, 198, 398, 199, 398, 200, 398, 201, 398, 202, 398, 203, 398, 204, 398, 205, 398, 206, 398, 207, 398, 208, 398, 209, 398, 210, 398, 211, 398, 212, 398, 220, 398, 221, 398, 222, 398, 223, 398],
	#[65535, 403, 7, 403],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -455, 37, 105, 42, -159, 63, 440, 72, -161, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 330, 117, 330, 63, 330, 7, 330, 95, 330, 102, 330, 103, 330, 110, 330, 123, 330, 97, 330, 105, 330, 124, 330, 122, 330, 101, 330, 106, 330, 116, 330, 100, 330, 109, 330, 99, 330, 114, 330, 108, 330, 112, 330, 104, 330, 115, 330, 111, 330, 121, 330, 107, 330, 113, 330, 118, 330, 125, 330, 126, 330, 127, 330, 129, 330, 130, 330, 131, 330, 132, 330, 133, 330, 136, 330, 137, 330, 138, 330, 139, 330, 140, 330, 141, 330, 188, 330, 163, 330, 164, 330, 165, 330, 166, 330, 167, 330, 169, 330, 170, 330, 189, 330, 190, 330, 192, 330, 193, 330, 195, 330, 196, 330, 197, 330, 198, 330, 199, 330, 200, 330, 204, 330, 205, 330],
	#[225, -457],
	#[111, 441, 63, 441, 7, 441, 95, 441, 126, 441, 32, 441, 13, -459, 103, 441, 117, 441, 108, 441, 107, 441, 102, 441, 62, 441, 101, 441, 110, 441, 131, 441, 112, 441, 125, 441, 97, 441, 105, 441, 118, 441, 99, 441, 114, 441, 116, 441, 100, 441, 106, 441, 113, 441, 121, 441, 139, 441, 122, 441, 109, 441, 133, 441, 123, 441, 127, 441, 129, 441, 115, 441, 141, 441, 104, 441, 138, 441, 140, 441, 124, 441, 130, 441, 167, 441, 136, 441, 132, 441, 137, 441, 162, 441, 163, 441, 164, 441, 165, 441, 166, 441, 205, 441, 169, 441, 170, 441, 188, 441, 189, 441, 190, 441, 192, 441, 193, 441, 195, 441, 196, 441, 197, 441, 198, 441, 199, 441, 200, 441, 204, 441],
	#[226, -251],
	#[65535, 442, 7, 442, 13, 442, 32, 442, 95, 442, 62, 442, 110, 442, 101, 442, 115, 442, 109, 442, 111, 442, 102, 442, 99, 442, 106, 442, 63, 442, 113, 442, 116, 442, 100, 442, 104, 442, 114, 442, 123, 442, 124, 442, 108, 442, 97, 442, 105, 442, 121, 442, 107, 442, 112, 442, 103, 442, 122, 442, 117, 442, 118, 442, 125, 442, 126, 442, 127, 442, 129, 442, 130, 442, 131, 442, 132, 442, 133, 442, 136, 442, 137, 442, 138, 442, 139, 442, 140, 442, 141, 442, 188, 442, 162, 442, 163, 442, 164, 442, 165, 442, 166, 442, 167, 442, 169, 442, 170, 442, 189, 442, 190, 442, 192, 442, 193, 442, 195, 442, 196, 442, 197, 442, 198, 442, 199, 442, 200, 442, 204, 442, 205, 442],
	#[225, -460],
	#[226, -251],
	#[65535, 443, 7, 443, 13, 443, 32, 443, 62, 443, 63, 443, 95, 443, 97, 443, 99, 443, 100, 443, 101, 443, 102, 443, 103, 443, 104, 443, 105, 443, 106, 443, 107, 443, 108, 443, 109, 443, 110, 443, 111, 443, 112, 443, 113, 443, 114, 443, 115, 443, 116, 443, 117, 443, 118, 443, 121, 443, 122, 443, 123, 443, 124, 443, 125, 443, 126, 443, 127, 443, 129, 443, 130, 443, 131, 443, 132, 443, 133, 443, 136, 443, 137, 443, 138, 443, 139, 443, 140, 443, 141, 443, 162, 443, 163, 443, 164, 443, 165, 443, 166, 443, 167, 443, 169, 443, 170, 443, 188, 443, 189, 443, 190, 443, 192, 443, 193, 443, 195, 443, 196, 443, 197, 443, 198, 443, 199, 443, 200, 443, 204, 443, 205, 443],
	#[65535, 362, 207, 362, 208, 362, 209, 362, 210, 362, 211, 362, 212, 362],
	#[207, -464, 208, -465, 209, -466, 210, -467, 211, -469, 212, -470],
	#[65535, 363, 110, 363, 13, 363, 109, 363, 7, 363, 95, 363, 118, 363, 127, 363, 103, 363, 167, 363, 111, 363, 116, 363, 133, 363, 131, 363, 107, 363, 63, 363, 123, 363, 99, 363, 102, 363, 126, 363, 132, 363, 130, 363, 140, 363, 117, 363, 122, 363, 114, 363, 141, 363, 100, 363, 97, 363, 105, 363, 124, 363, 108, 363, 101, 363, 121, 363, 139, 363, 112, 363, 115, 363, 104, 363, 129, 363, 125, 363, 113, 363, 138, 363, 106, 363, 136, 363, 137, 363, 163, 363, 164, 363, 165, 363, 166, 363, 205, 363, 169, 363, 170, 363, 188, 363, 189, 363, 190, 363, 192, 363, 193, 363, 195, 363, 196, 363, 197, 363, 198, 363, 199, 363, 200, 363, 204, 363],
	#[65535, 364, 7, 364, 13, 364, 63, 364, 95, 364, 124, 364, 111, 364, 102, 364, 113, 364, 123, 364, 99, 364, 101, 364, 110, 364, 116, 364, 100, 364, 112, 364, 114, 364, 109, 364, 121, 364, 106, 364, 108, 364, 97, 364, 105, 364, 104, 364, 107, 364, 115, 364, 103, 364, 122, 364, 117, 364, 118, 364, 125, 364, 126, 364, 127, 364, 129, 364, 130, 364, 131, 364, 132, 364, 133, 364, 136, 364, 137, 364, 138, 364, 139, 364, 140, 364, 141, 364, 188, 364, 163, 364, 164, 364, 165, 364, 166, 364, 167, 364, 169, 364, 170, 364, 189, 364, 190, 364, 192, 364, 193, 364, 195, 364, 196, 364, 197, 364, 198, 364, 199, 364, 200, 364, 204, 364, 205, 364],
	#[65535, 365, 7, 365, 13, 365, 63, 365, 95, 365, 103, 365, 116, 365, 109, 365, 102, 365, 123, 365, 99, 365, 106, 365, 101, 365, 110, 365, 121, 365, 100, 365, 122, 365, 113, 365, 112, 365, 124, 365, 108, 365, 97, 365, 105, 365, 104, 365, 107, 365, 115, 365, 114, 365, 111, 365, 117, 365, 118, 365, 125, 365, 126, 365, 127, 365, 129, 365, 130, 365, 131, 365, 132, 365, 133, 365, 136, 365, 137, 365, 138, 365, 139, 365, 140, 365, 141, 365, 188, 365, 163, 365, 164, 365, 165, 365, 166, 365, 167, 365, 169, 365, 170, 365, 189, 365, 190, 365, 192, 365, 193, 365, 195, 365, 196, 365, 197, 365, 198, 365, 199, 365, 200, 365, 204, 365, 205, 365],
	#[65535, 366, 63, 366, 100, 366, 95, 366, 102, 366, 126, 366, 111, 366, 101, 366, 99, 366, 110, 366, 13, 366, 109, 366, 7, 366, 123, 366, 121, 366, 127, 366, 139, 366, 117, 366, 103, 366, 108, 366, 107, 366, 167, 366, 113, 366, 116, 366, 115, 366, 105, 366, 122, 366, 118, 366, 104, 366, 124, 366, 106, 366, 141, 366, 129, 366, 97, 366, 114, 366, 112, 366, 130, 366, 131, 366, 125, 366, 138, 366, 133, 366, 140, 366, 136, 366, 132, 366, 137, 366, 163, 366, 164, 366, 165, 366, 166, 366, 205, 366, 169, 366, 170, 366, 188, 366, 189, 366, 190, 366, 192, 366, 193, 366, 195, 366, 196, 366, 197, 366, 198, 366, 199, 366, 200, 366, 204, 366],
	#[7, 440, 13, -455, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 367, 7, 367, 13, 367, 63, 367, 95, 367, 97, 367, 99, 367, 100, 367, 101, 367, 102, 367, 103, 367, 104, 367, 105, 367, 106, 367, 107, 367, 108, 367, 109, 367, 110, 367, 111, 367, 112, 367, 113, 367, 114, 367, 115, 367, 116, 367, 117, 367, 118, 367, 121, 367, 122, 367, 123, 367, 124, 367, 125, 367, 126, 367, 127, 367, 129, 367, 130, 367, 131, 367, 132, 367, 133, 367, 136, 367, 137, 367, 138, 367, 139, 367, 140, 367, 141, 367, 163, 367, 164, 367, 165, 367, 166, 367, 167, 367, 169, 367, 170, 367, 188, 367, 189, 367, 190, 367, 192, 367, 193, 367, 195, 367, 196, 367, 197, 367, 198, 367, 199, 367, 200, 367, 204, 367, 205, 367],
	#[65535, 368, 7, 368, 13, 368, 63, 368, 110, 368, 103, 368, 100, 368, 115, 368, 109, 368, 111, 368, 124, 368, 108, 368, 99, 368, 101, 368, 113, 368, 116, 368, 95, 368, 123, 368, 112, 368, 106, 368, 104, 368, 97, 368, 105, 368, 121, 368, 107, 368, 102, 368, 114, 368, 122, 368, 117, 368, 118, 368, 125, 368, 126, 368, 127, 368, 129, 368, 130, 368, 131, 368, 132, 368, 133, 368, 136, 368, 137, 368, 138, 368, 139, 368, 140, 368, 141, 368, 188, 368, 163, 368, 164, 368, 165, 368, 166, 368, 167, 368, 169, 368, 170, 368, 189, 368, 190, 368, 192, 368, 193, 368, 195, 368, 196, 368, 197, 368, 198, 368, 199, 368, 200, 368, 204, 368, 205, 368],
	#[65535, 349, 7, 349, 102, 349, 63, 349, 95, 349, 108, 349, 110, 349, 103, 349, 105, 349, 106, 349, 109, 349, 111, 349, 124, 349, 113, 349, 123, 349, 99, 349, 97, 349, 101, 349, 104, 349, 116, 349, 100, 349, 115, 349, 112, 349, 121, 349, 107, 349, 114, 349, 122, 349, 117, 349, 118, 349, 125, 349, 126, 349, 127, 349, 129, 349, 130, 349, 131, 349, 132, 349, 133, 349, 136, 349, 137, 349, 138, 349, 139, 349, 140, 349, 141, 349, 163, 349, 164, 349, 165, 349, 166, 349, 167, 349, 169, 349, 170, 349, 188, 349, 189, 349, 190, 349, 192, 349, 193, 349, 195, 349, 196, 349, 197, 349, 198, 349, 199, 349, 200, 349, 204, 349, 205, 349],
	#[11, -473, 37, 105, 42, -159, 72, -161],
	#[84, -247, 2, -227, 7, -13, 10, -223, 11, -250, 92, -241, 33, -256, 87, -257, 31, -228, 119, -33, 102, -236, 63, -232, 90, -225, 95, -100, 38, -253, 91, -252, 86, -242, 110, -65, 101, -255, 103, -254, 88, -248, 100, -235, 115, -70, 109, -80, 111, -56, 124, -76, 108, -89, 123, -105, 104, -229, 99, -245, 89, -233, 106, -231, 105, -246, 113, -94, 116, -117, 85, -259, 97, -237, 112, -97, 121, -58, 107, -118, 114, -72, 122, -107, 117, -119, 118, -36, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -476, 63, 440, 101, 440, 103, 440, 100, 440, 109, 440, 111, 440, 102, 440, 123, 440, 99, 440, 106, 440, 115, 440, 110, 440, 116, 440, 95, 440, 113, 440, 121, 440, 124, 440, 108, 440, 97, 440, 105, 440, 104, 440, 107, 440, 112, 440, 114, 440, 122, 440, 117, 440, 118, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 188, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 336, 111, 336, 7, 336, 115, 336, 109, 336, 103, 336, 110, 336, 106, 336, 100, 336, 102, 336, 63, 336, 108, 336, 117, 336, 101, 336, 122, 336, 99, 336, 113, 336, 116, 336, 95, 336, 107, 336, 123, 336, 124, 336, 104, 336, 97, 336, 105, 336, 121, 336, 112, 336, 114, 336, 118, 336, 125, 336, 126, 336, 127, 336, 129, 336, 130, 336, 131, 336, 132, 336, 133, 336, 136, 336, 137, 336, 138, 336, 139, 336, 140, 336, 141, 336, 188, 336, 163, 336, 164, 336, 165, 336, 166, 336, 167, 336, 169, 336, 170, 336, 189, 336, 190, 336, 192, 336, 193, 336, 195, 336, 196, 336, 197, 336, 198, 336, 199, 336, 200, 336, 204, 336, 205, 336],
	#[11, -477, 225, -457],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -479],
	#[2, -227, 7, -13, 10, -223, 11, -250, 103, -254, 91, -252, 87, -257, 31, -228, 33, -256, 38, -253, 124, -76, 86, -242, 92, -241, 101, -255, 85, -259, 100, -235, 84, -247, 109, -80, 88, -248, 102, -236, 123, -105, 99, -245, 119, -33, 63, -232, 110, -65, 116, -117, 95, -100, 122, -107, 90, -225, 104, -229, 105, -246, 113, -94, 106, -231, 108, -89, 89, -233, 97, -237, 112, -97, 121, -58, 107, -118, 115, -70, 114, -72, 111, -56, 117, -119, 118, -36, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[12, -481],
	#[65535, 418, 7, 418, 13, 418, 63, 418, 101, 418, 95, 418, 97, 418, 99, 418, 100, 418, 102, 418, 103, 418, 104, 418, 105, 418, 106, 418, 107, 418, 108, 418, 109, 418, 110, 418, 111, 418, 112, 418, 113, 418, 114, 418, 115, 418, 116, 418, 117, 418, 118, 418, 121, 418, 122, 418, 123, 418, 124, 418, 125, 418, 126, 418, 127, 418, 129, 418, 130, 418, 131, 418, 132, 418, 133, 418, 136, 418, 137, 418, 138, 418, 139, 418, 140, 418, 141, 418, 163, 418, 164, 418, 165, 418, 166, 418, 167, 418, 169, 418, 170, 418, 188, 418, 189, 418, 190, 418, 192, 418, 193, 418, 195, 418, 196, 418, 197, 418, 198, 418, 199, 418, 200, 418, 204, 418, 205, 418],
	#[13, -483],
	#[7, -13, 2, -227, 10, -223, 11, -250, 90, -225, 38, -253, 86, -242, 33, -256, 141, -120, 31, -228, 100, -235, 102, -236, 139, -77, 92, -241, 103, -254, 126, -115, 87, -257, 119, -33, 63, -232, 116, -117, 95, -100, 84, -247, 125, -62, 127, -102, 118, -36, 129, -81, 91, -252, 132, -93, 101, -255, 138, -104, 106, -231, 85, -259, 99, -245, 110, -65, 88, -248, 89, -233, 97, -237, 108, -89, 131, -110, 121, -58, 107, -118, 133, -88, 123, -105, 117, -119, 113, -94, 122, -107, 112, -97, 114, -72, 104, -229, 105, -246, 140, -64, 124, -76, 130, -57, 109, -80, 111, -56, 115, -70, 136, -87, 137, -59, 226, -251],
	#[12, -485],
	#[65535, 417, 7, 417, 13, 417, 63, 417, 95, 417, 97, 417, 99, 417, 100, 417, 101, 417, 102, 417, 103, 417, 104, 417, 105, 417, 106, 417, 107, 417, 108, 417, 109, 417, 110, 417, 111, 417, 112, 417, 113, 417, 114, 417, 115, 417, 116, 417, 117, 417, 118, 417, 121, 417, 122, 417, 123, 417, 124, 417, 125, 417, 126, 417, 127, 417, 129, 417, 130, 417, 131, 417, 132, 417, 133, 417, 136, 417, 137, 417, 138, 417, 139, 417, 140, 417, 141, 417, 163, 417, 164, 417, 165, 417, 166, 417, 167, 417, 169, 417, 170, 417, 188, 417, 189, 417, 190, 417, 192, 417, 193, 417, 195, 417, 196, 417, 197, 417, 198, 417, 199, 417, 200, 417, 204, 417, 205, 417],
	#[31, -128, 35, -123, 66, -126, 68, -133, 65, -122, 63, -121, 7, -13, 11, -127, 36, -129, 70, -140, 67, -137, 69, -136, 33, -139, 64, -141, 73, -138, 71, -130, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -489],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -492, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 423, 111, 423, 63, 423, 7, 423, 110, 423, 109, 423, 99, 423, 97, 423, 108, 423, 95, 423, 117, 423, 116, 423, 100, 423, 101, 423, 102, 423, 103, 423, 104, 423, 105, 423, 106, 423, 107, 423, 112, 423, 113, 423, 114, 423, 115, 423, 118, 423, 121, 423, 122, 423, 123, 423, 124, 423, 125, 423, 126, 423, 127, 423, 129, 423, 130, 423, 131, 423, 132, 423, 133, 423, 136, 423, 137, 423, 138, 423, 139, 423, 140, 423, 141, 423, 163, 423, 164, 423, 165, 423, 166, 423, 167, 423, 169, 423, 170, 423, 188, 423, 189, 423, 190, 423, 192, 423, 193, 423, 195, 423, 196, 423, 197, 423, 198, 423, 199, 423, 200, 423, 204, 423, 205, 423],
	#[93, -494, 225, -457],
	#[65535, 346, 7, 346, 63, 346, 95, 346, 97, 346, 99, 346, 100, 346, 101, 346, 102, 346, 103, 346, 104, 346, 105, 346, 106, 346, 107, 346, 108, 346, 109, 346, 110, 346, 111, 346, 112, 346, 113, 346, 114, 346, 115, 346, 116, 346, 117, 346, 118, 346, 121, 346, 122, 346, 123, 346, 124, 346, 125, 346, 126, 346, 127, 346, 129, 346, 130, 346, 131, 346, 132, 346, 133, 346, 136, 346, 137, 346, 138, 346, 139, 346, 140, 346, 141, 346, 163, 346, 164, 346, 165, 346, 166, 346, 167, 346, 169, 346, 170, 346, 188, 346, 189, 346, 190, 346, 192, 346, 193, 346, 195, 346, 196, 346, 197, 346, 198, 346, 199, 346, 200, 346, 204, 346, 205, 346],
	#[38, -495],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 424, 103, 424, 63, 424, 108, 424, 7, 424, 110, 424, 122, 424, 101, 424, 166, 424, 100, 424, 102, 424, 126, 424, 105, 424, 95, 424, 127, 424, 107, 424, 113, 424, 104, 424, 123, 424, 99, 424, 97, 424, 118, 424, 138, 424, 167, 424, 109, 424, 111, 424, 199, 424, 140, 424, 124, 424, 133, 424, 114, 424, 198, 424, 139, 424, 106, 424, 115, 424, 130, 424, 116, 424, 129, 424, 136, 424, 112, 424, 121, 424, 137, 424, 141, 424, 163, 424, 117, 424, 131, 424, 125, 424, 170, 424, 165, 424, 164, 424, 132, 424, 169, 424, 197, 424, 195, 424, 204, 424, 188, 424, 189, 424, 190, 424, 192, 424, 193, 424, 200, 424, 196, 424, 205, 424],
	#[7, 440, 13, -492, 63, 440, 95, 440, 103, 440, 109, 440, 102, 440, 123, 440, 99, 440, 106, 440, 101, 440, 110, 440, 116, 440, 100, 440, 122, 440, 105, 440, 113, 440, 121, 440, 124, 440, 108, 440, 97, 440, 112, 440, 104, 440, 107, 440, 115, 440, 114, 440, 111, 440, 117, 440, 118, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 188, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 343, 7, 343, 63, 343, 95, 343, 110, 343, 103, 343, 109, 343, 111, 343, 102, 343, 99, 343, 106, 343, 101, 343, 113, 343, 116, 343, 100, 343, 104, 343, 105, 343, 123, 343, 124, 343, 108, 343, 97, 343, 112, 343, 121, 343, 107, 343, 115, 343, 114, 343, 122, 343, 117, 343, 118, 343, 125, 343, 126, 343, 127, 343, 129, 343, 130, 343, 131, 343, 132, 343, 133, 343, 136, 343, 137, 343, 138, 343, 139, 343, 140, 343, 141, 343, 188, 343, 163, 343, 164, 343, 165, 343, 166, 343, 167, 343, 169, 343, 170, 343, 189, 343, 190, 343, 192, 343, 193, 343, 195, 343, 196, 343, 197, 343, 198, 343, 199, 343, 200, 343, 204, 343, 205, 343],
	#[36, -129, 11, -127, 31, -128, 63, -121, 33, -139, 7, -13, 35, -123, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 118, -36, 73, -138],
	#[2, -227, 31, -228, 7, -13, 10, -223, 11, -250, 42, -159, 33, -256, 37, 105, 38, -253, 63, -232, 72, -161, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -502],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 329, 7, 329, 63, 329, 97, 329, 101, 329, 100, 329, 99, 329, 103, 329, 102, 329, 104, 329, 105, 329, 107, 329, 95, 329, 106, 329, 108, 329, 109, 329, 110, 329, 111, 329, 112, 329, 113, 329, 114, 329, 115, 329, 116, 329, 117, 329, 118, 329, 121, 329, 122, 329, 123, 329, 124, 329, 125, 329, 126, 329, 127, 329, 129, 329, 130, 329, 131, 329, 132, 329, 133, 329, 136, 329, 137, 329, 138, 329, 139, 329, 140, 329, 141, 329, 163, 329, 164, 329, 165, 329, 166, 329, 167, 329, 169, 329, 170, 329, 188, 329, 189, 329, 190, 329, 192, 329, 193, 329, 195, 329, 196, 329, 197, 329, 198, 329, 199, 329, 200, 329, 204, 329, 205, 329],
	#[2, -227, 7, -13, 10, -223, 11, -250, 86, -242, 110, -65, 92, -241, 31, -228, 33, -256, 63, -232, 95, -100, 38, -253, 91, -252, 103, -254, 85, -259, 84, -247, 109, -80, 102, -236, 90, -225, 87, -257, 99, -245, 119, -33, 106, -231, 101, -255, 113, -94, 116, -117, 100, -235, 122, -107, 88, -248, 104, -229, 105, -246, 123, -105, 124, -76, 108, -89, 89, -233, 97, -237, 112, -97, 121, -58, 107, -118, 115, -70, 114, -72, 111, -56, 117, -119, 118, -36, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -507],
	#[69, -508],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[11, -510],
	#[12, -513, 36, -511],
	#[2, -227, 11, -250, 31, -228, 33, -256, 95, -100, 38, -253, 91, -252, 86, -242, 110, -65, 101, -255, 103, -254, 85, -259, 109, -80, 111, -56, 102, -236, 87, -257, 99, -245, 89, -233, 106, -231, 88, -248, 113, -94, 116, -117, 100, -235, 84, -247, 90, -225, 104, -229, 105, -246, 123, -105, 124, -76, 108, -89, 92, -241, 97, -237, 112, -97, 121, -58, 107, -118, 115, -70, 114, -72, 122, -107, 117, -119, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[12, -516, 36, -515],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 284, 7, 284, 32, 284, 62, 284, 63, 284, 95, 284, 97, 284, 99, 284, 100, 284, 101, 284, 102, 284, 103, 284, 104, 284, 105, 284, 106, 284, 107, 284, 108, 284, 109, 284, 110, 284, 111, 284, 112, 284, 113, 284, 114, 284, 115, 284, 116, 284, 117, 284, 118, 284, 121, 284, 122, 284, 123, 284, 124, 284, 125, 284, 126, 284, 127, 284, 129, 284, 130, 284, 131, 284, 132, 284, 133, 284, 136, 284, 137, 284, 138, 284, 139, 284, 140, 284, 141, 284, 162, 284, 163, 284, 164, 284, 165, 284, 166, 284, 167, 284, 169, 284, 170, 284, 188, 284, 189, 284, 190, 284, 192, 284, 193, 284, 195, 284, 196, 284, 197, 284, 198, 284, 199, 284, 200, 284, 204, 284, 205, 284],
	#[103, -254, 84, -247, 2, -227, 85, -259, 102, -236, 11, -250, 33, -256, 87, -257, 31, -228, 38, -253, 91, -252, 86, -242, 101, -255, 90, -225, 92, -241, 88, -248, 100, -235, 107, -118, 99, -245, 104, -229, 89, -233, 97, -237, 105, -246, 95, -100, 106, -231, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 283, 7, 283, 32, 283, 62, 283, 63, 283, 95, 283, 97, 283, 99, 283, 100, 283, 101, 283, 102, 283, 103, 283, 104, 283, 105, 283, 106, 283, 107, 283, 108, 283, 109, 283, 110, 283, 111, 283, 112, 283, 113, 283, 114, 283, 115, 283, 116, 283, 117, 283, 118, 283, 121, 283, 122, 283, 123, 283, 124, 283, 125, 283, 126, 283, 127, 283, 129, 283, 130, 283, 131, 283, 132, 283, 133, 283, 136, 283, 137, 283, 138, 283, 139, 283, 140, 283, 141, 283, 162, 283, 163, 283, 164, 283, 165, 283, 166, 283, 167, 283, 169, 283, 170, 283, 188, 283, 189, 283, 190, 283, 192, 283, 193, 283, 195, 283, 196, 283, 197, 283, 198, 283, 199, 283, 200, 283, 204, 283, 205, 283],
	#[13, -519],
	#[69, -520],
	#[7, -13, 84, -247, 10, -223, 2, -227, 86, -242, 31, -228, 119, -33, 63, -232, 92, -241, 95, -100, 85, -259, 102, -236, 11, -250, 33, -256, 116, -117, 38, -253, 103, -254, 123, -105, 124, -76, 88, -248, 91, -252, 87, -257, 100, -235, 109, -80, 122, -107, 90, -225, 114, -72, 108, -89, 105, -246, 101, -255, 112, -97, 106, -231, 89, -233, 97, -237, 121, -58, 99, -245, 115, -70, 110, -65, 107, -118, 113, -94, 104, -229, 111, -56, 117, -119, 118, -36, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 316, 12, 316, 36, 316],
	#[13, -523],
	#[69, -524],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 63, -232, 95, -100, 38, -253, 91, -252, 86, -242, 110, -65, 103, -254, 85, -259, 109, -80, 111, -56, 89, -233, 102, -236, 87, -257, 99, -245, 119, -33, 106, -231, 88, -248, 113, -94, 116, -117, 100, -235, 84, -247, 90, -225, 104, -229, 105, -246, 123, -105, 124, -76, 108, -89, 92, -241, 97, -237, 112, -97, 121, -58, 107, -118, 101, -255, 115, -70, 114, -72, 122, -107, 117, -119, 118, -36, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 315, 12, 315, 36, 315],
	#[13, -527],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -529],
	#[31, -128, 36, -129, 7, -13, 11, -127, 35, -123, 33, -139, 66, -126, 65, -122, 64, -141, 63, -121, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 334, 7, 334, 63, 334, 103, 334, 102, 334, 101, 334, 100, 334, 99, 334, 104, 334, 97, 334, 105, 334, 107, 334, 95, 334, 106, 334, 108, 334, 109, 334, 110, 334, 111, 334, 112, 334, 113, 334, 114, 334, 115, 334, 116, 334, 117, 334, 118, 334, 121, 334, 122, 334, 123, 334, 124, 334, 125, 334, 126, 334, 127, 334, 129, 334, 130, 334, 131, 334, 132, 334, 133, 334, 136, 334, 137, 334, 138, 334, 139, 334, 140, 334, 141, 334, 163, 334, 164, 334, 165, 334, 166, 334, 167, 334, 169, 334, 170, 334, 188, 334, 189, 334, 190, 334, 192, 334, 193, 334, 195, 334, 196, 334, 197, 334, 198, 334, 199, 334, 200, 334, 204, 334, 205, 334],
	#[2, -227, 31, -228, 7, -13, 10, -223, 11, -250, 38, -253, 33, -256, 63, -232, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 103, 440, 102, 440, 101, 440, 100, 440, 99, 440, 104, 440, 97, 440, 105, 440, 107, 440, 95, 440, 106, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 281, 7, 281, 32, 281, 62, 281, 63, 281, 101, 281, 95, 281, 97, 281, 99, 281, 100, 281, 102, 281, 103, 281, 104, 281, 105, 281, 106, 281, 107, 281, 108, 281, 109, 281, 110, 281, 111, 281, 112, 281, 113, 281, 114, 281, 115, 281, 116, 281, 117, 281, 118, 281, 121, 281, 122, 281, 123, 281, 124, 281, 125, 281, 126, 281, 127, 281, 129, 281, 130, 281, 131, 281, 132, 281, 133, 281, 136, 281, 137, 281, 138, 281, 139, 281, 140, 281, 141, 281, 162, 281, 163, 281, 164, 281, 165, 281, 166, 281, 167, 281, 169, 281, 170, 281, 188, 281, 189, 281, 190, 281, 192, 281, 193, 281, 195, 281, 196, 281, 197, 281, 198, 281, 199, 281, 200, 281, 204, 281, 205, 281],
	#[13, -537],
	#[69, -538],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -540],
	#[69, -541],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 103, 440, 102, 440, 101, 440, 100, 440, 99, 440, 104, 440, 97, 440, 105, 440, 107, 440, 95, 440, 106, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 282, 7, 282, 32, 282, 62, 282, 63, 282, 95, 282, 97, 282, 99, 282, 100, 282, 101, 282, 102, 282, 103, 282, 104, 282, 105, 282, 106, 282, 107, 282, 108, 282, 109, 282, 110, 282, 111, 282, 112, 282, 113, 282, 114, 282, 115, 282, 116, 282, 117, 282, 118, 282, 121, 282, 122, 282, 123, 282, 124, 282, 125, 282, 126, 282, 127, 282, 129, 282, 130, 282, 131, 282, 132, 282, 133, 282, 136, 282, 137, 282, 138, 282, 139, 282, 140, 282, 141, 282, 162, 282, 163, 282, 164, 282, 165, 282, 166, 282, 167, 282, 169, 282, 170, 282, 188, 282, 189, 282, 190, 282, 192, 282, 193, 282, 195, 282, 196, 282, 197, 282, 198, 282, 199, 282, 200, 282, 204, 282, 205, 282],
	#[31, 419, 36, 419, 33, 419, 7, 419, 11, 419, 35, 419, 65, 419, 70, 419, 69, 419, 67, 419, 63, 419, 64, 419, 66, 419, 68, 419, 71, 419, 73, 419, 118, 419, 195, -545],
	#[65535, 420, 7, 420, 11, 420, 31, 420, 33, 420, 35, 420, 36, 420, 63, 420, 64, 420, 65, 420, 66, 420, 67, 420, 68, 420, 69, 420, 70, 420, 71, 420, 73, 420, 118, 420, 121, 420, 123, 420, 139, 420, 140, 420, 141, 420, 213, 420, 214, 420, 215, 420, 216, 420, 217, 420, 218, 420],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -492, 103, 440, 63, 440, 102, 440, 101, 440, 100, 440, 107, 440, 99, 440, 104, 440, 97, 440, 105, 440, 95, 440, 106, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 341, 7, 341, 63, 341, 95, 341, 97, 341, 99, 341, 100, 341, 101, 341, 102, 341, 103, 341, 104, 341, 105, 341, 106, 341, 107, 341, 108, 341, 109, 341, 110, 341, 111, 341, 112, 341, 113, 341, 114, 341, 115, 341, 116, 341, 117, 341, 118, 341, 121, 341, 122, 341, 123, 341, 124, 341, 125, 341, 126, 341, 127, 341, 129, 341, 130, 341, 131, 341, 132, 341, 133, 341, 136, 341, 137, 341, 138, 341, 139, 341, 140, 341, 141, 341, 163, 341, 164, 341, 165, 341, 166, 341, 167, 341, 169, 341, 170, 341, 188, 341, 189, 341, 190, 341, 192, 341, 193, 341, 195, 341, 196, 341, 197, 341, 198, 341, 199, 341, 200, 341, 204, 341, 205, 341],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[209, 361, 208, 361, 206, -462, 207, 361, 210, 361, 211, 361, 212, 361],
	#[209, -466, 208, -465, 207, -464, 210, -467, 211, -469, 212, -470],
	#[7, 440, 127, 440, 13, -492, 63, 440, 101, 440, 103, 440, 102, 440, 111, 440, 100, 440, 99, 440, 104, 440, 97, 440, 105, 440, 107, 440, 95, 440, 106, 440, 108, 440, 109, 440, 110, 440, 122, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 141, 440, 123, 440, 124, 440, 125, 440, 126, 440, 138, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 167, 440, 139, 440, 140, 440, 169, 440, 166, 440, 165, 440, 164, 440, 170, 440, 163, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 342, 7, 342, 167, 342, 63, 342, 103, 342, 102, 342, 126, 342, 101, 342, 95, 342, 97, 342, 99, 342, 100, 342, 104, 342, 105, 342, 106, 342, 107, 342, 108, 342, 109, 342, 110, 342, 111, 342, 112, 342, 113, 342, 114, 342, 115, 342, 116, 342, 117, 342, 118, 342, 121, 342, 122, 342, 123, 342, 124, 342, 125, 342, 170, 342, 127, 342, 129, 342, 130, 342, 131, 342, 132, 342, 133, 342, 136, 342, 137, 342, 138, 342, 139, 342, 140, 342, 141, 342, 166, 342, 165, 342, 164, 342, 169, 342, 163, 342, 188, 342, 189, 342, 190, 342, 192, 342, 193, 342, 195, 342, 196, 342, 197, 342, 198, 342, 199, 342, 200, 342, 204, 342, 205, 342],
	#[65535, 71, 7, 71, 11, 71, 31, 71, 33, 71, 35, 71, 36, 71, 57, 71, 58, 71, 59, 71, 60, 71, 63, 71, 64, 71, 65, 71, 66, 71, 67, 71, 68, 71, 69, 71, 70, 71, 71, 71, 73, 71, 118, 71],
	#[65535, 65, 7, 65, 69, 65, 70, 65, 11, 65, 35, 65, 33, 65, 67, 65, 31, 65, 59, 65, 36, 65, 60, 65, 58, 65, 57, 65, 65, 65, 73, 65, 66, 65, 63, 65, 64, 65, 68, 65, 71, 65, 118, 65],
	#[38, -601],
	#[65535, 73, 7, 73, 11, 73, 31, 73, 33, 73, 35, 73, 36, 73, 57, 73, 58, 73, 59, 73, 60, 73, 63, 73, 64, 73, 65, 73, 66, 73, 67, 73, 68, 73, 69, 73, 70, 73, 71, 73, 73, 73, 118, 73],
	#[65535, 60, 7, 60, 11, 60, 36, 60, 33, 60, 67, 60, 57, 60, 31, 60, 59, 60, 35, 60, 58, 60, 60, 60, 70, 60, 69, 60, 65, 60, 73, 60, 66, 60, 63, 60, 64, 60, 68, 60, 71, 60, 118, 60],
	#[65535, 66, 7, 66, 11, 66, 31, 66, 33, 66, 35, 66, 36, 66, 57, 66, 58, 66, 59, 66, 60, 66, 63, 66, 64, 66, 65, 66, 66, 66, 67, 66, 68, 66, 69, 66, 70, 66, 71, 66, 73, 66, 118, 66],
	#[65535, 67, 7, 67, 11, 67, 59, 67, 57, 67, 31, 67, 33, 67, 35, 67, 36, 67, 60, 67, 58, 67, 65, 67, 66, 67, 63, 67, 64, 67, 67, 67, 68, 67, 69, 67, 70, 67, 71, 67, 73, 67, 118, 67],
	#[65535, 61, 7, 61, 11, 61, 31, 61, 33, 61, 35, 61, 36, 61, 57, 61, 58, 61, 59, 61, 60, 61, 63, 61, 64, 61, 65, 61, 66, 61, 67, 61, 68, 61, 69, 61, 70, 61, 71, 61, 73, 61, 118, 61],
	#[65535, 68, 7, 68, 11, 68, 31, 68, 33, 68, 35, 68, 36, 68, 57, 68, 58, 68, 59, 68, 60, 68, 63, 68, 64, 68, 65, 68, 66, 68, 67, 68, 68, 68, 69, 68, 70, 68, 71, 68, 73, 68, 118, 68],
	#[65535, 62, 69, 62, 7, 62, 11, 62, 59, 62, 58, 62, 67, 62, 31, 62, 33, 62, 35, 62, 36, 62, 60, 62, 70, 62, 57, 62, 65, 62, 73, 62, 66, 62, 63, 62, 64, 62, 68, 62, 71, 62, 118, 62],
	#[65535, 69, 7, 69, 11, 69, 31, 69, 33, 69, 35, 69, 36, 69, 57, 69, 58, 69, 59, 69, 60, 69, 63, 69, 64, 69, 65, 69, 66, 69, 67, 69, 68, 69, 69, 69, 70, 69, 71, 69, 73, 69, 118, 69],
	#[65535, 63, 7, 63, 11, 63, 31, 63, 33, 63, 35, 63, 36, 63, 57, 63, 58, 63, 59, 63, 60, 63, 63, 63, 64, 63, 65, 63, 66, 63, 67, 63, 68, 63, 69, 63, 70, 63, 71, 63, 73, 63, 118, 63],
	#[65535, 70, 31, 70, 7, 70, 11, 70, 36, 70, 33, 70, 67, 70, 69, 70, 59, 70, 35, 70, 58, 70, 60, 70, 70, 70, 57, 70, 65, 70, 73, 70, 66, 70, 63, 70, 64, 70, 68, 70, 71, 70, 118, 70],
	#[65535, 64, 7, 64, 11, 64, 31, 64, 33, 64, 35, 64, 36, 64, 57, 64, 58, 64, 59, 64, 60, 64, 63, 64, 64, 64, 65, 64, 66, 64, 67, 64, 68, 64, 69, 64, 70, 64, 71, 64, 73, 64, 118, 64],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 70, -140, 69, -136, 65, -122, 67, -137, 57, -575, 58, -573, 59, -569, 60, -574, 63, -121, 64, -141, 66, -126, 68, -133, 71, -130, 73, -570, 118, -36],
	#[65535, 77, 7, 77, 11, 77, 31, 77, 33, 77, 35, 77, 36, 77, 57, 77, 58, 77, 59, 77, 60, 77, 63, 77, 64, 77, 65, 77, 66, 77, 67, 77, 68, 77, 69, 77, 70, 77, 71, 77, 73, 77, 118, 77],
	#[2, 196, 7, 196, 10, 196, 11, 196, 31, 196, 33, 196, 38, 196, 42, -151, 63, 196, 84, 196, 85, 196, 86, 196, 87, 196, 88, 196, 89, 196, 90, 196, 91, 196, 92, 196, 95, 196, 97, 196, 99, 196, 100, 196, 101, 196, 102, 196, 103, 196, 104, 196, 105, 196, 106, 196, 107, 196, 108, 196, 109, 196, 110, 196, 111, 196, 112, 196, 113, 196, 114, 196, 115, 196, 116, 196, 117, 196, 118, 196, 119, 196, 121, 196, 122, 196, 123, 196, 124, 196, 125, 196, 126, 196, 127, 196, 129, 196, 130, 196, 131, 196, 132, 196, 133, 196, 136, 196, 137, 196, 138, 196, 139, 196, 140, 196, 141, 196, 226, 196],
	#[65535, 74, 68, 74, 11, 74, 63, 74, 7, 74, 36, 74, 66, 74, 35, 74, 71, 74, 59, 74, 58, 74, 60, 74, 67, 74, 57, 74, 31, 74, 33, 74, 65, 74, 64, 74, 69, 74, 70, 74, 73, 74, 118, 74],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 76, 7, 76, 11, 76, 31, 76, 33, 76, 35, 76, 36, 76, 57, 76, 58, 76, 59, 76, 60, 76, 63, 76, 64, 76, 65, 76, 66, 76, 67, 76, 68, 76, 69, 76, 70, 76, 71, 76, 73, 76, 118, 76],
	#[65535, 78, 11, 78, 63, 78, 7, 78, 66, 78, 33, 78, 68, 78, 36, 78, 35, 78, 58, 78, 60, 78, 67, 78, 57, 78, 31, 78, 59, 78, 65, 78, 71, 78, 64, 78, 69, 78, 70, 78, 73, 78, 118, 78],
	#[65535, 75, 7, 75, 11, 75, 31, 75, 33, 75, 35, 75, 36, 75, 57, 75, 58, 75, 59, 75, 60, 75, 63, 75, 64, 75, 65, 75, 66, 75, 67, 75, 68, 75, 69, 75, 70, 75, 71, 75, 73, 75, 118, 75],
	#[7, 195, 11, 195, 10, 195, 2, 195, 31, 195, 33, 195, 63, 195, 37, 105, 38, 195, 86, 195, 42, -159, 91, 195, 102, 195, 101, 195, 72, -161, 103, 195, 90, 195, 89, 195, 100, 195, 84, 195, 85, 195, 99, 195, 87, 195, 88, 195, 104, 195, 97, 195, 105, 195, 92, 195, 107, 195, 95, 195, 106, 195, 108, 195, 109, 195, 110, 195, 111, 195, 112, 195, 113, 195, 114, 195, 115, 195, 116, 195, 117, 195, 118, 195, 119, 195, 121, 195, 122, 195, 123, 195, 124, 195, 125, 195, 126, 195, 127, 195, 129, 195, 130, 195, 131, 195, 132, 195, 133, 195, 136, 195, 137, 195, 138, 195, 139, 195, 140, 195, 141, 195, 226, 195],
	#[42, -578],
	#[7, -13, 11, -127, 13, 319, 31, -128, 33, -139, 35, -123, 36, -129, 43, 319, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -589, 43, -588],
	#[7, -13, 11, -250, 10, -223, 2, -227, 31, -228, 33, -256, 38, -253, 60, -209, 63, -232, 102, -236, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 93, -211, 95, -100, 97, -237, 99, -245, 100, -235, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -582, 109, -581, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 220, -207, 221, -210, 222, -212, 223, -214, 226, -251],
	#[2, 410, 7, 410, 10, 410, 11, 410, 121, 410, 131, 410, 31, 410, 33, 410, 38, 410, 42, 183, 60, 410, 63, 410, 84, 410, 85, 410, 86, 410, 87, 410, 88, 410, 89, 410, 90, 410, 91, 410, 92, 410, 93, 410, 95, 410, 97, 410, 99, 410, 100, 410, 101, 410, 102, 410, 103, 410, 104, 410, 105, 410, 106, 410, 107, 410, 108, 410, 109, 410, 110, 410, 111, 410, 112, 410, 113, 410, 114, 410, 115, 410, 116, 410, 117, 410, 118, 410, 119, 410, 122, 410, 123, 410, 124, 410, 125, 410, 126, 410, 127, 410, 129, 410, 130, 410, 132, 410, 133, 410, 136, 410, 137, 410, 138, 410, 139, 410, 140, 410, 141, 410, 222, 410, 223, 410, 226, 410, 220, 410, 221, 410],
	#[103, 409, 38, 409, 86, 409, 100, 409, 87, 409, 2, 409, 63, 409, 7, 409, 60, 409, 10, 409, 11, 409, 33, 409, 84, 409, 93, 409, 89, 409, 139, 409, 92, 409, 101, 409, 118, 409, 141, 409, 106, 409, 125, 409, 31, 409, 119, 409, 109, 409, 95, 409, 124, 409, 108, 409, 42, 182, 130, 409, 127, 409, 105, 409, 129, 409, 91, 409, 90, 409, 126, 409, 111, 409, 138, 409, 123, 409, 85, 409, 99, 409, 110, 409, 88, 409, 140, 409, 97, 409, 133, 409, 117, 409, 116, 409, 122, 409, 136, 409, 102, 409, 114, 409, 104, 409, 131, 409, 107, 409, 112, 409, 113, 409, 121, 409, 115, 409, 137, 409, 132, 409, 222, 409, 223, 409, 226, 409, 220, 409, 221, 409],
	#[65535, 322, 13, 322, 43, 322, 60, 322, 59, 322, 58, 322, 57, 322, 93, 322, 108, 322, 109, 322, 220, 322, 221, 322, 222, 322, 223, 322],
	#[13, 320, 43, 320, 58, -167, 93, -211, 59, -169, 108, -208, 60, -209, 57, -170, 109, -213, 222, -212, 223, -214, 220, -207, 221, -210],
	#[65535, 404, 13, 404, 43, 404, 60, 404, 57, 404, 58, 404, 59, 404, 93, 404, 108, 404, 109, 404, 220, 404, 221, 404, 222, 404, 223, 404],
	#[65535, 323, 13, 323, 43, 323, 57, 323, 58, 323, 59, 323, 60, 323, 93, 323, 108, 323, 109, 323, 220, 323, 221, 323, 222, 323, 223, 323],
	#[65535, 405, 13, 405, 43, 405, 57, 405, 58, 405, 59, 405, 60, 405, 93, 405, 108, 405, 109, 405, 220, 405, 221, 405, 222, 405, 223, 405],
	#[65535, 211, 57, 211, 58, 211, 59, 211, 96, 211, 171, 211, 172, 211, 173, 211, 174, 211, 175, 211, 176, 211, 177, 211, 178, 211, 179, 211, 180, 211, 181, 211, 182, 211, 183, 211, 184, 211, 185, 211, 186, 211, 187, 211],
	#[63, -121, 7, -13, 35, -123, 11, -127, 68, -133, 36, -129, 66, -126, 33, -139, 67, -137, 31, -128, 65, -122, 71, -130, 64, -141, 69, -136, 70, -140, 73, -138, 118, -36],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 60, -209, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 93, -211, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -582, 109, -581, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 220, -207, 221, -210, 222, -212, 223, -214, 226, -251],
	#[65535, 322, 13, 322, 43, 322, 57, 322, 58, 322, 59, 322, 60, 322, 93, 322, 108, 322, 109, 322, 220, 322, 221, 322, 222, 322, 223, 322],
	#[13, 321, 43, 321, 60, -209, 57, -170, 58, -167, 59, -169, 93, -211, 108, -208, 109, -213, 222, -212, 221, -210, 220, -207, 223, -214],
	#[57, -192, 58, -175, 59, -179, 96, -594, 171, -180, 172, -191, 173, -182, 174, -184, 175, -178, 176, -185, 177, -193, 178, -176, 179, -181, 180, -183, 181, -186, 182, -189, 183, -174, 184, -177, 185, -187, 186, -190, 187, -173],
	#[69, -595],
	#[2, -227, 33, -256, 84, -247, 31, -228, 63, -232, 7, -13, 10, -223, 11, -250, 122, -107, 90, -225, 38, -253, 103, -254, 86, -242, 123, -105, 108, -89, 92, -241, 101, -255, 89, -233, 131, -110, 85, -259, 87, -257, 125, -62, 129, -81, 109, -80, 95, -100, 124, -76, 130, -57, 127, -102, 106, -231, 91, -252, 121, -58, 126, -115, 114, -72, 111, -56, 97, -237, 105, -246, 113, -94, 99, -245, 110, -65, 88, -248, 140, -64, 119, -33, 118, -36, 117, -119, 116, -117, 100, -235, 136, -87, 102, -236, 141, -120, 104, -229, 138, -104, 107, -118, 112, -97, 139, -77, 133, -88, 115, -70, 137, -59, 132, -93, 226, -251],
	#[168, -597],
	#[69, -598],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 287, 7, 287, 32, 287, 63, 287, 62, 287, 101, 287, 102, 287, 95, 287, 97, 287, 127, 287, 99, 287, 100, 287, 103, 287, 104, 287, 105, 287, 106, 287, 107, 287, 108, 287, 109, 287, 110, 287, 111, 287, 112, 287, 113, 287, 114, 287, 115, 287, 116, 287, 117, 287, 118, 287, 121, 287, 122, 287, 123, 287, 124, 287, 125, 287, 126, 287, 129, 287, 130, 287, 131, 287, 132, 287, 133, 287, 136, 287, 137, 287, 138, 287, 139, 287, 140, 287, 141, 287, 162, 287, 163, 287, 164, 287, 165, 287, 166, 287, 167, 287, 169, 287, 170, 287, 188, 287, 189, 287, 190, 287, 192, 287, 193, 287, 195, 287, 196, 287, 197, 287, 198, 287, 199, 287, 200, 287, 204, 287, 205, 287],
	#[65535, 72, 7, 72, 11, 72, 31, 72, 33, 72, 35, 72, 36, 72, 57, 72, 58, 72, 59, 72, 60, 72, 63, 72, 64, 72, 65, 72, 66, 72, 67, 72, 68, 72, 69, 72, 70, 72, 71, 72, 73, 72, 118, 72],
	#[65535, 288, 7, 288, 32, 288, 62, 288, 63, 288, 95, 288, 97, 288, 99, 288, 100, 288, 101, 288, 102, 288, 103, 288, 104, 288, 105, 288, 106, 288, 107, 288, 108, 288, 109, 288, 110, 288, 111, 288, 112, 288, 113, 288, 114, 288, 115, 288, 116, 288, 117, 288, 118, 288, 121, 288, 122, 288, 123, 288, 124, 288, 125, 288, 126, 288, 127, 288, 129, 288, 130, 288, 131, 288, 132, 288, 133, 288, 136, 288, 137, 288, 138, 288, 139, 288, 140, 288, 141, 288, 162, 288, 163, 288, 164, 288, 165, 288, 166, 288, 167, 288, 169, 288, 170, 288, 188, 288, 189, 288, 190, 288, 192, 288, 193, 288, 195, 288, 196, 288, 197, 288, 198, 288, 199, 288, 200, 288, 204, 288, 205, 288],
	#[37, 105, 42, -159, 72, -161, 191, -604],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[102, 354, 63, 354, 7, 354, 111, 354, 13, 354, 103, 354, 99, 354, 101, 354, 100, 354, 109, 354, 126, 354, 105, 354, 95, 354, 108, 354, 118, 354, 107, 354, 124, 354, 106, 354, 115, 354, 110, 354, 125, 354, 97, 354, 114, 354, 129, 354, 166, 354, 130, 354, 113, 354, 131, 354, 140, 354, 137, 354, 136, 354, 127, 354, 116, 354, 122, 354, 141, 354, 104, 354, 138, 354, 121, 354, 165, 354, 167, 354, 123, 354, 163, 354, 112, 354, 139, 354, 133, 354, 170, 354, 117, 354, 164, 354, 132, 354, 169, 354, 188, 354, 189, 354, 190, 354, 192, 354, 193, 354, 195, 354, 196, 354, 197, 354, 198, 354, 199, 354, 200, 354, 201, -606, 202, 354, 203, 354, 204, 354, 205, 354],
	#[65535, 355, 7, 355, 13, 355, 63, 355, 102, 355, 101, 355, 95, 355, 97, 355, 99, 355, 100, 355, 103, 355, 104, 355, 105, 355, 106, 355, 107, 355, 108, 355, 109, 355, 110, 355, 111, 355, 112, 355, 113, 355, 114, 355, 115, 355, 116, 355, 117, 355, 118, 355, 121, 355, 122, 355, 123, 355, 124, 355, 125, 355, 126, 355, 127, 355, 129, 355, 130, 355, 131, 355, 132, 355, 133, 355, 136, 355, 137, 355, 138, 355, 139, 355, 140, 355, 141, 355, 163, 355, 164, 355, 165, 355, 166, 355, 167, 355, 169, 355, 170, 355, 188, 355, 189, 355, 190, 355, 192, 355, 193, 355, 195, 355, 196, 355, 197, 355, 198, 355, 199, 355, 200, 355, 202, 355, 203, 355, 204, 355, 205, 355],
	#[65535, 356, 7, 356, 13, 356, 63, 356, 102, 356, 101, 356, 95, 356, 97, 356, 99, 356, 100, 356, 103, 356, 104, 356, 105, 356, 106, 356, 107, 356, 108, 356, 109, 356, 110, 356, 111, 356, 112, 356, 113, 356, 114, 356, 115, 356, 116, 356, 117, 356, 118, 356, 121, 356, 122, 356, 123, 356, 124, 356, 125, 356, 126, 356, 127, 356, 129, 356, 130, 356, 131, 356, 132, 356, 133, 356, 136, 356, 137, 356, 138, 356, 139, 356, 140, 356, 141, 356, 163, 356, 164, 356, 165, 356, 166, 356, 167, 356, 169, 356, 170, 356, 188, 356, 189, 356, 190, 356, 192, 356, 193, 356, 195, 356, 196, 356, 197, 356, 198, 356, 199, 356, 200, 356, 202, 356, 203, 356, 204, 356, 205, 356],
	#[7, 440, 13, -455, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 202, -610, 203, -609, 204, 440, 205, 440],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 337, 7, 337, 63, 337, 102, 337, 101, 337, 95, 337, 97, 337, 99, 337, 100, 337, 103, 337, 104, 337, 105, 337, 106, 337, 107, 337, 108, 337, 109, 337, 110, 337, 111, 337, 112, 337, 113, 337, 114, 337, 115, 337, 116, 337, 117, 337, 118, 337, 121, 337, 122, 337, 123, 337, 124, 337, 125, 337, 126, 337, 127, 337, 129, 337, 130, 337, 131, 337, 132, 337, 133, 337, 136, 337, 137, 337, 138, 337, 139, 337, 140, 337, 141, 337, 163, 337, 164, 337, 165, 337, 166, 337, 167, 337, 169, 337, 170, 337, 188, 337, 189, 337, 190, 337, 192, 337, 193, 337, 195, 337, 196, 337, 197, 337, 198, 337, 199, 337, 200, 337, 204, 337, 205, 337],
	#[65535, 357, 7, 357, 13, 357, 63, 357, 102, 357, 101, 357, 95, 357, 97, 357, 99, 357, 100, 357, 103, 357, 104, 357, 105, 357, 106, 357, 107, 357, 108, 357, 109, 357, 110, 357, 111, 357, 112, 357, 113, 357, 114, 357, 115, 357, 116, 357, 117, 357, 118, 357, 121, 357, 122, 357, 123, 357, 124, 357, 125, 357, 126, 357, 127, 357, 129, 357, 130, 357, 131, 357, 132, 357, 133, 357, 136, 357, 137, 357, 138, 357, 139, 357, 140, 357, 141, 357, 163, 357, 164, 357, 165, 357, 166, 357, 167, 357, 169, 357, 170, 357, 188, 357, 189, 357, 190, 357, 192, 357, 193, 357, 195, 357, 196, 357, 197, 357, 198, 357, 199, 357, 200, 357, 202, 357, 203, 357, 204, 357, 205, 357],
	#[65535, 358, 95, 358, 127, 358, 117, 358, 103, 358, 199, 358, 189, 358, 63, 358, 7, 358, 101, 358, 197, 358, 141, 358, 102, 358, 106, 358, 13, 358, 109, 358, 140, 358, 99, 358, 115, 358, 111, 358, 105, 358, 116, 358, 166, 358, 190, 358, 97, 358, 125, 358, 138, 358, 104, 358, 124, 358, 108, 358, 198, 358, 131, 358, 133, 358, 107, 358, 110, 358, 114, 358, 170, 358, 163, 358, 100, 358, 136, 358, 121, 358, 129, 358, 118, 358, 167, 358, 112, 358, 113, 358, 205, 358, 123, 358, 126, 358, 164, 358, 139, 358, 132, 358, 122, 358, 137, 358, 204, 358, 188, 358, 169, 358, 130, 358, 165, 358, 193, 358, 196, 358, 195, 358, 203, 358, 192, 358, 200, 358, 202, 358],
	#[13, -341],
	#[7, 440, 13, -617, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 352, 111, 352, 7, 352, 103, 352, 101, 352, 141, 352, 118, 352, 167, 352, 107, 352, 163, 352, 138, 352, 106, 352, 63, 352, 110, 352, 100, 352, 125, 352, 97, 352, 105, 352, 109, 352, 99, 352, 166, 352, 130, 352, 133, 352, 102, 352, 140, 352, 108, 352, 114, 352, 95, 352, 127, 352, 116, 352, 122, 352, 129, 352, 115, 352, 126, 352, 104, 352, 131, 352, 113, 352, 124, 352, 165, 352, 137, 352, 123, 352, 136, 352, 112, 352, 139, 352, 121, 352, 170, 352, 117, 352, 164, 352, 132, 352, 169, 352, 188, 352, 189, 352, 190, 352, 192, 352, 193, 352, 195, 352, 196, 352, 197, 352, 198, 352, 199, 352, 200, 352, 204, 352, 205, 352],
	#[38, -345, 225, -457],
	#[13, -619],
	#[35, -123, 68, -133, 36, -129, 63, -121, 7, -13, 11, -127, 71, -130, 67, -137, 66, -126, 65, -122, 33, -139, 31, -128, 64, -141, 69, -136, 70, -140, 73, -138, 118, -36],
	#[13, -621],
	#[68, -133, 36, -129, 118, -36, 63, -121, 7, -13, 35, -123, 11, -127, 71, -130, 67, -137, 66, -126, 65, -122, 33, -139, 31, -128, 64, -141, 69, -136, 70, -140, 73, -138],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 331, 7, 331, 102, 331, 63, 331, 101, 331, 95, 331, 97, 331, 99, 331, 100, 331, 103, 331, 104, 331, 105, 331, 106, 331, 107, 331, 108, 331, 109, 331, 110, 331, 111, 331, 112, 331, 113, 331, 114, 331, 115, 331, 116, 331, 117, 331, 118, 331, 121, 331, 122, 331, 123, 331, 124, 331, 125, 331, 126, 331, 127, 331, 129, 331, 130, 331, 131, 331, 132, 331, 133, 331, 136, 331, 137, 331, 138, 331, 139, 331, 140, 331, 141, 331, 163, 331, 164, 331, 165, 331, 166, 331, 167, 331, 169, 331, 170, 331, 188, 331, 189, 331, 190, 331, 192, 331, 193, 331, 195, 331, 196, 331, 197, 331, 198, 331, 199, 331, 200, 331, 204, 331, 205, 331],
	#[13, -625],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 66, -126, 65, -122, 64, -141, 63, -121, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -455, 37, 105, 42, -159, 63, 440, 72, -161, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 332, 117, 332, 63, 332, 7, 332, 95, 332, 126, 332, 116, 332, 100, 332, 103, 332, 123, 332, 140, 332, 139, 332, 108, 332, 101, 332, 118, 332, 131, 332, 166, 332, 125, 332, 129, 332, 109, 332, 115, 332, 138, 332, 112, 332, 130, 332, 127, 332, 124, 332, 102, 332, 167, 332, 114, 332, 111, 332, 198, 332, 106, 332, 121, 332, 110, 332, 165, 332, 97, 332, 164, 332, 163, 332, 99, 332, 122, 332, 136, 332, 137, 332, 141, 332, 104, 332, 105, 332, 107, 332, 199, 332, 113, 332, 133, 332, 170, 332, 132, 332, 169, 332, 197, 332, 195, 332, 205, 332, 204, 332, 188, 332, 189, 332, 190, 332, 192, 332, 193, 332, 200, 332, 196, 332],
	#[13, -629],
	#[11, -630],
	#[12, -632, 69, -633],
	#[12, -637, 13, -636],
	#[126, 440, 63, 440, 117, 440, 103, 440, 124, 440, 111, 440, 115, 440, 122, 440, 118, 440, 166, 440, 7, 440, 95, 440, 139, 440, 102, 440, 127, 440, 13, -455, 116, 440, 62, 440, 32, 440, 137, 440, 123, 440, 140, 440, 167, 440, 108, 440, 114, 440, 131, 440, 125, 440, 129, 440, 109, 440, 198, 440, 138, 440, 130, 440, 110, 440, 97, 440, 105, 440, 106, 440, 113, 440, 121, 440, 165, 440, 136, 440, 163, 440, 99, 440, 100, 440, 101, 440, 112, 440, 141, 440, 104, 440, 107, 440, 199, 440, 164, 440, 133, 440, 170, 440, 132, 440, 162, 440, 169, 440, 197, 440, 195, 440, 205, 440, 204, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 200, 440, 196, 440],
	#[7, -13, 11, -250, 10, -223, 2, -227, 31, -228, 33, -256, 38, -253, 63, -232, 102, -236, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 317, 12, 317, 13, 317],
	#[65535, 286, 63, 286, 95, 286, 117, 286, 170, 286, 7, 286, 141, 286, 126, 286, 198, 286, 103, 286, 133, 286, 140, 286, 139, 286, 108, 286, 111, 286, 199, 286, 118, 286, 131, 286, 166, 286, 106, 286, 32, 286, 115, 286, 62, 286, 127, 286, 110, 286, 97, 286, 105, 286, 99, 286, 102, 286, 167, 286, 107, 286, 101, 286, 100, 286, 123, 286, 130, 286, 125, 286, 165, 286, 109, 286, 164, 286, 163, 286, 116, 286, 122, 286, 136, 286, 137, 286, 114, 286, 104, 286, 138, 286, 124, 286, 112, 286, 113, 286, 121, 286, 129, 286, 132, 286, 162, 286, 169, 286, 197, 286, 195, 286, 205, 286, 204, 286, 188, 286, 189, 286, 190, 286, 192, 286, 193, 286, 200, 286, 196, 286],
	#[69, -639],
	#[95, 440, 131, 440, 125, 440, 63, 440, 7, 440, 99, 440, 126, 440, 13, -455, 109, 440, 103, 440, 140, 440, 32, 440, 108, 440, 111, 440, 118, 440, 97, 440, 166, 440, 106, 440, 100, 440, 102, 440, 115, 440, 62, 440, 107, 440, 110, 440, 127, 440, 105, 440, 129, 440, 122, 440, 117, 440, 167, 440, 114, 440, 101, 440, 198, 440, 123, 440, 113, 440, 130, 440, 112, 440, 137, 440, 165, 440, 124, 440, 163, 440, 116, 440, 139, 440, 136, 440, 204, 440, 141, 440, 104, 440, 138, 440, 121, 440, 199, 440, 164, 440, 133, 440, 170, 440, 132, 440, 162, 440, 169, 440, 197, 440, 195, 440, 205, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 200, 440, 196, 440],
	#[65535, 285, 7, 285, 32, 285, 62, 285, 63, 285, 95, 285, 97, 285, 99, 285, 100, 285, 101, 285, 102, 285, 103, 285, 104, 285, 105, 285, 106, 285, 107, 285, 108, 285, 109, 285, 110, 285, 111, 285, 112, 285, 113, 285, 114, 285, 115, 285, 116, 285, 117, 285, 118, 285, 121, 285, 122, 285, 123, 285, 124, 285, 125, 285, 126, 285, 127, 285, 129, 285, 130, 285, 131, 285, 132, 285, 133, 285, 136, 285, 137, 285, 138, 285, 139, 285, 140, 285, 141, 285, 162, 285, 163, 285, 164, 285, 165, 285, 166, 285, 167, 285, 169, 285, 170, 285, 188, 285, 189, 285, 190, 285, 192, 285, 193, 285, 195, 285, 196, 285, 197, 285, 198, 285, 199, 285, 200, 285, 204, 285, 205, 285],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[65535, 318, 13, 318, 12, 318],
	#[7, 419, 11, 419, 31, 419, 33, 419, 35, 419, 36, 419, 63, 419, 64, 419, 65, 419, 66, 419, 67, 419, 68, 419, 69, 419, 70, 419, 71, 419, 73, 419, 118, 419, 195, -545],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -644],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 66, -126, 65, -122, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -492, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 344, 7, 344, 63, 344, 102, 344, 101, 344, 95, 344, 97, 344, 99, 344, 100, 344, 103, 344, 104, 344, 105, 344, 106, 344, 107, 344, 108, 344, 109, 344, 110, 344, 111, 344, 112, 344, 113, 344, 114, 344, 115, 344, 116, 344, 117, 344, 118, 344, 121, 344, 122, 344, 123, 344, 124, 344, 125, 344, 126, 344, 127, 344, 129, 344, 130, 344, 131, 344, 132, 344, 133, 344, 136, 344, 137, 344, 138, 344, 139, 344, 140, 344, 141, 344, 163, 344, 164, 344, 165, 344, 166, 344, 167, 344, 169, 344, 170, 344, 188, 344, 189, 344, 190, 344, 192, 344, 193, 344, 195, 344, 196, 344, 197, 344, 198, 344, 199, 344, 200, 344, 204, 344, 205, 344],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 65, -122, 66, -126, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -649],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 65, -122, 66, -126, 63, -121, 64, -141, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[206, -462, 207, 361, 208, 361, 209, 361, 210, 361, 211, 361, 212, 361],
	#[207, -464, 208, -465, 209, -466, 210, -467, 211, -469, 212, -470],
	#[7, 440, 13, -492, 63, 440, 101, 440, 95, 440, 97, 440, 99, 440, 100, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 345, 63, 345, 7, 345, 95, 345, 126, 345, 100, 345, 99, 345, 114, 345, 124, 345, 101, 345, 103, 345, 118, 345, 129, 345, 102, 345, 106, 345, 105, 345, 132, 345, 122, 345, 125, 345, 127, 345, 123, 345, 140, 345, 108, 345, 97, 345, 104, 345, 107, 345, 111, 345, 138, 345, 141, 345, 113, 345, 112, 345, 109, 345, 131, 345, 133, 345, 130, 345, 117, 345, 116, 345, 121, 345, 110, 345, 136, 345, 139, 345, 115, 345, 137, 345, 163, 345, 164, 345, 165, 345, 166, 345, 167, 345, 169, 345, 170, 345, 188, 345, 189, 345, 190, 345, 192, 345, 193, 345, 195, 345, 196, 345, 197, 345, 198, 345, 199, 345, 200, 345, 204, 345, 205, 345],
	#[7, 440, 13, -656, 37, 105, 42, -159, 63, 440, 72, -161, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 339, 7, 339, 63, 339, 95, 339, 97, 339, 99, 339, 100, 339, 101, 339, 102, 339, 103, 339, 104, 339, 105, 339, 106, 339, 107, 339, 108, 339, 109, 339, 110, 339, 111, 339, 112, 339, 113, 339, 114, 339, 115, 339, 116, 339, 117, 339, 118, 339, 121, 339, 122, 339, 123, 339, 124, 339, 125, 339, 126, 339, 127, 339, 129, 339, 130, 339, 131, 339, 132, 339, 133, 339, 136, 339, 137, 339, 138, 339, 139, 339, 140, 339, 141, 339, 163, 339, 164, 339, 165, 339, 166, 339, 167, 339, 169, 339, 170, 339, 188, 339, 189, 339, 190, 339, 192, 339, 193, 339, 195, 339, 196, 339, 197, 339, 198, 339, 199, 339, 200, 339, 204, 339, 205, 339],
	#[36, -657, 93, -494, 225, -457],
	#[7, -13, 11, -250, 10, -223, 2, -227, 31, -228, 33, -256, 38, -253, 63, -232, 102, -236, 101, -255, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -492, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 340, 7, 340, 63, 340, 102, 340, 101, 340, 95, 340, 97, 340, 99, 340, 100, 340, 103, 340, 104, 340, 105, 340, 106, 340, 107, 340, 108, 340, 109, 340, 110, 340, 111, 340, 112, 340, 113, 340, 114, 340, 115, 340, 116, 340, 117, 340, 118, 340, 121, 340, 122, 340, 123, 340, 124, 340, 125, 340, 126, 340, 127, 340, 129, 340, 130, 340, 131, 340, 132, 340, 133, 340, 136, 340, 137, 340, 138, 340, 139, 340, 140, 340, 141, 340, 163, 340, 164, 340, 165, 340, 166, 340, 167, 340, 169, 340, 170, 340, 188, 340, 189, 340, 190, 340, 192, 340, 193, 340, 195, 340, 196, 340, 197, 340, 198, 340, 199, 340, 200, 340, 204, 340, 205, 340],
	#[121, -663, 123, -661, 139, -662, 140, -664, 141, -669, 213, -670, 214, -666, 215, -665, 216, -668, 217, -671, 218, -672],
	#[65535, 371, 7, 371, 11, 371, 31, 371, 33, 371, 35, 371, 36, 371, 66, 371, 65, 371, 63, 371, 64, 371, 67, 371, 68, 371, 69, 371, 70, 371, 71, 371, 73, 371, 118, 371],
	#[65535, 372, 7, 372, 11, 372, 31, 372, 33, 372, 35, 372, 36, 372, 63, 372, 64, 372, 65, 372, 66, 372, 67, 372, 68, 372, 69, 372, 70, 372, 71, 372, 73, 372, 118, 372],
	#[65535, 370, 7, 370, 11, 370, 31, 370, 33, 370, 35, 370, 36, 370, 63, 370, 64, 370, 65, 370, 66, 370, 67, 370, 68, 370, 69, 370, 70, 370, 71, 370, 73, 370, 118, 370],
	#[65535, 374, 7, 374, 11, 374, 36, 374, 67, 374, 31, 374, 33, 374, 35, 374, 65, 374, 66, 374, 63, 374, 64, 374, 68, 374, 69, 374, 70, 374, 71, 374, 73, 374, 118, 374],
	#[65535, 376, 7, 376, 11, 376, 31, 376, 33, 376, 35, 376, 36, 376, 63, 376, 64, 376, 65, 376, 66, 376, 67, 376, 68, 376, 69, 376, 70, 376, 71, 376, 73, 376, 118, 376],
	#[65535, 373, 7, 373, 11, 373, 31, 373, 33, 373, 35, 373, 36, 373, 63, 373, 64, 373, 65, 373, 66, 373, 67, 373, 68, 373, 69, 373, 70, 373, 71, 373, 73, 373, 118, 373],
	#[63, -121, 7, -13, 11, -127, 68, -133, 36, -129, 35, -123, 33, -139, 67, -137, 31, -128, 66, -126, 65, -122, 71, -130, 64, -141, 69, -136, 70, -140, 73, -138, 118, -36],
	#[65535, 377, 7, 377, 11, 377, 31, 377, 33, 377, 35, 377, 36, 377, 63, 377, 64, 377, 65, 377, 66, 377, 67, 377, 68, 377, 69, 377, 70, 377, 71, 377, 73, 377, 118, 377],
	#[65535, 375, 7, 375, 11, 375, 31, 375, 33, 375, 35, 375, 36, 375, 71, 375, 68, 375, 63, 375, 64, 375, 65, 375, 66, 375, 67, 375, 69, 375, 70, 375, 73, 375, 118, 375],
	#[65535, 369, 7, 369, 11, 369, 31, 369, 33, 369, 35, 369, 36, 369, 66, 369, 65, 369, 63, 369, 64, 369, 67, 369, 68, 369, 69, 369, 70, 369, 71, 369, 73, 369, 118, 369],
	#[65535, 378, 7, 378, 11, 378, 31, 378, 33, 378, 35, 378, 36, 378, 63, 378, 64, 378, 65, 378, 66, 378, 67, 378, 68, 378, 69, 378, 70, 378, 71, 378, 73, 378, 118, 378],
	#[65535, 379, 70, 379, 7, 379, 11, 379, 31, 379, 33, 379, 35, 379, 36, 379, 66, 379, 63, 379, 64, 379, 65, 379, 67, 379, 68, 379, 69, 379, 71, 379, 73, 379, 118, 379],
	#[13, -674],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[206, -462, 207, 361, 208, 361, 209, 361, 210, 361, 211, 361, 212, 361],
	#[207, -464, 208, -465, 209, -466, 210, -467, 211, -469, 212, -470],
	#[63, 440, 7, 440, 13, -455, 126, 440, 111, 440, 140, 440, 102, 440, 115, 440, 110, 440, 95, 440, 127, 440, 109, 440, 118, 440, 106, 440, 103, 440, 131, 440, 101, 440, 105, 440, 99, 440, 125, 440, 124, 440, 108, 440, 114, 440, 107, 440, 133, 440, 97, 440, 117, 440, 116, 440, 100, 440, 129, 440, 112, 440, 141, 440, 104, 440, 138, 440, 113, 440, 121, 440, 130, 440, 167, 440, 123, 440, 122, 440, 139, 440, 132, 440, 136, 440, 137, 440, 163, 440, 164, 440, 165, 440, 166, 440, 205, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440],
	#[65535, 348, 7, 348, 63, 348, 101, 348, 103, 348, 109, 348, 111, 348, 97, 348, 102, 348, 99, 348, 115, 348, 110, 348, 116, 348, 95, 348, 125, 348, 127, 348, 123, 348, 118, 348, 106, 348, 104, 348, 133, 348, 105, 348, 117, 348, 126, 348, 132, 348, 100, 348, 122, 348, 124, 348, 108, 348, 114, 348, 121, 348, 107, 348, 112, 348, 113, 348, 129, 348, 130, 348, 131, 348, 136, 348, 137, 348, 138, 348, 139, 348, 140, 348, 141, 348, 169, 348, 163, 348, 164, 348, 165, 348, 166, 348, 167, 348, 170, 348, 188, 348, 189, 348, 190, 348, 192, 348, 193, 348, 195, 348, 196, 348, 197, 348, 198, 348, 199, 348, 200, 348, 204, 348, 205, 348],
	#[13, -680],
	#[38, -681],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 353, 7, 353, 63, 353, 95, 353, 97, 353, 99, 353, 100, 353, 101, 353, 102, 353, 103, 353, 104, 353, 105, 353, 106, 353, 107, 353, 108, 353, 109, 353, 110, 353, 111, 353, 112, 353, 113, 353, 114, 353, 115, 353, 116, 353, 117, 353, 118, 353, 121, 353, 122, 353, 123, 353, 124, 353, 125, 353, 126, 353, 127, 353, 129, 353, 130, 353, 131, 353, 132, 353, 133, 353, 136, 353, 137, 353, 138, 353, 139, 353, 140, 353, 141, 353, 163, 353, 164, 353, 165, 353, 166, 353, 167, 353, 169, 353, 170, 353, 188, 353, 189, 353, 190, 353, 192, 353, 193, 353, 195, 353, 196, 353, 197, 353, 198, 353, 199, 353, 200, 353, 204, 353, 205, 353],
	#[7, 440, 13, -455, 32, 440, 62, 440, 63, 440, 95, 440, 103, 440, 114, 440, 106, 440, 97, 440, 127, 440, 99, 440, 100, 440, 101, 440, 102, 440, 126, 440, 104, 440, 105, 440, 125, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 133, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 129, 440, 130, 440, 131, 440, 132, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 289, 7, 289, 32, 289, 62, 289, 63, 289, 95, 289, 97, 289, 99, 289, 100, 289, 101, 289, 102, 289, 103, 289, 104, 289, 105, 289, 106, 289, 107, 289, 108, 289, 109, 289, 110, 289, 111, 289, 112, 289, 113, 289, 114, 289, 115, 289, 116, 289, 117, 289, 118, 289, 121, 289, 122, 289, 123, 289, 124, 289, 125, 289, 126, 289, 127, 289, 129, 289, 130, 289, 131, 289, 132, 289, 133, 289, 136, 289, 137, 289, 138, 289, 139, 289, 140, 289, 141, 289, 162, 289, 163, 289, 164, 289, 165, 289, 166, 289, 167, 289, 169, 289, 170, 289, 188, 289, 189, 289, 190, 289, 192, 289, 193, 289, 195, 289, 196, 289, 197, 289, 198, 289, 199, 289, 200, 289, 204, 289, 205, 289],
	#[13, -686],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 118, -36, 73, -138],
	#[13, -688],
	#[7, -13, 11, -127, 70, -140, 31, -128, 33, -139, 63, -121, 35, -123, 36, -129, 69, -136, 66, -126, 64, -141, 65, -122, 67, -137, 68, -133, 71, -130, 118, -36, 73, -138],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 335, 7, 335, 63, 335, 95, 335, 97, 335, 99, 335, 100, 335, 101, 335, 102, 335, 103, 335, 104, 335, 105, 335, 106, 335, 107, 335, 108, 335, 109, 335, 110, 335, 111, 335, 112, 335, 113, 335, 114, 335, 115, 335, 116, 335, 117, 335, 118, 335, 121, 335, 122, 335, 123, 335, 124, 335, 125, 335, 126, 335, 127, 335, 129, 335, 130, 335, 131, 335, 132, 335, 133, 335, 136, 335, 137, 335, 138, 335, 139, 335, 140, 335, 141, 335, 163, 335, 164, 335, 165, 335, 166, 335, 167, 335, 169, 335, 170, 335, 188, 335, 189, 335, 190, 335, 192, 335, 193, 335, 195, 335, 196, 335, 197, 335, 198, 335, 199, 335, 200, 335, 204, 335, 205, 335],
	#[13, -692],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 440, 13, -455, 63, 440, 103, 440, 111, 440, 126, 440, 115, 440, 110, 440, 95, 440, 127, 440, 118, 440, 124, 440, 108, 440, 102, 440, 114, 440, 101, 440, 106, 440, 113, 440, 99, 440, 125, 440, 109, 440, 105, 440, 131, 440, 107, 440, 133, 440, 97, 440, 117, 440, 116, 440, 100, 440, 129, 440, 112, 440, 141, 440, 104, 440, 138, 440, 140, 440, 121, 440, 130, 440, 167, 440, 123, 440, 122, 440, 139, 440, 132, 440, 136, 440, 137, 440, 163, 440, 164, 440, 165, 440, 166, 440, 205, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440],
	#[65535, 333, 7, 333, 63, 333, 95, 333, 97, 333, 99, 333, 100, 333, 101, 333, 102, 333, 103, 333, 104, 333, 105, 333, 106, 333, 107, 333, 108, 333, 109, 333, 110, 333, 111, 333, 112, 333, 113, 333, 114, 333, 115, 333, 116, 333, 117, 333, 118, 333, 121, 333, 122, 333, 123, 333, 124, 333, 125, 333, 126, 333, 127, 333, 129, 333, 130, 333, 131, 333, 132, 333, 133, 333, 136, 333, 137, 333, 138, 333, 139, 333, 140, 333, 141, 333, 163, 333, 164, 333, 165, 333, 166, 333, 167, 333, 169, 333, 170, 333, 188, 333, 189, 333, 190, 333, 192, 333, 193, 333, 195, 333, 196, 333, 197, 333, 198, 333, 199, 333, 200, 333, 204, 333, 205, 333],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 37, 105, 38, -253, 42, -159, 63, -232, 72, -161, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -697],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 326, 7, 326, 103, 326, 102, 326, 63, 326, 106, 326, 110, 326, 114, 326, 95, 326, 97, 326, 127, 326, 99, 326, 100, 326, 101, 326, 115, 326, 126, 326, 104, 326, 105, 326, 125, 326, 107, 326, 108, 326, 109, 326, 123, 326, 111, 326, 112, 326, 113, 326, 133, 326, 132, 326, 116, 326, 117, 326, 118, 326, 121, 326, 122, 326, 124, 326, 129, 326, 130, 326, 131, 326, 136, 326, 137, 326, 138, 326, 139, 326, 140, 326, 141, 326, 163, 326, 164, 326, 165, 326, 166, 326, 167, 326, 169, 326, 170, 326, 188, 326, 189, 326, 190, 326, 192, 326, 193, 326, 195, 326, 196, 326, 197, 326, 198, 326, 199, 326, 200, 326, 204, 326, 205, 326],
	#[65535, 311, 7, 311, 13, 311, 32, 311, 62, 311, 63, 311, 95, 311, 97, 311, 99, 311, 100, 311, 101, 311, 102, 311, 103, 311, 104, 311, 105, 311, 106, 311, 107, 311, 108, 311, 109, 311, 110, 311, 111, 311, 112, 311, 113, 311, 114, 311, 115, 311, 116, 311, 117, 311, 118, 311, 121, 311, 122, 311, 123, 311, 124, 311, 125, 311, 126, 311, 127, 311, 129, 311, 130, 311, 131, 311, 132, 311, 133, 311, 136, 311, 137, 311, 138, 311, 139, 311, 140, 311, 141, 311, 162, 311, 163, 311, 164, 311, 165, 311, 166, 311, 167, 311, 169, 311, 170, 311, 188, 311, 189, 311, 190, 311, 192, 311, 193, 311, 195, 311, 196, 311, 197, 311, 198, 311, 199, 311, 200, 311, 204, 311, 205, 311],
	#[167, 440, 102, 440, 127, 440, 13, -705, 103, 440, 166, 440, 164, 440, 32, 440, 100, 440, 105, 440, 111, 440, 199, 440, 163, 440, 99, 440, 62, 440, 108, 440, 113, 440, 165, 440, 63, 440, 107, 440, 7, 440, 95, 440, 118, 440, 133, 440, 115, 440, 162, 440, 101, 440, 126, 440, 129, 440, 198, 440, 125, 440, 114, 440, 170, 440, 139, 440, 116, 440, 122, 440, 117, 440, 104, 440, 123, 440, 106, 440, 141, 440, 112, 440, 132, 440, 131, 440, 121, 440, 138, 440, 110, 440, 97, 440, 140, 440, 124, 440, 137, 440, 136, 440, 130, 440, 109, 440, 169, 440, 197, 440, 196, 440, 195, 440, 204, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 200, 440, 205, 440],
	#[7, 440, 13, -455, 32, 440, 63, 440, 62, 440, 127, 440, 42, -151, 115, 440, 106, 440, 109, 440, 111, 440, 102, 440, 117, 440, 101, 440, 110, 440, 116, 440, 95, 440, 125, 440, 97, 440, 123, 440, 118, 440, 99, 440, 103, 440, 133, 440, 105, 440, 104, 440, 126, 440, 132, 440, 100, 440, 122, 440, 124, 440, 108, 440, 114, 440, 121, 440, 107, 440, 112, 440, 113, 440, 129, 440, 130, 440, 131, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 162, 440, 169, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 280, 7, 280, 32, 280, 62, 280, 63, 280, 95, 280, 97, 280, 99, 280, 100, 280, 101, 280, 102, 280, 103, 280, 104, 280, 105, 280, 106, 280, 107, 280, 108, 280, 109, 280, 110, 280, 111, 280, 112, 280, 113, 280, 114, 280, 115, 280, 116, 280, 117, 280, 118, 280, 121, 280, 122, 280, 123, 280, 124, 280, 125, 280, 126, 280, 127, 280, 129, 280, 130, 280, 131, 280, 132, 280, 133, 280, 136, 280, 137, 280, 138, 280, 139, 280, 140, 280, 141, 280, 162, 280, 163, 280, 164, 280, 165, 280, 166, 280, 167, 280, 169, 280, 170, 280, 188, 280, 189, 280, 190, 280, 192, 280, 193, 280, 195, 280, 196, 280, 197, 280, 198, 280, 199, 280, 200, 280, 204, 280, 205, 280],
	#[65535, 279, 7, 279, 32, 279, 62, 279, 63, 279, 103, 279, 102, 279, 107, 279, 101, 279, 100, 279, 99, 279, 97, 279, 105, 279, 104, 279, 95, 279, 106, 279, 108, 279, 109, 279, 110, 279, 111, 279, 112, 279, 113, 279, 114, 279, 115, 279, 116, 279, 117, 279, 118, 279, 121, 279, 122, 279, 123, 279, 124, 279, 125, 279, 126, 279, 127, 279, 129, 279, 130, 279, 131, 279, 132, 279, 133, 279, 136, 279, 137, 279, 138, 279, 139, 279, 140, 279, 141, 279, 162, 279, 163, 279, 164, 279, 165, 279, 166, 279, 167, 279, 169, 279, 170, 279, 188, 279, 189, 279, 190, 279, 192, 279, 193, 279, 195, 279, 196, 279, 197, 279, 198, 279, 199, 279, 200, 279, 204, 279, 205, 279],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36, 225, -457],
	#[65535, 312, 7, 312, 13, 312, 32, 312, 62, 312, 63, 312, 95, 312, 97, 312, 99, 312, 100, 312, 101, 312, 102, 312, 103, 312, 104, 312, 105, 312, 106, 312, 107, 312, 108, 312, 109, 312, 110, 312, 111, 312, 112, 312, 113, 312, 114, 312, 115, 312, 116, 312, 117, 312, 118, 312, 121, 312, 122, 312, 123, 312, 124, 312, 125, 312, 126, 312, 127, 312, 129, 312, 130, 312, 131, 312, 132, 312, 133, 312, 136, 312, 137, 312, 138, 312, 139, 312, 140, 312, 141, 312, 162, 312, 163, 312, 164, 312, 165, 312, 166, 312, 167, 312, 169, 312, 170, 312, 188, 312, 189, 312, 190, 312, 192, 312, 193, 312, 195, 312, 196, 312, 197, 312, 198, 312, 199, 312, 200, 312, 204, 312, 205, 312],
	#[7, -13, 11, -127, 70, -140, 31, -128, 33, -139, 35, -123, 36, -129, 66, -126, 63, -121, 64, -141, 65, -122, 67, -137, 68, -133, 69, -136, 71, -130, 73, -138, 118, -36],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 37, 105, 38, -253, 42, -159, 63, -232, 72, -161, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 114, -72, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -710],
	#[2, -227, 7, -13, 10, -223, 11, -250, 91, -252, 33, -256, 87, -257, 31, -228, 89, -233, 63, -232, 38, -253, 124, -76, 86, -242, 90, -225, 99, -245, 109, -80, 111, -56, 102, -236, 108, -89, 117, -119, 119, -33, 101, -255, 110, -65, 116, -117, 95, -100, 84, -247, 125, -62, 127, -102, 123, -105, 118, -36, 106, -231, 103, -254, 92, -241, 133, -88, 122, -107, 104, -229, 126, -115, 132, -93, 100, -235, 85, -259, 88, -248, 115, -70, 97, -237, 105, -246, 114, -72, 121, -58, 107, -118, 112, -97, 113, -94, 129, -81, 130, -57, 131, -110, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 328, 167, 328, 63, 328, 123, 328, 164, 328, 116, 328, 140, 328, 124, 328, 190, 328, 7, 328, 127, 328, 165, 328, 109, 328, 100, 328, 103, 328, 133, 328, 139, 328, 108, 328, 199, 328, 118, 328, 138, 328, 125, 328, 114, 328, 111, 328, 198, 328, 141, 328, 121, 328, 110, 328, 115, 328, 97, 328, 170, 328, 107, 328, 95, 328, 136, 328, 117, 328, 99, 328, 122, 328, 101, 328, 102, 328, 126, 328, 104, 328, 105, 328, 106, 328, 112, 328, 113, 328, 129, 328, 131, 328, 130, 328, 166, 328, 137, 328, 163, 328, 132, 328, 169, 328, 197, 328, 204, 328, 188, 328, 189, 328, 200, 328, 192, 328, 193, 328, 195, 328, 196, 328, 205, 328],
	#[13, -714],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -341],
	#[7, 440, 13, -617, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 351, 101, 351, 63, 351, 204, 351, 197, 351, 127, 351, 103, 351, 114, 351, 189, 351, 116, 351, 102, 351, 126, 351, 111, 351, 188, 351, 163, 351, 99, 351, 141, 351, 124, 351, 115, 351, 113, 351, 109, 351, 205, 351, 7, 351, 95, 351, 123, 351, 133, 351, 192, 351, 190, 351, 117, 351, 106, 351, 105, 351, 130, 351, 108, 351, 198, 351, 125, 351, 167, 351, 165, 351, 138, 351, 199, 351, 137, 351, 100, 351, 97, 351, 112, 351, 118, 351, 104, 351, 166, 351, 169, 351, 139, 351, 129, 351, 164, 351, 122, 351, 132, 351, 131, 351, 170, 351, 110, 351, 140, 351, 136, 351, 121, 351, 107, 351, 200, 351, 193, 351, 195, 351, 196, 351],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 37, 105, 38, -253, 42, -159, 63, -232, 95, -100, 72, -161, 103, -254, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 92, -241, 106, -231, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 114, -72, 104, -229, 105, -246, 113, -94, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[13, -720],
	#[2, -227, 7, -13, 10, -223, 11, -250, 31, -228, 33, -256, 38, -253, 63, -232, 92, -241, 84, -247, 85, -259, 86, -242, 87, -257, 88, -248, 89, -233, 90, -225, 91, -252, 114, -72, 95, -100, 97, -237, 99, -245, 100, -235, 101, -255, 102, -236, 103, -254, 104, -229, 105, -246, 106, -231, 107, -118, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 133, -88, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[7, 440, 13, -455, 63, 440, 95, 440, 97, 440, 99, 440, 100, 440, 101, 440, 102, 440, 103, 440, 104, 440, 105, 440, 106, 440, 107, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 114, 440, 115, 440, 116, 440, 117, 440, 118, 440, 121, 440, 122, 440, 123, 440, 124, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 169, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 327, 7, 327, 63, 327, 95, 327, 97, 327, 99, 327, 100, 327, 101, 327, 102, 327, 103, 327, 104, 327, 105, 327, 106, 327, 107, 327, 108, 327, 109, 327, 110, 327, 111, 327, 112, 327, 113, 327, 114, 327, 115, 327, 116, 327, 117, 327, 118, 327, 121, 327, 122, 327, 123, 327, 124, 327, 125, 327, 126, 327, 127, 327, 129, 327, 130, 327, 131, 327, 132, 327, 133, 327, 136, 327, 137, 327, 138, 327, 139, 327, 140, 327, 141, 327, 163, 327, 164, 327, 165, 327, 166, 327, 167, 327, 169, 327, 170, 327, 188, 327, 189, 327, 190, 327, 192, 327, 193, 327, 195, 327, 196, 327, 197, 327, 198, 327, 199, 327, 200, 327, 204, 327, 205, 327],
	#[65535, 73, 33, 73, 68, 73, 70, 73, 31, 73, 63, 73, 7, 73, 57, 73, 35, 73, 11, 73, 69, 73, 67, 73, 73, 73, 36, 73, 59, 73, 66, 73, 58, 73, 60, 73, 65, 73, 71, 73, 64, 73, 118, 73],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 57, -575, 58, -573, 59, -569, 60, -574, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -570, 118, -36],
	#[2, -227, 7, -13, 10, -223, 11, -250, 92, -241, 31, -228, 33, -256, 63, -232, 38, -253, 102, -236, 86, -242, 95, -100, 103, -254, 89, -233, 91, -252, 90, -225, 101, -255, 100, -235, 84, -247, 85, -259, 99, -245, 87, -257, 88, -248, 104, -229, 97, -237, 105, -246, 114, -72, 107, -118, 133, -88, 106, -231, 108, -89, 109, -80, 110, -65, 111, -56, 112, -97, 113, -94, 115, -70, 116, -117, 117, -119, 118, -36, 119, -33, 121, -58, 122, -107, 123, -105, 124, -76, 125, -62, 126, -115, 127, -102, 129, -81, 130, -57, 131, -110, 132, -93, 136, -87, 137, -59, 138, -104, 139, -77, 140, -64, 141, -120, 226, -251],
	#[42, -727],
	#[7, -13, 11, -127, 13, 319, 31, -128, 33, -139, 35, -123, 36, -129, 43, 319, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -589, 43, -729],
	#[65535, 211, 7, 211, 13, 211, 57, 211, 58, 211, 59, 211, 63, 211, 95, 211, 97, 211, 99, 211, 100, 211, 101, 211, 102, 211, 103, 211, 104, 211, 105, 211, 106, 211, 107, 211, 108, 211, 109, 211, 110, 211, 111, 211, 112, 211, 113, 211, 114, 211, 115, 211, 116, 211, 117, 211, 118, 211, 121, 211, 122, 211, 123, 211, 124, 211, 125, 211, 126, 211, 127, 211, 129, 211, 130, 211, 131, 211, 132, 211, 133, 211, 136, 211, 137, 211, 138, 211, 139, 211, 140, 211, 141, 211, 163, 211, 164, 211, 165, 211, 166, 211, 167, 211, 169, 211, 170, 211, 171, 211, 172, 211, 173, 211, 174, 211, 175, 211, 176, 211, 177, 211, 178, 211, 179, 211, 180, 211, 181, 211, 182, 211, 183, 211, 184, 211, 185, 211, 186, 211, 187, 211, 188, 211, 189, 211, 190, 211, 192, 211, 193, 211, 195, 211, 196, 211, 197, 211, 198, 211, 199, 211, 200, 211, 204, 211, 205, 211],
	#[7, 440, 13, -455, 103, 440, 63, 440, 58, -175, 116, 440, 111, 440, 102, 440, 57, -192, 59, -179, 117, 440, 101, 440, 100, 440, 118, 440, 99, 440, 104, 440, 121, 440, 167, 440, 97, 440, 115, 440, 124, 440, 105, 440, 114, 440, 107, 440, 95, 440, 106, 440, 108, 440, 109, 440, 110, 440, 122, 440, 112, 440, 113, 440, 123, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 164, 440, 169, 440, 163, 440, 186, -190, 165, 440, 166, 440, 178, -176, 170, 440, 171, -180, 172, -191, 173, -182, 174, -184, 175, -178, 176, -185, 177, -193, 185, -187, 179, -181, 180, -183, 181, -186, 182, -189, 183, -174, 184, -177, 187, -173, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 338, 63, 338, 101, 338, 7, 338, 126, 338, 198, 338, 97, 338, 111, 338, 199, 338, 118, 338, 141, 338, 166, 338, 100, 338, 102, 338, 95, 338, 106, 338, 130, 338, 103, 338, 110, 338, 116, 338, 127, 338, 105, 338, 140, 338, 99, 338, 115, 338, 122, 338, 117, 338, 167, 338, 107, 338, 138, 338, 137, 338, 104, 338, 109, 338, 108, 338, 129, 338, 133, 338, 163, 338, 113, 338, 121, 338, 136, 338, 114, 338, 170, 338, 131, 338, 124, 338, 112, 338, 164, 338, 123, 338, 125, 338, 139, 338, 165, 338, 132, 338, 169, 338, 197, 338, 195, 338, 204, 338, 188, 338, 189, 338, 190, 338, 192, 338, 193, 338, 200, 338, 196, 338, 205, 338],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 219, 7, 219, 13, 219, 111, 219, 63, 219, 116, 219, 118, 219, 103, 219, 102, 219, 100, 219, 107, 219, 95, 219, 97, 219, 117, 219, 99, 219, 121, 219, 101, 219, 115, 219, 114, 219, 104, 219, 105, 219, 106, 219, 124, 219, 108, 219, 109, 219, 110, 219, 122, 219, 112, 219, 113, 219, 123, 219, 125, 219, 126, 219, 127, 219, 129, 219, 130, 219, 131, 219, 132, 219, 133, 219, 136, 219, 137, 219, 138, 219, 139, 219, 140, 219, 141, 219, 169, 219, 163, 219, 164, 219, 165, 219, 166, 219, 167, 219, 170, 219, 188, 219, 189, 219, 190, 219, 192, 219, 193, 219, 195, 219, 196, 219, 197, 219, 198, 219, 199, 219, 200, 219, 204, 219, 205, 219],
	#[7, 440, 13, -736, 111, 440, 63, 440, 95, 440, 99, 440, 103, 440, 102, 440, 100, 440, 107, 440, 118, 440, 97, 440, 117, 440, 116, 440, 121, 440, 101, 440, 115, 440, 114, 440, 104, 440, 105, 440, 106, 440, 124, 440, 108, 440, 109, 440, 110, 440, 122, 440, 112, 440, 113, 440, 123, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 169, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 350, 7, 350, 63, 350, 95, 350, 97, 350, 99, 350, 100, 350, 101, 350, 102, 350, 103, 350, 104, 350, 105, 350, 106, 350, 107, 350, 108, 350, 109, 350, 110, 350, 111, 350, 112, 350, 113, 350, 114, 350, 115, 350, 116, 350, 117, 350, 118, 350, 121, 350, 122, 350, 123, 350, 124, 350, 125, 350, 126, 350, 127, 350, 129, 350, 130, 350, 131, 350, 132, 350, 133, 350, 136, 350, 137, 350, 138, 350, 139, 350, 140, 350, 141, 350, 163, 350, 164, 350, 165, 350, 166, 350, 167, 350, 169, 350, 170, 350, 188, 350, 189, 350, 190, 350, 192, 350, 193, 350, 195, 350, 196, 350, 197, 350, 198, 350, 199, 350, 200, 350, 204, 350, 205, 350],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 70, -140, 69, -136, 65, -122, 67, -137, 63, -121, 64, -141, 66, -126, 68, -133, 71, -130, 73, -138, 118, -36, 225, -457],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[13, -739],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 70, -140, 69, -136, 65, -122, 67, -137, 63, -121, 64, -141, 66, -126, 68, -133, 71, -130, 118, -36, 73, -138],
	#[13, -741],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 70, -140, 69, -136, 65, -122, 67, -137, 63, -121, 64, -141, 66, -126, 68, -133, 71, -130, 73, -138, 118, -36],
	#[206, -462, 207, 361, 208, 361, 209, 361, 210, 361, 211, 361, 212, 361],
	#[207, -464, 208, -465, 209, -466, 210, -467, 211, -469, 212, -470],
	#[7, 440, 13, -455, 103, 440, 63, 440, 116, 440, 97, 440, 102, 440, 117, 440, 101, 440, 95, 440, 99, 440, 104, 440, 122, 440, 121, 440, 100, 440, 115, 440, 124, 440, 105, 440, 114, 440, 107, 440, 118, 440, 106, 440, 108, 440, 109, 440, 110, 440, 111, 440, 112, 440, 113, 440, 123, 440, 125, 440, 126, 440, 127, 440, 129, 440, 130, 440, 131, 440, 132, 440, 133, 440, 136, 440, 137, 440, 138, 440, 139, 440, 140, 440, 141, 440, 169, 440, 163, 440, 164, 440, 165, 440, 166, 440, 167, 440, 170, 440, 188, 440, 189, 440, 190, 440, 192, 440, 193, 440, 195, 440, 196, 440, 197, 440, 198, 440, 199, 440, 200, 440, 204, 440, 205, 440],
	#[65535, 347, 7, 347, 63, 347, 95, 347, 97, 347, 99, 347, 100, 347, 101, 347, 102, 347, 103, 347, 104, 347, 105, 347, 106, 347, 107, 347, 108, 347, 109, 347, 110, 347, 111, 347, 112, 347, 113, 347, 114, 347, 115, 347, 116, 347, 117, 347, 118, 347, 121, 347, 122, 347, 123, 347, 124, 347, 125, 347, 126, 347, 127, 347, 129, 347, 130, 347, 131, 347, 132, 347, 133, 347, 136, 347, 137, 347, 138, 347, 139, 347, 140, 347, 141, 347, 163, 347, 164, 347, 165, 347, 166, 347, 167, 347, 169, 347, 170, 347, 188, 347, 189, 347, 190, 347, 192, 347, 193, 347, 195, 347, 196, 347, 197, 347, 198, 347, 199, 347, 200, 347, 204, 347, 205, 347],
	#[65535, 359, 31, 359, 7, 359, 11, 359, 47, 359, 36, 359, 45, 359, 54, 359, 33, 359, 35, 359, 58, 359, 51, 359, 50, 359, 44, 359, 46, 359, 70, 359, 48, 359, 49, 359, 69, 359, 65, 359, 52, 359, 53, 359, 67, 359, 55, 359, 56, 359, 57, 359, 59, 359, 60, 359, 63, 359, 64, 359, 66, 359, 68, 359, 71, 359, 73, 359, 118, 359],
	#[65535, 200, 103, 200, 116, 200, 141, 200, 109, 200, 189, 200, 95, 200, 97, 200, 127, 200, 99, 200, 100, 200, 101, 200, 102, 200, 114, 200, 104, 200, 105, 200, 106, 200, 107, 200, 108, 200, 167, 200, 110, 200, 111, 200, 112, 200, 113, 200, 115, 200, 137, 200, 117, 200, 121, 200, 122, 200, 123, 200, 124, 200, 125, 200, 126, 200, 165, 200, 129, 200, 130, 200, 131, 200, 132, 200, 133, 200, 136, 200, 163, 200, 138, 200, 139, 200, 140, 200, 169, 200, 192, 200, 166, 200, 190, 200, 164, 200, 170, 200, 188, 200, 193, 200, 195, 200, 196, 200, 197, 200, 198, 200, 199, 200, 200, 200, 204, 200, 205, 200],
	#[65535, 203, 30, 203, 95, 203, 97, 203, 99, 203, 100, 203, 101, 203, 102, 203, 103, 203, 104, 203, 105, 203, 106, 203, 107, 203, 108, 203, 109, 203, 110, 203, 111, 203, 112, 203, 113, 203, 114, 203, 115, 203, 116, 203, 117, 203, 121, 203, 122, 203, 123, 203, 124, 203, 125, 203, 126, 203, 127, 203, 129, 203, 130, 203, 131, 203, 132, 203, 133, 203, 136, 203, 137, 203, 138, 203, 139, 203, 140, 203, 141, 203, 163, 203, 164, 203, 165, 203, 166, 203, 167, 203, 169, 203, 170, 203, 188, 203, 189, 203, 190, 203, 192, 203, 193, 203, 195, 203, 196, 203, 197, 203, 198, 203, 199, 203, 200, 203, 204, 203, 205, 203],
	#[65535, 278, 7, 278, 103, 278, 111, 278, 63, 278, 116, 278, 97, 278, 102, 278, 117, 278, 101, 278, 118, 278, 99, 278, 104, 278, 121, 278, 100, 278, 115, 278, 124, 278, 105, 278, 114, 278, 107, 278, 95, 278, 106, 278, 108, 278, 109, 278, 110, 278, 122, 278, 112, 278, 113, 278, 123, 278, 125, 278, 126, 278, 127, 278, 129, 278, 130, 278, 131, 278, 132, 278, 133, 278, 136, 278, 137, 278, 138, 278, 139, 278, 140, 278, 141, 278, 169, 278, 163, 278, 164, 278, 165, 278, 166, 278, 167, 278, 170, 278, 188, 278, 189, 278, 190, 278, 192, 278, 193, 278, 195, 278, 196, 278, 197, 278, 198, 278, 199, 278, 200, 278, 204, 278, 205, 278],
	#[65535, 198, 7, 198, 32, 198, 62, 198, 63, 198, 118, 198, 99, 198, 103, 198, 102, 198, 122, 198, 124, 198, 121, 198, 95, 198, 97, 198, 117, 198, 116, 198, 100, 198, 101, 198, 115, 198, 114, 198, 104, 198, 105, 198, 106, 198, 107, 198, 108, 198, 109, 198, 110, 198, 111, 198, 112, 198, 113, 198, 123, 198, 125, 198, 126, 198, 127, 198, 129, 198, 130, 198, 131, 198, 132, 198, 133, 198, 136, 198, 137, 198, 138, 198, 139, 198, 140, 198, 141, 198, 162, 198, 169, 198, 163, 198, 164, 198, 165, 198, 166, 198, 167, 198, 170, 198, 188, 198, 189, 198, 190, 198, 192, 198, 193, 198, 195, 198, 196, 198, 197, 198, 198, 198, 199, 198, 200, 198, 204, 198, 205, 198],
	#[7, 274, 62, -752, 63, 274, 95, 274, 97, 274, 99, 274, 100, 274, 101, 274, 102, 274, 103, 274, 104, 274, 105, 274, 106, 274, 107, 274, 108, 274, 109, 274, 110, 274, 111, 274, 112, 274, 113, 274, 114, 274, 115, 274, 116, 274, 117, 274, 118, 274, 121, 274, 122, 274, 123, 274, 124, 274, 125, 274, 126, 274, 127, 274, 129, 274, 130, 274, 131, 274, 132, 274, 133, 274, 136, 274, 137, 274, 138, 274, 139, 274, 140, 274, 141, 274, 162, -47, 163, 274, 164, 274, 165, 274, 166, 274, 167, 274, 169, 274, 170, 274, 188, 274, 189, 274, 190, 274, 192, 274, 193, 274, 195, 274, 196, 274, 197, 274, 198, 274, 199, 274, 200, 274, 204, 274, 205, 274],
	#[65535, 79, 0, 79, 1, 79, 4, 79, 7, 79, 9, 79, 10, 79, 14, 79, 15, 79, 16, 79, 17, 79, 18, 79, 19, 79, 20, 79, 21, 79, 22, 79, 23, 79, 24, 79, 25, 79, 26, 79, 27, 79, 28, 79, 29, 79, 41, 79, 40, 79, 76, 79, 63, 79, 82, 79, 74, 79, 75, 79, 77, 79, 83, 79, 119, 79, 118, 79, 225, 79, 226, 79],
	#[65535, 116, 3, 116, 7, 116, 11, 116, 31, 116, 33, 116, 35, 116, 36, 116, 44, 116, 45, 116, 46, 116, 47, 116, 48, 116, 49, 116, 50, 116, 51, 116, 52, 116, 53, 116, 54, 116, 55, 116, 56, 116, 57, 116, 58, 116, 59, 116, 60, 116, 63, 116, 64, 116, 65, 116, 66, 116, 67, 116, 68, 116, 69, 116, 70, 116, 71, 116, 72, 116, 73, 116, 77, 116, 81, 116, 82, 116, 83, 116, 118, 116],
	#[65535, 114, 3, 114, 7, 114, 11, 114, 47, 114, 36, 114, 31, 114, 33, 114, 35, 114, 58, 114, 71, 114, 44, 114, 45, 114, 46, 114, 48, 114, 49, 114, 50, 114, 51, 114, 52, 114, 53, 114, 54, 114, 55, 114, 56, 114, 57, 114, 59, 114, 60, 114, 63, 114, 64, 114, 65, 114, 66, 114, 67, 114, 68, 114, 69, 114, 70, 114, 82, 114, 72, 114, 73, 114, 77, 114, 81, 114, 83, 114, 118, 114],
	#[7, 59, 11, 59, 47, -565, 31, 59, 33, 59, 35, 59, 36, 59, 44, -558, 45, -561, 46, -563, 58, 59, 48, -567, 49, -555, 50, -559, 51, -560, 52, -562, 53, -564, 54, -566, 55, -554, 56, -556, 57, 59, 59, 59, 60, 59, 63, 59, 64, 59, 65, 59, 66, 59, 67, 59, 68, 59, 69, 59, 70, 59, 71, 59, 73, 59, 118, 59],
	#[65535, 115, 3, 115, 7, 115, 11, 115, 31, 115, 33, 115, 35, 115, 36, 115, 44, 115, 45, 115, 46, 115, 47, 115, 48, 115, 49, 115, 50, 115, 51, 115, 52, 115, 53, 115, 54, 115, 55, 115, 56, 115, 57, 115, 58, 115, 59, 115, 60, 115, 63, 115, 64, 115, 65, 115, 66, 115, 67, 115, 68, 115, 69, 115, 70, 115, 71, 115, 72, 115, 73, 115, 77, 115, 81, 115, 82, 115, 83, 115, 118, 115],
	#[65535, 52, 61, 52, 31, 52],
	#[65535, 73, 7, 73, 11, 73, 31, 73, 33, 73, 35, 73, 36, 73, 58, 73, 57, 73, 59, 73, 60, 73, 63, 73, 64, 73, 65, 73, 66, 73, 67, 73, 68, 73, 69, 73, 70, 73, 71, 73, 73, 73, 118, 73],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 57, -575, 58, -573, 59, -569, 60, -574, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -570, 118, -36],
	#[10, -762, 119, -33],
	#[65535, 57, 42, 57],
	#[65535, 58, 42, 58],
	#[42, -764],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 43, 205, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 98, -766, 118, -36],
	#[109, -213, 93, -211, 108, -208, 43, 324, 118, -36, 7, -13, 13, 324, 60, -209, 222, -212, 223, -214, 220, -207, 221, -210],
	#[65535, 208, 43, 208],
	#[43, -774],
	#[13, -769, 43, 206],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 98, -771, 118, -36],
	#[7, -13, 13, 324, 43, 324, 60, -209, 93, -211, 108, -208, 109, -213, 118, -36, 221, -210, 220, -207, 222, -212, 223, -214],
	#[65535, 207, 43, 207],
	#[65535, 325, 13, 325, 43, 325],
	#[65535, 210, 13, 210, 43, 210],
	#[0, 123, 1, 123, 4, 123, 7, 123, 9, 123, 10, 123, 14, 123, 15, 123, 16, 123, 17, 123, 18, 123, 19, 123, 20, 123, 21, 123, 22, 123, 23, 123, 24, 123, 25, 123, 26, 123, 27, 123, 28, 123, 29, 123, 31, 123, 40, 123, 41, 123, 58, 123, 93, 123, 57, 123, 59, 123, 61, 123, 63, 123, 82, 123, 74, 123, 75, 123, 76, 123, 77, 123, 81, -776, 83, 123, 119, 123, 94, 123, 118, 123, 120, 123, 179, 123, 171, 123, 172, 123, 173, 123, 174, 123, 175, 123, 176, 123, 177, 123, 178, 123, 180, 123, 181, 123, 182, 123, 183, 123, 184, 123, 185, 123, 186, 123, 187, 123, 225, 123, 226, 123],
	#[65535, 211, 0, 211, 1, 211, 4, 211, 7, 211, 9, 211, 10, 211, 14, 211, 15, 211, 16, 211, 17, 211, 18, 211, 19, 211, 20, 211, 21, 211, 22, 211, 23, 211, 24, 211, 25, 211, 26, 211, 27, 211, 28, 211, 29, 211, 31, 211, 40, 211, 41, 211, 57, 211, 58, 211, 59, 211, 61, 211, 63, 211, 74, 211, 75, 211, 76, 211, 77, 211, 82, 211, 83, 211, 93, 211, 94, 211, 118, 211, 119, 211, 120, 211, 171, 211, 172, 211, 173, 211, 174, 211, 175, 211, 176, 211, 177, 211, 178, 211, 179, 211, 180, 211, 181, 211, 182, 211, 183, 211, 184, 211, 185, 211, 186, 211, 187, 211, 225, 211, 226, 211],
	#[65535, 124, 0, 124, 1, 124, 4, 124, 7, 124, 9, 124, 10, 124, 14, 124, 15, 124, 16, 124, 17, 124, 18, 124, 19, 124, 20, 124, 21, 124, 22, 124, 23, 124, 24, 124, 25, 124, 26, 124, 27, 124, 28, 124, 29, 124, 31, 124, 58, 124, 40, 124, 41, 124, 93, 124, 57, 124, 59, 124, 61, 124, 63, 124, 82, 124, 72, 124, 74, 124, 75, 124, 76, 124, 77, 124, 83, 124, 119, 124, 94, 124, 118, 124, 120, 124, 179, 124, 171, 124, 172, 124, 173, 124, 174, 124, 175, 124, 176, 124, 177, 124, 178, 124, 180, 124, 181, 124, 182, 124, 183, 124, 184, 124, 185, 124, 186, 124, 187, 124, 225, 124, 226, 124],
	#[0, 213, 1, 213, 4, 213, 7, 213, 9, 213, 10, 213, 14, 213, 15, 213, 16, 213, 17, 213, 18, 213, 19, 213, 20, 213, 21, 213, 22, 213, 23, 213, 24, 213, 25, 213, 26, 213, 27, 213, 28, 213, 29, 213, 31, 213, 40, 213, 41, 213, 57, -192, 58, -175, 59, -179, 61, 213, 63, 213, 74, 213, 75, 213, 76, 213, 77, 213, 82, 213, 83, 213, 93, 213, 94, -780, 118, 213, 119, 213, 120, 213, 171, -180, 172, -191, 173, -182, 174, -184, 175, -178, 176, -185, 177, -193, 178, -176, 179, -181, 180, -183, 181, -186, 182, -189, 183, -174, 184, -177, 185, -187, 186, -190, 187, -173, 225, 213, 226, 213],
	#[65535, 214, 76, 214, 7, 214, 61, 214, 14, 214, 23, 214, 10, 214, 0, 214, 1, 214, 31, 214, 20, 214, 4, 214, 63, 214, 19, 214, 18, 214, 74, 214, 9, 214, 29, 214, 28, 214, 27, 214, 15, 214, 16, 214, 17, 214, 25, 214, 24, 214, 40, 214, 21, 214, 22, 214, 75, 214, 26, 214, 41, 214, 83, 214, 119, 214, 77, 214, 82, 214, 93, 214, 118, 214, 120, 214, 226, 214, 225, 214],
	#[0, 215, 1, 215, 4, 215, 7, 215, 9, 215, 10, 215, 14, 215, 15, 215, 16, 215, 17, 215, 18, 215, 19, 215, 20, 215, 21, 215, 22, 215, 23, 215, 24, 215, 25, 215, 26, 215, 27, 215, 28, 215, 29, 215, 31, 215, 40, 215, 41, 215, 74, 215, 61, 215, 63, 215, 82, 215, 93, -783, 75, 215, 76, 215, 77, 215, 83, 215, 119, 215, 118, 215, 120, 215, 225, 215, 226, 215],
	#[7, -781],
	#[65535, 151, 27, 151, 0, 151, 1, 151, 4, 151, 19, 151, 7, 151, 9, 151, 10, 151, 28, 151, 13, 151, 14, 151, 15, 151, 16, 151, 17, 151, 18, 151, 24, 151, 20, 151, 21, 151, 22, 151, 23, 151, 25, 151, 26, 151, 41, 151, 29, 151, 31, 151, 63, 151, 40, 151, 83, 151, 77, 151, 76, 151, 61, 151, 75, 151, 74, 151, 82, 151, 93, 151, 118, 151, 119, 151, 120, 151, 225, 151, 226, 151],
	#[0, 217, 1, 217, 4, 217, 7, 217, 9, 217, 10, 217, 14, 217, 15, 217, 16, 217, 17, 217, 18, 217, 19, 217, 20, 217, 21, 217, 22, 217, 23, 217, 24, 217, 25, 217, 26, 217, 27, 217, 28, 217, 29, 217, 31, 217, 40, 217, 41, 217, 61, 217, 63, 217, 82, 217, 74, 217, 75, 217, 76, 217, 77, 217, 83, 217, 119, 217, 118, 217, 120, -785, 225, 217, 226, 217],
	#[38, -784],
	#[65535, 216, 0, 216, 1, 216, 4, 216, 7, 216, 9, 216, 10, 216, 14, 216, 15, 216, 16, 216, 17, 216, 18, 216, 19, 216, 20, 216, 21, 216, 22, 216, 23, 216, 24, 216, 25, 216, 26, 216, 27, 216, 28, 216, 29, 216, 31, 216, 41, 216, 40, 216, 76, 216, 61, 216, 63, 216, 74, 216, 75, 216, 77, 216, 82, 216, 83, 216, 118, 216, 119, 216, 120, 216, 225, 216, 226, 216],
	#[7, -787],
	#[65535, 54, 7, 54, 14, 54, 23, 54, 10, 54, 0, 54, 1, 54, 9, 54, 20, 54, 4, 54, 63, 54, 19, 54, 18, 54, 24, 54, 29, 54, 28, 54, 27, 54, 15, 54, 16, 54, 17, 54, 25, 54, 74, 54, 40, 54, 21, 54, 22, 54, 61, 54, 26, 54, 41, 54, 31, 54, 75, 54, 77, 54, 76, 54, 83, 54, 82, 54, 118, 54, 119, 54, 226, 54, 225, 54],
	#[65535, 218, 0, 218, 1, 218, 4, 218, 7, 218, 9, 218, 10, 218, 14, 218, 15, 218, 16, 218, 17, 218, 18, 218, 19, 218, 20, 218, 21, 218, 22, 218, 23, 218, 24, 218, 25, 218, 26, 218, 27, 218, 28, 218, 29, 218, 31, 218, 40, 218, 41, 218, 61, 218, 63, 218, 82, 218, 74, 218, 75, 218, 76, 218, 77, 218, 83, 218, 119, 218, 118, 218, 225, 218, 226, 218],
	#[65535, 209, 13, 209, 43, 209],
	#[11, -791],
	#[65535, 11, 0, 11, 1, 11, 4, 11, 7, 11, 9, 11, 10, 11, 14, 11, 15, 11, 16, 11, 17, 11, 18, 11, 19, 11, 20, 11, 21, 11, 22, 11, 23, 11, 24, 11, 25, 11, 26, 11, 27, 11, 28, 11, 29, 11, 40, 11, 41, 11, 63, 11, 82, 11, 74, 11, 75, 11, 76, 11, 77, 11, 83, 11, 119, 11, 118, 11, 225, 11, 226, 11],
	#[7, -792, 12, 16, 13, 16],
	#[65535, 17, 12, 17, 13, 17],
	#[12, -795, 13, -794],
	#[7, -796],
	#[65535, 15, 0, 15, 1, 15, 4, 15, 63, 15, 7, 15, 9, 15, 10, 15, 14, 15, 15, 15, 16, 15, 17, 15, 18, 15, 19, 15, 20, 15, 21, 15, 22, 15, 23, 15, 24, 15, 25, 15, 26, 15, 27, 15, 28, 15, 29, 15, 41, 15, 75, 15, 77, 15, 40, 15, 74, 15, 82, 15, 83, 15, 76, 15, 118, 15, 119, 15, 225, 15, 226, 15],
	#[65535, 18, 12, 18, 13, 18],
	#[226, -798],
	#[31, -799],
	#[32, -800, 226, -803],
	#[65535, 425, 0, 425, 1, 425, 4, 425, 7, 425, 9, 425, 10, 425, 14, 425, 15, 425, 16, 425, 17, 425, 18, 425, 19, 425, 20, 425, 21, 425, 22, 425, 23, 425, 24, 425, 25, 425, 26, 425, 27, 425, 28, 425, 29, 425, 77, 425, 40, 425, 41, 425, 63, 425, 83, 425, 82, 425, 74, 425, 75, 425, 76, 425, 119, 425, 118, 425, 225, 425, 226, 425],
	#[13, -805, 32, -804],
	#[65535, 429, 13, 429, 32, 429],
	#[38, -319],
	#[65535, 426, 0, 426, 1, 426, 4, 426, 7, 426, 9, 426, 10, 426, 14, 426, 15, 426, 16, 426, 17, 426, 18, 426, 19, 426, 20, 426, 21, 426, 22, 426, 23, 426, 24, 426, 25, 426, 26, 426, 27, 426, 28, 426, 29, 426, 40, 426, 41, 426, 63, 426, 74, 426, 75, 426, 76, 426, 77, 426, 82, 426, 83, 426, 118, 426, 119, 426, 225, 426, 226, 426],
	#[226, -803],
	#[65535, 430, 13, 430, 32, 430],
	#[72, 113, 74, -754, 75, -756, 76, -753, 77, 113, 81, 113, 82, 113, 83, 113],
	#[3, 113, 82, 113, 74, -754, 75, -756, 77, 113, 83, 113, 72, 113, 76, -753, 81, 113],
	#[3, -810, 72, 117, 77, -812, 81, 117, 82, 117, 83, 117],
	#[19, -38, 7, 19, 11, 19, 14, -9, 15, -30, 16, -14, 17, -29, 18, -15, 36, 19, 20, -23, 21, -10, 22, -31, 23, -16, 24, -3, 25, -26, 26, -28, 31, 19, 33, 19, 35, 19, 70, 19, 69, 19, 67, 19, 65, 19, 63, 19, 64, 19, 66, 19, 68, 19, 71, 19, 73, 19, 97, 19, 95, 19, 118, 19],
	#[82, 105, 72, -161, 81, 105, 83, 105],
	#[42, -813, 72, 118, 81, 118, 82, 118, 83, 118],
	#[79, -817, 78, -816, 80, -814],
	#[65535, 122, 43, 122],
	#[43, -818],
	#[65535, 120, 43, 120],
	#[65535, 121, 43, 121],
	#[65535, 119, 72, 119, 81, 119, 82, 119, 83, 119],
	#[82, 123, 83, 123, 81, -776],
	#[82, -821, 83, -822],
	#[65535, 125, 63, 125, 7, 125, 11, 125, 68, 125, 36, 125, 35, 125, 70, 125, 33, 125, 31, 125, 66, 125, 64, 125, 65, 125, 67, 125, 69, 125, 71, 125, 73, 125, 118, 125],
	#[65535, 126, 7, 126, 11, 126, 31, 126, 33, 126, 35, 126, 36, 126, 63, 126, 64, 126, 65, 126, 66, 126, 67, 126, 68, 126, 69, 126, 70, 126, 71, 126, 73, 126, 118, 126],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[0, 147, 1, 147, 4, 147, 7, 147, 9, 147, 10, 147, 13, -825, 14, 147, 15, 147, 16, 147, 17, 147, 18, 147, 19, 147, 20, 147, 21, 147, 22, 147, 23, 147, 24, 147, 25, 147, 26, 147, 27, 147, 28, 147, 29, 147, 41, 147, 40, 147, 76, 147, 63, 147, 74, 147, 75, 147, 77, 147, 82, 147, 83, 147, 118, 147, 119, 147, 225, 147, 226, 147],
	#[93, -828, 94, -780],
	#[65535, 4, 24, 4, 0, 4, 1, 4, 4, 4, 7, 4, 9, 4, 10, 4, 14, 4, 15, 4, 16, 4, 17, 4, 18, 4, 19, 4, 20, 4, 21, 4, 22, 4, 23, 4, 25, 4, 26, 4, 27, 4, 28, 4, 29, 4, 40, 4, 41, 4, 76, 4, 63, 4, 74, 4, 75, 4, 77, 4, 82, 4, 83, 4, 119, 4, 118, 4, 225, 4, 226, 4],
	#[65535, 149, 0, 149, 1, 149, 4, 149, 7, 149, 9, 149, 10, 149, 13, 149, 14, 149, 15, 149, 16, 149, 17, 149, 18, 149, 19, 149, 20, 149, 21, 149, 22, 149, 23, 149, 24, 149, 25, 149, 26, 149, 27, 149, 28, 149, 29, 149, 40, 149, 41, 149, 63, 149, 74, 149, 75, 149, 76, 149, 77, 149, 82, 149, 83, 149, 118, 149, 119, 149, 225, 149, 226, 149],
	#[38, -831],
	#[0, 147, 1, 147, 4, 147, 7, 147, 9, 147, 10, 147, 13, -825, 14, 147, 15, 147, 16, 147, 17, 147, 18, 147, 19, 147, 20, 147, 21, 147, 22, 147, 23, 147, 24, 147, 25, 147, 26, 147, 27, 147, 28, 147, 29, 147, 41, 147, 40, 147, 76, 147, 63, 147, 82, 147, 74, 147, 75, 147, 77, 147, 83, 147, 118, 147, 119, 147, 225, 147, 226, 147],
	#[65535, 148, 0, 148, 1, 148, 4, 148, 7, 148, 9, 148, 10, 148, 14, 148, 15, 148, 16, 148, 17, 148, 18, 148, 19, 148, 20, 148, 21, 148, 22, 148, 23, 148, 24, 148, 25, 148, 26, 148, 27, 148, 28, 148, 29, 148, 40, 148, 41, 148, 63, 148, 74, 148, 75, 148, 76, 148, 77, 148, 82, 148, 83, 148, 118, 148, 119, 148, 225, 148, 226, 148],
	#[65535, 150, 0, 150, 1, 150, 4, 150, 7, 150, 9, 150, 10, 150, 13, 150, 14, 150, 15, 150, 16, 150, 17, 150, 18, 150, 19, 150, 20, 150, 21, 150, 22, 150, 23, 150, 24, 150, 25, 150, 26, 150, 27, 150, 28, 150, 29, 150, 40, 150, 41, 150, 63, 150, 74, 150, 75, 150, 76, 150, 77, 150, 82, 150, 83, 150, 118, 150, 119, 150, 225, 150, 226, 150],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 70, -140, 69, -136, 67, -137, 65, -122, 63, -121, 64, -141, 66, -126, 68, -133, 71, -130, 73, -570, 97, -834, 95, -833, 118, -36],
	#[42, -843],
	#[42, 421, 224, -362],
	#[7, -13, 10, -223, 63, -232, 118, -36, 119, -33],
	#[65535, 8, 0, 8, 1, 8, 4, 8, 7, 8, 9, 8, 10, 8, 14, 8, 15, 8, 16, 8, 17, 8, 18, 8, 19, 8, 20, 8, 21, 8, 22, 8, 23, 8, 24, 8, 25, 8, 26, 8, 27, 8, 28, 8, 29, 8, 40, 8, 41, 8, 63, 8, 74, 8, 75, 8, 76, 8, 77, 8, 82, 8, 83, 8, 118, 8, 119, 8, 225, 8, 226, 8],
	#[65535, 152, 15, 152, 118, 152, 22, 152, 76, 152, 74, 152, 14, 152, 0, 152, 29, 152, 119, 152, 27, 152, 63, 152, 28, 152, 7, 152, 1, 152, 96, 152, 77, 152, 21, 152, 75, 152, 18, 152, 20, 152, 4, 152, 10, 152, 17, 152, 25, 152, 16, 152, 83, 152, 19, 152, 23, 152, 40, 152, 41, 152, 9, 152, 82, 152, 24, 152, 26, 152, 226, 152, 225, 152],
	#[42, -839],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[65535, 219, 13, 219, 43, 219],
	#[13, -367, 43, -842],
	#[65535, 154, 29, 154, 7, 154, 15, 154, 1, 154, 14, 154, 23, 154, 22, 154, 0, 154, 77, 154, 20, 154, 4, 154, 63, 154, 19, 154, 18, 154, 9, 154, 10, 154, 27, 154, 26, 154, 16, 154, 17, 154, 25, 154, 74, 154, 21, 154, 75, 154, 24, 154, 76, 154, 41, 154, 28, 154, 119, 154, 40, 154, 82, 154, 118, 154, 83, 154, 96, 154, 226, 154, 225, 154],
	#[63, -121, 7, -13, 11, -127, 33, -139, 31, -128, 66, -126, 68, -133, 35, -123, 36, -129, 71, -130, 65, -122, 73, -570, 70, -140, 69, -136, 67, -137, 64, -141, 95, -833, 97, -834, 118, -36],
	#[96, -845],
	#[7, -13, 11, -127, 71, -130, 68, -133, 31, -128, 33, -139, 63, -121, 35, -123, 36, -129, 65, -122, 73, -138, 66, -126, 70, -140, 69, -136, 67, -137, 64, -141, 118, -36],
	#[37, 105, 42, -159, 43, -847, 72, -161],
	#[65535, 153, 0, 153, 1, 153, 4, 153, 7, 153, 9, 153, 10, 153, 14, 153, 15, 153, 16, 153, 17, 153, 18, 153, 19, 153, 20, 153, 21, 153, 22, 153, 23, 153, 24, 153, 25, 153, 26, 153, 27, 153, 28, 153, 29, 153, 40, 153, 41, 153, 63, 153, 74, 153, 75, 153, 76, 153, 77, 153, 82, 153, 83, 153, 96, 153, 118, 153, 119, 153, 225, 153, 226, 153],
	#[77, -812, 82, 117, 81, 117, 83, 117, 72, 117],
	#[72, 123, 81, -776, 82, 123, 83, 123],
	#[83, 105, 82, 105, 72, -161],
	#[82, -821, 83, -822],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[4, 147, 15, 147, 118, 147, 23, 147, 63, 147, 7, 147, 29, 147, 42, -159, 10, 147, 13, -825, 41, 147, 28, 147, 37, 105, 27, 147, 77, 147, 21, 147, 40, 147, 16, 147, 19, 147, 14, 147, 82, 147, 226, 147, 119, 147, 74, 147, 22, 147, 72, -161, 1, 147, 17, 147, 25, 147, 75, 147, 18, 147, 20, 147, 0, 147, 76, 147, 83, 147, 26, 147, 24, 147, 9, 147, 225, 147],
	#[65535, 6, 0, 6, 1, 6, 4, 6, 7, 6, 9, 6, 10, 6, 14, 6, 15, 6, 16, 6, 17, 6, 18, 6, 19, 6, 20, 6, 21, 6, 22, 6, 23, 6, 24, 6, 25, 6, 26, 6, 27, 6, 28, 6, 29, 6, 40, 6, 41, 6, 63, 6, 74, 6, 75, 6, 76, 6, 77, 6, 82, 6, 83, 6, 118, 6, 119, 6, 225, 6, 226, 6],
	#[65535, 14, 3, 14, 14, 14, 15, 14, 16, 14, 17, 14, 18, 14, 19, 14, 20, 14, 21, 14, 22, 14, 23, 14, 24, 14, 25, 14, 26, 14, 27, 14, 28, 14, 29, 14, 72, 14, 74, 14, 75, 14, 76, 14, 77, 14, 81, 14, 82, 14, 83, 14],
	#[7, -858],
	#[65535, 3, 0, 3, 1, 3, 4, 3, 7, 3, 9, 3, 10, 3, 14, 3, 15, 3, 16, 3, 17, 3, 18, 3, 19, 3, 20, 3, 21, 3, 22, 3, 23, 3, 24, 3, 25, 3, 26, 3, 27, 3, 28, 3, 29, 3, 40, 3, 41, 3, 63, 3, 74, 3, 75, 3, 76, 3, 77, 3, 82, 3, 83, 3, 118, 3, 119, 3, 225, 3, 226, 3],
	#[65535, 12, 29, 12, 0, 12, 1, 12, 4, 12, 63, 12, 19, 12, 7, 12, 9, 12, 10, 12, 14, 12, 15, 12, 16, 12, 17, 12, 18, 12, 74, 12, 20, 12, 21, 12, 22, 12, 23, 12, 24, 12, 25, 12, 26, 12, 27, 12, 28, 12, 41, 12, 75, 12, 82, 12, 77, 12, 40, 12, 83, 12, 76, 12, 118, 12, 119, 12, 226, 12, 225, 12],
	#[6, -860],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[37, 105, 42, -159, 72, -161, 226, -862],
	#[31, -863],
	#[7, -13, 11, -127, 31, -128, 32, -864, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 85, -323, 118, -36],
	#[65535, 427, 14, 427, 29, 427, 0, 427, 1, 427, 4, 427, 63, 427, 7, 427, 9, 427, 10, 427, 27, 427, 15, 427, 16, 427, 17, 427, 18, 427, 19, 427, 20, 427, 21, 427, 22, 427, 23, 427, 24, 427, 25, 427, 26, 427, 82, 427, 28, 427, 74, 427, 41, 427, 75, 427, 77, 427, 40, 427, 83, 427, 76, 427, 118, 427, 119, 427, 226, 427, 225, 427],
	#[13, -325, 32, -866],
	#[65535, 428, 0, 428, 1, 428, 4, 428, 19, 428, 7, 428, 9, 428, 10, 428, 28, 428, 14, 428, 15, 428, 16, 428, 17, 428, 18, 428, 24, 428, 20, 428, 21, 428, 22, 428, 23, 428, 25, 428, 26, 428, 27, 428, 29, 428, 41, 428, 40, 428, 83, 428, 63, 428, 82, 428, 74, 428, 75, 428, 76, 428, 77, 428, 118, 428, 119, 428, 225, 428, 226, 428],
	#[77, -812, 82, 117, 83, 117],
	#[82, -821, 83, -822],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 71, -130, 64, -141, 63, -121, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 73, -138, 118, -36],
	#[0, 147, 1, 147, 4, 147, 7, 147, 9, 147, 10, 147, 13, -825, 14, 147, 15, 147, 16, 147, 17, 147, 18, 147, 19, 147, 20, 147, 21, 147, 22, 147, 23, 147, 24, 147, 25, 147, 26, 147, 27, 147, 28, 147, 29, 147, 40, 147, 41, 147, 63, 147, 74, 147, 75, 147, 76, 147, 77, 147, 82, 147, 83, 147, 118, 147, 119, 147, 225, 147, 226, 147],
	#[65535, 5, 0, 5, 1, 5, 4, 5, 7, 5, 9, 5, 10, 5, 14, 5, 15, 5, 16, 5, 17, 5, 18, 5, 19, 5, 20, 5, 21, 5, 22, 5, 23, 5, 24, 5, 25, 5, 26, 5, 27, 5, 28, 5, 29, 5, 40, 5, 41, 5, 63, 5, 74, 5, 75, 5, 76, 5, 77, 5, 82, 5, 83, 5, 118, 5, 119, 5, 225, 5, 226, 5],
	#[65535, 56, 31, 56, 47, 56, 55, 56, 11, 56, 36, 56, 63, 56, 7, 56, 54, 56, 66, 56, 58, 56, 53, 56, 68, 56, 52, 56, 33, 56, 35, 56, 46, 56, 51, 56, 60, 56, 44, 56, 48, 56, 57, 56, 74, 56, 56, 56, 59, 56, 75, 56, 50, 56, 49, 56, 45, 56, 64, 56, 65, 56, 67, 56, 69, 56, 70, 56, 71, 56, 73, 56, 76, 56, 118, 56],
	#[65535, 50, 0, 50, 1, 50, 4, 50, 7, 50, 9, 50, 10, 50, 14, 50, 15, 50, 16, 50, 17, 50, 18, 50, 19, 50, 20, 50, 21, 50, 22, 50, 23, 50, 24, 50, 25, 50, 26, 50, 27, 50, 28, 50, 29, 50, 40, 50, 41, 50, 63, 50, 74, 50, 75, 50, 76, 50, 77, 50, 82, 50, 83, 50, 118, 50, 119, 50, 225, 50, 226, 50],
	#[65535, 55, 31, 55, 60, 55, 55, 55, 11, 55, 52, 55, 36, 55, 63, 55, 7, 55, 35, 55, 75, 55, 66, 55, 53, 55, 68, 55, 47, 55, 33, 55, 57, 55, 46, 55, 51, 55, 56, 55, 44, 55, 48, 55, 74, 55, 54, 55, 59, 55, 58, 55, 50, 55, 49, 55, 45, 55, 64, 55, 65, 55, 67, 55, 69, 55, 70, 55, 71, 55, 73, 55, 76, 55, 118, 55],
	#[46, 113, 63, 113, 7, 113, 54, 113, 35, 113, 11, 113, 69, 113, 71, 113, 53, 113, 47, 113, 66, 113, 75, -756, 70, 113, 55, 113, 44, 113, 49, 113, 31, 113, 33, 113, 68, 113, 52, 113, 36, 113, 51, 113, 74, -754, 76, -753, 60, 113, 65, 113, 45, 113, 59, 113, 58, 113, 48, 113, 64, 113, 50, 113, 67, 113, 56, 113, 57, 113, 73, 113, 118, 113],
	#[7, 59, 11, 59, 31, 59, 33, 59, 35, 59, 36, 59, 44, -558, 45, -561, 46, -563, 47, -565, 48, -567, 49, -555, 50, -559, 51, -560, 52, -562, 53, -564, 54, -566, 55, -554, 56, -556, 57, 59, 58, 59, 59, 59, 60, 59, 63, 59, 64, 59, 65, 59, 66, 59, 67, 59, 68, 59, 69, 59, 70, 59, 71, 59, 73, 59, 118, 59],
	#[65535, 53, 0, 53, 1, 53, 4, 53, 7, 53, 9, 53, 10, 53, 14, 53, 15, 53, 16, 53, 17, 53, 18, 53, 19, 53, 20, 53, 21, 53, 22, 53, 23, 53, 24, 53, 25, 53, 26, 53, 27, 53, 28, 53, 29, 53, 41, 53, 40, 53, 76, 53, 63, 53, 74, 53, 75, 53, 77, 53, 82, 53, 83, 53, 119, 53, 118, 53, 225, 53, 226, 53],
	#[6, -882],
	#[6, -880],
	#[7, -881],
	#[65535, 9, 0, 9, 1, 9, 4, 9, 7, 9, 9, 9, 10, 9, 14, 9, 15, 9, 16, 9, 17, 9, 18, 9, 19, 9, 20, 9, 21, 9, 22, 9, 23, 9, 24, 9, 25, 9, 26, 9, 27, 9, 28, 9, 29, 9, 40, 9, 41, 9, 63, 9, 74, 9, 75, 9, 76, 9, 77, 9, 82, 9, 83, 9, 118, 9, 119, 9, 225, 9, 226, 9],
	#[7, -883],
	#[65535, 10, 0, 10, 1, 10, 4, 10, 7, 10, 9, 10, 10, 10, 14, 10, 15, 10, 16, 10, 17, 10, 18, 10, 19, 10, 20, 10, 21, 10, 22, 10, 23, 10, 24, 10, 25, 10, 26, 10, 27, 10, 28, 10, 29, 10, 40, 10, 41, 10, 63, 10, 74, 10, 75, 10, 76, 10, 77, 10, 82, 10, 83, 10, 118, 10, 119, 10, 225, 10, 226, 10],
	#[77, -812, 82, 117, 83, 117],
	#[82, -821, 83, -822],
	#[7, -13, 11, -127, 31, -128, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[0, 147, 1, 147, 4, 147, 19, 147, 7, 147, 9, 147, 10, 147, 28, 147, 13, -825, 14, 147, 15, 147, 16, 147, 17, 147, 18, 147, 24, 147, 20, 147, 21, 147, 22, 147, 23, 147, 25, 147, 26, 147, 27, 147, 29, 147, 42, -159, 41, 147, 37, 105, 77, 147, 40, 147, 63, 147, 83, 147, 82, 147, 72, -161, 74, 147, 75, 147, 76, 147, 118, 147, 119, 147, 225, 147, 226, 147],
	#[65535, 7, 0, 7, 1, 7, 4, 7, 7, 7, 9, 7, 10, 7, 14, 7, 15, 7, 16, 7, 17, 7, 18, 7, 19, 7, 20, 7, 21, 7, 22, 7, 23, 7, 24, 7, 25, 7, 26, 7, 27, 7, 28, 7, 29, 7, 40, 7, 41, 7, 63, 7, 82, 7, 74, 7, 75, 7, 76, 7, 77, 7, 83, 7, 118, 7, 119, 7, 225, 7, 226, 7],
	#[7, -13, 11, -893, 31, -892, 33, -897, 35, -896, 36, -894, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[7, 42, 14, 42, 17, 42, 10, 42, 0, 42, 1, 42, 9, 42, 20, 42, 4, 42, 19, 42, 18, 42, 42, 87, 29, 42, 28, 42, 27, 42, 15, 42, 16, 42, 37, 87, 24, 42, 77, 42, 21, 42, 22, 42, 23, 42, 25, 42, 26, 42, 41, 42, 72, 87, 63, 42, 75, 42, 40, 42, 74, 42, 82, 42, 83, 42, 76, 42, 118, 42, 119, 42, 226, 42, 225, 42],
	#[37, 105, 42, -159, 72, -161],
	#[7, -13, 11, -127, 31, -128, 32, -913, 33, -139, 35, -123, 36, -129, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 71, -130, 73, -138, 118, -36],
	#[38, -908],
	#[0, 41, 1, 41, 4, 41, 7, 41, 9, 41, 10, 41, 14, 41, 15, 41, 16, 41, 17, 41, 18, 41, 19, 41, 20, 41, 21, 41, 22, 41, 23, 41, 24, 41, 25, 41, 26, 41, 27, 41, 28, 41, 29, 41, 37, 86, 40, 41, 41, 41, 42, 86, 63, 41, 72, 86, 74, 41, 75, 41, 76, 41, 77, 41, 82, 41, 83, 41, 118, 41, 119, 41, 225, 41, 226, 41],
	#[77, 46, 0, 46, 1, 46, 4, 46, 63, 46, 19, 46, 7, 46, 9, 46, 10, 46, 28, 46, 14, 46, 15, 46, 16, 46, 17, 46, 18, 46, 24, 46, 20, 46, 21, 46, 22, 46, 23, 46, 25, 46, 26, 46, 27, 46, 29, 46, 74, 46, 42, 84, 41, 46, 75, 46, 37, 84, 40, 46, 83, 46, 72, 84, 82, 46, 76, 46, 118, 46, 119, 46, 225, 46, 226, 46],
	#[18, 40, 24, 40, 0, 40, 1, 40, 4, 40, 7, 40, 9, 40, 10, 40, 28, 40, 14, 40, 15, 40, 16, 40, 17, 40, 19, 40, 20, 40, 21, 40, 22, 40, 23, 40, 25, 40, 26, 40, 27, 40, 29, 40, 63, 40, 37, 85, 40, 40, 41, 40, 42, 85, 72, 85, 83, 40, 82, 40, 74, 40, 75, 40, 76, 40, 77, 40, 118, 40, 119, 40, 225, 40, 226, 40],
	#[31, -899, 38, -898],
	#[39, -905],
	#[7, -13, 11, -127, 31, -128, 32, -901, 33, -139, 35, -123, 36, -129, 71, -130, 63, -121, 64, -141, 65, -122, 66, -126, 67, -137, 68, -133, 69, -136, 70, -140, 73, -138, 118, -36],
	#[13, -149, 32, -903],
	#[34, -902],
	#[118, 39, 20, 39, 42, 96, 15, 39, 9, 39, 37, 96, 72, 96, 21, 39, 23, 39, 10, 39, 19, 39, 14, 39, 29, 39, 119, 39, 22, 39, 28, 39, 7, 39, 83, 39, 16, 39, 77, 39, 82, 39, 75, 39, 40, 39, 74, 39, 4, 39, 27, 39, 17, 39, 25, 39, 76, 39, 1, 39, 0, 39, 18, 39, 24, 39, 26, 39, 41, 39, 63, 39, 226, 39, 225, 39],
	#[34, -904],
	#[29, 38, 1, 38, 0, 38, 77, 38, 119, 38, 4, 38, 63, 38, 7, 38, 9, 38, 10, 38, 28, 38, 72, 95, 14, 38, 15, 38, 16, 38, 17, 38, 18, 38, 19, 38, 20, 38, 21, 38, 22, 38, 23, 38, 24, 38, 25, 38, 26, 38, 27, 38, 74, 38, 42, 95, 41, 38, 37, 95, 40, 38, 83, 38, 75, 38, 76, 38, 82, 38, 118, 38, 226, 38, 225, 38],
	#[68, -133, 66, -126, 36, -129, 7, -13, 11, -127, 67, -137, 31, -128, 33, -139, 63, -121, 35, -123, 65, -122, 73, -138, 71, -130, 70, -140, 69, -136, 64, -141, 118, -36],
	#[37, 105, 34, -907, 42, -159, 72, -161],
	#[24, 45, 0, 45, 1, 45, 4, 45, 7, 45, 9, 45, 10, 45, 14, 45, 15, 45, 16, 45, 17, 45, 18, 45, 19, 45, 20, 45, 21, 45, 22, 45, 23, 45, 25, 45, 26, 45, 27, 45, 28, 45, 29, 45, 37, 92, 40, 45, 41, 45, 42, 92, 63, 45, 83, 45, 72, 92, 74, 45, 75, 45, 76, 45, 77, 45, 82, 45, 119, 45, 118, 45, 225, 45, 226, 45],
	#[39, -909],
	#[68, -133, 7, -13, 11, -127, 36, -129, 70, -140, 118, -36, 67, -137, 31, -128, 33, -139, 63, -121, 35, -123, 71, -130, 66, -126, 64, -141, 69, -136, 65, -122, 73, -138],
	#[12, -911, 37, 105, 42, -159, 72, -161],
	#[0, 44, 1, 44, 4, 44, 7, 44, 9, 44, 10, 44, 14, 44, 15, 44, 16, 44, 17, 44, 18, 44, 19, 44, 20, 44, 21, 44, 22, 44, 23, 44, 24, 44, 25, 44, 26, 44, 27, 44, 28, 44, 29, 44, 37, 91, 40, 44, 41, 44, 42, 91, 75, 44, 63, 44, 72, 91, 74, 44, 76, 44, 77, 44, 82, 44, 83, 44, 119, 44, 118, 44, 225, 44, 226, 44],
	#[13, -149, 32, -914],
	#[0, 37, 1, 37, 4, 37, 19, 37, 7, 37, 9, 37, 10, 37, 14, 37, 15, 37, 16, 37, 17, 37, 18, 37, 24, 37, 20, 37, 21, 37, 22, 37, 23, 37, 25, 37, 26, 37, 27, 37, 28, 37, 29, 37, 41, 37, 37, 94, 40, 37, 42, 94, 119, 37, 63, 37, 72, 94, 74, 37, 75, 37, 76, 37, 77, 37, 82, 37, 83, 37, 118, 37, 226, 37, 225, 37],
	#[76, 36, 77, 36, 63, 36, 22, 36, 118, 36, 37, 93, 72, 93, 18, 36, 74, 36, 23, 36, 10, 36, 19, 36, 14, 36, 20, 36, 75, 36, 82, 36, 1, 36, 119, 36, 27, 36, 7, 36, 26, 36, 21, 36, 15, 36, 4, 36, 24, 36, 0, 36, 29, 36, 83, 36, 28, 36, 9, 36, 17, 36, 25, 36, 42, 93, 40, 36, 16, 36, 41, 36, 226, 36, 225, 36],
	#[37, -916],
	#[24, 43, 0, 43, 1, 43, 4, 43, 7, 43, 9, 43, 10, 43, 28, 43, 14, 43, 15, 43, 16, 43, 17, 43, 18, 43, 19, 43, 20, 43, 21, 43, 22, 43, 23, 43, 25, 43, 26, 43, 27, 43, 29, 43, 41, 43, 37, 88, 40, 43, 42, 88, 63, 43, 83, 43, 82, 43, 72, 88, 74, 43, 75, 43, 76, 43, 77, 43, 118, 43, 119, 43, 226, 43, 225, 43],
	#[65535, 47, 30, 47],
	#[65535, 13, 3, 13, 19, 13, 14, 13, 15, 13, 16, 13, 17, 13, 18, 13, 24, 13, 20, 13, 21, 13, 22, 13, 23, 13, 25, 13, 26, 13, 27, 13, 28, 13, 29, 13, 83, 13, 75, 13, 72, 13, 74, 13, 76, 13, 77, 13, 81, 13, 82, 13]],
  goto-table:
      #[#[228, 1],
	#[277, 3, 276, 36, 275, 4, 236, 6, 234, 19, 229, 23, 226, 20, 225, 31, 235, 10, 231, 26, 119, 32, 118, 35, 63, 7, 41, 18, 40, 38, 22, 30, 21, 9, 20, 22, 24, 2, 17, 28, 16, 13, 26, 27, 27, 33, 28, 17, 9, 34, 19, 37, 25, 25, 1, 21, 0, 16, 10, 24, 23, 15, 4, 11, 15, 29, 18, 14, 7, 12, 29, 5, 14, 8],
	#[],
	#[6, 917],
	#[6, 916],
	#[],
	#[30, 888],
	#[6, 747],
	#[],
	#[],
	#[254, 883, 75, 755, 74, 753, 76, 752],
	#[8, 877, 5, 878],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[239, 872, 234, 873, 241, 874, 235, 871, 29, 5, 26, 27, 24, 2, 22, 30, 21, 9, 20, 22, 19, 37, 18, 14, 17, 28, 16, 13, 15, 29, 14, 8, 25, 25, 27, 33, 28, 17, 23, 15],
	#[254, 866, 74, 753, 76, 752, 75, 755],
	#[38, 858],
	#[2, 855],
	#[],
	#[],
	#[6, 854],
	#[],
	#[26, 27, 14, 8, 24, 2, 20, 22, 16, 13, 22, 30, 28, 17, 15, 29, 17, 28, 235, 806, 234, 807, 25, 25, 21, 9, 18, 14, 29, 5, 27, 33, 19, 37, 23, 15],
	#[],
	#[],
	#[],
	#[],
	#[6, 796],
	#[],
	#[],
	#[6, 788],
	#[],
	#[],
	#[],
	#[238, 41, 234, 40, 237, 39, 25, 25, 24, 2, 23, 15, 22, 30, 21, 9, 20, 22, 19, 37, 17, 28, 16, 13, 14, 8, 18, 14, 26, 27, 15, 29],
	#[],
	#[76, 752, 75, 755, 74, 753, 254, 754],
	#[246, 42, 61, 43, 31, 44],
	#[],
	#[290, 47, 273, 45, 247, 750, 162, 46],
	#[290, 47, 273, 45, 247, 48, 162, 46],
	#[],
	#[],
	#[291, 52, 276, 51, 275, 53, 274, 54, 118, 35, 63, 7, 7, 12],
	#[290, 47, 273, 49, 162, 46, 32, 50],
	#[],
	#[],
	#[305, 66, 302, 748, 292, 749, 287, 67, 285, 73, 270, 115, 204, 59, 200, 82, 199, 113, 198, 83, 197, 60, 196, 85, 195, 111, 193, 100, 192, 84, 190, 95, 189, 112, 188, 90, 205, 62, 170, 81, 169, 97, 167, 98, 166, 89, 164, 102, 163, 72, 165, 107, 141, 119, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 127, 101, 126, 114, 125, 61, 129, 80, 117, 118, 116, 116, 115, 69, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 121, 57, 140, 63, 105, 94, 104, 78, 101, 108, 99, 91, 97, 65, 106, 68, 107, 117, 114, 71, 124, 75, 113, 93, 100, 70, 102, 77, 122, 106, 95, 99, 123, 104, 103, 105],
	#[],
	#[6, 746],
	#[305, 66, 302, 74, 292, 110, 287, 67, 285, 73, 270, 115, 205, 62, 204, 59, 200, 82, 199, 113, 198, 83, 197, 60, 196, 85, 195, 111, 193, 100, 192, 84, 190, 95, 189, 112, 188, 90, 170, 81, 169, 97, 167, 98, 166, 89, 165, 107, 164, 102, 163, 72, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 68, 105, 94, 104, 78, 103, 105, 101, 108, 100, 70, 99, 91, 97, 65, 95, 99, 102, 77],
	#[],
	#[],
	#[286, 422, 135, 332, 134, 331],
	#[128, 416],
	#[205, 745],
	#[315, 736, 195, 544],
	#[286, 415, 135, 332, 134, 331],
	#[],
	#[],
	#[],
	#[316, 731, 224, 361],
	#[243, 722, 50, 558, 48, 566, 56, 555, 45, 560, 49, 554, 54, 565, 51, 559, 52, 561, 46, 562, 55, 553, 53, 563, 44, 557, 47, 564],
	#[275, 133, 249, 123, 248, 717, 250, 131, 252, 124, 118, 35, 70, 139, 73, 137, 66, 125, 67, 136, 64, 140, 65, 121, 36, 128, 35, 122, 63, 120, 33, 138, 31, 127, 68, 132, 69, 135, 11, 126, 71, 129, 7, 12],
	#[294, 712, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 36, 128, 64, 140, 69, 135, 35, 122, 66, 125, 71, 129, 68, 132, 65, 121, 33, 138, 73, 137, 67, 136, 63, 120, 11, 126, 70, 139, 7, 12, 31, 127],
	#[],
	#[288, 706, 151, 373, 150, 382, 149, 380, 148, 378, 147, 376, 145, 374, 144, 383, 143, 381, 146, 375, 142, 379],
	#[],
	#[295, 700, 294, 699, 275, 133, 252, 124, 249, 123, 248, 134, 250, 131, 118, 35, 65, 121, 64, 140, 11, 126, 73, 701, 66, 125, 67, 136, 35, 122, 33, 138, 36, 128, 7, 12, 63, 120, 31, 127, 68, 132, 69, 135, 70, 139, 71, 129],
	#[7, 12, 36, 128, 275, 133, 252, 124, 250, 131, 249, 123, 248, 694, 73, 137, 118, 35, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 35, 122, 33, 138, 31, 127, 11, 126],
	#[],
	#[],
	#[],
	#[294, 690, 275, 133, 249, 123, 248, 134, 250, 131, 252, 124, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 64, 140, 65, 121, 36, 128, 35, 122, 63, 120, 33, 138, 31, 127, 11, 126, 7, 12],
	#[294, 684, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 64, 140, 71, 129, 65, 121, 66, 125, 63, 120, 68, 132, 11, 126, 69, 135, 33, 138, 67, 136, 35, 122, 7, 12, 31, 127, 36, 128, 73, 137, 70, 139],
	#[],
	#[128, 359],
	#[294, 682, 275, 133, 249, 123, 248, 134, 252, 124, 250, 131, 118, 35, 66, 125, 64, 140, 73, 137, 71, 129, 69, 135, 11, 126, 65, 121, 7, 12, 68, 132, 33, 138, 35, 122, 67, 136, 63, 120, 36, 128, 70, 139, 31, 127],
	#[294, 678, 275, 133, 252, 124, 250, 131, 248, 134, 249, 123, 118, 35, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 73, 137, 36, 128, 33, 138, 11, 126, 35, 122, 7, 12, 31, 127],
	#[315, 659, 195, 544],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 653, 73, 137, 118, 35, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 63, 120, 64, 140, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[194, 640, 315, 641, 195, 544],
	#[286, 358, 135, 332, 134, 331],
	#[],
	#[],
	#[294, 627, 275, 133, 252, 124, 250, 131, 248, 134, 249, 123, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 33, 138, 35, 122, 11, 126, 7, 12, 31, 127],
	#[294, 623, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 70, 139, 11, 126, 7, 12],
	#[294, 617, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[],
	#[],
	#[252, 124, 250, 131, 249, 123, 248, 134, 275, 133, 294, 613, 118, 35, 69, 135, 73, 137, 35, 122, 68, 132, 65, 121, 64, 140, 63, 120, 66, 125, 11, 126, 71, 129, 33, 138, 67, 136, 36, 128, 70, 139, 31, 127, 7, 12],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 602, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[],
	#[322, 601, 13, 454],
	#[243, 556, 56, 555, 55, 553, 54, 565, 53, 563, 52, 561, 51, 559, 50, 558, 49, 554, 48, 566, 47, 564, 46, 562, 45, 560, 44, 557],
	#[],
	#[315, 545, 195, 544, 194, 543],
	#[128, 336],
	#[69, 532, 36, 531],
	#[128, 335],
	#[286, 330, 135, 332, 134, 331],
	#[294, 525, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[],
	#[36, 504],
	#[289, 498, 161, 297, 160, 293, 159, 291, 158, 294, 157, 303, 156, 301, 155, 290, 154, 296, 153, 292, 152, 289, 151, 287, 150, 302, 149, 299, 148, 295, 91, 298, 90, 300],
	#[],
	#[],
	#[196, 486, 193, 485],
	#[275, 133, 249, 123, 248, 471, 250, 131, 252, 124, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 35, 122, 33, 138, 36, 128, 11, 126, 7, 12, 31, 127],
	#[306, 462, 206, 461],
	#[],
	#[294, 130, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 33, 138, 35, 122, 36, 128, 11, 126, 7, 12, 31, 127],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[38, 199],
	#[275, 133, 267, 196, 250, 131, 249, 123, 248, 143, 73, 137, 118, 35, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 64, 140, 252, 124, 65, 121, 66, 125, 33, 138, 11, 126, 32, 197, 36, 128, 7, 12, 63, 120, 31, 127, 35, 122],
	#[],
	#[],
	#[96, 451],
	#[],
	#[],
	#[],
	#[321, 257, 320, 223, 309, 248, 296, 243, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 251, 159, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 111, 55, 110, 64, 109, 79, 108, 88, 112, 96, 113, 93, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 99, 244, 97, 236, 106, 230, 107, 117, 92, 240, 91, 251, 119, 32, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 100, 234, 90, 224, 72, 160, 95, 99, 63, 231, 42, 158, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[],
	#[],
	#[42, 150],
	#[38, 141, 31, 142],
	#[],
	#[],
	#[39, 219],
	#[275, 133, 267, 144, 252, 124, 250, 131, 249, 123, 248, 143, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 32, 145, 31, 127, 11, 126, 7, 12],
	#[251, 159, 72, 160, 42, 158],
	#[32, 147, 13, 148],
	#[34, 146],
	#[],
	#[34, 218],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 149, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[251, 159, 72, 160, 42, 158],
	#[275, 133, 266, 153, 265, 154, 253, 151, 252, 152, 250, 131, 249, 156, 248, 157, 118, 35, 98, 155, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[313, 214, 223, 213, 220, 206, 221, 209, 222, 211, 109, 212, 108, 207, 60, 208, 93, 210],
	#[],
	#[13, 203],
	#[43, 216],
	#[],
	#[312, 167, 57, 169, 59, 168, 58, 166],
	#[251, 159, 72, 160, 42, 158],
	#[275, 133, 266, 153, 265, 165, 253, 151, 252, 152, 250, 131, 249, 156, 248, 157, 118, 35, 98, 155, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 33, 138, 35, 122, 11, 126, 7, 12, 36, 128, 31, 127],
	#[37, 164],
	#[42, 161],
	#[38, 162],
	#[43, 163],
	#[],
	#[],
	#[43, 170],
	#[],
	#[],
	#[],
	#[],
	#[280, 171],
	#[293, 187, 187, 172, 186, 189, 185, 186, 184, 176, 183, 173, 182, 188, 181, 185, 180, 182, 179, 180, 177, 192, 176, 184, 175, 177, 174, 183, 173, 181, 172, 190, 171, 179, 178, 175, 59, 178, 57, 191, 58, 174],
	#[],
	#[],
	#[],
	#[],
	#[42, 193],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[38, 194],
	#[43, 195],
	#[],
	#[32, 198, 13, 148],
	#[],
	#[],
	#[39, 200],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 201, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[251, 159, 72, 160, 42, 158, 12, 202],
	#[],
	#[275, 133, 253, 204, 252, 152, 250, 131, 249, 156, 248, 157, 118, 35, 98, 205, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[313, 214, 221, 209, 222, 211, 223, 213, 220, 206, 109, 212, 108, 207, 93, 210, 60, 208],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[38, 215],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[280, 217],
	#[293, 187, 187, 172, 186, 189, 185, 186, 184, 176, 183, 173, 182, 188, 181, 185, 180, 182, 179, 180, 178, 175, 177, 192, 176, 184, 175, 177, 174, 183, 173, 181, 172, 190, 171, 179, 59, 178, 58, 174, 57, 191],
	#[],
	#[11, 126, 7, 12, 275, 133, 252, 124, 250, 131, 249, 123, 248, 220, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127],
	#[251, 159, 72, 160, 42, 158, 34, 221],
	#[],
	#[],
	#[],
	#[],
	#[42, 417],
	#[310, 446, 219, 445, 184, 444],
	#[275, 133, 268, 441, 259, 259, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 64, 140, 68, 132, 69, 135, 70, 139, 66, 125, 71, 129, 67, 136, 65, 121, 36, 128, 35, 122, 63, 120, 33, 138, 32, 442, 11, 126, 7, 12, 31, 127],
	#[42, 405],
	#[],
	#[42, 394],
	#[],
	#[],
	#[42, 389],
	#[288, 377, 151, 373, 150, 382, 149, 380, 148, 378, 147, 376, 146, 375, 145, 374, 144, 383, 143, 381, 142, 379],
	#[42, 368],
	#[316, 360, 224, 361],
	#[],
	#[42, 400],
	#[],
	#[42, 436],
	#[],
	#[],
	#[],
	#[42, 346],
	#[42, 338],
	#[7, 435],
	#[],
	#[],
	#[275, 133, 268, 432, 259, 259, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 65, 121, 64, 140, 63, 120, 66, 125, 36, 128, 35, 122, 33, 138, 31, 127, 12, 433, 11, 126, 7, 12],
	#[38, 318, 31, 317, 7, 316],
	#[],
	#[],
	#[42, 309],
	#[289, 288, 161, 297, 160, 293, 159, 291, 158, 294, 157, 303, 156, 301, 155, 290, 154, 296, 153, 292, 152, 289, 151, 287, 150, 302, 149, 299, 148, 295, 91, 298, 90, 300],
	#[275, 133, 268, 261, 252, 124, 250, 131, 249, 123, 248, 262, 259, 259, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 260, 11, 126, 7, 12],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 268, 427, 267, 196, 259, 259, 252, 124, 250, 131, 249, 123, 248, 286, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 32, 428, 31, 127, 7, 12, 11, 126],
	#[34, 426, 13, 281],
	#[321, 257, 320, 264, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 278, 264, 267, 251, 159, 226, 250, 141, 119, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 140, 63, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 123, 104, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 269, 91, 274, 90, 265, 89, 268, 88, 272, 110, 64, 86, 270, 85, 275, 84, 271, 72, 160, 63, 231, 87, 277, 42, 158, 38, 263, 33, 276, 31, 266, 11, 273, 10, 222, 7, 12],
	#[],
	#[],
	#[],
	#[275, 133, 259, 259, 249, 123, 248, 262, 252, 124, 268, 412, 250, 131, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 32, 413, 31, 127, 11, 126, 7, 12],
	#[],
	#[],
	#[42, 353],
	#[],
	#[7, 337],
	#[],
	#[275, 133, 268, 327, 252, 124, 250, 131, 249, 123, 259, 259, 248, 262, 70, 139, 68, 132, 73, 137, 65, 121, 36, 128, 64, 140, 66, 125, 69, 135, 35, 122, 71, 129, 33, 138, 11, 126, 7, 12, 63, 120, 67, 136, 12, 328, 118, 35, 31, 127],
	#[],
	#[],
	#[275, 133, 268, 280, 252, 124, 250, 131, 249, 123, 248, 262, 259, 259, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 73, 137, 65, 121, 64, 140, 63, 120, 66, 125, 71, 129, 36, 128, 35, 122, 33, 138, 31, 279, 11, 126, 7, 12],
	#[],
	#[],
	#[275, 133, 268, 284, 267, 196, 259, 259, 252, 124, 250, 131, 249, 123, 248, 286, 118, 35, 71, 129, 70, 139, 69, 135, 67, 136, 73, 137, 64, 140, 63, 120, 66, 125, 68, 132, 65, 121, 36, 128, 33, 138, 32, 285, 35, 122, 31, 127, 11, 126, 7, 12],
	#[34, 282, 13, 281],
	#[275, 133, 259, 283, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[],
	#[],
	#[32, 424, 13, 281],
	#[34, 423],
	#[321, 257, 320, 264, 270, 225, 269, 278, 285, 233, 264, 267, 275, 242, 287, 238, 277, 237, 226, 250, 251, 159, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 125, 61, 124, 75, 123, 104, 119, 32, 117, 118, 116, 116, 115, 69, 121, 57, 112, 96, 122, 106, 110, 64, 109, 79, 108, 88, 107, 117, 113, 93, 105, 245, 104, 228, 126, 114, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 106, 230, 114, 71, 91, 274, 90, 265, 89, 268, 88, 272, 87, 277, 86, 270, 85, 275, 84, 271, 111, 55, 92, 269, 103, 253, 118, 35, 72, 160, 63, 231, 42, 158, 38, 263, 95, 99, 33, 276, 31, 266, 10, 222, 7, 12, 11, 273],
	#[],
	#[42, 304],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 259, 305, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 306],
	#[275, 133, 259, 307, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 308],
	#[],
	#[275, 133, 259, 310, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 66, 125, 65, 121, 64, 140, 63, 120, 67, 136, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 311],
	#[275, 133, 259, 312, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 11, 126, 7, 12, 31, 127],
	#[13, 313],
	#[275, 133, 259, 314, 252, 124, 248, 262, 250, 131, 249, 123, 118, 35, 70, 139, 69, 135, 67, 136, 68, 132, 64, 140, 73, 137, 66, 125, 65, 121, 36, 128, 35, 122, 33, 138, 31, 127, 71, 129, 11, 126, 63, 120, 7, 12],
	#[43, 315],
	#[],
	#[],
	#[319, 321, 275, 133, 259, 320, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 85, 322, 67, 136, 73, 137, 65, 121, 64, 140, 63, 120, 66, 125, 68, 132, 70, 139, 71, 129, 36, 128, 35, 122, 33, 138, 32, 319, 69, 135, 11, 126, 7, 12, 31, 127],
	#[],
	#[],
	#[],
	#[32, 323, 13, 324],
	#[],
	#[],
	#[275, 133, 259, 325, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 85, 326, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 33, 138, 35, 122, 11, 126, 7, 12, 31, 127],
	#[],
	#[],
	#[13, 281, 12, 329],
	#[],
	#[],
	#[],
	#[135, 334],
	#[134, 333],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 259, 339, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 35, 122, 33, 138, 31, 127, 36, 128, 11, 126, 7, 12],
	#[271, 341, 13, 340],
	#[38, 345],
	#[43, 342, 13, 343],
	#[],
	#[38, 344],
	#[],
	#[],
	#[275, 133, 259, 347, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 348],
	#[275, 133, 259, 349, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 350],
	#[275, 133, 250, 131, 248, 262, 249, 123, 259, 351, 252, 124, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 352],
	#[],
	#[277, 354, 119, 32],
	#[13, 355],
	#[275, 356, 118, 35, 7, 12],
	#[43, 357],
	#[],
	#[],
	#[],
	#[42, 362],
	#[],
	#[275, 133, 259, 363, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[284, 364],
	#[43, 365, 13, 366],
	#[],
	#[275, 133, 294, 367, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[],
	#[275, 133, 259, 369, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 71, 129, 67, 136, 65, 121, 64, 140, 63, 120, 73, 137, 66, 125, 68, 132, 69, 135, 70, 139, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 370],
	#[275, 133, 259, 371, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 70, 139, 73, 137, 65, 121, 67, 136, 68, 132, 64, 140, 66, 125, 71, 129, 35, 122, 63, 120, 33, 138, 69, 135, 36, 128, 11, 126, 7, 12, 31, 127],
	#[43, 372],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[42, 384],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 259, 385, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 70, 139, 69, 135, 68, 132, 73, 137, 65, 121, 64, 140, 66, 125, 67, 136, 35, 122, 33, 138, 31, 127, 71, 129, 63, 120, 36, 128, 11, 126, 7, 12],
	#[13, 386],
	#[275, 133, 252, 124, 250, 131, 249, 123, 259, 387, 248, 262, 118, 35, 67, 136, 73, 137, 65, 121, 64, 140, 63, 120, 66, 125, 68, 132, 69, 135, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 70, 139, 11, 126, 7, 12],
	#[43, 388],
	#[],
	#[275, 133, 259, 390, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 391],
	#[275, 133, 248, 262, 249, 123, 250, 131, 259, 392, 252, 124, 118, 35, 66, 125, 65, 121, 64, 140, 63, 120, 73, 137, 67, 136, 68, 132, 69, 135, 70, 139, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 393],
	#[],
	#[275, 133, 250, 131, 249, 123, 252, 124, 248, 262, 259, 395, 118, 35, 70, 139, 68, 132, 67, 136, 73, 137, 65, 121, 64, 140, 63, 120, 66, 125, 69, 135, 35, 122, 33, 138, 36, 128, 71, 129, 11, 126, 7, 12, 31, 127],
	#[13, 396],
	#[275, 133, 259, 397, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[271, 398, 13, 340],
	#[43, 399, 13, 343],
	#[],
	#[275, 133, 259, 401, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 402],
	#[275, 133, 259, 403, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 404],
	#[],
	#[252, 124, 248, 262, 259, 406, 250, 131, 249, 123, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 275, 133, 65, 121, 64, 140, 63, 120, 118, 35, 66, 125, 67, 136, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 407],
	#[275, 133, 259, 408, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 409],
	#[275, 133, 259, 410, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 411],
	#[],
	#[32, 414, 13, 281],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 259, 418, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[96, 419],
	#[275, 133, 249, 123, 250, 131, 252, 124, 248, 420, 118, 35, 70, 139, 69, 135, 67, 136, 73, 137, 64, 140, 65, 121, 68, 132, 71, 129, 36, 128, 63, 120, 66, 125, 33, 138, 35, 122, 11, 126, 31, 127, 7, 12],
	#[251, 159, 72, 160, 43, 421, 42, 158],
	#[],
	#[],
	#[],
	#[34, 425],
	#[],
	#[],
	#[32, 430, 13, 281],
	#[34, 429],
	#[],
	#[34, 431],
	#[],
	#[13, 281, 12, 434],
	#[],
	#[],
	#[],
	#[277, 437, 119, 32],
	#[13, 438],
	#[7, 12, 275, 439, 118, 35],
	#[43, 440],
	#[],
	#[32, 443, 13, 281],
	#[],
	#[],
	#[],
	#[184, 450],
	#[7, 447],
	#[13, 448],
	#[7, 449],
	#[],
	#[],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 452, 118, 35, 71, 129, 70, 139, 68, 132, 67, 136, 73, 137, 65, 121, 64, 140, 63, 120, 66, 125, 69, 135, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[322, 453, 251, 159, 72, 160, 42, 158, 13, 454],
	#[],
	#[323, 455, 225, 456],
	#[13, 458],
	#[320, 457, 321, 257, 226, 250],
	#[],
	#[225, 459],
	#[321, 257, 320, 460, 226, 250],
	#[],
	#[],
	#[307, 467, 212, 469, 211, 468, 210, 466, 209, 465, 208, 464, 207, 463],
	#[],
	#[],
	#[],
	#[],
	#[322, 470, 13, 454],
	#[],
	#[],
	#[],
	#[314, 473, 251, 159, 72, 160, 42, 158, 11, 472],
	#[321, 257, 320, 223, 309, 248, 296, 481, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 136, 86, 133, 87, 131, 109, 137, 58, 129, 80, 125, 61, 124, 75, 123, 104, 122, 106, 130, 56, 118, 35, 117, 118, 116, 116, 132, 92, 121, 57, 112, 96, 111, 55, 109, 79, 108, 88, 107, 117, 113, 93, 105, 245, 104, 228, 126, 114, 115, 69, 101, 254, 100, 234, 99, 244, 127, 101, 97, 236, 106, 230, 114, 71, 91, 251, 89, 232, 88, 247, 110, 64, 86, 241, 85, 258, 84, 246, 102, 235, 92, 240, 103, 253, 90, 224, 38, 252, 95, 99, 63, 231, 33, 255, 119, 32, 31, 227, 87, 256, 11, 249, 10, 222, 7, 12, 2, 226],
	#[322, 474, 13, 475],
	#[],
	#[323, 455, 225, 456, 11, 476],
	#[321, 257, 320, 223, 309, 248, 296, 477, 285, 233, 275, 242, 270, 225, 264, 239, 287, 238, 269, 229, 277, 237, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 124, 75, 119, 32, 118, 35, 117, 118, 116, 116, 132, 92, 121, 57, 113, 93, 112, 96, 111, 55, 123, 104, 109, 79, 108, 88, 107, 117, 125, 61, 105, 245, 104, 228, 114, 71, 115, 69, 101, 254, 100, 234, 99, 244, 97, 236, 106, 230, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 110, 64, 86, 241, 85, 258, 84, 246, 102, 235, 122, 106, 103, 253, 95, 99, 63, 231, 87, 256, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[13, 478],
	#[296, 479, 321, 257, 277, 237, 320, 223, 309, 248, 270, 225, 285, 233, 264, 239, 287, 238, 269, 229, 275, 242, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[12, 480],
	#[],
	#[13, 482],
	#[320, 223, 321, 257, 309, 248, 296, 483, 287, 238, 285, 233, 277, 237, 270, 225, 264, 239, 275, 242, 269, 229, 226, 250, 140, 63, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 138, 103, 125, 61, 124, 75, 123, 104, 141, 119, 121, 57, 130, 56, 118, 35, 117, 118, 116, 116, 129, 80, 114, 71, 139, 76, 112, 96, 122, 106, 110, 64, 109, 79, 108, 88, 107, 117, 113, 93, 105, 245, 104, 228, 126, 114, 115, 69, 101, 254, 100, 234, 99, 244, 127, 101, 97, 236, 106, 230, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 111, 55, 95, 99, 63, 231, 119, 32, 102, 235, 103, 253, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[12, 484],
	#[],
	#[294, 496, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[294, 487, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 11, 126, 31, 127, 7, 12],
	#[13, 488],
	#[294, 489, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[322, 490, 317, 492, 13, 491],
	#[],
	#[323, 455, 225, 456, 93, 493],
	#[],
	#[38, 494],
	#[322, 495, 13, 454],
	#[],
	#[322, 490, 317, 497, 13, 491],
	#[],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 499, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 63, 120, 64, 140, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[321, 257, 320, 223, 296, 500, 309, 248, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 251, 159, 226, 250, 136, 86, 133, 87, 121, 57, 131, 109, 138, 103, 122, 106, 137, 58, 130, 56, 105, 245, 104, 228, 115, 69, 101, 254, 139, 76, 116, 116, 132, 92, 123, 104, 95, 99, 129, 80, 92, 240, 91, 251, 88, 247, 87, 256, 86, 241, 113, 93, 100, 234, 114, 71, 126, 114, 117, 118, 112, 96, 97, 236, 89, 232, 108, 88, 140, 63, 127, 101, 125, 61, 72, 160, 85, 258, 141, 119, 110, 64, 63, 231, 106, 230, 119, 32, 99, 244, 90, 224, 124, 75, 107, 117, 111, 55, 109, 79, 118, 35, 103, 253, 38, 252, 33, 255, 31, 227, 42, 158, 84, 246, 11, 249, 102, 235, 7, 12, 2, 226, 10, 222],
	#[13, 501],
	#[321, 257, 320, 223, 309, 248, 296, 502, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[322, 503, 13, 454],
	#[],
	#[320, 223, 321, 257, 309, 248, 296, 505, 287, 238, 285, 233, 275, 242, 270, 225, 269, 229, 264, 239, 277, 237, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 124, 75, 119, 32, 118, 35, 117, 118, 116, 116, 132, 92, 121, 57, 113, 93, 112, 96, 111, 55, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 114, 71, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 110, 64, 86, 241, 85, 258, 84, 246, 122, 106, 115, 69, 103, 253, 123, 104, 125, 61, 38, 252, 63, 231, 33, 255, 31, 227, 87, 256, 11, 249, 10, 222, 7, 12, 2, 226],
	#[13, 506],
	#[69, 507],
	#[320, 223, 321, 257, 309, 248, 296, 508, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 138, 103, 126, 114, 125, 61, 123, 104, 129, 80, 121, 57, 130, 56, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 112, 96, 102, 235, 101, 254, 122, 106, 113, 93, 127, 101, 97, 236, 106, 230, 107, 117, 92, 240, 108, 88, 124, 75, 104, 228, 88, 247, 110, 64, 99, 244, 85, 258, 105, 245, 100, 234, 111, 55, 90, 224, 91, 251, 103, 253, 89, 232, 84, 246, 86, 241, 119, 32, 87, 256, 109, 79, 38, 252, 95, 99, 63, 231, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[11, 509],
	#[297, 511, 36, 510, 12, 512],
	#[321, 257, 320, 223, 309, 521, 287, 238, 285, 233, 270, 225, 269, 229, 226, 250, 141, 119, 140, 63, 139, 76, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 138, 103, 126, 114, 125, 61, 124, 75, 123, 104, 129, 80, 121, 57, 130, 56, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 99, 244, 127, 101, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 84, 246, 100, 234, 101, 254, 122, 106, 85, 258, 38, 252, 33, 255, 31, 227, 11, 249, 2, 226],
	#[36, 514, 12, 515],
	#[322, 513, 13, 454],
	#[],
	#[320, 223, 321, 257, 309, 517, 287, 238, 285, 233, 270, 225, 269, 229, 226, 250, 140, 63, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 139, 76, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 117, 118, 137, 58, 115, 69, 114, 71, 113, 93, 111, 55, 110, 64, 109, 79, 108, 88, 112, 96, 106, 230, 138, 103, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 105, 245, 97, 236, 95, 99, 107, 117, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 86, 241, 85, 258, 141, 119, 116, 116, 87, 256, 126, 114, 84, 246, 38, 252, 33, 255, 31, 227, 11, 249, 2, 226],
	#[322, 516, 13, 454],
	#[],
	#[13, 518],
	#[69, 519],
	#[321, 257, 320, 223, 309, 248, 296, 520, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[],
	#[13, 522],
	#[69, 523],
	#[320, 223, 321, 257, 296, 524, 287, 238, 277, 237, 270, 225, 269, 229, 264, 239, 285, 233, 226, 250, 275, 242, 309, 248, 141, 119, 140, 63, 139, 76, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 138, 103, 126, 114, 125, 61, 124, 75, 123, 104, 129, 80, 121, 57, 130, 56, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 112, 96, 102, 235, 101, 254, 122, 106, 113, 93, 127, 101, 99, 244, 85, 258, 84, 246, 91, 251, 105, 245, 92, 240, 104, 228, 111, 55, 109, 79, 100, 234, 88, 247, 90, 224, 103, 253, 110, 64, 108, 88, 38, 252, 97, 236, 89, 232, 119, 32, 33, 255, 106, 230, 86, 241, 107, 117, 95, 99, 7, 12, 63, 231, 10, 222, 11, 249, 2, 226, 31, 227, 87, 256],
	#[],
	#[13, 526],
	#[294, 527, 275, 133, 252, 124, 250, 131, 248, 134, 249, 123, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 528],
	#[294, 529, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[322, 530, 13, 454],
	#[],
	#[321, 257, 320, 223, 309, 248, 296, 535, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[321, 257, 320, 223, 309, 248, 296, 533, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 264, 239, 269, 229, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[322, 534, 13, 454],
	#[],
	#[13, 536],
	#[69, 537],
	#[320, 223, 309, 248, 287, 238, 321, 257, 285, 233, 277, 237, 296, 538, 226, 250, 269, 229, 264, 239, 270, 225, 275, 242, 140, 63, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 138, 103, 139, 76, 125, 61, 124, 75, 123, 104, 141, 119, 121, 57, 130, 56, 118, 35, 117, 118, 116, 116, 129, 80, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 126, 114, 102, 235, 101, 254, 99, 244, 127, 101, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 100, 234, 122, 106, 115, 69, 119, 32, 103, 253, 38, 252, 63, 231, 33, 255, 31, 227, 11, 249, 10, 222, 2, 226, 7, 12],
	#[13, 539],
	#[69, 540],
	#[320, 223, 321, 257, 287, 238, 296, 541, 269, 229, 264, 239, 285, 233, 275, 242, 309, 248, 226, 250, 270, 225, 277, 237, 141, 119, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 140, 63, 122, 106, 121, 57, 119, 32, 118, 35, 116, 116, 114, 71, 113, 93, 112, 96, 111, 55, 123, 104, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 110, 64, 85, 258, 115, 69, 84, 246, 117, 118, 87, 256, 38, 252, 63, 231, 33, 255, 31, 227, 10, 222, 7, 12, 2, 226, 11, 249, 86, 241],
	#[322, 542, 13, 454],
	#[],
	#[315, 548, 195, 544],
	#[],
	#[294, 546, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[322, 490, 317, 547, 13, 491],
	#[],
	#[294, 549, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[306, 550, 206, 461],
	#[307, 551, 212, 469, 211, 468, 210, 466, 209, 465, 207, 463, 208, 464],
	#[322, 490, 317, 552, 13, 491],
	#[],
	#[],
	#[],
	#[38, 600],
	#[244, 567],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[275, 133, 272, 571, 252, 124, 250, 131, 249, 123, 248, 575, 245, 570, 118, 35, 73, 569, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 58, 572, 57, 574, 59, 568, 60, 573, 36, 128, 33, 138, 35, 122, 11, 126, 7, 12, 31, 127],
	#[],
	#[42, 150],
	#[],
	#[321, 257, 320, 223, 296, 576, 309, 248, 287, 238, 275, 242, 270, 225, 269, 229, 285, 233, 264, 239, 226, 250, 277, 237, 140, 63, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 138, 103, 125, 61, 124, 75, 137, 58, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 100, 234, 97, 236, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 86, 241, 85, 258, 141, 119, 101, 254, 126, 114, 95, 99, 63, 231, 139, 76, 99, 244, 123, 104, 102, 235, 84, 246, 127, 101, 38, 252, 33, 255, 31, 227, 103, 253, 11, 249, 87, 256, 10, 222, 2, 226, 7, 12],
	#[],
	#[],
	#[],
	#[251, 159, 72, 160, 42, 158],
	#[42, 577],
	#[299, 578, 275, 133, 253, 579, 252, 152, 250, 131, 249, 156, 248, 157, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[43, 587, 13, 588],
	#[321, 257, 320, 223, 313, 214, 309, 248, 296, 582, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 222, 211, 221, 209, 220, 206, 226, 250, 223, 213, 140, 63, 136, 86, 133, 87, 132, 92, 131, 109, 137, 58, 129, 80, 138, 103, 124, 75, 123, 104, 141, 119, 115, 69, 112, 96, 130, 56, 121, 57, 125, 61, 105, 245, 104, 228, 114, 71, 102, 235, 101, 254, 100, 234, 113, 93, 117, 118, 97, 236, 118, 35, 107, 117, 93, 210, 92, 240, 89, 232, 88, 247, 116, 116, 126, 114, 90, 224, 91, 251, 139, 76, 106, 230, 109, 580, 127, 101, 84, 246, 85, 258, 110, 64, 86, 241, 122, 106, 108, 581, 33, 255, 11, 249, 10, 222, 63, 231, 119, 32, 99, 244, 60, 208, 2, 226, 95, 99, 111, 55, 7, 12, 38, 252, 31, 227, 87, 256, 103, 253],
	#[],
	#[],
	#[300, 583],
	#[313, 586, 312, 584, 311, 585, 220, 206, 222, 211, 223, 213, 221, 209, 109, 212, 108, 207, 93, 210, 60, 208, 59, 168, 58, 166, 57, 169],
	#[],
	#[],
	#[],
	#[280, 592],
	#[275, 133, 253, 589, 252, 152, 250, 131, 249, 156, 248, 157, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 7, 12, 11, 126],
	#[321, 257, 320, 223, 313, 214, 309, 248, 296, 590, 287, 238, 277, 237, 275, 242, 270, 225, 269, 229, 285, 233, 264, 239, 223, 213, 222, 211, 221, 209, 220, 206, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 580, 108, 581, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 93, 210, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 60, 208, 38, 252, 33, 255, 11, 249, 10, 222, 7, 12, 2, 226, 31, 227],
	#[300, 591],
	#[313, 586, 312, 584, 311, 585, 222, 211, 220, 206, 223, 213, 221, 209, 109, 212, 108, 207, 93, 210, 60, 208, 59, 168, 58, 166, 57, 169],
	#[293, 187, 187, 172, 186, 189, 185, 186, 184, 176, 183, 173, 182, 188, 181, 185, 180, 182, 179, 180, 178, 175, 177, 192, 176, 184, 175, 177, 174, 183, 173, 181, 172, 190, 171, 179, 96, 593, 59, 178, 58, 174, 57, 191],
	#[69, 594],
	#[321, 257, 320, 223, 309, 248, 296, 595, 287, 238, 285, 233, 275, 242, 270, 225, 269, 229, 264, 239, 277, 237, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 87, 256, 86, 241, 85, 258, 84, 246, 63, 231, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[168, 596],
	#[69, 597],
	#[321, 257, 320, 223, 309, 248, 296, 598, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 105, 245, 104, 228, 103, 253, 102, 235, 101, 254, 100, 234, 99, 244, 97, 236, 95, 99, 92, 240, 91, 251, 90, 224, 89, 232, 88, 247, 85, 258, 84, 246, 63, 231, 87, 256, 86, 241, 38, 252, 33, 255, 31, 227, 11, 249, 10, 222, 7, 12, 2, 226],
	#[322, 599, 13, 454],
	#[],
	#[],
	#[],
	#[251, 159, 191, 603, 72, 160, 42, 158],
	#[294, 604, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 66, 125, 65, 121, 64, 140, 63, 120, 67, 136, 68, 132, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[303, 606, 201, 605],
	#[],
	#[304, 607],
	#[322, 610, 203, 608, 202, 609, 13, 454],
	#[249, 123, 259, 612, 248, 262, 252, 124, 275, 133, 250, 131, 118, 35, 73, 137, 71, 129, 70, 139, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 69, 135, 36, 128, 33, 138, 35, 122, 11, 126, 31, 127, 7, 12],
	#[294, 611, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 66, 125, 64, 140, 65, 121, 67, 136, 68, 132, 36, 128, 35, 122, 31, 127, 33, 138, 11, 126, 7, 12, 63, 120],
	#[],
	#[],
	#[],
	#[271, 614, 13, 340],
	#[322, 615, 13, 616],
	#[],
	#[323, 455, 225, 456, 38, 344],
	#[13, 618],
	#[294, 619, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 70, 139, 69, 135, 68, 132, 67, 136, 65, 121, 64, 140, 63, 120, 73, 137, 66, 125, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 7, 12, 11, 126, 118, 35],
	#[13, 620],
	#[294, 621, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 73, 137, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[322, 622, 13, 454],
	#[],
	#[13, 624],
	#[275, 133, 252, 124, 249, 123, 250, 131, 118, 35, 71, 129, 69, 135, 67, 136, 66, 125, 64, 140, 63, 120, 65, 121, 248, 625, 73, 137, 36, 128, 33, 138, 70, 139, 35, 122, 68, 132, 11, 126, 31, 127, 7, 12],
	#[322, 626, 251, 159, 72, 160, 42, 158, 13, 454],
	#[],
	#[13, 628],
	#[11, 629],
	#[298, 630, 69, 632, 12, 631],
	#[13, 635, 12, 636],
	#[322, 634, 13, 454],
	#[321, 257, 320, 223, 296, 633, 309, 248, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 124, 75, 119, 32, 118, 35, 116, 116, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 125, 61, 105, 245, 104, 228, 114, 71, 102, 235, 101, 254, 121, 57, 113, 93, 117, 118, 97, 236, 106, 230, 107, 117, 92, 240, 90, 224, 88, 247, 122, 106, 115, 69, 89, 232, 103, 253, 99, 244, 140, 63, 123, 104, 84, 246, 85, 258, 91, 251, 63, 231, 86, 241, 38, 252, 95, 99, 100, 234, 31, 227, 87, 256, 33, 255, 10, 222, 11, 249, 2, 226, 7, 12],
	#[],
	#[],
	#[69, 638],
	#[322, 637, 13, 454],
	#[],
	#[321, 257, 320, 223, 309, 248, 287, 238, 285, 233, 277, 237, 296, 639, 264, 239, 226, 250, 136, 86, 270, 225, 129, 80, 104, 228, 130, 56, 121, 57, 113, 93, 125, 61, 132, 92, 105, 245, 115, 69, 119, 32, 89, 232, 137, 58, 138, 103, 114, 71, 110, 64, 10, 222, 112, 96, 139, 76, 108, 88, 90, 224, 106, 230, 97, 236, 131, 109, 127, 101, 88, 247, 2, 226, 133, 87, 102, 235, 63, 231, 101, 254, 84, 246, 100, 234, 111, 55, 33, 255, 109, 79, 124, 75, 107, 117, 85, 258, 141, 119, 91, 251, 11, 249, 7, 12, 99, 244, 118, 35, 31, 227, 140, 63, 86, 241, 38, 252, 123, 104, 275, 242, 92, 240, 95, 99, 122, 106, 117, 118, 269, 229, 116, 116, 126, 114, 103, 253, 87, 256],
	#[],
	#[315, 646, 195, 544],
	#[275, 133, 294, 642, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 70, 139, 69, 135, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 33, 138, 11, 126, 71, 129, 35, 122, 7, 12, 31, 127, 68, 132, 73, 137, 36, 128],
	#[13, 643],
	#[118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12, 294, 644, 275, 133, 248, 134, 249, 123, 250, 131, 252, 124],
	#[322, 490, 317, 645, 13, 491],
	#[],
	#[294, 647, 249, 123, 275, 133, 248, 134, 250, 131, 252, 124, 118, 35, 70, 139, 69, 135, 68, 132, 65, 121, 64, 140, 63, 120, 73, 137, 67, 136, 66, 125, 36, 128, 35, 122, 33, 138, 11, 126, 7, 12, 71, 129, 31, 127],
	#[13, 648],
	#[294, 649, 275, 133, 249, 123, 248, 134, 250, 131, 252, 124, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[306, 650, 206, 461],
	#[307, 651, 212, 469, 211, 468, 210, 466, 209, 465, 208, 464, 207, 463],
	#[322, 490, 317, 652, 13, 491],
	#[],
	#[322, 490, 317, 654, 251, 159, 72, 160, 42, 158, 13, 655],
	#[],
	#[323, 455, 225, 456, 93, 493, 36, 656],
	#[320, 223, 321, 257, 309, 248, 296, 657, 287, 238, 277, 237, 275, 242, 226, 250, 269, 229, 264, 239, 270, 225, 137, 58, 136, 86, 106, 230, 121, 57, 114, 71, 140, 63, 88, 247, 125, 61, 130, 56, 85, 258, 100, 234, 129, 80, 90, 224, 105, 245, 97, 236, 104, 228, 124, 75, 118, 35, 109, 79, 127, 101, 110, 64, 63, 231, 112, 96, 92, 240, 102, 235, 91, 251, 113, 93, 141, 119, 101, 254, 89, 232, 138, 103, 285, 233, 139, 76, 99, 244, 123, 104, 132, 92, 33, 255, 111, 55, 133, 87, 86, 241, 131, 109, 122, 106, 115, 69, 7, 12, 87, 256, 10, 222, 11, 249, 116, 116, 38, 252, 2, 226, 107, 117, 108, 88, 126, 114, 117, 118, 84, 246, 95, 99, 119, 32, 103, 253, 31, 227],
	#[13, 491, 322, 490, 317, 658],
	#[],
	#[308, 666, 218, 671, 216, 667, 214, 665, 213, 669, 217, 670, 215, 664, 141, 668, 140, 663, 139, 661, 121, 662, 123, 660],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[294, 672, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 66, 125, 65, 121, 63, 120, 64, 140, 67, 136, 33, 138, 31, 127, 35, 122, 11, 126, 36, 128, 7, 12],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[13, 673],
	#[294, 674, 275, 133, 248, 134, 249, 123, 250, 131, 252, 124, 118, 35, 70, 139, 69, 135, 65, 121, 64, 140, 73, 137, 67, 136, 66, 125, 35, 122, 63, 120, 33, 138, 36, 128, 68, 132, 11, 126, 7, 12, 31, 127, 71, 129],
	#[306, 675, 206, 461],
	#[307, 676, 212, 469, 209, 465, 207, 463, 208, 464, 210, 466, 211, 468],
	#[322, 677, 13, 454],
	#[],
	#[13, 679],
	#[38, 680],
	#[322, 681, 13, 454],
	#[],
	#[322, 683, 13, 454],
	#[],
	#[13, 685],
	#[294, 686, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 687],
	#[294, 688, 252, 124, 249, 123, 250, 131, 248, 134, 275, 133, 118, 35, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 33, 138, 31, 127, 35, 122, 36, 128, 11, 126, 7, 12],
	#[322, 689, 13, 454],
	#[],
	#[13, 691],
	#[294, 692, 275, 133, 249, 123, 248, 134, 252, 124, 250, 131, 118, 35, 70, 139, 69, 135, 68, 132, 66, 125, 65, 121, 64, 140, 73, 137, 67, 136, 71, 129, 36, 128, 33, 138, 35, 122, 11, 126, 7, 12, 31, 127, 63, 120],
	#[322, 693, 13, 454],
	#[],
	#[321, 257, 320, 223, 309, 248, 296, 695, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 251, 159, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 125, 61, 124, 75, 123, 104, 122, 106, 121, 57, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 106, 230, 104, 228, 92, 240, 105, 245, 90, 224, 88, 247, 103, 253, 99, 244, 72, 160, 100, 234, 97, 236, 89, 232, 87, 256, 102, 235, 84, 246, 86, 241, 38, 252, 95, 99, 63, 231, 33, 255, 31, 227, 85, 258, 91, 251, 101, 254, 42, 158, 11, 249, 10, 222, 7, 12, 2, 226],
	#[13, 696],
	#[321, 257, 320, 223, 309, 248, 296, 697, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 136, 86, 126, 114, 124, 75, 130, 56, 118, 35, 117, 118, 137, 58, 129, 80, 121, 57, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 104, 228, 125, 61, 86, 241, 85, 258, 106, 230, 89, 232, 114, 71, 90, 224, 88, 247, 105, 245, 97, 236, 103, 253, 99, 244, 107, 117, 84, 246, 91, 251, 132, 92, 63, 231, 119, 32, 92, 240, 133, 87, 102, 235, 131, 109, 115, 69, 100, 234, 116, 116, 123, 104, 127, 101, 38, 252, 122, 106, 33, 255, 31, 227, 87, 256, 101, 254, 11, 249, 10, 222, 95, 99, 7, 12, 2, 226],
	#[322, 698, 13, 454],
	#[],
	#[],
	#[322, 703, 13, 704],
	#[322, 702, 42, 150, 13, 454],
	#[],
	#[],
	#[252, 124, 225, 456, 248, 134, 294, 705, 249, 123, 250, 131, 323, 455, 275, 133, 73, 137, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 63, 120, 64, 140, 71, 129, 36, 128, 33, 138, 35, 122, 7, 12, 11, 126, 31, 127],
	#[],
	#[252, 124, 249, 123, 250, 131, 248, 707, 275, 133, 118, 35, 65, 121, 67, 136, 73, 137, 68, 132, 69, 135, 64, 140, 70, 139, 36, 128, 35, 122, 63, 120, 33, 138, 7, 12, 31, 127, 66, 125, 11, 126, 71, 129],
	#[321, 257, 309, 248, 320, 223, 296, 708, 287, 238, 285, 233, 277, 237, 275, 242, 270, 225, 269, 229, 264, 239, 251, 159, 226, 250, 141, 119, 140, 63, 139, 76, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 126, 114, 124, 75, 119, 32, 118, 35, 117, 118, 116, 116, 115, 69, 121, 57, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 125, 61, 105, 245, 104, 228, 114, 71, 102, 235, 101, 254, 100, 234, 99, 244, 86, 241, 85, 258, 106, 230, 89, 232, 88, 247, 97, 236, 92, 240, 103, 253, 72, 160, 123, 104, 84, 246, 95, 99, 91, 251, 63, 231, 87, 256, 122, 106, 38, 252, 33, 255, 90, 224, 42, 158, 11, 249, 10, 222, 7, 12, 2, 226, 31, 227],
	#[13, 709],
	#[321, 257, 320, 223, 309, 248, 287, 238, 296, 710, 277, 237, 275, 242, 264, 239, 269, 229, 285, 233, 226, 250, 270, 225, 141, 119, 140, 63, 138, 103, 137, 58, 136, 86, 133, 87, 132, 92, 131, 109, 130, 56, 129, 80, 127, 101, 139, 76, 124, 75, 123, 104, 121, 57, 119, 32, 118, 35, 117, 118, 114, 71, 113, 93, 112, 96, 111, 55, 110, 64, 109, 79, 108, 88, 107, 117, 125, 61, 104, 228, 92, 240, 105, 245, 90, 224, 89, 232, 88, 247, 86, 241, 85, 258, 106, 230, 100, 234, 101, 254, 126, 114, 102, 235, 122, 106, 97, 236, 103, 253, 99, 244, 87, 256, 91, 251, 115, 69, 116, 116, 38, 252, 63, 231, 84, 246, 31, 227, 11, 249, 33, 255, 2, 226, 95, 99, 10, 222, 7, 12],
	#[322, 711, 13, 454],
	#[],
	#[13, 713],
	#[294, 714, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 65, 121, 64, 140, 63, 120, 66, 125, 73, 137, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[271, 715, 13, 340],
	#[322, 716, 13, 616],
	#[],
	#[321, 257, 320, 223, 309, 248, 296, 718, 277, 237, 275, 242, 270, 225, 269, 229, 285, 233, 264, 239, 287, 238, 251, 159, 226, 250, 136, 86, 110, 64, 137, 58, 121, 57, 113, 93, 130, 56, 106, 230, 109, 79, 90, 224, 72, 160, 91, 251, 104, 228, 63, 231, 97, 236, 92, 240, 111, 55, 129, 80, 117, 118, 89, 232, 124, 75, 140, 63, 114, 71, 99, 244, 125, 61, 105, 245, 101, 254, 115, 69, 86, 241, 131, 109, 132, 92, 112, 96, 127, 101, 10, 222, 139, 76, 95, 99, 123, 104, 33, 255, 118, 35, 2, 226, 122, 106, 133, 87, 84, 246, 100, 234, 116, 116, 138, 103, 88, 247, 38, 252, 107, 117, 108, 88, 126, 114, 102, 235, 11, 249, 42, 158, 103, 253, 85, 258, 141, 119, 7, 12, 119, 32, 31, 227, 87, 256],
	#[13, 719],
	#[321, 257, 320, 223, 309, 248, 287, 238, 277, 237, 264, 239, 285, 233, 275, 242, 226, 250, 269, 229, 270, 225, 296, 720, 140, 63, 136, 86, 133, 87, 131, 109, 130, 56, 129, 80, 127, 101, 125, 61, 124, 75, 137, 58, 122, 106, 110, 64, 109, 79, 108, 88, 121, 57, 106, 230, 105, 245, 104, 228, 114, 71, 102, 235, 101, 254, 100, 234, 132, 92, 97, 236, 118, 35, 138, 103, 113, 93, 91, 251, 90, 224, 89, 232, 139, 76, 86, 241, 85, 258, 111, 55, 126, 114, 88, 247, 95, 99, 116, 116, 119, 32, 99, 244, 92, 240, 112, 96, 141, 119, 107, 117, 38, 252, 84, 246, 33, 255, 117, 118, 103, 253, 123, 104, 87, 256, 63, 231, 31, 227, 10, 222, 2, 226, 115, 69, 7, 12, 11, 249],
	#[322, 721, 13, 454],
	#[],
	#[244, 723],
	#[275, 133, 272, 724, 252, 124, 249, 123, 248, 575, 245, 570, 250, 131, 73, 569, 70, 139, 67, 136, 64, 140, 60, 573, 59, 568, 66, 125, 57, 574, 71, 129, 65, 121, 33, 138, 58, 572, 35, 122, 36, 128, 68, 132, 11, 126, 7, 12, 69, 135, 118, 35, 63, 120, 31, 127],
	#[320, 223, 321, 257, 296, 725, 226, 250, 264, 239, 270, 225, 287, 238, 309, 248, 277, 237, 129, 80, 125, 61, 124, 75, 130, 56, 118, 35, 117, 118, 139, 76, 112, 96, 110, 64, 104, 228, 137, 58, 136, 86, 121, 57, 113, 93, 132, 92, 97, 236, 106, 230, 138, 103, 114, 71, 108, 88, 88, 247, 87, 256, 111, 55, 275, 242, 122, 106, 105, 245, 285, 233, 89, 232, 103, 253, 99, 244, 140, 63, 109, 79, 90, 224, 84, 246, 85, 258, 91, 251, 63, 231, 119, 32, 269, 229, 133, 87, 100, 234, 116, 116, 92, 240, 86, 241, 38, 252, 126, 114, 123, 104, 31, 227, 131, 109, 101, 254, 115, 69, 33, 255, 107, 117, 127, 101, 10, 222, 95, 99, 2, 226, 11, 249, 141, 119, 102, 235, 7, 12],
	#[42, 726],
	#[299, 727, 253, 579, 275, 133, 249, 156, 252, 152, 250, 131, 248, 157, 73, 137, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 71, 129, 36, 128, 33, 138, 35, 122, 11, 126, 7, 12, 31, 127],
	#[43, 728, 13, 588],
	#[280, 729],
	#[322, 730, 185, 186, 183, 173, 181, 185, 176, 184, 175, 177, 173, 181, 172, 190, 171, 179, 177, 192, 184, 176, 178, 175, 179, 180, 182, 188, 180, 182, 59, 178, 58, 174, 174, 183, 57, 191, 13, 454, 186, 189, 187, 172, 293, 187],
	#[],
	#[294, 732, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 118, 35, 70, 139, 68, 132, 64, 140, 63, 120, 73, 137, 66, 125, 67, 136, 69, 135, 65, 121, 71, 129, 36, 128, 35, 122, 33, 138, 11, 126, 7, 12, 31, 127],
	#[284, 733],
	#[322, 734, 13, 735],
	#[],
	#[323, 455, 294, 367, 275, 133, 252, 124, 250, 131, 225, 456, 248, 134, 249, 123, 69, 135, 67, 136, 66, 125, 64, 140, 65, 121, 73, 137, 36, 128, 63, 120, 31, 127, 70, 139, 35, 122, 33, 138, 71, 129, 11, 126, 68, 132, 118, 35, 7, 12],
	#[294, 737, 252, 124, 249, 123, 250, 131, 248, 134, 275, 133, 118, 35, 73, 137, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[13, 738],
	#[275, 133, 252, 124, 249, 123, 294, 739, 250, 131, 248, 134, 118, 35, 71, 129, 64, 140, 73, 137, 66, 125, 11, 126, 69, 135, 33, 138, 36, 128, 67, 136, 65, 121, 35, 122, 7, 12, 63, 120, 70, 139, 68, 132, 31, 127],
	#[13, 740],
	#[294, 741, 275, 133, 252, 124, 250, 131, 249, 123, 248, 134, 73, 137, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 71, 129, 36, 128, 35, 122, 33, 138, 31, 127, 11, 126, 7, 12],
	#[306, 742, 206, 461],
	#[210, 466, 211, 468, 208, 464, 209, 465, 307, 743, 207, 463, 212, 469],
	#[322, 744, 13, 454],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[],
	#[273, 49, 290, 47, 162, 46, 62, 751],
	#[],
	#[],
	#[],
	#[240, 756, 243, 757, 56, 555, 54, 565, 53, 563, 48, 566, 45, 560, 49, 554, 50, 558, 55, 553, 51, 559, 52, 561, 46, 562, 44, 557, 47, 564],
	#[],
	#[],
	#[244, 758],
	#[275, 133, 252, 124, 250, 131, 249, 123, 272, 759, 248, 575, 245, 570, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 66, 125, 65, 121, 63, 120, 59, 568, 58, 572, 57, 574, 64, 140, 73, 569, 71, 129, 33, 138, 60, 573, 35, 122, 11, 126, 7, 12, 36, 128, 31, 127],
	#[242, 762, 277, 760, 10, 761, 119, 32],
	#[],
	#[],
	#[42, 763],
	#[279, 767, 275, 133, 253, 764, 252, 152, 250, 131, 249, 156, 248, 157, 278, 766, 118, 35, 98, 765, 73, 137, 71, 129, 70, 139, 69, 135, 68, 132, 65, 121, 64, 140, 63, 120, 66, 125, 67, 136, 36, 128, 35, 122, 33, 138, 11, 126, 7, 12, 31, 127],
	#[313, 214, 301, 787, 222, 211, 275, 771, 223, 213, 220, 206, 108, 207, 93, 210, 109, 212, 60, 208, 7, 12, 221, 209, 118, 35],
	#[],
	#[43, 773],
	#[13, 768],
	#[252, 152, 253, 769, 248, 157, 249, 156, 275, 133, 250, 131, 73, 137, 118, 35, 71, 129, 68, 132, 64, 140, 65, 121, 69, 135, 70, 139, 66, 125, 67, 136, 36, 128, 35, 122, 98, 770, 33, 138, 11, 126, 7, 12, 63, 120, 31, 127],
	#[313, 214, 301, 772, 275, 771, 220, 206, 221, 209, 222, 211, 223, 213, 60, 208, 93, 210, 118, 35, 108, 207, 109, 212, 7, 12],
	#[],
	#[],
	#[],
	#[257, 774, 81, 775],
	#[280, 776],
	#[],
	#[293, 187, 184, 176, 178, 175, 180, 182, 179, 180, 176, 184, 183, 173, 262, 777, 187, 172, 186, 189, 177, 192, 174, 183, 172, 190, 281, 778, 173, 181, 171, 179, 182, 188, 181, 185, 185, 186, 94, 779, 58, 174, 57, 191, 59, 178, 175, 177],
	#[],
	#[282, 781, 93, 782],
	#[7, 780],
	#[],
	#[283, 785, 120, 784],
	#[38, 783],
	#[],
	#[7, 786],
	#[],
	#[],
	#[],
	#[232, 789, 11, 790],
	#[],
	#[233, 792, 7, 791],
	#[],
	#[13, 793, 12, 794],
	#[7, 795],
	#[],
	#[],
	#[226, 797],
	#[31, 798],
	#[321, 801, 318, 800, 32, 799, 226, 802],
	#[],
	#[32, 803, 13, 804],
	#[],
	#[38, 318],
	#[],
	#[321, 805, 226, 802],
	#[],
	#[254, 847, 75, 755, 74, 753, 76, 752],
	#[254, 808, 76, 752, 75, 755, 74, 753],
	#[255, 810, 77, 811, 3, 809],
	#[234, 831, 26, 27, 25, 25, 24, 2, 23, 15, 22, 30, 21, 9, 20, 22, 18, 14, 17, 28, 16, 13, 19, 37, 14, 8, 15, 29],
	#[251, 818, 72, 160],
	#[42, 812],
	#[256, 814, 78, 815, 80, 813, 79, 816],
	#[],
	#[43, 817],
	#[],
	#[],
	#[],
	#[257, 819, 81, 775],
	#[258, 822, 83, 821, 82, 820],
	#[],
	#[],
	#[275, 133, 252, 124, 249, 123, 248, 262, 259, 823, 250, 131, 73, 137, 118, 35, 71, 129, 70, 139, 69, 135, 68, 132, 66, 125, 65, 121, 64, 140, 67, 136, 36, 128, 63, 120, 33, 138, 11, 126, 35, 122, 31, 127, 7, 12],
	#[260, 825, 13, 824],
	#[262, 826, 261, 828, 93, 827, 94, 779],
	#[],
	#[],
	#[38, 830],
	#[260, 829, 13, 824],
	#[],
	#[],
	#[272, 834, 263, 835, 275, 133, 248, 575, 249, 123, 250, 131, 252, 124, 118, 35, 97, 833, 95, 832, 70, 139, 64, 140, 66, 125, 67, 136, 73, 569, 65, 121, 36, 128, 35, 122, 33, 138, 31, 127, 68, 132, 71, 129, 69, 135, 11, 126, 7, 12, 63, 120],
	#[42, 842],
	#[224, 361, 316, 837],
	#[277, 237, 275, 242, 264, 836, 119, 32, 118, 35, 7, 12, 63, 231, 10, 222],
	#[],
	#[],
	#[42, 838],
	#[275, 133, 259, 839, 252, 124, 250, 131, 249, 123, 248, 262, 118, 35, 73, 137, 71, 129, 70, 139, 64, 140, 65, 121, 67, 136, 68, 132, 69, 135, 66, 125, 36, 128, 11, 126, 35, 122, 33, 138, 31, 127, 7, 12, 63, 120],
	#[284, 840],
	#[43, 841, 13, 366],
	#[],
	#[272, 834, 263, 843, 275, 133, 252, 124, 250, 131, 248, 575, 249, 123, 97, 833, 71, 129, 70, 139, 67, 136, 66, 125, 73, 569, 68, 132, 36, 128, 35, 122, 63, 120, 33, 138, 31, 127, 65, 121, 64, 140, 69, 135, 11, 126, 95, 832, 118, 35, 7, 12],
	#[96, 844],
	#[275, 133, 252, 124, 250, 131, 248, 845, 249, 123, 118, 35, 70, 139, 69, 135, 68, 132, 67, 136, 65, 121, 64, 140, 73, 137, 66, 125, 71, 129, 36, 128, 35, 122, 63, 120, 33, 138, 11, 126, 7, 12, 31, 127],
	#[251, 159, 72, 160, 43, 846, 42, 158],
	#[],
	#[255, 848, 77, 811],
	#[257, 849, 81, 775],
	#[251, 850, 72, 160],
	#[258, 851, 82, 820, 83, 821],
	#[249, 123, 248, 852, 275, 133, 68, 132, 63, 120, 252, 124, 66, 125, 11, 126, 250, 131, 7, 12, 73, 137, 65, 121, 64, 140, 33, 138, 35, 122, 118, 35, 69, 135, 67, 136, 36, 128, 70, 139, 71, 129, 31, 127],
	#[260, 853, 251, 159, 72, 160, 13, 824, 42, 158],
	#[],
	#[],
	#[230, 856, 7, 857],
	#[],
	#[],
	#[6, 859],
	#[275, 133, 252, 124, 250, 131, 248, 860, 249, 123, 118, 35, 73, 137, 70, 139, 69, 135, 66, 125, 64, 140, 65, 121, 71, 129, 67, 136, 36, 128, 35, 122, 33, 138, 68, 132, 7, 12, 11, 126, 63, 120, 31, 127],
	#[251, 159, 226, 861, 72, 160, 42, 158],
	#[31, 862],
	#[118, 35, 85, 322, 73, 137, 71, 129, 70, 139, 69, 135, 67, 136, 66, 125, 65, 121, 64, 140, 63, 120, 68, 132, 36, 128, 33, 138, 32, 863, 35, 122, 11, 126, 31, 127, 7, 12, 319, 864, 275, 133, 252, 124, 250, 131, 249, 123, 248, 262, 259, 320],
	#[],
	#[32, 865, 13, 324],
	#[],
	#[255, 867, 77, 811],
	#[258, 868, 83, 821, 82, 820],
	#[275, 133, 259, 869, 252, 124, 250, 131, 248, 262, 249, 123, 118, 35, 73, 137, 71, 129, 65, 121, 64, 140, 66, 125, 67, 136, 35, 122, 63, 120, 68, 132, 7, 12, 31, 127, 70, 139, 33, 138, 11, 126, 69, 135, 36, 128],
	#[260, 870, 13, 824],
	#[],
	#[],
	#[],
	#[],
	#[75, 755, 74, 753, 76, 752, 254, 875],
	#[240, 876, 243, 757, 48, 566, 49, 554, 56, 555, 51, 559, 44, 557, 50, 558, 54, 565, 47, 564, 53, 563, 45, 560, 52, 561, 46, 562, 55, 553],
	#[],
	#[6, 881],
	#[6, 879],
	#[7, 880],
	#[],
	#[7, 882],
	#[],
	#[255, 884, 77, 811],
	#[258, 885, 82, 820, 83, 821],
	#[275, 133, 252, 124, 250, 131, 249, 123, 248, 886, 118, 35, 70, 139, 68, 132, 67, 136, 65, 121, 64, 140, 11, 126, 66, 125, 36, 128, 33, 138, 71, 129, 7, 12, 63, 120, 31, 127, 69, 135, 35, 122, 73, 137],
	#[251, 159, 260, 887, 72, 160, 42, 158, 13, 824],
	#[],
	#[33, 896, 67, 136, 250, 889, 70, 139, 64, 140, 71, 129, 248, 890, 66, 125, 249, 123, 35, 895, 36, 893, 65, 121, 68, 132, 73, 137, 11, 892, 118, 35, 69, 135, 275, 133, 7, 12, 63, 120, 31, 891, 252, 894],
	#[],
	#[251, 914, 72, 160, 42, 158],
	#[252, 124, 250, 131, 249, 123, 248, 143, 267, 911, 275, 133, 118, 35, 70, 139, 68, 132, 64, 140, 63, 120, 73, 137, 67, 136, 69, 135, 36, 128, 35, 122, 65, 121, 32, 912, 66, 125, 7, 12, 33, 138, 11, 126, 31, 127, 71, 129],
	#[38, 907],
	#[],
	#[],
	#[],
	#[38, 897, 31, 898],
	#[39, 904],
	#[275, 133, 267, 899, 252, 124, 250, 131, 249, 123, 248, 143, 118, 35, 64, 140, 67, 136, 65, 121, 69, 135, 70, 139, 33, 138, 11, 126, 35, 122, 32, 900, 73, 137, 71, 129, 66, 125, 7, 12, 63, 120, 36, 128, 68, 132, 31, 127],
	#[32, 902, 13, 148],
	#[34, 901],
	#[],
	#[34, 903],
	#[],
	#[275, 133, 252, 124, 248, 905, 250, 131, 249, 123, 118, 35, 69, 135, 64, 140, 36, 128, 63, 120, 66, 125, 65, 121, 67, 136, 33, 138, 73, 137, 35, 122, 68, 132, 11, 126, 7, 12, 70, 139, 71, 129, 31, 127],
	#[251, 159, 72, 160, 42, 158, 34, 906],
	#[],
	#[39, 908],
	#[249, 123, 275, 133, 250, 131, 248, 909, 252, 124, 118, 35, 73, 137, 65, 121, 64, 140, 66, 125, 67, 136, 68, 132, 69, 135, 71, 129, 36, 128, 35, 122, 63, 120, 33, 138, 31, 127, 7, 12, 70, 139, 11, 126],
	#[251, 159, 72, 160, 12, 910, 42, 158],
	#[],
	#[13, 148, 32, 913],
	#[],
	#[],
	#[37, 915],
	#[],
	#[],
	#[]],
  action-function-table:
	 vector(llvm-parser-action0,
		llvm-parser-action1,
		llvm-parser-action0,
		llvm-parser-action3,
		llvm-parser-action4,
		llvm-parser-action5,
		llvm-parser-action6,
		llvm-parser-action7,
		llvm-parser-action8,
		llvm-parser-action9,
		llvm-parser-action10,
		llvm-parser-action11,
		llvm-parser-action12,
		llvm-parser-action13,
		llvm-parser-action13,
		llvm-parser-action15,
		llvm-parser-action16,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action19,
		llvm-parser-action20,
		llvm-parser-action21,
		llvm-parser-action22,
		llvm-parser-action23,
		llvm-parser-action24,
		llvm-parser-action25,
		llvm-parser-action26,
		llvm-parser-action27,
		llvm-parser-action28,
		llvm-parser-action29,
		llvm-parser-action30,
		llvm-parser-action31,
		llvm-parser-action32,
		llvm-parser-action33,
		llvm-parser-action34,
		llvm-parser-action35,
		llvm-parser-action36,
		llvm-parser-action37,
		llvm-parser-action38,
		llvm-parser-action39,
		llvm-parser-action40,
		llvm-parser-action41,
		llvm-parser-action42,
		llvm-parser-action43,
		llvm-parser-action44,
		llvm-parser-action45,
		llvm-parser-action46,
		llvm-parser-action13,
		llvm-parser-action12,
		llvm-parser-action0,
		llvm-parser-action0,
		llvm-parser-action51,
		llvm-parser-action52,
		llvm-parser-action53,
		llvm-parser-action54,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action59,
		llvm-parser-action60,
		llvm-parser-action61,
		llvm-parser-action62,
		llvm-parser-action63,
		llvm-parser-action64,
		llvm-parser-action65,
		llvm-parser-action66,
		llvm-parser-action67,
		llvm-parser-action68,
		llvm-parser-action69,
		llvm-parser-action70,
		llvm-parser-action71,
		llvm-parser-action72,
		llvm-parser-action73,
		llvm-parser-action74,
		llvm-parser-action75,
		llvm-parser-action76,
		llvm-parser-action77,
		llvm-parser-action78,
		llvm-parser-action79,
		llvm-parser-action79,
		llvm-parser-action81,
		llvm-parser-action82,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action85,
		llvm-parser-action86,
		llvm-parser-action87,
		llvm-parser-action88,
		llvm-parser-action89,
		llvm-parser-action90,
		llvm-parser-action91,
		llvm-parser-action92,
		llvm-parser-action93,
		llvm-parser-action94,
		llvm-parser-action95,
		llvm-parser-action96,
		llvm-parser-action97,
		llvm-parser-action98,
		llvm-parser-action99,
		llvm-parser-action100,
		llvm-parser-action101,
		llvm-parser-action102,
		llvm-parser-action103,
		llvm-parser-action104,
		llvm-parser-action105,
		llvm-parser-action106,
		llvm-parser-action107,
		llvm-parser-action108,
		llvm-parser-action109,
		llvm-parser-action109,
		llvm-parser-action111,
		llvm-parser-action112,
		llvm-parser-action113,
		llvm-parser-action114,
		llvm-parser-action115,
		llvm-parser-action116,
		llvm-parser-action1,
		llvm-parser-action118,
		llvm-parser-action106,
		llvm-parser-action120,
		llvm-parser-action121,
		llvm-parser-action122,
		llvm-parser-action1,
		llvm-parser-action118,
		llvm-parser-action125,
		llvm-parser-action126,
		llvm-parser-action127,
		llvm-parser-action128,
		llvm-parser-action129,
		llvm-parser-action127,
		llvm-parser-action127,
		llvm-parser-action128,
		llvm-parser-action133,
		llvm-parser-action134,
		llvm-parser-action135,
		llvm-parser-action136,
		llvm-parser-action137,
		llvm-parser-action137,
		llvm-parser-action135,
		llvm-parser-action140,
		llvm-parser-action140,
		llvm-parser-action142,
		llvm-parser-action143,
		llvm-parser-action144,
		llvm-parser-action145,
		llvm-parser-action137,
		llvm-parser-action147,
		llvm-parser-action148,
		llvm-parser-action149,
		llvm-parser-action150,
		llvm-parser-action72,
		llvm-parser-action137,
		llvm-parser-action153,
		llvm-parser-action154,
		llvm-parser-action155,
		llvm-parser-action156,
		llvm-parser-action157,
		llvm-parser-action158,
		llvm-parser-action159,
		llvm-parser-action160,
		llvm-parser-action161,
		llvm-parser-action162,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action169,
		llvm-parser-action154,
		llvm-parser-action171,
		llvm-parser-action172,
		llvm-parser-action173,
		llvm-parser-action174,
		llvm-parser-action175,
		llvm-parser-action176,
		llvm-parser-action177,
		llvm-parser-action178,
		llvm-parser-action179,
		llvm-parser-action180,
		llvm-parser-action181,
		llvm-parser-action182,
		llvm-parser-action183,
		llvm-parser-action184,
		llvm-parser-action185,
		llvm-parser-action186,
		llvm-parser-action187,
		llvm-parser-action188,
		llvm-parser-action189,
		llvm-parser-action190,
		llvm-parser-action191,
		llvm-parser-action192,
		llvm-parser-action193,
		llvm-parser-action18,
		llvm-parser-action12,
		llvm-parser-action196,
		llvm-parser-action197,
		llvm-parser-action197,
		llvm-parser-action1,
		llvm-parser-action13,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action13,
		llvm-parser-action12,
		llvm-parser-action159,
		llvm-parser-action160,
		llvm-parser-action161,
		llvm-parser-action162,
		llvm-parser-action209,
		llvm-parser-action210,
		llvm-parser-action73,
		llvm-parser-action74,
		llvm-parser-action1,
		llvm-parser-action12,
		llvm-parser-action1,
		llvm-parser-action72,
		llvm-parser-action1,
		llvm-parser-action72,
		llvm-parser-action16,
		llvm-parser-action18,
		llvm-parser-action221,
		llvm-parser-action222,
		llvm-parser-action223,
		llvm-parser-action224,
		llvm-parser-action225,
		llvm-parser-action226,
		llvm-parser-action227,
		llvm-parser-action228,
		llvm-parser-action229,
		llvm-parser-action230,
		llvm-parser-action231,
		llvm-parser-action232,
		llvm-parser-action233,
		llvm-parser-action234,
		llvm-parser-action16,
		llvm-parser-action236,
		llvm-parser-action237,
		llvm-parser-action238,
		llvm-parser-action238,
		llvm-parser-action240,
		llvm-parser-action241,
		llvm-parser-action242,
		llvm-parser-action243,
		llvm-parser-action244,
		llvm-parser-action245,
		llvm-parser-action246,
		llvm-parser-action247,
		llvm-parser-action248,
		llvm-parser-action249,
		llvm-parser-action250,
		llvm-parser-action251,
		llvm-parser-action252,
		llvm-parser-action253,
		llvm-parser-action254,
		llvm-parser-action255,
		llvm-parser-action256,
		llvm-parser-action257,
		llvm-parser-action258,
		llvm-parser-action259,
		llvm-parser-action260,
		llvm-parser-action261,
		llvm-parser-action262,
		llvm-parser-action263,
		llvm-parser-action264,
		llvm-parser-action265,
		llvm-parser-action266,
		llvm-parser-action267,
		llvm-parser-action254,
		llvm-parser-action255,
		llvm-parser-action256,
		llvm-parser-action257,
		llvm-parser-action272,
		llvm-parser-action273,
		llvm-parser-action274,
		llvm-parser-action275,
		llvm-parser-action276,
		llvm-parser-action277,
		llvm-parser-action277,
		llvm-parser-action279,
		llvm-parser-action280,
		llvm-parser-action281,
		llvm-parser-action282,
		llvm-parser-action283,
		llvm-parser-action284,
		llvm-parser-action285,
		llvm-parser-action286,
		llvm-parser-action287,
		llvm-parser-action288,
		llvm-parser-action289,
		llvm-parser-action290,
		llvm-parser-action291,
		llvm-parser-action75,
		llvm-parser-action76,
		llvm-parser-action77,
		llvm-parser-action295,
		llvm-parser-action296,
		llvm-parser-action297,
		llvm-parser-action298,
		llvm-parser-action299,
		llvm-parser-action300,
		llvm-parser-action301,
		llvm-parser-action302,
		llvm-parser-action303,
		llvm-parser-action304,
		llvm-parser-action305,
		llvm-parser-action306,
		llvm-parser-action307,
		llvm-parser-action308,
		llvm-parser-action309,
		llvm-parser-action310,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action315,
		llvm-parser-action316,
		llvm-parser-action0,
		llvm-parser-action318,
		llvm-parser-action16,
		llvm-parser-action320,
		llvm-parser-action321,
		llvm-parser-action73,
		llvm-parser-action74,
		llvm-parser-action1,
		llvm-parser-action12,
		llvm-parser-action326,
		llvm-parser-action327,
		llvm-parser-action328,
		llvm-parser-action329,
		llvm-parser-action330,
		llvm-parser-action331,
		llvm-parser-action332,
		llvm-parser-action333,
		llvm-parser-action334,
		llvm-parser-action335,
		llvm-parser-action336,
		llvm-parser-action337,
		llvm-parser-action338,
		llvm-parser-action339,
		llvm-parser-action340,
		llvm-parser-action341,
		llvm-parser-action342,
		llvm-parser-action343,
		llvm-parser-action344,
		llvm-parser-action345,
		llvm-parser-action346,
		llvm-parser-action347,
		llvm-parser-action348,
		llvm-parser-action349,
		llvm-parser-action350,
		llvm-parser-action351,
		llvm-parser-action352,
		llvm-parser-action353,
		llvm-parser-action1,
		llvm-parser-action118,
		llvm-parser-action16,
		llvm-parser-action357,
		llvm-parser-action357,
		llvm-parser-action359,
		llvm-parser-action360,
		llvm-parser-action361,
		llvm-parser-action362,
		llvm-parser-action363,
		llvm-parser-action364,
		llvm-parser-action365,
		llvm-parser-action366,
		llvm-parser-action367,
		llvm-parser-action368,
		llvm-parser-action369,
		llvm-parser-action370,
		llvm-parser-action371,
		llvm-parser-action372,
		llvm-parser-action373,
		llvm-parser-action374,
		llvm-parser-action375,
		llvm-parser-action376,
		llvm-parser-action377,
		llvm-parser-action378,
		llvm-parser-action379,
		llvm-parser-action380,
		llvm-parser-action380,
		llvm-parser-action382,
		llvm-parser-action383,
		llvm-parser-action384,
		llvm-parser-action385,
		llvm-parser-action386,
		llvm-parser-action385,
		llvm-parser-action388,
		llvm-parser-action388,
		llvm-parser-action390,
		llvm-parser-action391,
		llvm-parser-action388,
		llvm-parser-action390,
		llvm-parser-action394,
		llvm-parser-action395,
		llvm-parser-action396,
		llvm-parser-action12,
		llvm-parser-action398,
		llvm-parser-action12,
		llvm-parser-action147,
		llvm-parser-action401,
		llvm-parser-action402,
		llvm-parser-action403,
		llvm-parser-action12,
		llvm-parser-action12,
		llvm-parser-action76,
		llvm-parser-action77,
		llvm-parser-action75,
		llvm-parser-action76,
		llvm-parser-action77,
		llvm-parser-action411,
		llvm-parser-action78,
		llvm-parser-action413,
		llvm-parser-action414,
		llvm-parser-action415,
		llvm-parser-action416,
		llvm-parser-action417,
		llvm-parser-action418,
		llvm-parser-action1,
		llvm-parser-action118,
		llvm-parser-action1,
		llvm-parser-action118,
		llvm-parser-action423,
		llvm-parser-action424,
		llvm-parser-action425,
		llvm-parser-action426,
		llvm-parser-action427,
		llvm-parser-action428,
		llvm-parser-action17,
		llvm-parser-action18,
		llvm-parser-action17,
		llvm-parser-action432,
		llvm-parser-action18,
		llvm-parser-action434,
		llvm-parser-action12,
		llvm-parser-action436,
		llvm-parser-action437,
		llvm-parser-action438,
		llvm-parser-action439,
		llvm-parser-action16,
		llvm-parser-action72,
		llvm-parser-action442,
		llvm-parser-action443),
  action-nargs-table: #[2, 0, 2, 3, 9, 6, 9, 6, 6, 4, 4, 3, 1, 2, 2, 3, 0, 1, 3, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 5, 4, 7, 6, 3, 3, 3, 5, 7, 7, 3, 2, 1, 2, 2, 2, 3, 3, 12, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 0, 2, 1, 1, 1, 1, 3, 3, 1, 2, 1, 1, 1, 1, 1, 3, 1, 1, 5, 5, 3, 2, 5, 4, 1, 1, 1, 1, 1, 1, 1, 1, 0, 4, 5, 5, 1, 1, 2, 2, 0, 1, 1, 1, 0, 1, 4, 1, 1, 1, 0, 1, 1, 1, 4, 3, 3, 4, 4, 3, 6, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 7, 2, 0, 3, 1, 2, 2, 2, 6, 6, 1, 1, 1, 1, 0, 1, 3, 1, 1, 3, 1, 3, 1, 3, 6, 6, 8, 6, 6, 7, 7, 6, 8, 8, 5, 7, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 1, 1, 3, 3, 0, 2, 1, 1, 2, 1, 0, 1, 3, 1, 2, 4, 0, 2, 0, 1, 0, 2, 0, 2, 0, 3, 2, 1, 2, 1, 2, 1, 1, 2, 1, 2, 1, 1, 1, 1, 0, 1, 1, 2, 2, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 2, 2, 2, 3, 3, 4, 10, 10, 9, 7, 6, 16, 2, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 1, 1, 1, 2, 1, 3, 1, 1, 5, 6, 2, 4, 0, 3, 5, 0, 2, 0, 1, 6, 6, 7, 7, 5, 7, 5, 5, 7, 7, 4, 7, 10, 3, 6, 4, 7, 4, 6, 9, 6, 10, 9, 4, 5, 6, 4, 5, 0, 1, 0, 3, 3, 2, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 2, 2, 3, 2, 5, 4, 6, 1, 5, 1, 0, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 5, 7, 0, 1, 0, 1, 1, 4, 5, 6, 7, 8, 1, 3, 1, 1, 3, 3, 1, 3, 4, 2, 2, 0, 2, 2, 4],
  action-nt-table: #[227, 228, 228, 229, 229, 229, 229, 229, 229, 229, 229, 229, 230, 231, 231, 232, 233, 233, 233, 234, 234, 234, 234, 234, 234, 234, 234, 234, 234, 234, 234, 234, 234, 235, 235, 235, 229, 229, 229, 229, 229, 229, 229, 229, 229, 229, 229, 236, 236, 229, 229, 237, 238, 239, 240, 241, 241, 242, 242, 243, 243, 243, 243, 243, 243, 243, 243, 243, 243, 243, 243, 243, 243, 244, 244, 245, 245, 245, 245, 246, 246, 247, 247, 248, 248, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 250, 250, 250, 250, 250, 250, 250, 250, 251, 251, 252, 252, 253, 253, 253, 253, 254, 254, 254, 254, 255, 255, 255, 256, 256, 256, 257, 257, 258, 258, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 259, 260, 260, 261, 261, 262, 263, 263, 263, 264, 264, 264, 264, 265, 265, 265, 265, 266, 266, 267, 267, 268, 268, 269, 269, 269, 269, 269, 269, 269, 269, 269, 269, 269, 269, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 270, 271, 271, 272, 272, 273, 273, 274, 274, 275, 275, 276, 277, 278, 278, 278, 278, 279, 279, 280, 280, 281, 281, 282, 282, 283, 283, 284, 284, 285, 285, 285, 285, 285, 285, 285, 285, 285, 285, 285, 285, 285, 285, 286, 286, 286, 286, 286, 287, 287, 287, 287, 287, 287, 287, 287, 288, 288, 288, 288, 288, 288, 288, 288, 288, 288, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 289, 290, 290, 290, 291, 291, 292, 292, 292, 292, 292, 292, 292, 292, 292, 292, 292, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 293, 294, 295, 295, 296, 296, 297, 297, 298, 298, 299, 299, 299, 300, 300, 301, 301, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 302, 303, 303, 304, 304, 304, 305, 305, 306, 306, 307, 307, 307, 307, 307, 307, 308, 308, 308, 308, 308, 308, 308, 308, 308, 308, 308, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 309, 310, 310, 310, 310, 311, 311, 312, 312, 312, 313, 313, 313, 313, 313, 313, 313, 313, 314, 314, 315, 315, 316, 316, 317, 317, 229, 229, 229, 229, 318, 318, 319, 319, 319, 319, 320, 320, 320, 320, 321, 322, 322, 323, 323]);
